﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _07_Epsilon
    {
        public int NbFrames = 63;//84;
        public int BarSize = 32;
        public int NbBarsPerBranches = 8; // must be div 2
        public int NbBranches = 4;
        public double Scale = 4.32*0.78;

        public _07_Epsilon()
        {
            try
            {
                CreateRaster(   Program.ToDataFolder("07_Epsilon/Rasters.png"),
                                Program.ToZ80Folder("07_Epsilon/_Rasters.png.z80"));

                CreateVerticalBars(
                    96*4, 
                    NbFrames*2, // *2 because R9=1
                    Program.ToDataFolder("07_Epsilon/Bars.png"),
                    Program.ToZ80Folder("07_Epsilon/_Epsilon.bmp")
                    );

                CreateSinMoves(
                    Program.ToZ80Folder("07_Epsilon/_SinXDisto1.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinXDisto2.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinYDisto1.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinYDisto2.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinXFrameMove.bin")
                    );
                
                CreatePalettes(
                    Program.ToDataFolder("07_Epsilon/Bars2.png"),
                    Program.ToZ80Folder("07_Epsilon/_Palettes.bin")
                    );
            }
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}
        }

        private void CreateRaster(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }

        private Framework.Math3D.Vector3D[] CreateGeometry(double scale, out double minZ, out double maxZ)
        {
            var vertices = new Framework.Math3D.Vector3D[NbBarsPerBranches * NbBranches];
            var iVert = 0;

            int iSkipFirst = 1;

            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D((iSkipFirst + iNbBarPerBranch) * scale, 0, 0);
                iVert++;
            }
            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D((-iSkipFirst - iNbBarPerBranch) * scale, 0, 0);
                iVert++;
            }
            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D(0, 0, (iSkipFirst + iNbBarPerBranch) * scale);
                iVert++;
            }
            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D(0, 0, (-iSkipFirst - iNbBarPerBranch) * scale);
                iVert++;
            }

            minZ = (double)(-(iSkipFirst + (NbBarsPerBranches - 1)) * scale);
            maxZ = (double)((iSkipFirst + (NbBarsPerBranches - 1)) * scale);

            return vertices;
        }

        // zoomScale = 0 mininum  1 maximum
        private void DrawBar(Bitmap barsBitmap, Bitmap bitmap, int x, int y, double zoomScale) 
        {
            var max = 255 / NbBarsPerBranches;

            if(zoomScale<0)
            {
                zoomScale = 0;
            }
            else if (zoomScale > 1)
            {
                zoomScale = 1;
            }

            zoomScale = 1 - zoomScale;

            int iBar = (int)((zoomScale) * (16.9999));
            if (iBar<0)
            {
                iBar = 0;
            }
            else if (iBar>16)
            {
                iBar = 16;
            }
            var edgePixelsToRemove = iBar;
            iBar *= 36;

            for(int i=0; i<BarSize-(edgePixelsToRemove*2); i++)
            {
                var iX = x + i + edgePixelsToRemove;
                if ((iX>=0)&&(iX<bitmap.Width))
                {
                    bitmap.SetPixel(iX, y, barsBitmap.GetPixel(i, iBar));
                    bitmap.SetPixel(iX, y+1, barsBitmap.GetPixel(i, iBar+1));
                }
            }
        }

        public void CreateVerticalBars(int width, int height, string barsFilename, string filename)
        {
            var barsBitmap = new Bitmap(Bitmap.FromFile(barsFilename));
            var bitmap = new Bitmap(width, height);

            for (int iFrame = 0; iFrame < NbFrames; iFrame++)
            {
                double yRot = (360.0 / NbFrames / 4) * iFrame;

                double minZ; // -7.0 ..
                double maxZ; // 7.0 ...
                var vertices = CreateGeometry(Scale, out minZ, out maxZ);

                vertices = Math3D.RotateY(vertices, yRot);

                vertices = Math3D.DepthSort(vertices);
                
                for (int i = 0; i < vertices.Length; i++)
                {
                    var vec = vertices[i];

                    var x2D = (double)(-vec.X/(0.1));

                    x2D += (width / 2)-(barsBitmap.Width/2);

                    var ratio = vec.Z - minZ;
                    var length = maxZ - minZ;

                    double zoomScale = 0;
                    double limit = 3.3;
                    if (ratio>(length/ limit))
                    {
                        ratio -= length / limit;
                        ratio *= limit;
                        zoomScale = (double)(ratio / length);
                    }

                    //zoomScale = (double)((vec.Z - minZ) /(maxZ-minZ));

                    DrawBar(barsBitmap, bitmap, (int)(x2D), iFrame*2, zoomScale);
                }
            }

            bitmap.Save(filename, System.Drawing.Imaging.ImageFormat.Bmp);

            CreateVerticalBars_CPC(bitmap, filename);
        }

        private void CreateVerticalBars_CPC(Bitmap bitmap, string bmpFilename)
        {
            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 48,
                R4 = 21,
                R9 = 1,
                StartVLC = 0,
                StartR12 = 0x00
            });
            
            for (var iPage = 0; iPage < image.Pages.Length; iPage++)
            {
                foreach (var VLC in image.Pages[iPage].VLCs)
                {
                    File.WriteAllBytes(bmpFilename + ".page" + iPage + ".vlc" + VLC.Index, VLC.CPCBytes);
                }
            }

            var r12r13 = new List<byte>();
            for(int i=0; i<image.R12R13.Length; i+=4)
            {
                r12r13.Add(image.R12R13[i]);
                r12r13.Add(image.R12R13[i+1]);
            }
            File.WriteAllBytes(bmpFilename + ".r12r13", r12r13.ToArray());

            File.WriteAllText(
                bmpFilename + ".pal.z80",
                image.Palette.ToZ80TableString(Program.FilenameToKeyWordName(bmpFilename))
            );
        }

        public void CreateSinMoves( string sinXDisto1Filename, 
                                    string sinXDisto2Filename, 
                                    string sinYDisto1Filename, 
                                    string sinYDisto2Filename, 
                                    string sinXFrameMove)
        {
            File.WriteAllBytes(sinXDisto1Filename, SinCurve.CreateCos16Bits(256, 14));
            File.WriteAllBytes(sinXDisto2Filename, SinCurve.CreateSin16Bits(256, 11, 2));

            File.WriteAllBytes(sinYDisto1Filename, SinCurve.CreateSin16Bits(256, 5));
            File.WriteAllBytes(sinYDisto2Filename, SinCurve.CreateCos16Bits(256, 2));

            var moveXMax = -6;
            File.WriteAllBytes(sinXFrameMove, SinCurve.CreateSin(256, 34-moveXMax, 60+moveXMax));
            var incMove = new List<byte>();
            for(int i=0; i<256; i++)
            {
                incMove.Add((byte)i);
            }
            File.WriteAllBytes(sinXFrameMove, incMove.ToArray());
        }

        public void CreatePalettes(string barsFilename, string filename)
        {
            var barsBitmap = new Bitmap(Bitmap.FromFile(barsFilename));

            var palettes = new List<byte>();

            for (int iSubBitmap = 0; iSubBitmap < barsBitmap.Height / 2; iSubBitmap++)
            {
                var subBitmap = new Bitmap(barsBitmap.Width, 2);

                for (int y = iSubBitmap * 2; y < (iSubBitmap * 2) + 2; y++)
                {
                    for (int x = 0; x < barsBitmap.Width; x++)
                    {
                        subBitmap.SetPixel(x, y - iSubBitmap * 2, barsBitmap.GetPixel(x, y));
                    }
                }

                var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = subBitmap,
                    Mode = 1,
                    R1 = 8,
                    R4 = 0,
                    R9 = 1,
                    StartVLC = 0,
                    StartR12 = 0x00
                });

                var gateArrayPalette = image.Palette.ToGateArrayByteArray();
                for (int iRepeat = 0; iRepeat < 1; iRepeat++)
                {
                    for (int i = 1; i < 4; i++)
                    {
                        palettes.Add(gateArrayPalette[i]);
                    }
                }
            }

            File.WriteAllBytes(filename, palettes.ToArray());
        }
    }
}
