﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _09_SineScroll
    {
        private const int CharHeight = 21;

        public _09_SineScroll()
        {
            try
            {
                CreateBackgroundRasters(
                    Program.ToDataFolder("09_SineScroll/BackgroundRasters.png"),
                    Program.ToDataFolder("09_SineScroll/BackgroundRasters.png"),
                    Program.ToZ80Folder("09_SineScroll/_BackgroundRasters.png.z80"));

                CreateForegroundRasters(
                    Program.ToDataFolder("09_SineScroll/ForegroundRasters.png"),
                    Program.ToDataFolder("09_SineScroll/ForegroundRasters.png"),
                    Program.ToZ80Folder("09_SineScroll/_ForegroundRasters.png.z80"));

                CreateFont(     Program.ToDataFolder("09_SineScroll/ProportionalFont_x21_Fixed.png"),
                                Program.ToZ80Folder("09_SineScroll/_ScrollText.z80"),
                                Program.ToZ80Folder("09_SineScroll/_VertSlices.bin"),
                                " ABCDEFGHIJKLMNOPQRSTUVWXYZ!:.?()1234567890\\/@",
                                "                                  WARNING ... WHAT IS SEEN CANNOT BE UNSEEN !!!                                 "
                                );

                CreateSinMoves(
                    Program.ToZ80Folder("09_SineScroll/_BackgroundYSine.bin"),
                    Program.ToZ80Folder("09_SineScroll/_ForegroundYSine.bin"),
                    Program.ToZ80Folder("09_SineScroll/_PrecaSine.bin")
                    );
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        Bitmap bmp1;
        Bitmap bmp2;
        int bmpY;
        private void GetNextColors(out string cpcColorLabel1, out string cpcColorLabel2)
        {
            var y = bmpY;

            var pixel1 = bmp1.GetPixel(0, y);
            var palette1 = new Palette();
            palette1.AddColor(pixel1);

            var pixel2 = bmp2.GetPixel(0, y);
            var palette2 = new Palette();
            palette2.AddColor(pixel2);

            var gateArrayColor1 = palette1.ToFirmwareByteArray()[0];
            var gateArrayColor2 = palette2.ToFirmwareByteArray()[0];

            cpcColorLabel1 = "COLOR" + gateArrayColor1;
            cpcColorLabel2 = "COLOR" + gateArrayColor2;

            bmpY++;
        }

        private void CreateBackgroundRasters(string imageFilename1, string imageFilename2, string destZ80)
        {
            bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));
            bmpY = 0; //96;

            string c0 = "";
            string c1 = "";
            var z80 = "";

            for (var iRepeat1 = 0; iRepeat1 < 2 * 2 * 4 * 2; iRepeat1++)
            {
                for (var iRepeat2 = 0; iRepeat2 < 2; iRepeat2++)
                {
                    c0 = "COLOR1";
                    c1 = "COLOR2";

                    z80 += "\tdb BACKGROUND_PEN0, " + c0 + ", BACKGROUND_PEN1, " + c1 + "\n";
                    z80 += "\tdb BACKGROUND_PEN1, " + c1 + ", BACKGROUND_PEN0, " + c0 + "\n";
                }
                /*for (var iRepeat3 = 0; iRepeat3 < 1; iRepeat3++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    c0 = "COLOR11";
                    c1 = "COLOR13";
                    z80 += "\tdb BACKGROUND_PEN0, " + c1 + ", BACKGROUND_PEN1, " + c0 + "\n";
                    z80 += "\tdb BACKGROUND_PEN1, " + c0 + ", BACKGROUND_PEN0, " + c1 + "\n";
                }*/
            }

            File.WriteAllText(destZ80, z80);
        }

        private void CreateForegroundRasters(string imageFilename1, string imageFilename2, string destZ80)
        {
            bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));
            bmpY = 35;

            string c0 = "";
            string c1 = "";
            var z80 = "";

            for (var iRepeat1 = 0; iRepeat1 < 4; iRepeat1++)
            {
                for (var iRepeat2 = 0; iRepeat2 < 8; iRepeat2++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN0, " + c0 + ", FOREGROUND_PEN1, " + c1 + "\n";
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN1, " + c1 + ", FOREGROUND_PEN0, " + c0 + "\n";
                }
                for (var iRepeat3 = 0; iRepeat3 < 8; iRepeat3++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN0, " + c1 + ", FOREGROUND_PEN1, " + c0 + "\n";
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN1, " + c0 + ", FOREGROUND_PEN0, " + c1 + "\n";
                }
            }

            File.WriteAllText(destZ80, z80);
        }

        private byte[] DoubleValues(byte[] src)
        {
            for (int i = 0; i < src.Length; i++)
            {
                src[i] = (byte)(src[i] * 4);
            }

            return src;
        }

        public void CreateSinMoves(string backgroundYSineFilename,
                                    string foregroundYSineFilename,
                                    string precalculatedSineFilename)
        {
            File.WriteAllBytes(backgroundYSineFilename, DoubleValues(SinCurve.CreateSinAbs(256, 0, 30, 1)));
            File.WriteAllBytes(foregroundYSineFilename, DoubleValues(SinCurve.CreateCos(256, 0, 50, 1)));

            // scroll anim
            var bgNbFrames = 140;

            var frameBytes = new List<byte>();
            for (int iFrame = 0; iFrame < bgNbFrames; iFrame++)
            {
                double animProgress = (double)iFrame/(double)bgNbFrames;

                var iFrameSine1 = Math.Sin(2 * Math.PI * animProgress);
                var iFrameSine2 = Math.Cos(2 * Math.PI * animProgress);

                var cubeWidth = 20;
                var cubeX = Math.Abs(-Math.Cos(2 * Math.PI * animProgress) * (1.0*(96- (double)cubeWidth)));

                var cubeWidth2 = 20;
                var cubeX2 = 30;

                for (int iX = 0; iX < 96; iX++)
                {
                    double floatValue =
                            (29.0f * Math.Sin(iFrameSine1))
                        + (23.0f * Math.Sin(-iFrameSine2));

                    iFrameSine1 += 0.037f;
                    iFrameSine2 -= 0.027f;

                    int b = (int)floatValue;
                    while (b >= CharHeight)
                    {
                        b -= CharHeight;
                    }
                    while (b<0)
                    {
                        b += CharHeight;
                    }

                    // flipBackground and shadeBackground are mutually EXCLUSIVE
                    var flipBackground = false;
                    var shadeBackground = false;

                    var flipForeground = false;

                    if ((Math.Abs((iX - iFrame) % 40) < 20))
                    {
                        flipBackground = true;
                    }

                    if ((Math.Abs((iX - iFrame) % 40) < 10))
                    {
                        shadeBackground = true;
                    }
                    shadeBackground = true;

                    /*if((iX>=cubeX)&&(iX<(cubeX + cubeWidth)))
                    {
                        shadeBackground = false;
                        flipBackground = false;
                    }*/
                   /* if ((iX >= cubeX2) && (iX < (cubeX2 + cubeWidth2)))
                    {
                        shadeBackground = true;
                        flipBackground = false;
                    }*/


                    if ((Math.Abs((iX - iFrame) % 40) < 20))
                    {
                        flipForeground = true;
                    }

                    if (shadeBackground)
                    {
                        b += (CharHeight + CharHeight) * 1;
                    }
                    if (flipForeground)
                    {
                        b += (CharHeight + CharHeight) *2;
                    }
                    frameBytes.Add((byte)b);
                    //frameBytes.Add((byte)0);
                }
            }

            File.WriteAllBytes(precalculatedSineFilename, frameBytes.ToArray());
        }

        private class CharBitmap
        {
            public byte CPCCharWidthInBytes;
            public char Letter;
            public byte[] Data;
        }

        private void CreateFont(string srcImage, string dstZ80, string dstSliceFilename, string charset, string scrollText)
        {
            var bitmap = new Bitmap(srcImage);

            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 1000,
                R4 = 38,
                R9 = 7,
                StartVLC = 0,
                FirstPixelInPaletteX = 0,
                FirstPixelInPaletteY = 0
            }
            );

            var charBitmaps = new List<CharBitmap>();
            int iCharsetIndex = 0;

            //var cpcCharWidth = image.CPCRawBuffer_Width / 8;
            int baseY = 0;
            var isFirst = true;
            int cpcCharWidth = 0;
            int baseX = 0;
            for (int iChar=0; iChar<charset.Length; iChar++)
            {
                baseX += cpcCharWidth;

                char letter = charset[iCharsetIndex];

                if (isFirst)
                {
                    isFirst = false;
                    cpcCharWidth = 16 / 4; // 1st character (space) is 16 pixels width
                }
                else
                {
                    // search for complete empty vertical line to determine width of a character
                    byte p0, p1, p2, p3;
                    var emptyPixel = image.CPCRawBuffer[0];                    
                    Pixel.FromMode1Pixel(emptyPixel, out p0, out p1, out p2, out p3);
                    byte emptyBitPixelP3 = p3;

                    cpcCharWidth = 0;
                    var vertLineFound = false;
                    while(!vertLineFound)
                    {
                        int wrongPixelsCount = 0;
                        for (int y=0; y<image.CPCRawBuffer_Height; y++)
                        {
                            var currentPixel = image.CPCRawBuffer[(y*image.CPCRawBuffer_Width)+baseX+cpcCharWidth];
                            Pixel.FromMode1Pixel(currentPixel, out p0, out p1, out p2, out p3);

                            if(p3 != emptyBitPixelP3)
                            {
                                wrongPixelsCount++;
                            }
                        }

                        if(wrongPixelsCount>0)
                        {
                            cpcCharWidth++;
                        }
                        else
                        {
                            vertLineFound = true;
                        }
                    }
                    cpcCharWidth++;
                }

                var data = new List<byte>();

                // WARNING!!! Vertical slice of cpc-pixel-data!
                for (int x = 0; x < cpcCharWidth; x++)
                {
                    #region CASE 1: BACKGROUND NORMAL - FOREGROUND NORMAL
                    for (var y=0; y<CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];
                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];
                        data.Add(b);
                    }
                    #endregion

                    #region CASE 3: BACKGROUND SHADE - FOREGROUND NORMAL
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN0 to PEN2 SHADE
                        if((y&7)==7)
                        {
                            if (p1 == 0) { p1 = 2; };
                        }
                        else if ((y&3)<2)
                        {
                            if (p0 == 0) { p0 = 2; };
                            if (p2 == 0) { p2 = 2; };
                        }
                        else
                        {
                            if (p1 == 0) { p1 = 2; };
                            if (p3 == 0) { p3 = 2; };
                        }

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN0 to PEN2 SHADE
                        if ((y & 7) == 7)
                        {
                            if (p1 == 0) { p1 = 2; };
                        }
                        else if ((y & 3) < 2)
                        {
                            if (p0 == 0) { p0 = 2; };
                            if (p2 == 0) { p2 = 2; };
                        }
                        else
                        {
                            if (p1 == 0) { p1 = 2; };
                            if (p3 == 0) { p3 = 2; };
                        }

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    #endregion

                    #region CASE 4: BACKGROUND NORMAL - FOREGROUND REVERSE
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    #endregion

                    #region CASE 6: BACKGROUND SHADE - FOREGROUND REVERSE
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);


                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        // PEN0 to PEN2 SHADE
                        if ((y & 7) == 7)
                        {
                            if (p1 == 0) { p1 = 2; };
                        }
                        else if ((y & 3) < 2)
                        {
                            if (p0 == 0) { p0 = 2; };
                            if (p2 == 0) { p2 = 2; };
                        }
                        else
                        {
                            if (p1 == 0) { p1 = 2; };
                            if (p3 == 0) { p3 = 2; };
                        }

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);


                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        // PEN0 to PEN2 SHADE
                        if ((y & 7) == 7)
                        {
                            if (p1 == 0) { p1 = 2; };
                        }
                        else if ((y & 3) < 2)
                        {
                            if (p0 == 0) { p0 = 2; };
                            if (p2 == 0) { p2 = 2; };
                        }
                        else
                        {
                            if (p1 == 0) { p1 = 2; };
                            if (p3 == 0) { p3 = 2; };
                        }

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    #endregion
                }

                var charBitmap = new CharBitmap
                {
                    CPCCharWidthInBytes = (byte)cpcCharWidth,
                    Letter = letter,
                    Data = data.ToArray()
                };
                charBitmaps.Add(charBitmap);

                iCharsetIndex++;
            }

            // parse scroll text and generate bitmap data
            var scrollText_ScrollTextSlicesData = new List<int>();
            var vertSlices_Data = new List<byte[]>();
            var sliceByteSize = (CharHeight + CharHeight) * 4;

            foreach (var letter in scrollText)
            {
                int foundIndex = -1;
                for (int i = 0; i < charBitmaps.Count; i++)
                {
                    var charBitmap = charBitmaps[i];

                    if (charBitmap.Letter == letter)
                    {
                        foundIndex = i;
                        break;
                    }
                }
                if (foundIndex == -1)
                {
                    // letter not found in charset
                    System.Diagnostics.Debugger.Break();
                }

                var foundCharBitmap = charBitmaps[foundIndex];

                cpcCharWidth = foundCharBitmap.CPCCharWidthInBytes;

                int iOffset = 0;
                for (int iSlice = 0; iSlice < cpcCharWidth; iSlice++)
                {
                    var slice = new List<byte>();
                    for (int i = 0; i < sliceByteSize; i++)
                    {
                        slice.Add(foundCharBitmap.Data[iOffset]);
                        iOffset++;
                    }

                    var vertSliceFoundIndex = -1;
                    for (int iVertSliceIndex = 0; iVertSliceIndex < vertSlices_Data.Count(); iVertSliceIndex++)
                    {
                        var found = true;
                        var iVertSlice = vertSlices_Data[iVertSliceIndex];
                        for (int i = 0; i < sliceByteSize; i++)
                        {
                            if (iVertSlice[i] != slice[i])
                            {
                                found = false;
                            }
                        }
                        if (found)
                        {
                            vertSliceFoundIndex = iVertSliceIndex;
                            break;
                        }
                    }

                    if (vertSliceFoundIndex != -1)
                    {
                        scrollText_ScrollTextSlicesData.Add(vertSliceFoundIndex);
                    }
                    else
                    {
                        var newIndex = vertSlices_Data.Count();
                        vertSlices_Data.Add(slice.ToArray());

                        scrollText_ScrollTextSlicesData.Add(newIndex);
                    }
                }
            }

            var vertSlices_ByteData = new List<byte>();
            foreach (var sliceData in vertSlices_Data)
            {
                foreach (var b in sliceData)
                {
                    vertSlices_ByteData.Add(b);
                }
            }
            File.WriteAllBytes(dstSliceFilename, vertSlices_ByteData.ToArray());

            var keyword = Program.FilenameToKeyWordName(dstZ80);

            var z80 = ";\n";
            foreach (var sliceIndex in scrollText_ScrollTextSlicesData)
            {
                var offset = sliceIndex * sliceByteSize;

                z80 += "\tdw SineScroll_VertSlices - " + (0x7C * 256) + " + " + offset + " ;; " + sliceIndex + "\n";
            }
            
            File.WriteAllText(dstZ80, z80);
        }
    }
}
