﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _10_RotatingRasters
    {
        private int NbAngles;

        public _10_RotatingRasters()
        {
            try
            {
                CreateSinMoves(Program.ToZ80Folder("10_RotatingRasters/_Sine.bin"));

                var palette = new Palette();
                NbAngles = 40;
                CreateHorizRotateFrames(ref palette,
                                        Program.ToDataFolder("10_RotatingRasters/Rasters_32_20Empty.png"),
                                        Program.ToZ80Folder("10_RotatingRasters/_Frames/"),
                                        Program.ToZ80Folder("10_RotatingRasters/_LineStackPointer.z80"), // pointers to stack for each lines
                                        Program.ToZ80Folder("10_RotatingRasters/_LineIndices.z80"), // stack 96 entries
                                        Program.ToZ80Folder("10_RotatingRasters/_SliceOpcodes"), // the opcodes for each unique vertical slices
                                        Program.ToZ80Folder("10_RotatingRasters/_Palette.z80"));

                NbAngles = 52;
                CreateVertRotateFrames(ref palette,
                                        Program.ToDataFolder("10_RotatingRasters/VertSpriteBitmap.png"),
                                        Program.ToZ80Folder("10_RotatingRasters/_Frames/"),
                                        Program.ToZ80Folder("10_RotatingRasters/_Frames/FrameBitmap" + (NbAngles-1).ToString("00") + ".png"),
                                        Program.ToZ80Folder("10_RotatingRasters/_VertLinePointers.z80"),
                                        Program.ToZ80Folder("10_RotatingRasters/_VertLineIndices.z80"),
                                        Program.ToZ80Folder("10_RotatingRasters/_VertLineR12R13Slices.z80"),
                                        Program.ToZ80Folder("10_RotatingRasters/_VertLines_VLC")
                                        );
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        #region Helpers

        private byte[] DoubleValues(byte[] src)
        {
            for (int i = 0; i < src.Length; i++)
            {
                src[i] = (byte)(src[i] * 2);
            }

            return src;
        }

        public void CreateSinMoves(string sineFilename)
        {
            File.WriteAllBytes(sineFilename, DoubleValues(SinCurve.CreateCos(256, 0, NbAngles, 2)));
        }

        private void ClearFrame(Graphics graphics, int width, int height, Color clearColor)
        {
            var brush = new SolidBrush(clearColor);
            var rect = new Rectangle(0, 0, width, height);
            graphics.FillRectangle(brush, rect);
        }

        private static double DegreeToRadian(double angle)
        {
            return Math.PI * angle / 180.0;
        }

        #endregion

        #region Vertical

        private Bitmap CreateSprBitmap(int sprWidth, string frameBitmap0Filename)
        {
            var frameBitmap0 = Bitmap.FromFile(frameBitmap0Filename);
            var bmp = new Bitmap(frameBitmap0);

            var newBitmap = new Bitmap(sprWidth, 1);
            var graphics = Graphics.FromImage(newBitmap);

            graphics.SmoothingMode = SmoothingMode.None;
            graphics.PixelOffsetMode = PixelOffsetMode.None;

            for (int x = 0; x < sprWidth; x++)
            {
                var srcPixel = bmp.GetPixel(x, 0);
                var brush = new SolidBrush(srcPixel);
                graphics.FillRectangle(brush, x, 0, 1, 1);
            }

            graphics.Dispose();
            frameBitmap0.Dispose();

            return newBitmap;
        }

        private static Bitmap MakeVertCycleBitmap(System.Drawing.Image lineImage, int destWidth, int destHeight, out List<int> destYOffsets)
        {
            destYOffsets = new List<int>();

            destYOffsets.Clear();

            var lineBitmap = new Bitmap(lineImage);

            var newBitmap = new Bitmap(destWidth, destHeight);
            var graphics = Graphics.FromImage(newBitmap);

            var clearColor = Color.FromArgb(0, 0, 0);
            var clearBrush = new SolidBrush(clearColor);
            var rect = new Rectangle(0, 0, destWidth, destHeight);
            graphics.FillRectangle(clearBrush, rect);

            var lineColor = Color.FromArgb(255, 255, 255, 255);
            var lineBrush = new SolidBrush(lineColor);

            var srcX = (lineBitmap.Width - destWidth) / 4;
            var srcY = (lineBitmap.Height - destHeight) / 2;
            var dstY = 0;
            for (int y = srcY; y < srcY + destHeight; y++)
            {
                var found = false;
                int x = srcX;
                while (!found)
                {
                    var color = lineBitmap.GetPixel(x, y);
                    if (color == clearColor)
                    {
                        x++;
                    }
                    else
                    {
                        found = true;
                    }
                }

                //var dstX = x/2;
                var dstX = (int)(((float)(x)*1.0f));
                while (dstX >= destWidth)
                {
                    dstX -= destWidth;
                }

                destYOffsets.Add(dstX);

                graphics.FillRectangle(lineBrush, dstX, dstY, 1, 1);
                dstY++;
            }

            graphics.Dispose();

            return newBitmap;
        }

        private static Bitmap MakeVertFrameBitmap(double zoomRatio, List<int> destXOffsets, System.Drawing.Image rasterImage, int destWidth, int destHeight)
        {
            var rasterBitmap = new Bitmap(rasterImage);

            var newBitmap = new Bitmap(destWidth, destHeight);
            var graphics = Graphics.FromImage(newBitmap);

            graphics.SmoothingMode = SmoothingMode.None;
            graphics.PixelOffsetMode = PixelOffsetMode.None;

            int dstY = 0;
            foreach (var xOffset in destXOffsets)
            {
                var srcY = 0;

                for (int iX = 0; iX < rasterImage.Width; iX++)
                {
                    var srcX = (int)((((float)(xOffset + (iX))))* zoomRatio);
                    srcX += 14;
                    while (srcX >= rasterImage.Width)
                    {
                        srcX -= rasterImage.Width;
                    }

                    var pixel = rasterBitmap.GetPixel(srcX, srcY);
                    var brush = new SolidBrush(pixel);
                    graphics.FillRectangle(brush, iX, dstY, 1, 1);
                }

                dstY++;
            }

            graphics.Dispose();

            return newBitmap;
        }

        public class VertSlice
        {
            public byte[] CPCData = null;

            public VertSlice()
            {
            }
        }

        // returns -1 if not found
        private int GetIndexInVertSlices(ref List<VertSlice> slices, ref byte[] cpcData)
        {
            for (int iSliceIndex = 0; iSliceIndex < slices.Count; iSliceIndex++)
            {
                var slice = slices[iSliceIndex];

                var equals = true;
                for (var i = 0; i < cpcData.Length; i++)
                {
                    if (cpcData[i] != slice.CPCData[i])
                    {
                        equals = false;
                    }
                }

                if (equals)
                {
                    return iSliceIndex;
                }
            }

            return -1;
        }

        private List<int> CreateVertSlices(ref Palette palette, ref List<VertSlice> slices, System.Drawing.Image frameImage)
        {
            var indices = new List<int>();

            var frameBitmap = new Bitmap(frameImage);

            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = frameBitmap,
                Mode = 0,
                R1 = frameImage.Width / 2,
                R4 = frameImage.Height / 8,
                R9 = 7,
                StartVLC = 0,
                Palette = palette
            }
            );

            for (int iCpcY = 0; iCpcY < image.CPCRawBuffer_Height; iCpcY++)
            {
                var cpcData = new byte[image.CPCRawBuffer_Width];
                for (int iCpcX = 0; iCpcX < image.CPCRawBuffer_Width; iCpcX++)
                {
                    cpcData[iCpcX] = image.CPCRawBuffer[(iCpcY * image.CPCRawBuffer_Width) + iCpcX];
                }

                var sliceIndex = GetIndexInVertSlices(ref slices, ref cpcData);

                if (sliceIndex == -1)
                {
                    sliceIndex = slices.Count;

                    var newSlice = new VertSlice();
                    newSlice.CPCData = cpcData;
                    slices.Add(newSlice);
                }

                indices.Add(sliceIndex);
            }

            return indices;
        }

        private string CreateVertLinePointers(List<List<int>> lineIndices)
        {
            var z80 = "";

            z80 += ";\n";
            z80 += "VERT_ANGLE_COUNT equ " + lineIndices.Count + "\n";
            z80 += "VERT_ANGLE_POINTERS:\n";

            for (var iLine = 0; iLine < lineIndices.Count; iLine++)
            {
                z80 += "\tdw VERT_ANGLE_" + iLine + "\n";
                //z80 += "\tdw VERT_ANGLE_" + 0 + "\n";
                //z80 += "\tdw VERT_ANGLE_" + (lineIndices.Count-1) + "\n";
            }

            return z80;
        }

        private string CreateVertLineIndices(List<List<int>> lineIndices)
        {
            var z80 = "";

            for (var iLine = 0; iLine < lineIndices.Count; iLine++)
            {
                var indices = lineIndices[iLine];

                z80 += "; index in VERT_SLICE_R12R13 , already multiplied by 2 in data\n";
                z80 += "VERT_ANGLE_" + iLine + ":\n";

                foreach (var index in indices)
                {
                    z80 += "\tdb " + (index*2) + "\n";
                }

                z80 += "\n";
            }

            return z80;
        }

        private string CreateVertR12R13Slices(ref List<VertSlice> slices, int cpcScrWidthInBytes, string vlcPrefixFilename)
        {
            var z80 = "";

            z80 += "; R12,R13\n";
            z80 += "VERT_SLICE_R12R13:\n";

            var R12 = 0x20; // &8000
            var vlc = new byte[0x0800];
            var offsetInVLC = 0;

            for (int iSlice = 0; iSlice < slices.Count; iSlice++)
            {
                var slice = slices[iSlice];
                
                var cpcLine = new byte[cpcScrWidthInBytes];
                var srcX = 0;
                for (int i = 0; i < cpcScrWidthInBytes; i++)
                {
                    cpcLine[i] = slice.CPCData[srcX];
                    srcX++;
                    if (srcX == slice.CPCData.Length)
                    {
                        srcX = 0;
                    }
                }

                ushort cpcOffset = (ushort)(offsetInVLC);
                byte finalR12 = (byte)(R12 + ((cpcOffset / 2) >> 8));
                byte finalR13 = (byte)((cpcOffset / 2) & 255);

                z80 += "\tdb &" + finalR12.ToString("X2") + ", &" + finalR13.ToString("X2") + "\n";

                for (int i = 0; i < cpcScrWidthInBytes; i++)
                {
                    vlc[offsetInVLC] = cpcLine[i];
                    offsetInVLC++;
                }

                if (iSlice == (21 - 1))
                {
                    File.WriteAllBytes(vlcPrefixFilename + "0.bin", vlc);

                    R12 = 0x30; // &C000
                    vlc = new byte[0x0800];
                    offsetInVLC = 0;
                }
            }

            if(R12!=0x30)
            {
                File.WriteAllBytes(vlcPrefixFilename + "0.bin", vlc);
            }
            else
            {
                File.WriteAllBytes(vlcPrefixFilename + "1.bin", vlc);
            }
            
            return z80;
        }

        private void CreateVertRotateFrames(ref Palette palette, string vertSpriteBitmapFilename, string outFolder, string frameBitmap0Filename, string vertLinePointersFilename, string vertLineIndicesFilename, string vertR12R13SlicesFilename, string vlcPrefixFilename)
        {
            var sprWidth = 32;
            var cpcScrWidth = 96*2; // mode 0 width
            var cpcScrHeight = 268; // visible screen height

            //var sprBitmap = CreateSprBitmap(sprWidth, frameBitmap0Filename);
            //sprBitmap.Save(outFolder + "VertSpriteBitmap.png", ImageFormat.Png);

            var sprBitmap = Bitmap.FromFile(vertSpriteBitmapFilename);

            var nbFrames = NbAngles;

            var slices = new List<VertSlice>();
            var lineIndices = new List<List<int>>();

            for (int iFrame = 0; iFrame < nbFrames; iFrame++)
            {
                double progress = ((double)iFrame) / ((double)nbFrames);
                double angle = 90 + -45 + progress * 90;

                double zoomRatio = 1.0;
                var maxZoomRatio = 0.0;
                if (progress < 0.5f)
                {
                    double p = progress * 2; // range 0-1
                    zoomRatio = ((double)1) + (maxZoomRatio * p);
                }
                else
                {
                    double p = 1 - ((progress - 0.5) * 2); // range 0-1
                    zoomRatio = ((double)1) + (maxZoomRatio * p);
                }

                var lineBitmap = MakeLineBitmap(angle,6);
                //lineBitmap.Save(outFolder + "VertLineBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var rotatedLineBitmap = RotateImageByAngle(lineBitmap, angle);
                //rotatedLineBitmap.Save(outFolder + "VertRotatedLineBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                List<int> destXOffsets = null;

                var cycleLineBitmap = MakeVertCycleBitmap(rotatedLineBitmap, sprWidth, cpcScrHeight, out destXOffsets);
                //cycleLineBitmap.Save(outFolder + "VertCycleLineBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var frameBitmap = MakeVertFrameBitmap(zoomRatio, destXOffsets, sprBitmap, sprWidth, cpcScrHeight);
                frameBitmap.Save(outFolder + "VertFrameBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var indices = CreateVertSlices(ref palette, ref slices, frameBitmap);
                lineIndices.Add(indices);

                lineBitmap.Dispose();
                rotatedLineBitmap.Dispose();
                cycleLineBitmap.Dispose();
            }

            sprBitmap.Dispose();

            var z80LinePointers = CreateVertLinePointers(lineIndices);
            File.WriteAllText(vertLinePointersFilename, z80LinePointers);

            var z80LineIndices = CreateVertLineIndices(lineIndices);
            File.WriteAllText(vertLineIndicesFilename, z80LineIndices);

            var cpcScrWidthInBytes = cpcScrWidth / 2;
            var z80R12R13Slices = CreateVertR12R13Slices(ref slices, cpcScrWidthInBytes, vlcPrefixFilename);
            File.WriteAllText(vertR12R13SlicesFilename, z80R12R13Slices);
        }

        #endregion

        #region Horizontal

        private static Bitmap MakeLineBitmap(double degAngle, int ratio)
        {
            var width = 400*ratio;
            var height = 400*ratio;

            var newBitmap = new Bitmap(width, height);
            var graphics = Graphics.FromImage(newBitmap);
            graphics.SmoothingMode = SmoothingMode.HighSpeed;
            graphics.PixelOffsetMode = PixelOffsetMode.HighSpeed;

            var clearColor = Color.FromArgb(0, 0, 0);
            var clearBrush = new SolidBrush(clearColor);
            var rect = new Rectangle(0, 0, width, height);
            graphics.FillRectangle(clearBrush, rect);

            var lineColor = Color.FromArgb(255, 255, 255);
            var lineBrush = new SolidBrush(lineColor);
            
            float y = (float)(height / 2);

            graphics.DrawLine(new Pen(lineBrush), 0, y, width, y);

            graphics.Dispose();

            return newBitmap;
        }
        
        private static Bitmap RotateImageByAngle(System.Drawing.Image oldBitmap, double angle)
        {
            var newBitmap = new Bitmap(oldBitmap.Width, oldBitmap.Height);
            var graphics = Graphics.FromImage(newBitmap);
            graphics.SmoothingMode = SmoothingMode.HighSpeed;
            graphics.PixelOffsetMode = PixelOffsetMode.HighSpeed;
            graphics.TranslateTransform((float)oldBitmap.Width / 2, (float)oldBitmap.Height / 2);
            graphics.RotateTransform((float)angle);
            graphics.TranslateTransform(-(float)oldBitmap.Width / 2, -(float)oldBitmap.Height / 2);
            graphics.DrawImage(oldBitmap, new Point(0, 0));
            graphics.Dispose();
            return newBitmap;
        }

        private static Bitmap MakeCycleBitmap(System.Drawing.Image lineImage, int destWidth, int destHeight, out List<int> destYOffsets)
        {
            destYOffsets = new List<int>();

            destYOffsets.Clear();

            var lineBitmap = new Bitmap(lineImage);

            var newBitmap = new Bitmap(destWidth, destHeight);
            var graphics = Graphics.FromImage(newBitmap);

            var clearColor = Color.FromArgb(0, 0, 0);
            var clearBrush = new SolidBrush(clearColor);
            var rect = new Rectangle(0, 0, destWidth, destHeight);
            graphics.FillRectangle(clearBrush, rect);

            var lineColor = Color.FromArgb(255, 255, 255, 255);
            var lineBrush = new SolidBrush(lineColor);

            var srcX = (lineBitmap.Width - destWidth) / 2;
            var srcY = (lineBitmap.Height - destHeight) / 4;
            var dstX = 0;
            for (int x = srcX; x < srcX + destWidth; x++)
            {
                var found = false;
                int y = srcY;
                while (!found)
                {
                    var color = lineBitmap.GetPixel(x, y);
                    if (color == clearColor)
                    {
                        y++;
                    }
                    else
                    {
                        found = true;
                    }
                }

                var dstY = y;
                while (dstY >= destHeight)
                {
                    dstY -= destHeight;
                }

                destYOffsets.Add(dstY);

                graphics.FillRectangle(lineBrush, dstX, dstY, 1, 1);
                dstX++;
            }

            graphics.Dispose();

            return newBitmap;
        }

        private static Bitmap MakeFrameBitmap(List<int> destYOffsets, System.Drawing.Image rasterImage, int destWidth, int destHeight)
        {
            var rasterBitmap = new Bitmap(rasterImage);

            var newBitmap = new Bitmap(destWidth, destHeight);
            var graphics = Graphics.FromImage(newBitmap);

            graphics.SmoothingMode = SmoothingMode.None;
            graphics.PixelOffsetMode = PixelOffsetMode.None;

            int dstX = 0;
            foreach (var yOffset in destYOffsets)
            {
                var srcX = 0;

                for (int iY = 0; iY < rasterImage.Height; iY++)
                {
                    var srcY = (yOffset + iY);
                    while (srcY >= rasterImage.Height)
                    {
                        srcY -= rasterImage.Height;
                    }

                    var pixel = rasterBitmap.GetPixel(srcX, srcY);
                    var brush = new SolidBrush(pixel);
                    graphics.FillRectangle(brush, dstX, iY, 1, 1);
                }

                dstX++;
            }

            graphics.Dispose();

            return newBitmap;
        }

        public class Slice
        {
            public byte[] CPCData = null;

            public Slice()
            {
            }
        }

        // returns -1 if not found
        private int GetIndexInSlices(ref List<Slice> slices, ref byte[] cpcData)
        {
            for (int iSliceIndex = 0; iSliceIndex < slices.Count; iSliceIndex++)
            {
                var slice = slices[iSliceIndex];

                var equals = true;
                for (var i = 0; i < cpcData.Length; i++)
                {
                    if (cpcData[i] != slice.CPCData[i])
                    {
                        equals = false;
                    }
                }

                if (equals)
                {
                    return iSliceIndex;
                }
            }

            return -1;
        }

        private List<int> CreateSlices(ref Palette palette, ref List<Slice> slices, System.Drawing.Image frameImage)
        {
            var indices = new List<int>();

            var frameBitmap = new Bitmap(frameImage);

            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = frameBitmap,
                Mode = 0,
                R1 = frameImage.Width / 2,
                R4 = frameImage.Height / 8,
                R9 = 7,
                StartVLC = 0,
                Palette = palette
            }
            );

            for (int iCpcX = 0; iCpcX < image.CPCRawBuffer_Width; iCpcX++)
            {
                var cpcData = new byte[image.CPCRawBuffer_Height];
                for (int iCpcY = 0; iCpcY < image.CPCRawBuffer_Height; iCpcY++)
                {
                    cpcData[iCpcY] = image.CPCRawBuffer[(iCpcY * image.CPCRawBuffer_Width) + iCpcX];
                }

                var sliceIndex = GetIndexInSlices(ref slices, ref cpcData);

                if (sliceIndex == -1)
                {
                    sliceIndex = slices.Count;

                    var newSlice = new Slice();
                    newSlice.CPCData = cpcData;
                    slices.Add(newSlice);
                }

                indices.Add(sliceIndex);
            }

            return indices;
        }

        private string CreateLineStackPointer(List<List<int>> lineIndices)
        {
            var z80 = "";

            z80 += ";; " + lineIndices.Count + " horizontal angles \n";
            z80 += "HORIZ_ANGLE_COUNT equ " + lineIndices.Count + "\n";
            z80 += "HORIZONTAL_ANGLES_STACK_POINTERS:\n";

            for (int iLineIndices = 0; iLineIndices < lineIndices.Count; iLineIndices++)
            {
                z80 += "\tdw HORIZONTAL_ANGLE_" + iLineIndices + "\n";
                //z80 += "\tdw HORIZONTAL_ANGLE_" + (lineIndices.Count-1) + "\n";
                //z80 += "\tdw HORIZONTAL_ANGLE_" + 0 + "\n";
            }

            return z80;
        }

        private string CreateLineIndices(List<List<int>> lineIndices)
        {
            var z80 = "";

            for(var iLine=0; iLine<lineIndices.Count; iLine++)
            {
                var indices = lineIndices[iLine];

                z80 += ";\n";
                z80 += "HORIZONTAL_ANGLE_" + iLine + ":\n";

                foreach (var index in indices)
                {
                    z80 += "\tdw SLICE_OPCODES_" + index + "\n";
                }

                z80 += "\tdw DrawVertSlices_QuitJumpTable\n";
                z80 += "\n";
            }

            return z80;
        }
        
        private string CreateSliceOpcodes(ref List<int> clearCountList, List<Slice> slices, int start, int end)
        {
            var z80 = "";

            var globalClearCount = -1;

            var clearCpcValue = 0;

            for (var iSlice = start; iSlice < end; iSlice++)
            {
                var slice = slices[iSlice];

                var clearCount = 0;

                z80 += ";\n";
                z80 += "SLICE_OPCODES_" + iSlice + ":\n";

                for (int iBaseY = 0; iBaseY < slice.CPCData.Length; iBaseY+=8)
                {
                    var isFirstBlock = (iBaseY == 0);
                    var isLastBlock = (iBaseY == (slice.CPCData.Length - 8));

                    if (isFirstBlock)
                    {
                        var offset = -7 * 8 * 256 + 96;
                        ushort cpcOffset = (ushort)(offset);

                        z80 += "\tld a, h:ld de, " + cpcOffset  + "\n";
                    }

                    for (int iY=0; iY<8; iY++)
                    {
                        var cpcByte = slice.CPCData[iBaseY + iY];

                        if (iY == 0)
                        {
                            if (cpcByte == clearCpcValue)
                            {
                                z80 += "\tld (hl), c ;; clear\n";
                                clearCount++;
                            }
                            else
                            {
                                z80 += "\tld (hl), " + cpcByte + "\n";
                            }
                        }
                        else
                        {
                            if (cpcByte == clearCpcValue)
                            {
                                z80 += "\tadd a, b:ld h, a:ld (hl), c ;; clear\n";
                                clearCount++;
                            }
                            else
                            {
                                z80 += "\tadd a, b:ld h, a:ld (hl), " + cpcByte + "\n";
                            }
                        }

                        if (iY == 7)
                        {
                            if (isLastBlock)
                            {
                                var offset = -7 * 0x800 - 3 * 96 + 1;
                                ushort cpcOffset = (ushort)(offset);
                                z80 += "\tld de, " + offset + "\n";
                                z80 += "\tadd hl, de\n";
                            }
                            else
                            {
                                z80 += "\tadd hl, de:ld a, h\n";
                            }
                        }
                    }
                }
                
                if (!clearCountList.Contains(clearCount))
                {
                    clearCountList.Add(clearCount);
                }

                //Console.WriteLine("" + clearCount);
                if (clearCount >= 13)
                {
                    var nops = (clearCount - 13);
                    if (nops > 0)
                    {
                        z80 += "\t_NOPS " + nops + "\n";
                    }
                }

                z80 += "\tret\n";
                z80 += "\n";
            }

            return z80;
        }

        private void CreateHorizRotateFrames(ref Palette palette, string rasterFilename, string outFolder, string lineStackPointerFilename, string lineIndicesFilename, string sliceOpcodesFilename, string paletteFilename)
        {
            var width = 96 * 2; // MODE 0
            var height = 32;
            var nbFrames = NbAngles;

            var originalRasterBitmap = new Bitmap(rasterFilename);
            var slices = new List<Slice>();
            var lineIndices = new List<List<int>>();
            
            for (int iFrame = 0; iFrame < nbFrames; iFrame++)
            {
                double angle = ((double)iFrame) / ((double)nbFrames);
                angle = -45 + angle * 90;

                var lineBitmap = MakeLineBitmap(angle, 1);
                //lineBitmap.Save(outFolder + "LineBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var rotatedLineBitmap = RotateImageByAngle(lineBitmap, angle);
                //rotatedLineBitmap.Save(outFolder + "RotatedLineBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                List<int> destYOffsets = null;

                var cycleLineBitmap = MakeCycleBitmap(rotatedLineBitmap, width, height, out destYOffsets);
                //cycleLineBitmap.Save(outFolder + "CycleLineBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var frameBitmap = MakeFrameBitmap(destYOffsets, originalRasterBitmap, width, height);
                frameBitmap.Save(outFolder + "FrameBitmap" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var indices = CreateSlices(ref palette, ref slices, frameBitmap);
                lineIndices.Add(indices);

                lineBitmap.Dispose();
                rotatedLineBitmap.Dispose();
                cycleLineBitmap.Dispose();
            }

            originalRasterBitmap.Dispose();

            var z80LineStackPointer = CreateLineStackPointer(lineIndices);
            File.WriteAllText(lineStackPointerFilename, z80LineStackPointer);

            var z80LineIndices = CreateLineIndices(lineIndices);
            File.WriteAllText(lineIndicesFilename, z80LineIndices);

            var clearCountList = new List<int>();

            int iSliceFile = 0;
            int start = 0;
            int interval = 4000;
            while(start < slices.Count)
            {
                var end = start + interval;
                if (end >= slices.Count)
                {
                    end = slices.Count;
                }

                var z80SliceOpcodes = CreateSliceOpcodes(ref clearCountList, slices, start, end);

                File.WriteAllText(sliceOpcodesFilename + "_" + iSliceFile +".z80", z80SliceOpcodes);
                iSliceFile++;

                start += interval;
            }

            File.WriteAllText(
                paletteFilename,
                palette.ToZ80TableString(Program.FilenameToKeyWordName(paletteFilename))
            );
        }

        #endregion
    }
}
