﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _11_DeltaRasters
    {
        enum ECPCRoutineType
        {
            Unknown,
            XPlane_ToUp,
            XPlane_ToDown
        };

        public _11_DeltaRasters()
        {
            try
            {
                CreateKeftales(
                        Program.ToZ80Folder("11_DeltaRasters/")
                    );

                var rasterFilename = "11_DeltaRasters/Rasters_12_Width2.png";

                /*CreateVertSlicesScreenPtr(
                        0x0000, 0x0200,
                        Program.ToZ80Folder("11_DeltaRasters/")
                    );*/

                CreateXPlaneSlices(
                        Program.ToZ80Folder("11_DeltaRasters/XPlanes/")
                    );

                CreateCircleAnim(
                        Program.ToZ80Folder("11_DeltaRasters/_Anim.z80")
                    );

                CreateAnd11Table256(
                        Program.ToZ80Folder("11_DeltaRasters/_And11Table256.z80")
                    );
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        private void CreateAnd11Table256(string filename)
        {
            var z80 = ";\n";

            byte value = 0;
            for (int i=0;i<256;i++)
            {
                if(value==0)
                {
                    z80 += "\tdb &" + value.ToString("X2");
                }
                else
                {
                    z80 += ", &" + value.ToString("X2");
                }

                value++;
                if(value==12)
                {
                    z80 += "\n";
                    value = 0;
                }
            }
            
            z80 += "\n";

            File.WriteAllText(filename, z80);
        }

        /*private void CreateVertSlicesScreenPtr(ushort baseAddress_Screen0, ushort baseAddress_Screen1, string outFolder)
        {
            var z80_0 = "";
            var z80_1 = "";
            
            for(int iChar=0; iChar<96; iChar++)
            {
                ushort vccOffset = 0;
                for (int iVCC = 0; iVCC < 3; iVCC++)
                {
                    ushort vlcOffset = vccOffset;

                    for (int iVLC = 0; iVLC < 8; iVLC++)
                    {
                        z80_0 += "\tdw &" + (baseAddress_Screen0 + vlcOffset).ToString("X4") + "\n";
                        z80_1 += "\tdw &" + (baseAddress_Screen1 + vlcOffset).ToString("X4") + "\n";
                        vlcOffset += 0x800;
                    }

                    vccOffset += 96;
                }

                baseAddress_Screen0++;
                baseAddress_Screen1++;
            }
            
            File.WriteAllText(outFolder + "/_VertSlices_ScreenPtr0.z80", z80_0);
            File.WriteAllText(outFolder + "/_VertSlices_ScreenPtr1.z80", z80_1);
        }*/

        private Bitmap CreateVertSliceBitmap(int firmwareColor0, int firmwareColor1, int yOffset)
        {
            int vertSliceWidth = 2;
            int vertSliceHeight = 12;

            var color0 = Color.FromArgb(Pixel.CPCAsicPalette[firmwareColor0]);
            var color1 = Color.FromArgb(Pixel.CPCAsicPalette[firmwareColor1]);

            var newBitmap = new Bitmap(vertSliceWidth * 2, vertSliceHeight);

            for(int y=0; y<vertSliceHeight; y++)
            {
                var c0 = color0;
                if (y >= (vertSliceHeight / 2))
                {
                    c0 = color1;
                }

                int iY = y + yOffset;
                if(iY<0)
                {
                    iY += vertSliceHeight;
                }
                else if(iY >= vertSliceHeight)
                {
                    iY -= vertSliceHeight;
                }

                var c1 = color0;
                if (iY >= (vertSliceHeight / 2))
                {
                    c1 = color1;
                }

                newBitmap.SetPixel(0, y, c0);
                newBitmap.SetPixel(1, y, c1);
                newBitmap.SetPixel(2, y, c0);
                newBitmap.SetPixel(3, y, c1);
            }

            return newBitmap;
        }

        private string CreateVertSliceZ80Bitmap(ref Palette palette, int firmwareColor0, int firmwareColor1, int yOffset)
        {
            var srcBitmap = CreateVertSliceBitmap(firmwareColor0, firmwareColor1, yOffset);
            
            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = srcBitmap,
                    Mode = 0,
                    R1 = 1,
                    R4 = srcBitmap.Height / 8,
                    R9 = 7,
                    StartVLC = 0,
                    Palette = palette
                }
            );

            srcBitmap.Dispose();

            var z80 = "\tdb ";
            
            for (int y = 0; y < image.CPCRawBuffer_Height; y++)
            {
                var cpcByte = image.CPCRawBuffer[y * image.CPCRawBuffer_Width];

                if(y!=0)
                {
                    z80 += ", ";
                }

                z80 += "&" + cpcByte.ToString("X2");
            }

            z80 += "\n";
            return z80;
        }

        private void CreateXPlaneSlices(string outFolder)
        {
            // create palette with colors 0->15
            // palette will not be used in the end, 
            // it's just to make sure pen indices are correctly written into final data
            var palette = new Palette();
            for (int i=0; i<16; i++)
            {
                var firmwareColor = i;
                var color = Color.FromArgb(Pixel.CPCAsicPalette[firmwareColor]);
                palette.AddColor(color);
            }

            // iterate through CPC colors
            for (int iBg = 0; iBg < 12; iBg++)
            {
                var firmwareColor0 = iBg;
                var firmwareColor1 = iBg+4;

                var z80ToEqual = CreateVertSliceZ80Bitmap(ref palette, firmwareColor0, firmwareColor1, 0);
                var z80ToDown = CreateVertSliceZ80Bitmap(ref palette, firmwareColor0, firmwareColor1, -1);
                var z80ToUp = CreateVertSliceZ80Bitmap(ref palette, firmwareColor0, firmwareColor1, 1);

                File.WriteAllText(outFolder + "_ToDown_BG"+iBg+".z80", z80ToDown);
                File.WriteAllText(outFolder + "_ToEqual_BG" + iBg + ".z80", z80ToEqual);
                File.WriteAllText(outFolder + "_ToUp_BG" + iBg + ".z80", z80ToUp);
            }
        }

        private static string ToString(double value)
        {
            return value.ToString("0.############################");
        }

        private static ushort ToFixedPoint(double value)
        {
            var isNeg = value < 0 ? true : false;

            value = Math.Abs(value);

            int intPart = (int)(value);
            double floatPart = 256.0 * (value - ((double)(int)(value)));

            ushort fixedPointValue = (ushort)((intPart << 8) + (byte)floatPart);

            if (isNeg)
            {
                fixedPointValue = (ushort )(-fixedPointValue);
            }

            return fixedPointValue;
        }

        private void CreateCircleAnim(string animZ80Filename)
        {
            const int intNbFrames = 48;
            const double nbFrames = (double)(intNbFrames);
            const int cpcFXHeight = 12; // in pixels

            var animZ80 = "";

            animZ80 += ";\n";
            animZ80 += "DeltaRasters_AnimItemCount\tequ " + (int)nbFrames + "\n";
            animZ80 += "DeltaRasters_FXHeight\tequ " + cpcFXHeight + "\n";
            animZ80 += "\n";
            
            animZ80 += ";\n";
            animZ80 += "RoutineType_XPlane_ToUp     equ 0\n";
            animZ80 += "RoutineType_XPlane_ToDown   equ 2\n";
            animZ80 += "\n";

            animZ80 += ";\n";
            animZ80 += "DeltaRasters_AnimFrames:\n";
            animZ80 += "\n";
            
            double amplitude = 90;// 12; // max=90
            var rotateSinCurve = SinCurve.CreateSinDouble(intNbFrames, 0, (int)amplitude);
            var ySinCurveMax = 255;
            var ySinCurve = SinCurve.CreateCos(intNbFrames, 0, ySinCurveMax, 1);

            for (int iFrame = 0; iFrame<intNbFrames; iFrame++)                
            {
                double degAngle = -(amplitude / 2) + (double)(rotateSinCurve[iFrame]);
                //double degAngle = 10;

                double angle = degAngle * Math.PI / 180.0;
                var routineType = ECPCRoutineType.Unknown;

                var x1 = Math.Cos(angle);
                var y1 = Math.Sin(angle);
                var x2 = -x1;
                var y2 = -y1;

                if (x2 < x1)
                {
                    double t;
                    t = x1; x1 = x2; x2 = t;
                    t = y1; y1 = y2; y2 = t;
                }
                                                
                double slope = 0;
                if (x2 != x1)
                {
                    slope = (y2 - y1) / (x2 - x1);

                    if (slope >= 0)
                    {
                        routineType = ECPCRoutineType.XPlane_ToUp; // routine type 0
                    }
                    else
                    {
                        routineType = ECPCRoutineType.XPlane_ToDown; // routine type 1
                    }
                }

                double leftX = -1.0;

                var leftScreenYValue = 96 + (95.9999999999 * leftX * slope);

                var yValue = (int)ySinCurve[iFrame];
                while (yValue >= cpcFXHeight)
                {
                    yValue -= cpcFXHeight;
                }
                yValue = 0; // cancel value

                if ((routineType == ECPCRoutineType.XPlane_ToUp)
                    || (routineType == ECPCRoutineType.XPlane_ToDown))
                {
                }
                else
                {
                    while (leftScreenYValue >= (double)cpcFXHeight)
                    {
                        leftScreenYValue -= (double)cpcFXHeight;
                    }
                }

                leftScreenYValue += (double)((int)yValue);
                
                animZ80 += ";; x1=" + ToString(x1) + " y1=" + ToString(y1) + " x2=" + ToString(x2) + " y2=" + ToString(y2) + "\n";
                animZ80 += ";; leftScreenYValue=" + ToString(leftScreenYValue) + "\n";
                animZ80 += ";; slope=" + ToString(slope) + "\n";
                animZ80 += "DeltaRasters_AnimFrame" + iFrame + ":\n";

                //animZ80 += "\tdb RoutineType_" + routineType.ToString() + "\n";

                ushort cpcSlope = ToFixedPoint(slope);
                animZ80 += "\tdw &" + cpcSlope.ToString("X4") + " ;; slope\n";

                ushort cpcLeftScreenYValue = ToFixedPoint(leftScreenYValue);
                animZ80 += "\tdw &" + cpcLeftScreenYValue.ToString("X4") + " ;; cpcLeftScreenYValue\n";

                animZ80 += "\n";
            }

            File.WriteAllText(animZ80Filename, animZ80);
        }

        private void WriteIndexedPalette(string z80Filename, int[] colorsPart0, int[] colorsPart1)
        {
            var z80 = "";

            for(int iShift=0; iShift<4; iShift++)
            {
                z80 += "\t;; shift " + iShift + "\n";

                z80 += "\tdb ";
                for(int i=0; i<colorsPart0.Length;i++)
                {
                    if(i!=0)
                    {
                        z80 += ", ";
                    }

                    var indexInPalette = i + iShift;
                    if(indexInPalette>=colorsPart0.Length)
                    {
                        indexInPalette -= colorsPart0.Length;
                    }
                    if (indexInPalette <0)
                    {
                        indexInPalette += colorsPart0.Length;
                    }

                    z80 += "COLOR" + colorsPart0[indexInPalette];
                }
                z80 += "\n";
                
                z80 += "\tdb ";
                for (int i = 0; i < colorsPart1.Length; i++)
                {
                    if (i != 0)
                    {
                        z80 += ", ";
                    }

                    var indexInPalette = i + iShift;
                    if (indexInPalette >= colorsPart1.Length)
                    {
                        indexInPalette -= colorsPart1.Length;
                    }
                    if (indexInPalette < 0)
                    {
                        indexInPalette += colorsPart1.Length;
                    }

                    z80 += "COLOR" + colorsPart1[indexInPalette];
                }
                z80 += "\n";
                
                z80 += "\n";
            }

            File.WriteAllText(z80Filename, z80);
        }

        private void CreateAnimFrames(int nbFrames, string outFolder)
        {
            var z80TableFrames = ";\n";
            z80TableFrames += "MagicFramesCount equ " + nbFrames + "\n";
            z80TableFrames += "\n";
            z80TableFrames += ";\n";
            for (int i = 0; i < nbFrames; i++)
            {
                z80TableFrames += "\t_INCZ80 \"11_DeltaRasters\\Keftales\\_Horiz_" + i + ".z80\"\n";
                z80TableFrames += "\t_INCZ80 \"11_DeltaRasters\\Keftales\\_Vert_" + i + ".z80\"\n";
            }
            z80TableFrames += "\n";
            z80TableFrames += ";\n";
            z80TableFrames += "MagicFrames:\n";

            for (int i = 0; i < nbFrames; i++)
            {
                z80TableFrames += "\tdw HorizFrame" + i + ", VertFrame" + i + "\n";
            }
            z80TableFrames += "\tdw 0";

            File.WriteAllText(outFolder + "_KeftalesFrames.z80", z80TableFrames);

            /*var anim = new List<byte>();
            for(int i=0; i<256;i+=2)
            {
                double ratio = 0;

                if(i<128)
                {
                    ratio = ((double)(i) / 128.0) * ((double)(nbFrames - 1));
                }
                else
                {
                    ratio = ((double)(256-i) / 128.0) * ((double)(nbFrames - 1));
                }

                byte v = (byte)(ratio);

                anim.Add(v);
            }
            anim.AddRange(anim.ToArray()); // double the array
            File.WriteAllBytes(outFolder + "_KeftalesBounceAnim.bin", anim.ToArray());*/

            /*var bounceAnimSinCurve = SinCurve.CreateCos(256, 0, nbFrames, 1);
            File.WriteAllBytes(outFolder + "_KeftalesBounceAnim.bin", bounceAnimSinCurve);*/

            var bounceAnimSinCurve = new List<byte>();
            bounceAnimSinCurve.AddRange(SinCurve.CreateSin(64, 0, (nbFrames-0) / 2));
            bounceAnimSinCurve.AddRange(SinCurve.CreateSin(64, 0, (nbFrames-0) / 1));
            bounceAnimSinCurve.AddRange(SinCurve.CreateSin(64, 0, (nbFrames-0) / 2));
            bounceAnimSinCurve.AddRange(SinCurve.CreateSin(64, 0, (nbFrames-0) / 1));
            File.WriteAllBytes(outFolder + "_KeftalesBounceAnim.bin", bounceAnimSinCurve.ToArray());
        }

        private void CreateKeftales(string outFolder)
        {
            double width = 96;
            double height = 46;

            var bitmap = new Bitmap((int)width, (int)height);

            var firmwareColors_Part2 = new int[] { 24, 15, 6, 3 };
            var firmwareColors_Part3 = new int[] { 26, 17, 8, 5 };
            WriteIndexedPalette(outFolder + "_IndexedPalette0.z80", firmwareColors_Part2, firmwareColors_Part3);

            var firmwareColors_Part0 = new int[] { 21, 18, 9, 0 };
            var firmwareColors_Part1 = new int[] { 23, 20, 11, 2 };
            WriteIndexedPalette(outFolder + "_IndexedPalette1.z80", firmwareColors_Part0, firmwareColors_Part1);

            var firmwareColors_Part4 = new int[] { 24, 15, 12, 9 };
            //var firmwareColors_Part4 = new int[] { 5, 5, 4, 4 };
            var firmwareColors_Part5 = new int[] { 26, 17, 14, 11 };
            WriteIndexedPalette(outFolder + "_IndexedPalette2.z80", firmwareColors_Part4, firmwareColors_Part5);

            var firmwareColors_Part6 = new int[] { 15, 6, 3, 0 };
            //var firmwareColors_Part6 = new int[] { 2, 2, 1, 1 };
            var firmwareColors_Part7 = new int[] { 17, 8, 5, 2 };
            WriteIndexedPalette(outFolder + "_IndexedPalette3.z80", firmwareColors_Part6, firmwareColors_Part7);

            var firmwareColors = firmwareColors_Part0; // Part0 or Part1 is the same, it's used here to generate GFX shape, not encode colors
            var colors = new Color[firmwareColors.Length];
            for(int i=0; i<firmwareColors.Length; i++)
            {
                var firmwareColor = firmwareColors[i];
                var color = Color.FromArgb(Pixel.CPCAsicPalette[firmwareColor]);
                color = Color.FromArgb(255, color); // set alpha to 255
                colors[i] = color;
            }

            var iFrame = 0;

            // var ZOOM = 40; // 1000=lots of strides  10=1 giant stride
            for (double z = 46; z >= 0; z -= 0.72)
            {
                var maxIndex = (double)(firmwareColors.Length-0);

                //1st pass: compute frame horizontal PEN line
                var horizOffsetIndices = new List<int>();
                for (double x = 0; x < width; x++)
                {
                    var R = (double)0;

                    if (x < height)
                    {
                        R = Math.Abs(Math.Sin(z * (x / (height * 2) - 0.5) * -0.5));
                    }
                    else
                    {
                        R = Math.Abs(Math.Sin(z * (x / (height * 2) - 0.5) * -0.5));
                    }

                    //R += 80 * (Math.Sin(z * (0 / (height * 2) - 0.5) * -0.5));


                    var offsetIndex = (int)(R * maxIndex);
                    horizOffsetIndices.Add(offsetIndex);
                }

                //2nd pass: compute raster frame vertical (8 colors)
                var vertOffsetIndices = new List<int>();
                for (double y = 0; y < height; y++)
                {
                    var R = Math.Abs(Math.Sin(z* (y/height-0.5) * -0.5));

                   // R += 0.9*(Math.Sin(z * (0 / (height * 2) - 0.5) * -0.5));

                    //R *= 0.9; // make Y with better ratio

                    var offsetIndex = (int)(R * maxIndex);
                    vertOffsetIndices.Add(offsetIndex);
                }

                //3nd pass: create debug bitmap
                for (int y = 0; y < (int)height; y++)
                {
                    for (int x = 0; x < (int)width; x++)
                    {
                        var index = horizOffsetIndices[x];
                        index += vertOffsetIndices[y];

                        while (index >= firmwareColors.Length)
                        {
                            index -= firmwareColors.Length;
                        }

                        var color = colors[index];
                        bitmap.SetPixel(x, y, color);
                    }
                }
                
                // write frame horizontal
                var z80 = ";\n";
                z80 += "HorizFrame" + iFrame + ":\n";
                foreach (var horizIndex in horizOffsetIndices)
                {
                    string v = "";
                    switch(horizIndex)
                    {
                        case 0: v = "ToEqualBytes0/256"; break;
                        case 1: v = "ToEqualBytes1/256"; break;
                        case 2: v = "ToEqualBytes2/256"; break;
                        default:
                        case 3: v = "ToEqualBytes3/256"; break;
                    }

                    z80 += "\tdb " + v + "\n";
                }
                File.WriteAllText(outFolder + "Keftales/_Horiz_" + iFrame + ".z80", z80);

                // write frame vertical
                z80 = ";\n";
                z80 += "VertFrame" + iFrame + ":\n";
                for (int y = 0; y < (int)height; y++)
                {
                    var index = vertOffsetIndices[y];
                    while (index >= firmwareColors.Length)
                    {
                        index -= firmwareColors.Length;
                    }

                    // value is now 0,1,2,3
                    index *= 8; // to match _IndexPalette

                    z80 += "\tdb " + index + "\n";
                }
                File.WriteAllText(outFolder + "Keftales/_Vert_" + iFrame + ".z80", z80);
                
                // write debug bitmap
                bitmap.Save(outFolder + "Keftales/_" + iFrame + ".png", ImageFormat.Png);
                iFrame++;
            }
            
            var nbFrames = iFrame;
            CreateAnimFrames(nbFrames, outFolder);
        }
    }
}
