﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _12_CubeRotX
    {
        private const int NbFrames90Deg = 90;
        private const int BmpWidth = 48 * 8; // MODE 0
        private const int BmpHeight = 268;
        private const double CubeSize = 1.2;

        public const int LineHeight = 2;

        private int FrameYState = 0;
        private bool FrameYIsFlippedColors = false;

        public class FrameY
        {
            public Bitmap Bitmap;
            public List<Math3D.Vector2D> VerticallySortedDots;
            public string Z80;

            public bool IsFlippedColors;
            public Color Color0;
            public Color Color1;
        }

        public _12_CubeRotX()
        {
            CreateCube(
                    Program.ToZ80Folder("12_Cube/")
                );
        }

        private void CreateCube(string outFolder)
        {
            CreateRotX(outFolder);
        }

        private Math3D.Vector3D[] CreateGeometry()
        {
            var vertices = new Math3D.Vector3D[4 * 2];
            vertices[0] = new Math3D.Vector3D(-CubeSize, -CubeSize, -CubeSize);
            vertices[1] = new Math3D.Vector3D(CubeSize, -CubeSize, -CubeSize);
            vertices[2] = new Math3D.Vector3D(CubeSize, -CubeSize, CubeSize);
            vertices[3] = new Math3D.Vector3D(-CubeSize, -CubeSize, CubeSize);
            vertices[4] = new Math3D.Vector3D(-CubeSize, CubeSize, -CubeSize);
            vertices[5] = new Math3D.Vector3D(CubeSize, CubeSize, -CubeSize);
            vertices[6] = new Math3D.Vector3D(CubeSize, CubeSize, CubeSize);
            vertices[7] = new Math3D.Vector3D(-CubeSize, CubeSize, CubeSize);
            return vertices;
        }

        private FrameY CreateSequentialFrameY(int iFrame)
        {
            if (iFrame == 0)
            {
                FrameYState = 0;
                FrameYIsFlippedColors = false;
            }

            var vertices = CreateGeometry();

            double rotDeg = 90 * (((double)iFrame) / (double)(NbFrames90Deg));
            var rotatedVertices = Math3D.RotateX(vertices, rotDeg);

            double distance = 3.1; // near camera
            var translatedVertices = Math3D.Translate(rotatedVertices, new Math3D.Vector3D(0, 0, -CubeSize - distance));

            var projectedVertices = new List<Math3D.Vector2D>();
            foreach (var v in translatedVertices)
            {
                v.Z *= 0.5;
                var x2D = BmpWidth / 2 + (int)((double)((v.X / v.Z) * (BmpWidth / 2)));
                var y2D = BmpHeight / 2 + (int)((double)((v.Y / v.Z) * (BmpHeight / 2)));

                projectedVertices.Add(new Math3D.Vector2D(x2D, y2D));
            }

            var verticallySortedDots = projectedVertices.OrderBy(o => o.Y).ToList();

            if (FrameYState == 0)
            {
                if (verticallySortedDots[6].X <= verticallySortedDots[0].X)
                {
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                }
                else
                {
                    FrameYState = 1;
                }
            }
            if (FrameYState == 1)
            {
                if (verticallySortedDots[7].X > verticallySortedDots[0].X)
                {
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                }
                else
                {
                    FrameYState = 2;
                }
            }
            if (FrameYState == 2)
            {
                if (verticallySortedDots[0].X < verticallySortedDots[2].X)
                {
                    FrameYIsFlippedColors = true;

                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                }
                else if (verticallySortedDots[2].Y <=BmpHeight/2)
                {
                    verticallySortedDots.RemoveAt(4);
                    verticallySortedDots.RemoveAt(4);
                }
            }

            var frameY = new FrameY();
            frameY.VerticallySortedDots = verticallySortedDots;
            frameY.IsFlippedColors = FrameYIsFlippedColors;
            return frameY;
        }

        private Bitmap CreateTextureBitmap(Color color0, Color color1)
        {
            var bitmap = new Bitmap(2, 2);
            bitmap.SetPixel(0, 0, color0);
            bitmap.SetPixel(1, 0, color1);
            bitmap.SetPixel(0, 1, color1);
            bitmap.SetPixel(1, 1, color0);
            return bitmap;
        }

        private void CreateRotX(string outFolder)
        {
            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[1] | 0xFF000000));
            var color1 = Color.FromArgb((int)(Pixel.CPCAsicPalette[4] | 0xFF000000)); // color 3
            var color2 = Color.FromArgb((int)(Pixel.CPCAsicPalette[15] | 0xFF000000));
            var color3 = Color.FromArgb((int)(Pixel.CPCAsicPalette[26] | 0xFF000000));

            var color15 = Color.FromArgb((int)(Pixel.CPCAsicPalette[18] | 0xFF000000));

            var brushColor0 = new SolidBrush(color0);
            var brushColor1 = new SolidBrush(color1);
            var brushColor2 = new SolidBrush(color2);
            var brushColor3 = new SolidBrush(color3);

            var textureBitmap0 = CreateTextureBitmap(color1, color2);
            var textureBitmap1 = CreateTextureBitmap(color3, color2);
            var textureBrush0 = new TextureBrush(textureBitmap0);
            var textureBrush1 = new TextureBrush(textureBitmap1);

            var penColor1 = new Pen(color1);
            var penColor2 = new Pen(color2);

            var penColor15 = new Pen(color15);

            var frames = new List<FrameY>();

            // create frames and compute TOP/BOTTOM
            for (int iFrame = 0; iFrame < NbFrames90Deg; iFrame++)
            {
                var frameY = CreateSequentialFrameY(iFrame);
                frames.Add(frameY);

                // reorder vertices
                var reorderedDots = new List<Math3D.Vector2D>();
                if (frameY.VerticallySortedDots.Count == 4)
                {                    
                    for (int i = 0; i < 4; i++)
                    {
                        var index = i;
                        if (!frameY.IsFlippedColors)
                        {
                            if (index == 2)
                            {
                                index = 3;
                            }
                            else if (index == 3)
                            {
                                index = 2;
                            }
                        }

                        reorderedDots.Add(frameY.VerticallySortedDots[index]);
                    }
                }
                else if (frameY.VerticallySortedDots.Count == 6)
                {
                    for (int i = 0; i < 4; i++)
                    {
                        var index = i;
                        if (!frameY.IsFlippedColors)
                        {
                            if (index == 2)
                            {
                                index = 3;
                            }
                            else if (index == 3)
                            {
                                index = 2;
                            }
                        }
                        reorderedDots.Add(frameY.VerticallySortedDots[index]);
                    }
                    
                    for (int i = 4; i < 6; i++)
                    {
                        var index = i;
                        if (!frameY.IsFlippedColors)
                        {
                            if (index == 4)
                            {
                                index = 5;
                            }
                            else if (index == 5)
                            {
                                index = 4;
                            }
                        }
                        reorderedDots.Add(frameY.VerticallySortedDots[index]);
                    }
                }
                frameY.VerticallySortedDots = reorderedDots;
            }

            // draw CPC graphics (no debug layer like top/bottom zones etc)
            for (int iFrame = 0; iFrame < NbFrames90Deg; iFrame++)
            {
                var frameY = frames[iFrame];

                frameY.Bitmap = new Bitmap(BmpWidth, BmpHeight);
                var graphics = Graphics.FromImage(frameY.Bitmap);

                // cls
                graphics.FillRectangle(brushColor0, 0, 0, BmpWidth, BmpHeight);

                // draw dots
                /*foreach ( var dot in frameY.VerticallySortedDots)
                {
                    frameY.Bitmap.SetPixel((int)dot.X, (int)dot.Y, color1);
                }*/
                Brush b0;
                Brush b1;
                if (frameY.IsFlippedColors)
                {
                    b0 = brushColor2;
                    b1 = brushColor1;
                }
                else
                {
                    b0 = brushColor1;
                    b1 = brushColor2;
                }

                // draw polygons
                if(frameY.VerticallySortedDots.Count==4)
                {
                    var points = new List<Point>();
                    for (int i = 0; i < 4; i++) 
                    {
                        var p = frameY.VerticallySortedDots[i];
                        var point = new Point((int)p.X, (int)p.Y);
                        points.Add(point);
                    }
                    graphics.FillPolygon(b0, points.ToArray());
                }
                else if (frameY.VerticallySortedDots.Count == 6)
                {
                    var points = new List<Point>();
                    for(int i=0; i<4; i++)
                    {
                        var p = frameY.VerticallySortedDots[i];
                        var point = new Point((int)p.X, (int)p.Y);
                        points.Add(point);
                    }
                    graphics.FillPolygon(b0, points.ToArray());

                    points = new List<Point>();
                    for (int i = 2; i < 6; i++)
                    {
                        var p = frameY.VerticallySortedDots[i];
                        var point = new Point((int)p.X, (int)p.Y);
                        points.Add(point);
                    }
                    graphics.FillPolygon(b1, points.ToArray());
                }
                
                graphics.Dispose();
            }

            // draw debug layer like top/bottom zones etc
            for (int iFrame = 0; iFrame < NbFrames90Deg; iFrame++)
            {
                var frameY = frames[iFrame];

                var graphics = Graphics.FromImage(frameY.Bitmap);

                // draw top/bottom
                /*graphics.DrawLine(penColor15, 0, TopY1, BmpWidth, TopY1);
                graphics.DrawLine(penColor15, 0, TopY2, BmpWidth, TopY2);
                graphics.DrawLine(penColor15, 0, BottomY1, BmpWidth, BottomY1);
                graphics.DrawLine(penColor15, 0, BottomY2, BmpWidth, BottomY2);*/

                graphics.Dispose();
                frameY.Bitmap.Save(outFolder + "RotX\\RotX" + iFrame.ToString("00") + ".png", ImageFormat.Png);
            }
        }
    }
}
