﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _12_CubeRotY
    {
        private const int NbFrames90Deg = 90;
        private const int BmpWidth = 48 * 4 * 2; // MODE 1
        private const int BmpHeight = 268;
        private const double CubeSize = 1.128;
        private const double CubeZDiv = 0.5;
        private const double CubeDistance = 3.1;
        public const int TileWidth = 4; // must be multiple of 4
        public const int TileHeight = 4;

        private int FrameYState = 0;

        public class Tile
        {
            public Bitmap Bitmap;
            public byte[] CPCSprite;
            public byte[] CPCReversedSprite;
        }
        public List<Tile> UniqueTiles = new List<Tile>();

        public class TileFrame
        {
            public int PosX = -1;
            public int PosY = -1;
            public int UniqueTileIndex = -1;
        }
        public class FrameY
        {
            public Bitmap TopBitmap;
            public List<Math3D.Vector2D> VerticallySortedDots;
            public List<TileFrame> Tiles = new List<TileFrame>();
            public string Z80;
        }

        public _12_CubeRotY()
        {
            for (int i = 2; i < 10; i++)
            {
                CreateRaster(Program.ToDataFolder("12_Cube/Step" + i + ".png"),
                                Program.ToZ80Folder("12_Cube/_Step" + i + ".png.z80"));
            }

            var sinCurve1 = SinCurve.CreateCos(256, 0, 8, 8);
            for(int i=0; i<256; i++)
            {
                sinCurve1[i] = (byte)(sinCurve1[i] * 2);
            }
            File.WriteAllBytes(Program.ToZ80Folder("12_Cube/_SinCurve1.bin"), sinCurve1);

            CreateCube(
                    Program.ToZ80Folder("12_Cube/")
                );
        }

        private void CreateRaster(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }

        private void CreateCube(string outFolder)
        {
            CreateRotY(outFolder);
        }

        // compare / add to unique tile if necessary / returns index
        private int PushTileToBuffer(Bitmap tileBitmap)
        {
            int tileIndex = 0;
            for (; tileIndex < UniqueTiles.Count; tileIndex++)
            {
                var uniqueTile = UniqueTiles[tileIndex];

                int samePixelCount = 0;

                for (int y = 0; y < TileHeight; y++)
                {
                    for (int x = 0; x < TileWidth; x++)
                    {
                        if (uniqueTile.Bitmap.GetPixel(x, y) == tileBitmap.GetPixel(x, y))
                        {
                            samePixelCount++;
                        }
                    }
                }

                if (samePixelCount == TileWidth * TileHeight)
                {
                    return tileIndex;
                }
            }

            // not found, let's add the new tile
            tileIndex = UniqueTiles.Count;

            var tile = new Tile();
            tile.Bitmap = tileBitmap;
            UniqueTiles.Add(tile);

            return tileIndex;
        }

        private Math3D.Vector3D[] CreateGeometry()
        {
            var vertices = new Math3D.Vector3D[4 * 2];
            vertices[0] = new Math3D.Vector3D(-CubeSize, -CubeSize, -CubeSize);
            vertices[1] = new Math3D.Vector3D(CubeSize, -CubeSize, -CubeSize);
            vertices[2] = new Math3D.Vector3D(CubeSize, -CubeSize, CubeSize);
            vertices[3] = new Math3D.Vector3D(-CubeSize, -CubeSize, CubeSize);
            vertices[4] = new Math3D.Vector3D(-CubeSize, CubeSize, -CubeSize);
            vertices[5] = new Math3D.Vector3D(CubeSize, CubeSize, -CubeSize);
            vertices[6] = new Math3D.Vector3D(CubeSize, CubeSize, CubeSize);
            vertices[7] = new Math3D.Vector3D(-CubeSize, CubeSize, CubeSize);
            return vertices;
        }

        private List<Math3D.Vector2D> CreateSequentialFrameY(int iFrame)
        {
            if (iFrame == 0)
            {
                FrameYState = 0;
            }

            var vertices = CreateGeometry();

            double rotDeg = 90 * (((double)iFrame) / (double)(NbFrames90Deg));
            var rotatedVertices = Math3D.RotateY(vertices, rotDeg);

            var translatedVertices = Math3D.Translate(rotatedVertices, new Math3D.Vector3D(0, 0, -CubeSize - CubeDistance));

            var projectedVertices = new List<Math3D.Vector2D>();
            foreach (var v in translatedVertices)
            {
                v.Z *= CubeZDiv;
                var x2D = BmpWidth / 2 + (int)((double)((v.X / v.Z) * (BmpWidth / 2)));
                var y2D = BmpHeight / 2 + (int)((double)((v.Y / v.Z) * (BmpHeight / 2)));

                projectedVertices.Add(new Math3D.Vector2D(x2D, y2D));
            }

            var verticallySortedDots = projectedVertices.OrderBy(o => o.Y).ToList();

            if (FrameYState == 0)
            {
                if (verticallySortedDots[0].Y == verticallySortedDots[1].Y)
                {
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                }
                else if (verticallySortedDots[2].X < verticallySortedDots[0].X)
                {
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                }
                else
                {
                    verticallySortedDots.RemoveAt(3);
                    verticallySortedDots.RemoveAt(3);
                    FrameYState = 1;
                }
            }
            else if (FrameYState == 1)
            {
                verticallySortedDots.RemoveAt(3);
                verticallySortedDots.RemoveAt(3);

                if (verticallySortedDots[1].Y >= verticallySortedDots[2].Y)
                {
                    FrameYState = 2;
                }
            }
            else if (FrameYState == 2)
            {
                verticallySortedDots.RemoveAt(3);
                verticallySortedDots.RemoveAt(3);

                if (verticallySortedDots[2].X > verticallySortedDots[0].X)
                {
                    FrameYState = 3;
                    verticallySortedDots.RemoveAt(2);
                    verticallySortedDots.RemoveAt(2);
                }
            }
            else if (FrameYState == 3)
            {
                verticallySortedDots.RemoveAt(3);
                verticallySortedDots.RemoveAt(3);
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
            }

            return verticallySortedDots;
        }

        private Bitmap CreateTextureBitmap(Color color0, Color color1)
        {
            var bitmap = new Bitmap(2, 2);
            bitmap.SetPixel(0, 0, color1);
            bitmap.SetPixel(1, 0, color0);
            bitmap.SetPixel(0, 1, color0);
            bitmap.SetPixel(1, 1, color1);

            return bitmap;
        }

        private void CreateRotY(string outFolder)
        {
            #region Colors
            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));
            var color1 = Color.FromArgb((int)(Pixel.CPCAsicPalette[5] | 0xFF000000));
            var color2 = Color.FromArgb((int)(Pixel.CPCAsicPalette[7] | 0xFF000000));
            var color3 = Color.FromArgb((int)(Pixel.CPCAsicPalette[26] | 0xFF000000));

            var brushColor0 = new SolidBrush(color0);
            var brushColor1 = new SolidBrush(color1);
            var brushColor2 = new SolidBrush(color2);
            var brushColor3 = new SolidBrush(color3);
            
            var textureBitmap0 = CreateTextureBitmap(color1, color0);
            var textureBitmap1 = CreateTextureBitmap(color2, color0);
            var textureBrush0 = new TextureBrush(textureBitmap0);
            var textureBrush1 = new TextureBrush(textureBitmap1);

            var penColor0 = new Pen(color0);
            var penColor1 = new Pen(color1);
            var penColor2 = new Pen(color2);
            var penColor3 = new Pen(color3);
            #endregion

            #region Compute palette
            var palette = new Palette();
            palette.AddColor(color0);
            palette.AddColor(color1);
            palette.AddColor(color2);
            palette.AddColor(color3);

            // create palette
            File.WriteAllText(outFolder + "_CubePalette.z80", palette.ToZ80TableString(Program.FilenameToKeyWordName("Cube")));
            #endregion

            var frames = new List<FrameY>();

            int TopY1 = BmpHeight;
            int TopY2 = 0;
            int MinX = BmpWidth;
            int MaxX = 0;

            byte Reg1 = 43;

            #region create frames and compute TOP/BOTTOM
            for (int iFrame = 0; iFrame < NbFrames90Deg+1; iFrame++)
            {
                var frameY = new FrameY();
                frameY.VerticallySortedDots = CreateSequentialFrameY(iFrame);
                frames.Add(frameY);

                foreach (var dot in frameY.VerticallySortedDots)
                {
                    if (MinX > (int)dot.X)
                    {
                        MinX = (int)dot.X;
                    }
                    if (MaxX < (int)dot.X)
                    {
                        MaxX = (int)dot.X;
                    }
                }

                if (frameY.VerticallySortedDots.Count == 6)
                {
                    if (frameY.VerticallySortedDots[0].Y < TopY1)
                    {
                        TopY1 = (int)frameY.VerticallySortedDots[0].Y;
                    }
                    if (frameY.VerticallySortedDots[2].Y > TopY2)
                    {
                        TopY2 = (int)frameY.VerticallySortedDots[2].Y;
                    }
                }
            }
            #endregion

            // make sure last line has no data
            TopY2++;

            // readjust topY1 to be multiple of TileHeight
            var iTopY1 = TopY2;
            while (iTopY1 > TopY1)
            {
                iTopY1 -= TileHeight;
            }
            TopY1 = iTopY1;

            // readjust MaxX to be multiple of TileWidth
            var iMaxX = MinX;
            while (iMaxX < MaxX)
            {
                iMaxX += TileWidth;
            }
            MaxX = iMaxX;

            // raw buffer of tile (used to compare and make diff)
            var tileBufferWidth = MaxX - MinX;
            var tileBufferTileWidth = tileBufferWidth / TileWidth;
            var tileBufferHeight = TopY2 - TopY1;
            var tileBufferTileHeight = tileBufferHeight / TileHeight;
            var tileBufferIndex = new List<int>();
            for (int i = 0; i < tileBufferTileHeight * tileBufferTileWidth; i++)
            {
                tileBufferIndex.Add(-1);
            }

            int tileBufferCRTCReg4 = tileBufferTileHeight - 1;
            int tileBufferCRTCReg9 = TileHeight - 1;

            Console.WriteLine("TileHeight= " + TileHeight);
            Console.WriteLine("tileBufferWidth= " + tileBufferWidth);
            Console.WriteLine("tileBufferTileWidth= " + tileBufferTileWidth);
            Console.WriteLine("tileBufferHeight= " + tileBufferHeight);
            Console.WriteLine("tileBufferTileHeight= " + tileBufferTileHeight);
            Console.WriteLine("tileBufferCRTCReg4= " + tileBufferCRTCReg4);
            Console.WriteLine("tileBufferCRTCReg9= " + tileBufferCRTCReg9);

            // compute line by line height
            int lineByLineHeight = 268; // fullscreen
            lineByLineHeight -= 2; // remove top/bottom 1-line border
            lineByLineHeight -= tileBufferHeight * 2; // remove tile buffer
            Console.WriteLine("lineByLineHeight= " + lineByLineHeight);

            #region draw flat triangles (top only) + write png frame 
            for (int iFrame = 0; iFrame < NbFrames90Deg+1; iFrame++)
            {
                var frameY = frames[iFrame];

                frameY.TopBitmap = new Bitmap(tileBufferWidth, tileBufferHeight);
                var topGraphics = Graphics.FromImage(frameY.TopBitmap);
                topGraphics.FillRectangle(brushColor0, 0, 0, tileBufferWidth, tileBufferHeight);

                // draw flat triangles
                if (frameY.VerticallySortedDots.Count == 4)
                {
                    var b = textureBrush0;
                    if (frameY.VerticallySortedDots[1].X > BmpWidth / 2)
                    {
                        b = textureBrush1;
                    }
                    if (frameY.VerticallySortedDots[0].Y == frameY.VerticallySortedDots[1].Y)
                    {
                        b = textureBrush0;
                    }

                    if (iFrame == NbFrames90Deg)
                    {
                        b = textureBrush1;
                    }

                    var points = new List<Point>();
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X - MinX, (int)frameY.VerticallySortedDots[0].Y - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[1].X - MinX, (int)frameY.VerticallySortedDots[1].Y - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[1].X - MinX, TopY2 - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X - MinX, TopY2 - TopY1));
                    topGraphics.FillPolygon(b, points.ToArray());
                }
                else if (frameY.VerticallySortedDots.Count == 6)
                {
                    var b0 = textureBrush0;
                    var b1 = textureBrush1;
                    if (frameY.VerticallySortedDots[1].X > frameY.VerticallySortedDots[0].X)
                    {
                        b0 = textureBrush1;
                        b1 = textureBrush0;
                    }

                    // top
                    var points = new List<Point>();
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X - MinX, (int)frameY.VerticallySortedDots[0].Y - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[1].X - MinX, (int)frameY.VerticallySortedDots[1].Y - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[1].X - MinX, TopY2 - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X - MinX, TopY2 - TopY1));
                    topGraphics.FillPolygon(b0, points.ToArray());

                    points = new List<Point>();
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X - MinX, (int)frameY.VerticallySortedDots[0].Y - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[2].X - MinX, (int)frameY.VerticallySortedDots[2].Y - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[2].X - MinX, TopY2 - TopY1));
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X - MinX, TopY2 - TopY1));
                    topGraphics.FillPolygon(b1, points.ToArray());
                }

                topGraphics.Dispose();
                frameY.TopBitmap.Save(outFolder + "RotY\\RotY" + iFrame.ToString("00") + ".png", ImageFormat.Png);
            }
            #endregion

            #region create tiles
            int maxTilesPerFrame = -1;
            int minTilesPerFrame = 999999999;
            for (int iFrame = 0; iFrame < NbFrames90Deg+1; iFrame++)
            {
                if (iFrame == NbFrames90Deg)
                {
                    frames[0].Tiles.Clear();
                }

                for (int iY = 0; iY < tileBufferHeight; iY += TileHeight)
                {
                    for (int iX = 0; iX < tileBufferWidth; iX += TileWidth)
                    {
                        // get current tile bitmap
                        var tileBitmap = new Bitmap(TileWidth, TileHeight);
                        for (int y = 0; y < TileHeight; y++)
                        {
                            for (int x = 0; x < TileWidth; x++)
                            {
                                var color = frames[iFrame].TopBitmap.GetPixel(iX + x, iY + y);
                                tileBitmap.SetPixel(x, y, color);
                            }
                        }
                        // compare / add to unique tile if necessary / returns index
                        var tileIndex = PushTileToBuffer(tileBitmap);
                        //tileBitmap.Save(outFolder + "RotY\\Tile" + tileIndex.ToString("00") + ".png", ImageFormat.Png);

                        var tileBufferY = iY / TileHeight;
                        var tileBufferX = iX / TileWidth;

                        if (tileBufferIndex[(tileBufferY * tileBufferTileWidth) + tileBufferX] != tileIndex)
                        {
                            tileBufferIndex[(tileBufferY * tileBufferTileWidth) + tileBufferX] = tileIndex;

                            var tileFrame = new TileFrame();
                            tileFrame.PosX = tileBufferX;
                            tileFrame.PosY = tileBufferY;
                            tileFrame.UniqueTileIndex = tileIndex;

                            if (iFrame == NbFrames90Deg)
                            {
                                frames[0].Tiles.Add(tileFrame);
                            }
                            else
                            {
                                frames[iFrame].Tiles.Add(tileFrame);
                            }
                        }
                    }
                }

                if ((iFrame != 0) && (iFrame<NbFrames90Deg))
                {
                    //Console.WriteLine("Frame" + iFrame + "= " + frames[iFrame].Tiles.Count + " tiles");

                    if (maxTilesPerFrame < frames[iFrame].Tiles.Count)
                    {
                        maxTilesPerFrame = frames[iFrame].Tiles.Count;
                    }
                    if (minTilesPerFrame > frames[iFrame].Tiles.Count)
                    {
                        minTilesPerFrame = frames[iFrame].Tiles.Count;
                    }
                }
            }
            #endregion

            Console.WriteLine("Unique tile count= " + UniqueTiles.Count);
            Console.WriteLine("Max tiles per frame= " + maxTilesPerFrame);

            #region create first frame
            int firstFrameYSecondLine = -1;
            int firstFrameWidthInBytes = tileBufferWidth / 4; // mode 1
            {
                var frameY = frames[0];

                var firstPixel = frameY.TopBitmap.GetPixel(frameY.TopBitmap.Width / 2, 0);
                int y = 1;
                while (firstPixel == frameY.TopBitmap.GetPixel(frameY.TopBitmap.Width / 2, y))
                {
                    y++;
                }

                firstFrameYSecondLine = y;

                var firstFrameBitmap = new Bitmap(tileBufferWidth, 2);
                for (int x = 0; x < tileBufferWidth; x++)
                {
                    firstFrameBitmap.SetPixel(x, 0, frameY.TopBitmap.GetPixel(x, y - 1));
                    firstFrameBitmap.SetPixel(x, 1, frameY.TopBitmap.GetPixel(x, y));
                }
                firstFrameBitmap.Save(outFolder + "_FirstFrame.png", ImageFormat.Png);

                #region create cpc bitmap for first frame
                for (y = 0; y < 2; y++)
                {
                    var cpcBytes = new List<byte>();

                    for (int x = 0; x < firstFrameBitmap.Width; x += 4)
                    {
                        Color p0 = firstFrameBitmap.GetPixel(x + 0, y);
                        Color p1 = firstFrameBitmap.GetPixel(x + 1, y);
                        Color p2 = firstFrameBitmap.GetPixel(x + 2, y);
                        Color p3 = firstFrameBitmap.GetPixel(x + 3, y);

                        var paletteEntry0 = palette.AddColor(p0);
                        var paletteEntry1 = palette.AddColor(p1);
                        var paletteEntry2 = palette.AddColor(p2);
                        var paletteEntry3 = palette.AddColor(p3);

                        var cpcByte = Pixel.ToMode1Pixel(
                                    paletteEntry0.IndexInPalette,
                                    paletteEntry1.IndexInPalette,
                                    paletteEntry2.IndexInPalette,
                                    paletteEntry3.IndexInPalette
                                    );

                        cpcByte = 0; // to debug
                        cpcBytes.Add(cpcByte);
                    }

                    File.WriteAllBytes(outFolder + "_FirstFrameLine" + y + ".bin", cpcBytes.ToArray());
                }
                #endregion
            }
            #endregion

            Console.WriteLine("firstFrameYSecondLine= " + firstFrameYSecondLine);
            Console.WriteLine("firstFrameWidthInBytes= " + firstFrameWidthInBytes);

            #region compute CPC sprite for all tiles
            foreach (var tile in UniqueTiles)
            {
                var bitmap = tile.Bitmap;

                {
                    var cpcBytes = new List<byte>();

                    for (int y = 0; y < bitmap.Height; y++)
                    {
                        Color p0 = bitmap.GetPixel(0, y);
                        Color p1 = bitmap.GetPixel(1, y);
                        Color p2 = bitmap.GetPixel(2, y);
                        Color p3 = bitmap.GetPixel(3, y);

                        var paletteEntry0 = palette.AddColor(p0);
                        var paletteEntry1 = palette.AddColor(p1);
                        var paletteEntry2 = palette.AddColor(p2);
                        var paletteEntry3 = palette.AddColor(p3);

                        var cpcByte = Pixel.ToMode1Pixel(
                                    paletteEntry0.IndexInPalette,
                                    paletteEntry1.IndexInPalette,
                                    paletteEntry2.IndexInPalette,
                                    paletteEntry3.IndexInPalette
                                    );

                        cpcBytes.Add(cpcByte);
                    }

                    tile.CPCSprite = cpcBytes.ToArray();
                }
            }

            var flippedPalette = new Palette();
            flippedPalette.AddColor(color0);
            flippedPalette.AddColor(color2); // flipped
            flippedPalette.AddColor(color1); // flipped
            flippedPalette.AddColor(color3);
            foreach (var tile in UniqueTiles)
            {
                var bitmap = tile.Bitmap;

                {
                    var cpcBytes = new List<byte>();

                    for (int y = 0; y < bitmap.Height; y++)
                    {
                        Color p0 = bitmap.GetPixel(0, y);
                        Color p1 = bitmap.GetPixel(1, y);
                        Color p2 = bitmap.GetPixel(2, y);
                        Color p3 = bitmap.GetPixel(3, y);

                        var paletteEntry0 = flippedPalette.AddColor(p0);
                        var paletteEntry1 = flippedPalette.AddColor(p1);
                        var paletteEntry2 = flippedPalette.AddColor(p2);
                        var paletteEntry3 = flippedPalette.AddColor(p3);

                        var cpcByte = Pixel.ToMode1Pixel(
                                    paletteEntry0.IndexInPalette,
                                    paletteEntry1.IndexInPalette,
                                    paletteEntry2.IndexInPalette,
                                    paletteEntry3.IndexInPalette
                                    );

                        cpcBytes.Add(cpcByte);
                    }

                    tile.CPCReversedSprite = cpcBytes.ToArray();
                }
            }
            #endregion

            #region create cpc sprite tables
            var z80Lines = new List<string>();
            for (int y = 0; y < TileHeight; y++)
            {
                var z80Line = "\tdb 0 ;; dummy (shifted table for Z80 'next line' opti)\n";
                z80Line += "\tdb ";
                z80Lines.Add(z80Line);
            }
            for (int y = 0; y < TileHeight; y++)
            {
                bool first = true;

                int tileCount = 0;

                foreach (var tile in UniqueTiles)
                {
                    if (tileCount == 8)
                    {
                        tileCount = 0;
                        z80Lines[y] += "\n\tdb ";
                        first = true;
                    }

                    var cpcByte = tile.CPCSprite[y];
                    if (first)
                    {
                        z80Lines[y] += "&" + cpcByte.ToString("X2");
                        first = false;
                    }
                    else
                    {
                        z80Lines[y] += ",&" + cpcByte.ToString("X2");
                    }

                    tileCount++;
                }
            }
            var z80 = "";
            z80 += ";\n";
            for (int y = 0; y < TileHeight; y++)
            {
                z80 += "\n";
                if (y != 0)
                {
                    z80 += "\tALIGN 256\n";
                }
                z80 += z80Lines[y] + "\n";
            }
            File.WriteAllText(outFolder + "_TileData.z80", z80);
            #endregion

            #region create ColorReversed cpc sprite tables
            z80Lines = new List<string>();
            for (int y = 0; y < TileHeight; y++)
            {
                var z80Line = "\tdb 0 ;; dummy (shifted table for Z80 'next line' opti)\n";
                z80Line += "\tdb ";
                z80Lines.Add(z80Line);
            }
            for (int y = 0; y < TileHeight; y++)
            {
                bool first = true;

                int tileCount = 0;

                foreach (var tile in UniqueTiles)
                {
                    if (tileCount == 8)
                    {
                        tileCount = 0;
                        z80Lines[y] += "\n\tdb ";
                        first = true;
                    }

                    var cpcByte = tile.CPCReversedSprite[y];
                    if (first)
                    {
                        z80Lines[y] += "&" + cpcByte.ToString("X2");
                        first = false;
                    }
                    else
                    {
                        z80Lines[y] += ",&" + cpcByte.ToString("X2");
                    }

                    tileCount++;
                }
            }
            z80 = "";
            z80 += ";\n";
            for (int y = 0; y < TileHeight; y++)
            {
                z80 += "\n";
                if (y != 0)
                {
                    z80 += "\tALIGN 256\n";
                }
                z80 += z80Lines[y] + "\n";
            }
            File.WriteAllText(outFolder + "_TileColorReversedData.z80", z80);
            #endregion

            #region create cpc screen table
            var cpcTopScreenScrPtr = new List<ushort>();
            var cpcBottomScreenScrPtr = new List<ushort>();
            ushort cpcTopScreen = (ushort)0x0000;
            ushort cpcBottomScreen = (ushort)0x4000;
            for (int y = 0; y < tileBufferTileHeight; y++)
            {
                cpcTopScreenScrPtr.Add(cpcTopScreen);
                cpcBottomScreenScrPtr.Add(cpcBottomScreen);

                cpcTopScreen = (ushort)(cpcTopScreen + ((byte)Reg1 * 2));
                cpcBottomScreen = (ushort)(cpcBottomScreen + ((byte)Reg1 * 2));
            }
            #endregion

            #region create cube first frame top/bottom lines
            z80 = ";\n";
            int iTopLine = 0;
            for (int y = 0; y < tileBufferTileHeight; y++)
            {
                ushort cpcPtr = cpcTopScreenScrPtr[y];

                for (int iTileY = 0; iTileY < TileHeight; iTileY++)
                {
                    byte lineType = 0;
                    if (iTopLine >= firstFrameYSecondLine)
                    {
                        lineType = 1;
                    }

                    z80 += "\tdw &" + cpcPtr.ToString("X4") + ":db " + lineType + " ;; " + iTopLine + "\n";

                    cpcPtr = (ushort)(cpcPtr + 0x0800);
                    iTopLine++;
                }
            }
            z80 += "\n";
            z80 += ";\n";
            int iBottomLine = 0;
            for (int y = 0; y < tileBufferTileHeight; y++)
            {
                ushort cpcPtr = cpcBottomScreenScrPtr[y];

                for (int iTileY = 0; iTileY < TileHeight; iTileY++)
                {
                    byte lineType = 0;
                    if (iBottomLine < (tileBufferHeight - firstFrameYSecondLine))
                    {
                        lineType = 1;
                    }

                    z80 += "\tdw &" + cpcPtr.ToString("X4") + ":db " + lineType + " ;; " + iBottomLine + "\n";

                    cpcPtr = (ushort)(cpcPtr + 0x0800);
                    iBottomLine++;
                }
            }

            z80 = ";\n\tTOPBOTTOM_INITLINES_COUNT equ " + (iTopLine + iBottomLine) + "\n\n" + z80;

            File.WriteAllText(outFolder + "_CubeInitTopBottom.z80", z80);
            #endregion

            #region create frames
            z80 = "";
            var frameBytes = new List<byte>();
            for (int iFrame = 1; iFrame < NbFrames90Deg + 1; iFrame++)
            {
                var iFixedFrame = iFrame;
                if (iFixedFrame == NbFrames90Deg)
                {
                    iFixedFrame = 0;
                }

                var frameY = frames[iFixedFrame];

                z80 += "\n;\n";
                z80 += "CUBE_FRAME" + iFixedFrame + ": ;; " + frameY.Tiles.Count + " tiles\n";

                for (int y = 0; y < tileBufferTileHeight; y++)
                {
                    z80 += "CUBE_FRAME" + iFrame + "_LINE" + y + ":\n";
                    z80 += "\tdb";
                    int tileCount = 0;
                    foreach (var tile in frameY.Tiles)
                    {
                        if (y == tile.PosY)
                        {
                            ushort refCpcPtrPosXAt0 = (ushort)cpcTopScreenScrPtr[y];
                            byte refHighByte = (byte)(refCpcPtrPosXAt0 >> 8);

                            int tilePosX = tile.PosX;
                            ushort cpcPtr = (ushort)(cpcTopScreenScrPtr[y] + tilePosX);
                            byte highByte = (byte)(cpcPtr >> 8);
                            byte RegE = (byte)(cpcPtr & 255);
                            byte RegD = (byte)((tile.UniqueTileIndex));

                            if (refHighByte != highByte)
                            {
                                // problem with data address. Not using same high-byte for screen ptr
                                System.Diagnostics.Debugger.Break();
                            }

                            frameBytes.Add(RegD); // tile index
                            frameBytes.Add(RegE); // offset X

                            if (tileCount == 4)
                            {
                                tileCount = 0;
                                z80 += "\n\tdb";
                            }

                            if (tileCount != 0)
                            {
                                z80 += ",";
                            }

                            z80 += " &" + RegD.ToString("X2") + ",&" + RegE.ToString("X2");
                            tileCount++;
                        }
                    }

                    if (tileCount != 0)
                    {
                        z80 += ",";
                    }
                    z80 += " &FF\n";

                    frameBytes.Add((byte)0xFF);
                }
            }
            #endregion

            var framesFilenameBin = outFolder + "_Frames.bin";
            File.WriteAllBytes(framesFilenameBin, frameBytes.ToArray());
            var framesFilenameZ80 = outFolder + "_Frames.z80";
            File.WriteAllText(framesFilenameZ80, z80);
            Console.Write("DataLength= " + frameBytes.Count + " bytes\n");

            #region create DrawTiles function macros
            z80 = ";\n";
            for (int y = 0; y < tileBufferTileHeight; y++)
            {
                string nextJump;
                if (y < (tileBufferTileHeight - 1))
                {
                    nextJump = "DrawTopTiles_Line" + (y + 1);
                }
                else
                {
                    nextJump = "DrawDummyTopTiles";
                }

                z80 += "DrawTopTiles_Line" + y + ": MDRAW_TOPTILES DrawTopTiles_Line" + y + "_SkipFirstNops, " + nextJump + ", &" + ((byte)(cpcTopScreenScrPtr[y] >> 8)).ToString("X2") + "\n";
            }
            z80 += "DrawDummyTopTiles: MDRAW_DUMMYTILES\n";
            z80 += "\n;\n";
            for (int y = 0; y < tileBufferTileHeight; y++)
            {
                string nextJump;
                if (y < (tileBufferTileHeight - 1))
                {
                    nextJump = "DrawBottomTiles_Line" + (y + 1);
                }
                else
                {
                    nextJump = "DrawDummyBottomTiles";
                }

                // add padding because we use data from top, but offset differs
                byte topScrPtr = (byte)(cpcTopScreenScrPtr[y] & 255);
                byte bottomScrPtr = (byte)(cpcBottomScreenScrPtr[tileBufferTileHeight - y - 1] & 255);
                byte XOffset = (byte)(bottomScrPtr - topScrPtr);
                z80 += "DrawBottomTiles_Line" + y + ": MDRAW_BOTTOMTILES DrawBottomTiles_Line" + y + "_SkipFirstNops, " + nextJump + ", &" + ((byte)(cpcBottomScreenScrPtr[tileBufferTileHeight - 1 - y] >> 8)).ToString("X2") + ", &" + XOffset.ToString("X2") + "\n";
            }
            z80 += "DrawDummyBottomTiles: MDRAW_DUMMYTILES\n";
            File.WriteAllText(outFolder + "_DrawTiles.z80", z80);
            #endregion

            // create global info
            z80 = ";\n";
            z80 += "TileWidth equ " + TileWidth + "\n";
            z80 += "TileHeight equ " + TileHeight + "\n";
            z80 += "TileBufferWidth equ " + tileBufferWidth + "\n";
            z80 += "TileBufferHeight equ " + tileBufferHeight + "\n";
            z80 += "TileBufferTileWidth equ " + tileBufferTileWidth + "\n";
            z80 += "TileBufferTileHeight equ " + tileBufferTileHeight + "\n";
            z80 += "TileBufferCRTCReg4 equ " + tileBufferCRTCReg4 + "\n";
            z80 += "TileBufferCRTCReg9 equ " + tileBufferCRTCReg9 + "\n";
            z80 += "FirstFrameYSecondLine equ " + firstFrameYSecondLine + "\n";
            z80 += "FirstFrameWidthInBytes equ " + firstFrameWidthInBytes + "\n";
            z80 += "UniqueTilesCount equ " + UniqueTiles.Count + "\n";
            z80 += "RotYCrtcReg1 equ " + Reg1 + "\n";
            z80 += "LineByLineHeight equ " + lineByLineHeight + "\n";
            z80 += "MaxTilesPerFrame equ " + maxTilesPerFrame + "\n";
            z80 += "NbFrames90Deg equ " + NbFrames90Deg + "\n";
            File.WriteAllText(outFolder + "_Config.z80", z80);
        }
    }
}
