﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _13_Rebels
    {
        private const int BmpWidth = 48 * 2 * 2; // MODE 0
        private const int BmpHeight = 268;
        private const int AnimFrameCount = 16;
        private const int RVIMaxLines = 3*21*4;

        private const int BarMaxWidth = 32;
        private const int BarZoomCount = 8; //11;
        private const int BarColorCount = 18;
        private int BarTypeCount = 0; // will be incremented while creating bars

        public const byte R1 = 48;

        public class Bar
        {
            public int Width;
            public Bitmap Bitmap;
            public Palette Palette;
        }
        private List<Bar> Bars = new List<Bar>();

        public class FrameLine
        {
            public int yPos = -1;
            public int lineIndex = -1;
        }
        
        public class Frame
        {
            public int Index;
            public Bitmap Bitmap;
            public List<FrameLine> FrameLines = new List<FrameLine>();

            public Frame(int index)
            {
                Index = index;
                Bitmap = new Bitmap(BmpWidth, BmpHeight);

                var frameGraphics = Graphics.FromImage(Bitmap);
                var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));
                var brushColor0 = new SolidBrush(color0);
                frameGraphics.FillRectangle(brushColor0, 0, 0, BmpWidth, BmpHeight);
                frameGraphics.Dispose();
            }
        }
        private List<Frame> Frames = new List<Frame>();

        public class Line
        {
            public int Index;
            public Bitmap Bitmap;

            public Line(int index)
            {
                Index = index;
                Bitmap = new Bitmap(BmpWidth, 2);
            }
        }
        private List<Line> Lines = new List<Line>();

        private int FindLine(Bitmap lineBitmap)
        {
            foreach (var iLine in Lines)
            {
                var samePixelCount = 0;

                for (var y = 0; y < 2; y++)
                {
                    for (var x = 0; x < BmpWidth; x++)
                    {
                        if (lineBitmap.GetPixel(x, y) == iLine.Bitmap.GetPixel(x, y))
                        {
                            samePixelCount++;
                        }
                    }
                }

                if (samePixelCount == (2 * BmpWidth))
                {
                    return iLine.Index;
                }
            }

            return -1;
        }

        public _13_Rebels()
        {
            try
            {
                CreateSinMoves();

                CreateBackgroundRaster(Program.ToDataFolder("13_Rebels/Rasters268.png"),
                                Program.ToZ80Folder("13_Rebels/_Rasters268.png.z80"));

                CreateBackgroundRaster(Program.ToDataFolder("13_Rebels/Rasters512.png"),
                                Program.ToZ80Folder("13_Rebels/_Rasters512.png.z80"));

                CreateBars(Program.ToDataFolder("13_Rebels/"));
                
                CreateFrames(Program.ToZ80Folder("13_Rebels/"));
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public void CreateSinMoves()
        {
            // 5+5=10 .. table will be used in asm by adding 2 values from same table
            File.WriteAllBytes(Program.ToZ80Folder("13_Rebels/_SinDisto.bin"), SinCurve.CreateSin(256, 0, 128));
            File.WriteAllBytes(Program.ToZ80Folder("13_Rebels/_SinDisto2.bin"), SinCurve.CreateSin(256, 0, 64));
        }

        private void CreateBars(string barsFolder)
        {
            for (int i = 1; i <= 6; i++)
            {
                var barFilename = barsFolder + "M0_Bar" + i + ".png";

                var allBars = new Bitmap(System.Drawing.Image.FromFile(barFilename));

                for (int iBar = 0; iBar < BarZoomCount; iBar++)
                {
                    var barY = iBar * 3;

                    var bar = new Bar();

                    // init palette
                    var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));
                    var barPalette = new Palette();
                    barPalette.AddColor(color0);

                    // get bar width
                    var barWidth = 0;
                    for (var x = 0; x < BarMaxWidth; x++)
                    {
                        var pixel = allBars.GetPixel(x, barY);
                        if ((pixel.R + pixel.G + pixel.B) != 0)
                        {
                            barWidth = x;
                        }
                    }
                    bar.Width = barWidth + 1;

                    // create bar bitmap
                    var barBitmap = new Bitmap(bar.Width, 2);
                    for (int y = 0; y < 2; y++)
                    {
                        for (int x = 0; x < bar.Width; x++)
                        {
                            var pixel = allBars.GetPixel(x, barY + y);

                            barPalette.AddColor(pixel);

                            barBitmap.SetPixel(x, y, pixel);
                        }
                    }
                    bar.Palette = barPalette;
                    bar.Bitmap = barBitmap;

                    Bars.Add(bar);
                }

                BarTypeCount++;
            }
        }

        // displays CENTERED bar AROUND specified X pos.
        // barZoom from 0 to 1
        // barColorIndex from 0 to BarColorCount
        // returns UNIQUE line index
        private void AddBar(Bitmap destBitmap, double x, double y, double barZoom, int barType)
        {
            int barIndex = (int)((1-barZoom) * BarZoomCount);
            if (barIndex < 0)
            {
                barIndex = 0;
            }
            else if (barIndex >= BarZoomCount)
            {
                barIndex = BarZoomCount-1;
            }
            
            var bar = Bars[(barType*BarZoomCount)+barIndex];
            
            x -= ((double)bar.Width) / 2;

            var destX = (int)(x+0.5);
            var destY = (int)(y+0.5);

            for (int v = 0; v < bar.Bitmap.Height; v++)
            {
                for (int u = 0; u < bar.Width; u++)
                {
                    if (((destX + u) >= 0) && ((destX + u) < BmpWidth) &&
                        ((destY + v) >= 0) && ((destY + v) < BmpHeight))
                    {
                        var srcPixel = bar.Bitmap.GetPixel(u, v);

                        destBitmap.SetPixel(destX + u, destY + v, srcPixel);
                    }
                }
            }
        }

        private void CreateBackgroundRaster(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }

        private void CreateFrames(string outFolder)
        {
            double xInc = 4;
            double xSize = 48;
            double xOffset = 0;
            double xOffsetInc = xInc / (double)(AnimFrameCount);

            #region Create Frames
            var reuseCount = 0;
            for (int iFrame = 0; iFrame < AnimFrameCount; iFrame++)
            {
                var frame = new Frame(Frames.Count);

                var listY = new List<int>();

                const double distance = 4.0;
                const double focal = 2.4;
                const double yScale = 1.2;

                for (var iBarZoom = 0; iBarZoom < BarZoomCount ; iBarZoom++)
                {
                    double y = -(double)(BarZoomCount - iBarZoom);
                    double z = focal * (double)(distance+(double)iBarZoom);
                    var v3D = new Math3D.Vector3D(0, y, z);

                    double y2D = (double)(BmpHeight / 4) + (double)(((v3D.Y*yScale) / v3D.Z) * (BmpHeight / 4));
                    listY.Add(2*(int)(y2D + 0.5));
                }

                // duplicate last value + adjust to match 134 lines. (134x2=268 visible lines)
                listY.Add(BmpHeight/2);

                var listYIndex = 0;
                for (var iBarZoom=0; iBarZoom<BarZoomCount; iBarZoom++)
                {
                    for (var iBarZoom2 = BarZoomCount-1; iBarZoom2 >= iBarZoom; iBarZoom2--)
                    {
                        for (double x = -xSize * xInc; x < xSize * xInc; x += xInc)
                        {
                            double z = focal * (double)(distance + (double)iBarZoom2);

                            double x2D = (double)(BmpWidth / 2) + (double)((((x + xOffset)*2.1) / z) * (BmpWidth / 2));
                            
                            double barZoom = (double)(BarZoomCount - iBarZoom2) / (double)(BarZoomCount);

                            var barTypes = new int[] { 5, 4, 3, 2, 2, 1, 1, 1, 0, 0, 0 };
                            var barType = barTypes[iBarZoom2];

                            var startY = listY[listYIndex];
                            var endY = listY[listYIndex + 1];
                            for (int iY = startY; iY < endY; iY += 2)
                            {
                                AddBar(frame.Bitmap, x2D, iY, barZoom, barType);
                            }
                        }
                    }

                    listYIndex++;
                }

                var lastLineIndex = -1;
                for (var iListYIndex = 0; iListYIndex < listY.Count - 1; iListYIndex++)
                {
                    var yPos = listY[iListYIndex];
                    Bitmap lineBitmap = new Bitmap(BmpWidth, 2);

                    for (int y = 0; y < 2; y++)
                    {
                        for (int x = 0; x < BmpWidth; x++)
                        {
                            lineBitmap.SetPixel(x, y, frame.Bitmap.GetPixel(x, y + yPos));
                        }
                    }

                    var lineIndex = FindLine(lineBitmap);
                    if (lineIndex != -1)
                    {
                        reuseCount++;
                    }
                    else
                    {
                        var line = new Line(Lines.Count);
                        line.Bitmap = lineBitmap;
                        Lines.Add(line);
                        lineIndex = line.Index;
                    }

                    if (lineIndex != lastLineIndex)
                    {
                        var frameLine = new FrameLine();
                        frameLine.yPos = yPos;
                        frameLine.lineIndex = lineIndex;
                        frame.FrameLines.Add(frameLine);

                        lastLineIndex = lineIndex;
                    }
                }

                Frames.Add(frame);
                xOffset += xOffsetInc;

                Console.WriteLine("FrameIndex={0} LinesCount={1}", frame.Index, Lines.Count);
            }
            Console.WriteLine("ReuseCount={0}", reuseCount);
            #endregion

            /*
            #region Z80 Frame data
            var z80 = "";
            z80 += ";\n";
            z80 += "FRAMEDATA_FRAMECOUNT\tequ " + Frames.Count + "\n";
            z80 += "\n";
            var FrameDataByteSize = 0;
            foreach (var frame in Frames)
            {
                int lastYPos = 0;

                z80 += ";\n";
                z80 += "REBELS_FRAME" + frame.Index +":\n";

                int iLine = 0;

                // top
                foreach (var frameLine in frame.FrameLines)
                {
                    z80 += "REBELS_FRAME" + frame.Index + "_LINE" + iLine + ":\n";
                    z80 += "\tdb " + (frameLine.yPos-lastYPos) + " ;; Height=&" + (frameLine.yPos - lastYPos).ToString("X2") + " YPos=" + frameLine.yPos + "\n";
                    z80 += "\tdb " + frameLine.lineIndex + " ;; LineIndex=&" + frameLine.lineIndex.ToString("X2") + "\n";

                    lastYPos = frameLine.yPos;
                    iLine++;

                    FrameDataByteSize += 2;
                }

                z80 += ";\n";

                // reversed bottom
                int iFrameLine = frame.FrameLines.Count-1;
                for(int frameLineIndex= frame.FrameLines.Count; frameLineIndex< frame.FrameLines.Count*2; frameLineIndex++)
                {
                    var frameLine = frame.FrameLines[iFrameLine];

                    var yPos = ((BmpHeight/2)-frameLine.yPos)+ (BmpHeight / 2);

                    z80 += "REBELS_FRAME" + frame.Index + "_LINE" + iLine + ":\n";
                    z80 += "\tdb " + (yPos - lastYPos) + " ;; Height=&" + (yPos - lastYPos).ToString("X2") + " YPos=" + yPos + "\n";
                    z80 += "\tdb " + frameLine.lineIndex + " ;; LineIndex=&" + frameLine.lineIndex.ToString("X2") + "\n";

                    lastYPos = yPos;
                    iLine++;

                    FrameDataByteSize += 2;

                    iFrameLine--;
                }

                z80 += "\n";
            }
            File.WriteAllText(outFolder + "_FrameData.z80", z80);
            Console.WriteLine("FrameDataByteSize={0}", FrameDataByteSize);
            #endregion
            */

            #region Z80 VLC data
            var z80 = "";
            z80 += ";\n";
            z80 += "REBELS_R12R13_SCR0:\n";
            var baseOffsets = new List<ushort>();
            byte R12 = 0x00; // %00000000 &0000
            ushort offset = (ushort)(0);
            //while (offset <= 1024)
            var valueCount = 0;
            for (int iFrame = 0; iFrame < Frames[0].FrameLines.Count*2; iFrame++)
            {
                byte _r12 = (byte)(R12 + (offset >> 8));
                byte _r13 = (byte)(offset & 255);

                if (iFrame == Frames[0].FrameLines.Count)
                {
                    z80 += ";\n";
                    z80 += "REBELS_R12R13_SCR1:\n";
                }

                // duplicate 8times value, so we get 128 values
                for (int i = 0; i < 16; i++)
                {
                    z80 += "\tdb &" + _r12.ToString("X2") + ", &" + _r13.ToString("X2") + " ;; &" + offset.ToString("X4") + "\n";
                    valueCount += 2;
                }
                
                offset += (ushort)(R1);
            }
            File.WriteAllText(outFolder + "_R12R13_256.z80", z80);
            #endregion

            /*
            #region Z80 VLC MACRO data
            z80 = "";
            z80 += ";\n";
            //z80 += "REBELS_R12R13:\n";
            baseOffsets = new List<ushort>();
            R12 = 0x00; // %00000000 &0000
            offset = (ushort)(0);
            var iScr = 0;
            //while (offset <= 1024)
            int iBlock = 0;
            for (int iFrame = 0; iFrame < Frames[0].FrameLines.Count * 2; iFrame++)
            {
                byte _r12 = (byte)(R12 + (offset >> 8));
                byte _r13 = (byte)(offset & 255);

                if (iFrame == Frames[0].FrameLines.Count)
                {
                    z80 += ";\n";
                    iScr++;
                    iBlock = 0;
                }

                z80 += "\tREBELS_R12_" + iBlock + "_SCR" + iScr + "\tequ &" + _r12.ToString("X2") + "\n";
                z80 += "\tREBELS_R13_" + iBlock + "_SCR" + iScr + "\tequ &" + _r13.ToString("X2") + "\n";

                offset += (ushort)(R1);
                iBlock++;
            }
            File.WriteAllText(outFolder + "_R12R13_Macros.z80", z80);
            #endregion
            */

            #region Z80 Line Bitmap Pointers data
            z80 = "";
            z80 += ";\n";
            z80 += "\tALIGN 256 ;; array size is exactly 256 bytes\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += "REBELS_LINES_PTR:\n";
            for (int iVLC = 0; iVLC < 8; iVLC++)
            {
                z80 += "REBELS_LINES_VLC" + iVLC + ":\n";

                for (int iFrame = 0; iFrame < 16; iFrame++)
                {
                    var line = (iFrame * 8) + iVLC;
                    z80 += "\tdw REBELS_LINE" + line + "_VLC0\n";
                }
            }
            File.WriteAllText(outFolder + "_LineBitmapTablePtr256.z80", z80);
            #endregion

            #region Z80 Line Bitmap data
            Console.WriteLine("Creating Z80 line bitmap data..");

            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));
            var palette = new Palette();
            palette.AddColor(color0);

            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = Lines[0].Bitmap,
                Mode = 0,
                R1 = 48,
                R4 = 1,
                R9 = 2,
                StartVLC = 0,
                ApplyOffsetAtNextPage = 0,
                Palette = palette
            });
            var globalPalette = image.Palette;

            var LineBitmapByteSize = 0;
            z80 = "";
            foreach (var line in Lines)
            {
                if((((byte)line.Index)&7)==0)
                {
                    Console.WriteLine("" + line.Index + "/" + Lines.Count);
                }
                var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = line.Bitmap,
                    Mode = 0,
                    R1 = 48,
                    R4 = 1,
                    R9 = 2,
                    StartVLC = 0,
                    ApplyOffsetAtNextPage = 0,
                    Palette = globalPalette
                });

                z80 += ";\n";
                z80 += "REBELS_LINE" + line.Index + "_VLC0:\n";
                z80 += "\tdb ";

                var firstInLine = true;

                var bCount = 0;
                var lineCount = 0;
                for (int i = 0; i < cpcImage.CPCRawBuffer_Width*cpcImage.CPCRawBuffer_Height; i++)
                {
                    var b = cpcImage.CPCRawBuffer[i];
                    if (firstInLine)
                    {
                        firstInLine = false;
                    }
                    else
                    {
                        z80 += ", ";
                    }

                    z80 += "&" + b.ToString("X2");
                    LineBitmapByteSize++;

                    bCount++;
                    if (bCount == BmpWidth / 16)
                    {
                        bCount = 0;
                        lineCount++;

                        if (i == (cpcImage.CPCRawBuffer_Width - 1))
                        {
                            z80 += "\nREBELS_LINE" + line.Index + "_VLC1:";
                            z80 += "\n\tdb ";
                        }
                        else if (i != ((cpcImage.CPCRawBuffer_Width* cpcImage.CPCRawBuffer_Height)- 1))
                        {
                            z80 += "\n\tdb ";
                        }
                        firstInLine = true;
                    }
                }
                z80 += "\n";
            }
            Console.WriteLine("" + Lines.Count + "/" + Lines.Count);
            File.WriteAllText(outFolder + "_LineBitmap.z80", z80);
            Console.WriteLine("LineBitmapByteSize={0}", LineBitmapByteSize);
            #endregion

            #region Write global palette MODE 0
            File.WriteAllText(outFolder + "_Rebels.pal.z80",
                globalPalette.ToZ80TableString(Program.FilenameToKeyWordName("REBELS")));
            #endregion

            #region Create reversed bottom part in frame bitmaps
            foreach (var frame in Frames)
            {
                // draw reversed bottom part
                for (int y = 0; y < frame.Bitmap.Height / 2; y++)
                {
                    for (int x = 0; x < frame.Bitmap.Width; x++)
                    {
                        var pixel = frame.Bitmap.GetPixel(x, y);
                        frame.Bitmap.SetPixel(x, frame.Bitmap.Height - 1 - y, pixel);
                    }
                }
            }
            #endregion

            #region Write Bitmap files
            foreach (var frame in Frames)
            { 
                frame.Bitmap.Save(outFolder + "Frames\\Frame" + frame.Index.ToString("00") + ".png", ImageFormat.Png);
            }
            #endregion
        }
    }
}
