﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _14_MorphingBars
    {
        private const int AnimFrameCount = 64;
        private const int BmpWidth = (48/3) * 2 * 2; // MODE 0
        private const int BmpHeight = 1;
        
        private const int BarMaxWidth = 32;
        private const int BarZoomCount = 8*2; //11;
        private const int BarColorCount = 18;
        
        public class Bar
        {
            public int Width;
            public Bitmap Bitmap;
        }
        private List<Bar> Bars = new List<Bar>();

        public class Anim
        {
            public Bitmap Bitmap;
            
            public Anim()
            {
                Bitmap = new Bitmap(BmpWidth, AnimFrameCount);

                var frameGraphics = Graphics.FromImage(Bitmap);
                var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));
                var brushColor0 = new SolidBrush(color0);
                frameGraphics.FillRectangle(brushColor0, 0, 0, BmpWidth, AnimFrameCount);
                frameGraphics.Dispose();
            }
        }

        public _14_MorphingBars()
        {
            try
            {
                CreateSinMoves();

                CreateXMoveR12R13();

                CreateBackgroundRaster(Program.ToDataFolder("14_MorphingBars/Rasters268.png"),
                                Program.ToZ80Folder("14_MorphingBars/_Rasters268.png.z80"));

                CreateBackgroundRaster(Program.ToDataFolder("14_MorphingBars/Rasters512.png"),
                                Program.ToZ80Folder("14_MorphingBars/_Rasters512.png.z80"));

                CreateBars(Program.ToDataFolder("14_MorphingBars/"));
                
                CreateAnimFrames(Program.ToZ80Folder("14_MorphingBars/"));
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public void CreateSinMoves()
        {
            var bytes = new byte[256];
            for(int i=0; i<256; i++)
            {
                bytes[i] = (byte)(i & 63);
            }
            //File.WriteAllBytes(Program.ToZ80Folder("14_MorphingBars/_SinDisto.bin"), bytes);

            File.WriteAllBytes(Program.ToZ80Folder("14_MorphingBars/_SinDisto.bin"), SinCurve.CreateSin(256, 0, (AnimFrameCount)/2));
            File.WriteAllBytes(Program.ToZ80Folder("14_MorphingBars/_SinDisto2.bin"), SinCurve.CreateCos(256, 0, 20));
            File.WriteAllBytes(Program.ToZ80Folder("14_MorphingBars/_SinDisto3.bin"), bytes);// SinCurve.CreateSin(256, 0, 256, 64));
            //File.WriteAllBytes(Program.ToZ80Folder("14_MorphingBars/_SinDisto3.bin"), SinCurve.CreateSin(256, 0, 64));
        }

        public void CreateXMoveR12R13()
        {
            var z80 = ";\n";

            int iHoriz = 0;
            int iChar = 15;

            for (int i = 0; i < 256; i++)
            {
                if(iHoriz==0)
                {
                    z80 += "\tdb Anim0000_R12R13/256\n";
                }
                else if (iHoriz == 1)
                {
                    z80 += "\tdb Anim4000_R12R13/256\n";
                }
                else if (iHoriz == 2)
                {
                    z80 += "\tdb Anim8000_R12R13/256\n";
                }
                else if (iHoriz == 3)
                {
                    z80 += "\tdb AnimC000_R12R13/256\n";
                }

                z80 += "\tdb " + iChar + "\n";

                iHoriz++;
                if(iHoriz==4)
                {
                    iHoriz = 0;
                    iChar--;
                    if (iChar == -1)
                    {
                        iChar = 15;
                    }
                }
            }
            File.WriteAllText(Program.ToZ80Folder("14_MorphingBars/_XMoveR12R13.z80"), z80);
        }

        private void CreateBackgroundRaster(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }

        private void CreateBars(string barsFolder)
        {
            for (int i = 1; i <= 6; i++)
            {
                var barFilename = barsFolder + "M0_Bar" + i + ".png";

                var allBars = new Bitmap(System.Drawing.Image.FromFile(barFilename));

                for (int iBar = 0; iBar < BarZoomCount/2; iBar++)
                {
                    var barY = iBar * 3;

                    var bar = new Bar();

                    // get bar width
                    var barWidth = 0;
                    for (var x = 0; x < BarMaxWidth; x++)
                    {
                        var pixel = allBars.GetPixel(x, barY);
                        if ((pixel.R + pixel.G + pixel.B) != 0)
                        {
                            barWidth = x;
                        }
                    }
                    bar.Width = barWidth + 1;

                    // create bar bitmap
                    var barBitmap = new Bitmap(bar.Width, 1);
                    for (int y = 0; y < 1; y++)
                    {
                        for (int x = 0; x < bar.Width; x++)
                        {
                            var pixel = allBars.GetPixel(x, barY + y);
                            barBitmap.SetPixel(x, y, pixel);
                        }
                    }
                    bar.Bitmap = barBitmap;

                    Bars.Add(bar);

                    // second line
                    bar = new Bar();
                    bar.Width = barWidth;
                    barBitmap = new Bitmap(bar.Width, 1);
                    for (int y = 1; y < 2; y++)
                    {
                        for (int x = 0; x < bar.Width; x++)
                        {
                            var pixel = allBars.GetPixel(x, barY + y);
                            barBitmap.SetPixel(x, y-1, pixel);
                        }
                    }
                    bar.Bitmap = barBitmap;

                    Bars.Add(bar);
                }
            }
        }

        // displays CENTERED bar AROUND specified X pos.
        // barZoom from 0 to 1
        // barColorIndex from 0 to BarColorCount
        // returns UNIQUE line index
        private void AddBar(Bitmap destBitmap, double x, double y, double barZoom, int barType)
        {
            int barIndex = (int)((1-barZoom) * BarZoomCount);
            if (barIndex < 0)
            {
                barIndex = 0;
            }
            else if (barIndex >= BarZoomCount)
            {
                barIndex = BarZoomCount-1;
            }
            
            var bar = Bars[(barType*BarZoomCount)+barIndex];
            
            x -= ((double)bar.Width) / 2;

            var destX = (int)(x+0.5);
            var destY = (int)(y+0.5);

            for (int u = 0; u < bar.Width; u++)
            {
                if (((destX + u) >= 0) && ((destX + u) < BmpWidth) &&
                    ((destY) >= 0) && ((destY) < (AnimFrameCount)))
                {
                    var srcPixel = bar.Bitmap.GetPixel(u, 0);

                    destBitmap.SetPixel(destX + u, destY, srcPixel);
                }
            }
        }

        private void CreateAnimFrames(string outFolder)
        {
            var anim = new Anim();

            int minValue = (Bars[0].Width/2)+5;
            int maxValue = (16*2*2)- minValue;
            var sinX1 = SinCurve.CreateSinDouble(AnimFrameCount, minValue+2, maxValue);
            var sinX2 = SinCurve.CreateCosDouble(AnimFrameCount, minValue, maxValue);
            var sinX3 = SinCurve.CreateSinDouble(AnimFrameCount, minValue+18, maxValue, 1);
            var sinX4 = SinCurve.CreateSinDouble(AnimFrameCount, minValue-5, maxValue-25, 1);
            var sinX5 = SinCurve.CreateSinDouble(AnimFrameCount, minValue+18, maxValue +5, 1);

            int minZoomValue = 0;
            int maxZoomValue = 1;
            var sinZoom1 = SinCurve.CreateSinDouble(AnimFrameCount, minZoomValue, maxZoomValue, 1);
            var sinZoom2 = SinCurve.CreateSinDouble(AnimFrameCount, minZoomValue, maxZoomValue, 1);
            var sinZoom3 = SinCurve.CreateSinDouble(AnimFrameCount, minZoomValue, maxZoomValue, 2);
            var sinZoom4 = SinCurve.CreateCosDouble(AnimFrameCount, minZoomValue, maxZoomValue, 2);
            var sinZoom5 = SinCurve.CreateSinDouble(AnimFrameCount, minZoomValue, maxZoomValue, 2);

            for (int iFrame = 0; iFrame < AnimFrameCount; iFrame++)
            {
                double y = iFrame;

                var offset = AnimFrameCount/2;
                offset += iFrame + AnimFrameCount / 4;
                if (offset >= AnimFrameCount)
                {
                    offset -= AnimFrameCount;
                }
                double x1 = sinX1[offset];
                double zoom1 = sinZoom1[offset];
                int barType1 = 5;

                double x2 = sinX2[iFrame];
                int barType2 = 1;
                double zoom2 = sinZoom2[iFrame];

                var offset3 = 12;
                offset3 += iFrame*4;
                while (offset3 >= AnimFrameCount)
                {
                    offset3 -= AnimFrameCount;
                }
                while (offset3 < 0)
                {
                    offset3 += AnimFrameCount;
                }
                double x3 = sinX3[offset3];
                int barType3 = 4;
                var sinZoom3Offset = iFrame + 80;
                while (sinZoom3Offset >= AnimFrameCount)
                {
                    sinZoom3Offset -= AnimFrameCount;
                }
                double zoom3 = sinZoom3[sinZoom3Offset];

                var offset4 = 167;
                offset4 += iFrame * 4;
                while (offset4 >= AnimFrameCount)
                {
                    offset4 -= AnimFrameCount;
                }
                while (offset4 < 0)
                {
                    offset4 += AnimFrameCount;
                }
                double x4 = sinX4[offset4];
                int barType4 = 4;
                var sinZoom4Offset = iFrame + 55;
                while (sinZoom4Offset >= AnimFrameCount)
                {
                    sinZoom4Offset -= AnimFrameCount;
                }
                double zoom4 = sinZoom4[sinZoom4Offset];

                var offset5 = 167;
                offset5 += iFrame * 4;
                while (offset5 >= AnimFrameCount)
                {
                    offset5 -= AnimFrameCount;
                }
                while (offset5 < 0)
                {
                    offset5 += AnimFrameCount;
                }
                double x5 = sinX5[offset5];
                int barType5 = 2;
                var sinZoom5Offset = iFrame + 0;
                while (sinZoom5Offset >= AnimFrameCount)
                {
                    sinZoom5Offset -= AnimFrameCount;
                }
                double zoom5 = sinZoom5[sinZoom5Offset];

                //AddBar(anim.Bitmap, x3, y, zoom3, barType3);

                AddBar(anim.Bitmap, x4, y, zoom4, barType4);
                AddBar(anim.Bitmap, x5, y, zoom5, barType5);

                if (iFrame < AnimFrameCount / 2)
                {
                    AddBar(anim.Bitmap, x1, y, zoom1, barType1);
                    AddBar(anim.Bitmap, x2, y, zoom2, barType2);
                }
                else
                {
                    AddBar(anim.Bitmap, x2, y, zoom2, barType2);
                    AddBar(anim.Bitmap, x1, y, zoom1, barType1);
                }
            }
            
            anim.Bitmap.Save(outFolder + "_Anim.png", ImageFormat.Png);

            // init palette with black as first color
            // will make PEN 0 for rasters
            var emptyPalette = new Palette();
            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));
            emptyPalette.AddColor(color0);

            // for palette
            var cpcGlobalImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = anim.Bitmap,
                Mode = 0,
                R1 = 48 / 3,
                R4 = AnimFrameCount,
                R9 = 0,
                StartVLC = 0,
                FirstPixelInPaletteX = 0,
                FirstPixelInPaletteY = 0,
                StartR12 = 0x00,
                Palette = emptyPalette,
            }
            );
            var globalPalette = cpcGlobalImage.Palette;

            var baseBitmap = anim.Bitmap;

            for (int iShift = 0; iShift < 4; iShift++)
            {
                var bitmap = new Bitmap(BmpWidth, AnimFrameCount);

                //bitmap = ReverseBitmapOrder(bitmap);

                for (int y = 0; y < AnimFrameCount; y++)
                {
                    var srcX = 0;
                    for (int x = iShift; x < BmpWidth; x++)
                    {
                        var srcPixel = baseBitmap.GetPixel(srcX, y);
                        srcX++;

                        bitmap.SetPixel(x, y, srcPixel);
                    }

                    // take left pixels from previous line
                    if (y > 0)
                    {
                        for (int x = 0; x < iShift; x++)
                        {
                            var srcPixel = baseBitmap.GetPixel(BmpWidth - iShift + x, y - 1);
                            bitmap.SetPixel(x, y, srcPixel);
                        }
                    }

                    // treat top-left pixels
                    if (y == (AnimFrameCount - 1))
                    {
                        for (int x = 0; x < iShift; x++)
                        {
                            var srcPixel = baseBitmap.GetPixel(BmpWidth - iShift + x, y);
                            bitmap.SetPixel(x, 0, srcPixel);
                        }
                    }
                }

                var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = bitmap,
                    Mode = 0,
                    R1 = 48 / 3,
                    R4 = AnimFrameCount,
                    R9 = 0,
                    StartVLC = 0,
                    FirstPixelInPaletteX = 0,
                    FirstPixelInPaletteY = 0,
                    StartR12 = 0x40 * iShift,
                    Palette = globalPalette
                }
                );

                var memInFilename = "0000";
                if (iShift == 1)
                {
                    memInFilename = "4000";
                }
                else if (iShift == 2)
                {
                    memInFilename = "8000";
                }
                else if (iShift == 3)
                {
                    memInFilename = "C000";
                }

                bitmap.Save(outFolder + "_Anim" + memInFilename + ".png", ImageFormat.Png);
                File.WriteAllBytes(outFolder + "_Anim" + memInFilename + ".bin", cpcImage.CPCRawBuffer);
                File.WriteAllBytes(outFolder + "_Anim" + memInFilename + "_R12R13.bin", cpcImage.R12R13);
            }

            File.WriteAllText(outFolder + "_Anim.pal.z80", globalPalette.ToZ80TableString(Program.FilenameToKeyWordName("Anim")));
        }
    }

    /*private Bitmap ReverseBitmapOrder(Bitmap baseBitmap)
    {
        var bitmap = new Bitmap(baseBitmap.Width, baseBitmap.Height);

        var orderY = new int[baseBitmap.Height];
        for(int i=0; i<orderY.Length; i++)
        {
            orderY[i] = i;
            orderY[i] = i+3;
            orderY[i] = i+6;
        }

        for(int y=0; y<baseBitmap.Height; y++)
        {

        }

        return bitmap;
    }*/
}
