﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _15_Circle
    {
        const int BarSize = 140;
        const int BarSinHeight = 30;
        const int NbFrames = 24;
        
        public _15_Circle()
        {
            try
            {
                CreateSinMoves(Program.ToZ80Folder("15_Circle/"));

                var r12r13 = new List<byte>();
                CreateCircle(Program.ToZ80Folder("15_Circle/"), ref r12r13);

                CreateRasterAnim(Program.ToDataFolder("15_Circle/Bar420_1.png"),
                                Program.ToZ80Folder("15_Circle/"));

                CreateR12R13Anim(Program.ToZ80Folder("15_Circle/"), ref r12r13);
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public void CreateSinMoves(string outFolder)
        {
            File.WriteAllBytes(outFolder + "_SinDisto.bin", SinCurve.CreateSin(256, 0, 268-BarSize, 2));
            //File.WriteAllBytes(outFolder + "_SinDisto.bin", SinCurve.CreateSin(256, 0, 0, 2));

            File.WriteAllBytes(outFolder + "_SinDisto2.bin", SinCurve.CreateCos(256, 0, BarSinHeight/2, 3));
            //File.WriteAllBytes(outFolder + "_SinDisto2.bin", SinCurve.CreateCos(256, 0, 0, 3));
        }

        public void CreateCircle(string outFolder, ref List<byte> r12r13)
        {
            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[1] | 0xFF000000));
            var color1 = Color.FromArgb((int)(Pixel.CPCAsicPalette[24] | 0xFF000000));
            var color2 = Color.FromArgb((int)(Pixel.CPCAsicPalette[20] | 0xFF000000));
            var color3 = Color.FromArgb((int)(Pixel.CPCAsicPalette[6] | 0xFF000000));

            var pen0 = new SolidBrush(color0);
            var pen1 = new SolidBrush(color1);
            var pen2 = new SolidBrush(color2);
            var pen3 = new SolidBrush(color3);

            var pens = new List<SolidBrush>();
            pens.Add(pen0);
            pens.Add(pen1);
            pens.Add(pen2);
            pens.Add(pen3);
            pens.Add(pen2);
            pens.Add(pen1);

            var penIndex = 0;

            var bitmap = new Bitmap(48 * 8, 21*4);

            var graphics = Graphics.FromImage(bitmap);

            float zoom = 1;
            for (float iStep = 400; iStep >= 2; iStep -= (float)5)
            {
                zoom -= 0.0132f;

                float x = (float)((bitmap.Width / 2)) - (iStep*zoom);
                float y = (float)((bitmap.Height)) - (iStep * zoom);
                float width = (iStep * zoom) * 2;
                float height = (iStep * zoom) * 2;

                penIndex++;
                if(penIndex>=pens.Count)
                {
                    penIndex = 0;
                }
                var pen = pens[penIndex];

                graphics.FillEllipse(pen, x, y, width, height);
            }

            graphics.Dispose();

            bitmap.Save(outFolder + "_Circle.png", ImageFormat.Png);

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 48,
                R4 = 21,
                R9 = 0
            }
            );

            for (int iPage = 0; iPage < cpcImage.Pages.Length; iPage++)
            {
                var page = cpcImage.Pages[iPage];

                foreach (var VLC in cpcImage.Pages[iPage].VLCs)
                {
                    File.WriteAllBytes(outFolder + "_Circle.page" + iPage + ".vlc" + VLC.Index, VLC.CPCBytes);
                }
            }
                        
            for (int i = 0; i < cpcImage.R12R13.Length; i += 2)
            {
                r12r13.Add(cpcImage.R12R13[i]);
                r12r13.Add(cpcImage.R12R13[i + 1]);
            }
            //File.WriteAllBytes(outFolder + "_Circle.r12r13", r12r13.ToArray());
        }

        private void CreateRasterAnim(string imageFilename, string outFolder)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var raster = new List<byte>();

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);
                var gateArrayColor = palette.ToGateArrayByteArray()[0];
                raster.Add(gateArrayColor);
            }

            var z80Index = "";

            var frames = new List<byte>();

            for (var iFrame = 0; iFrame < 1; iFrame++)
            {
                var frame = new List<byte>();

                var delta = 6;
                var p0Index = delta * 0;
                var p1Index = delta * 1;
                var p2Index = delta * 2;
                var p3Index = delta * 3;

                for (int i = 0; i < BarSize*3; i++)
                {
                    byte p0 = Pixel.CPCGateArrayPalette[0];
                    byte p1 = Pixel.CPCGateArrayPalette[0];
                    byte p2 = Pixel.CPCGateArrayPalette[0];
                    byte p3 = Pixel.CPCGateArrayPalette[0];

                    /*p0 = Pixel.CPCGateArrayPalette[6];
                    p1 = Pixel.CPCGateArrayPalette[16];
                    p2 = Pixel.CPCGateArrayPalette[24];
                    p3 = Pixel.CPCGateArrayPalette[15];*/

                    
                    {
                        if (p0Index < raster.Count)
                        {
                            p0 = raster[p0Index];
                        }
                        if (p1Index < raster.Count)
                        {
                            p1 = raster[p1Index];
                        }
                        if (p2Index < raster.Count)
                        {
                            p2 = raster[p2Index];
                        }
                        if (p3Index < raster.Count)
                        {
                            p3 = raster[p3Index];
                        }

                        p0Index++;
                        p1Index++;
                        p2Index++;
                        p3Index++;
                    }

                    z80Index += "\tdw Circle_Rasters + " + frame.Count + "\n";
                    frame.Add(p0);
                    frame.Add(p1);
                    frame.Add(p2);
                    frame.Add(p3);
                }

                frames.AddRange(frame.ToArray());
            }

            File.WriteAllBytes(outFolder + "_RasterFrames.bin", frames.ToArray());
            File.WriteAllText(outFolder + "_RasterFrames.z80", z80Index);
        }

        private void CreateR12R13Anim(string outFolder, ref List<byte> r12r13)
        {
            var frames = new List<byte>();

            const double Radian360 = 6.283185307179586;
            
            var frameOffset = new List<int>();

            for (int iFrame = 0; iFrame < NbFrames; iFrame++)
            {
                frameOffset.Add(frames.Count);

                var frame = new List<byte>();

                double fFrame = (double)iFrame;
                double fNbFrames = (double)(NbFrames+1);
                double progress = ((fFrame) / fNbFrames); // 0 to 2

                double radAngle = (progress * Radian360);

                double frameDeltaSin1 = Math.Cos(radAngle);
                double frameDeltaSin2 = Math.Sin(radAngle);

                //Console.WriteLine(frameDeltaSin1);
               // int iValue = -1;

                bool isPair = true;
                for (int i = 0; i < 268+BarSinHeight; i++)
                {
                    frameDeltaSin1 += 0.045;
                    frameDeltaSin2 += 0.045;

                    var sin1 = (0.5 + Math.Sin(frameDeltaSin1) * 0.5);
                    var sin2 = (0.5 - Math.Cos(frameDeltaSin2) * 0.5);

                    //int iValue = (int)( ((sin1+sin2)*0.5)*((double)r12r13.Count/2) );
                    int iValue = (int)(((sin1+sin2) * 0.5) * ((double)(r12r13.Count) / 2));

                    //Console.WriteLine(sin1);

                    while (iValue >= 21*4)
                    {
                        iValue -= (21*4);
                    }
                    while (iValue < 0)
                    {
                        iValue += (21 * 4);
                    }

                    var r12 = r12r13[(iValue * 2) + 0];
                    var r13 = r12r13[(iValue * 2) + 1];

                    // order is R12, R13 then R13, R12
                    if (isPair)
                    {
                        frame.Add(r12);
                        frame.Add(r13);
                        isPair = false;
                    }
                    else
                    {
                        frame.Add(r13);
                        frame.Add(r12);
                        isPair = true;
                    }
                }

                frames.AddRange(frame.ToArray());
            }

            File.WriteAllBytes(outFolder + "_R12R13Frames.bin", frames.ToArray());

            int iFrameCount = 0;
            var z80Frame = "";
            var sinCurve = SinCurve.CreateSin(256, 0, NbFrames, 2);
            //var sinCurve = SinCurve.CreateSin(256, 0, 0, 2);
            int j = 0;
            for (int i = 0; i < 256; i++)
            {
                z80Frame += "\tdw Circle_R12R13 + " + frameOffset[sinCurve[i]] + "\n";
                //z80Frame += "\tdw Circle_R12R13 + " + frameOffset[j] + "\n";

                j++;
                if(j>=NbFrames)
                {
                    j = 0;
                }

                iFrameCount++;
            }
            /*   for (int i = NbFrames-2; i > 0; i--)
               {
                   z80Frame += "\tdw Circle_R12R13 + " + frameOffset[i] + "\n";
                   iFrameCount++;
               }*/

            var z80 = "";
            z80 += ";\n";
            z80 += "CircleNbFrames equ " + iFrameCount + "\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += z80Frame;
            
            File.WriteAllText(outFolder + "_R12R13Frames.z80", z80);
        }
    }
}
