﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Shared;

namespace Build
{
    public class FXSineLogo
    {
        public FXSineLogo()
        {
            ImageMaker.ImageMaker.CreateMode0Raw(
                DirHelper.SourceFolder + "Demo/FXSineLogo/Raw/",
                DirHelper.SourceFolder + "Demo/FXSineLogo/Compiled/",
                "CondenseLogo.png");

            CreateRVIData(
                DirHelper.SourceFolder + "Demo/FXSineLogo/Compiled/" + "RVIData.bin");

            CreateSinAnim(
                DirHelper.SourceFolder + "Demo/FXSineLogo/Compiled/" + "SinData.bin");
        }
        
        private const int VLCCount = 4;

        private List<ushort> RegItems = new List<ushort>();

        private void AddScreenBlock(ushort screenPtr, byte high, int nbChars)
        {
            ushort reg = (ushort)((ushort)(high) << 8);

            for (int iChar = 0; iChar < nbChars; iChar++)
            {
                for (int iVLC = 0; iVLC < VLCCount; iVLC++)
                {
                    int vlc = iVLC;
                    ushort CPCValue = (ushort)((((ushort)vlc) << 14) + ((ushort)reg));
                    RegItems.Add(CPCValue);
                }

                reg += 48;
                //reg += 64;
            }
        }

        float iProgress = 0;

        public List<int> CreateFrameFX(float t, float maxTime, int maxLine, int fxHeight)
        {
            var frameData = new List<int>();

            float progress = (t / maxTime);
            float degStep = 360.0f / maxTime;
            float degrees = degStep * progress;
            double angle = Math.PI * degrees / 180.0;
            double ratio = Math.Sin(angle * 0.5f);

            float iStep2 = (float)t * 0.01f;


            for (var i = 0; i < fxHeight; i++)
            {
                float iStep = (float)i;
                degStep = 130.0f / ((float)maxLine);
                degrees = degStep * (iStep);

                angle = Math.PI * degrees / 180.0f;
                double sinAngle = Math.Sin(angle);

                //sinAngle += Math.Sin(iStep2);
                iStep2 += 0.001f;

                sinAngle *= ratio;

                sinAngle *= maxLine;

                int iLine = (int)sinAngle;

                iLine += (int)iProgress;

                if (iLine < 0)
                {
                    iLine = -iLine;
                }
                while (iLine >= maxLine)
                {
                    iLine -= maxLine;
                }

                frameData.Add(iLine);
            }

            return frameData;
        }

        public void CreateSinAnim(string filename)
        {
            var data = new List<byte>();

            float maxValue = 14.0f;
            for (int i = 0; i < 256; i++)
            {
                float progress = ((float)i) / 255.0f;

                float v = (maxValue / 2) + ((float)Math.Sin(2 * Math.PI * progress) * (maxValue / 2));

                byte b = (byte)v;

                data.Add(b);
            }

            File.WriteAllBytes(filename, data.ToArray());
        }

        public void CreateRVIData(string filename)
        {
            AddScreenBlock(0x4000, 0x10, 20);
            AddScreenBlock(0x8000, 0x20, 20);
            AddScreenBlock(0xC000, 0x30, 20);

            var data = new List<byte>();

            int nbFrames = 0;

            float maxTime = 1.0f;
            for (float t = 0.0f; t < maxTime; t += 0.008f)
            {
                float iT = t;

                var frameData = CreateFrameFX(iT, maxTime, RegItems.Count, 220);

                byte iVLC = 1;

                int frameDataCount = frameData.Count;
                for (int iLine = 0; iLine < frameDataCount; iLine++)
                {
                    int y = iLine; // frameData[iLine];
                    while (y >= RegItems.Count)
                    {
                        y -= RegItems.Count;
                    }

                    byte r12 = (byte)(RegItems[y] >> 8);
                    byte r13 = (byte)(RegItems[y] & 255);

                    byte wishedVLC = (byte)(RegItems[y] >> 14);

                    r12 &= 0x3F; // 00111111

                    if (iLine >= (frameDataCount - 2))
                    {
                        r12 = (byte)((RegItems[0] >> 8) & 0x3F);
                        r13 = (byte)(RegItems[0] & 255);
                        wishedVLC = 0;
                    }

                    byte addVlc = 255;

                    /*
                    if ((iVLC == 0) && (wishedVLC == 0)) { addVlc = 2; }
                    else if ((iVLC == 0) && (wishedVLC == 1)) { addVlc = 3; }
                    else if ((iVLC == 0) && (wishedVLC == 2)) { addVlc = 4; }

                    if ((iVLC == 1) && (wishedVLC == 0)) { addVlc = 1; }
                    else if ((iVLC == 1) && (wishedVLC == 1)) { addVlc = 2; }
                    else if ((iVLC == 1) && (wishedVLC == 2)) { addVlc = 3; }

                    if ((iVLC == 2) && (wishedVLC == 0)) { addVlc = 3; }
                    else if ((iVLC == 2) && (wishedVLC == 1)) { addVlc = 1; }
                    else if ((iVLC == 2) && (wishedVLC == 2)) { addVlc = 2; }
                    */

                    addVlc = 4;

                    if (addVlc == 255)
                    {
                        System.Diagnostics.Debugger.Break();
                    }

                    iVLC = wishedVLC; // next VLC

                    addVlc--;
                    r12 = (byte)(r12 + (byte)(addVlc << 6));

                    byte r13_2bits = (byte)((r13 >> 6) & 0x03);

                    r12 = (byte)(r12 + (byte)(r13_2bits << 2));
                    data.Add(r12);
                }

                nbFrames++;
            }

            File.WriteAllBytes(filename, data.ToArray());
        }
    }
}
