﻿using System;
using Shared;
using ImageMaker;
using System.IO;
using System.Collections.Generic;

namespace Build
{
	public class FXTheRipper
	{
		public FXTheRipper()
		{
			const int R2_Shift = 4;

			CreateVertR12R13Table(
				"TheRipper_VertR12R13_",
				R2_Shift,
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/");

			CreateVertExit(
				"TheRipper_VertExit_",
				R2_Shift,
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/");

			CreateHoriz(
				"TheRipper_Horiz_",
				R2_Shift,
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/");

			CreateVertical(
				"TheRipper_Vertical_",
				R2_Shift,
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/");

			ImageMaker.ImageMaker.CreateMode0Raw(
                DirHelper.SourceFolder + "Demo/FXTheRipper/Raw/",
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/",
				"TheRipper.png");

			ImageMaker.ImageMaker.CreateTopBottomBabyR2(
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/",
                DirHelper.SourceFolder + "Demo/FXTheRipper/Compiled/",
				"TheRipper.png.cpcRaw",
				96+16);
		}

		public static void CreateHoriz(string prefix, int R2_Shift, string compiledFolder)
		{
			var baseOffsets = new List<ushort>();

			var shiftOffset = R2_Shift * 2;
			var leftShift = 0; // from 0 to 16
			byte R12 = 0x2C; // %00101100
			ushort offset = (ushort)(16 + leftShift);
			while (offset <= 1024 - 48 + shiftOffset)
			{
				byte _r12 = (byte)(R12 + (offset >> 8));
				byte _r13 = (byte)(offset & 255);

				ushort finalOffset = (ushort)((_r12 << 8) + _r13);
				baseOffsets.Add(finalOffset);

				offset += (ushort)(48 + shiftOffset);
			}

			R12 = 0x3C; // %00111100
			offset = (ushort)(0 + leftShift);
			while (offset <= 1024 - 48 + shiftOffset)
			{
				byte _r12 = (byte)(R12 + (offset >> 8));
				byte _r13 = (byte)(offset & 255);

				ushort finalOffset = (ushort)((_r12 << 8) + _r13);
				baseOffsets.Add(finalOffset);

				offset += (ushort)(48 + shiftOffset);
			}

			int NbFrames = 180;

			var content = ";\n";
			content += prefix + "R12R13_NBFRAMES equ " + NbFrames + "\n";
			content += "\n";

			float s1 = 10.0f;
			float s2 = 70.0f;

			var frameCoefDelta = Math.PI / ((float)NbFrames);
			var frameCoefCalc = -(Math.PI / 2);

			for (int iFrame = 0; iFrame < NbFrames; iFrame++)
			{
				var frameCoef = 1.0f - Math.Abs(Math.Sin(frameCoefCalc));
				frameCoefCalc += frameCoefDelta;

				frameCoef *= 3.2f;
				if (frameCoef > 1.0f)
				{
					frameCoef = 1.0f;
				}
				if (frameCoef < 0)
				{
					frameCoef = 0;
				}

				//Console.WriteLine("" + frameCoef);

				//s1 += (float)(((float)Math.PI*3.0f)/(float)NbFrames);
				//s2 -= (float)(((float)Math.PI*4.6f)/(float)NbFrames);

				s1 -= 0.22f;
				s2 += 0.27f;

				float localS1 = s1;
				float localS2 = s2;

				content += ";\n";
				content += prefix + "R12R13_" + iFrame + ":\n";

				foreach (var baseOffset in baseOffsets)
				{
					ushort _r12 = (ushort)((baseOffset >> 10) << 10);
					ushort curOffset = (ushort)(baseOffset & 1023);

					double s = Math.Sin(localS1) * 10.0f;
					s += Math.Sin(localS2) * 10.0f;

					s *= frameCoef;

					s += 14 + 8f; // center

					var sinOffset = (ushort)(s * 0.45f); // from 0 to 16

					localS1 += 0.19f;
					localS2 -= 0.25f;

					ushort final = (ushort)0;

					if ((sinOffset & 1) == 0)
					{
						sinOffset >>= 1;
						final = (ushort)(_r12 + curOffset + sinOffset);
					}
					else
					{
						sinOffset >>= 1;
						_r12 -= 0x2000;
						final = (ushort)(_r12 + curOffset + sinOffset);
					}

					content += "\t db &" + (final >> 8).ToString("X2") + ", &" + (final & 255).ToString("X2") + " ; sin offset= " + sinOffset + "\n";
				}
			}

			File.WriteAllText(compiledFolder + "R12R13_Horiz.z80", content);
		}

		public static void CreateVertical(string prefix, int R2_Shift, string compiledFolder)
		{
			var baseOffsets = new List<ushort>();

			var shiftOffset = R2_Shift * 2;
			var leftShift = 4; // from 0 to 16
			byte R12 = 0x2C; // %00101100
			ushort offset = (ushort)(16 + leftShift);
			while (offset <= 1024 - 48 + shiftOffset)
			{
				byte _r12 = (byte)(R12 + (offset >> 8));
				byte _r13 = (byte)(offset & 255);

				ushort finalOffset = (ushort)((_r12 << 8) + _r13);
				finalOffset -= 0x2000;
				baseOffsets.Add(finalOffset);

				finalOffset += 0x2000;
				baseOffsets.Add(finalOffset);

				offset += (ushort)(48 + shiftOffset);
			}

			R12 = 0x3C; // %00111100
			offset = (ushort)(0 + leftShift);
			while (offset <= 1024 - 48 + shiftOffset)
			{
				byte _r12 = (byte)(R12 + (offset >> 8));
				byte _r13 = (byte)(offset & 255);

				ushort finalOffset = (ushort)((_r12 << 8) + _r13);
				finalOffset -= 0x2000;
				baseOffsets.Add(finalOffset);

				finalOffset += 0x2000;
				baseOffsets.Add(finalOffset);

				offset += (ushort)(48 + shiftOffset);
			}

			// dont forget to update FrameManager, cf:
			// dw THERIPPER_VBLFrame_Vertical, 160
			int NbFrames = 180;

			var content = ";\n";
			content += prefix + "R12R13_NBFRAMES equ " + NbFrames + "\n";
			content += "\n";

			float frameSin1 = 0.0f;
			float frameSin2 = 0.0f;

			for (int iFrame = 0; iFrame < NbFrames; iFrame++)
			{
				content += ";\n";
				content += prefix + "R12R13_" + iFrame + ":\n";

				var nbBlockLines = baseOffsets.Count;

				var values = new byte[nbBlockLines];
				for (int i = 0; i < nbBlockLines; i++)
				{
					values[i] = 0;
				}

				float frameCoef = 1.0f;
				if (iFrame != 0)
				{
					var delta = 0.999f / ((float)NbFrames);
					frameCoef = ((float)iFrame) * delta;
					frameCoef = 0.999f - frameCoef;

					/*frameCoef *= 0.8f;
					if (frameCoef > 1.0f)
					{
						frameCoef = 1.0f;
					}*/
				}

				ushort currentFlatBlock = 0;

				frameSin1 += 0.06f;
				frameSin2 += 0.043f;

				var sin1 = frameSin1;
				var sin2 = frameSin2;

				for (int iBlock = 0; iBlock < nbBlockLines; iBlock++)
				{
					float yBlockCenter = (float)(nbBlockLines / 2);

					float distance = 50000;
					if (iFrame != 0)
					{
						distance = (float)(NbFrames) / (float)(iFrame);
					}

					float fY = -(yBlockCenter - 0.1f - (float)iBlock) * distance;
					fY += yBlockCenter;// * (0.5f + (0.5f*(float) Math.Abs(Math.Sin(distance*0.5f))));

					var plasmaDelta = (float)Math.Sin(sin1) + (float)Math.Sin(sin2);
					plasmaDelta *= 16;
					plasmaDelta *= frameCoef;

					fY += plasmaDelta;

					int y = (int)fY;
					if ((y >= 0) && (y < nbBlockLines))
					{
						values[y] = (byte)iBlock;

						if (currentFlatBlock == 0)
						{
							currentFlatBlock = (byte)iBlock;
						}
					}
				}

				for (int y = 0; y < nbBlockLines; y++)
				{
					if (values[y] == 0)
					{
						values[y] = (byte)currentFlatBlock;
					}
					else
					{
						currentFlatBlock = values[y];
					}
				}

				/*for (int y = 0; y < nbBlockLines; y++)
				{
					var baseOffset = baseOffsets[y];
					ushort _r12 = (ushort)((baseOffset>>10)<<10);
					ushort curOffset = (ushort)(baseOffset & 1023);
					ushort final = (ushort)(_r12 + curOffset);

					values[y] = final;
				}*/

				foreach (byte final in values)
				{
					content += "\tdb &" + (final * 2).ToString("X2") + "\n";
				}
			}

			File.WriteAllText(compiledFolder + "R12R13_Vertical.z80", content);
		}

		public static void CreateVertExit(string prefix, int R2_Shift, string compiledFolder)
		{
			// dont forget to update FrameManager, cf:
			// dw THERIPPER_VBLFrame_Vertical, 160
			int NbFrames = 100;
			int NbFrames2 = 70;
			int NbFrames3 = 49;

			var rasterOffset = 2;
			const int maxLineHeightClipping = 276 - 4;
			var maxLineHeight = 276 + 4;

			var frameCount = NbFrames + NbFrames2 + NbFrames3;

			var content = "\n";

			var frameYCoefDelta = Math.PI / ((float)NbFrames) / 2;
			double frameYCoefCalc = 0;

			content += "; nb lines (R9=0) / imageSize R4 / R9 compensate\n";

			var skipFirstCount = 10;

			for (int iFrame = 0; iFrame < NbFrames; iFrame++)
			{
				var frameYCoef = 1.0f - Math.Abs(Math.Cos(frameYCoefCalc));
				frameYCoefCalc += frameYCoefDelta;

				if (skipFirstCount > 0)
				{
					skipFirstCount--;
					iFrame--;
					NbFrames--;
					continue;
				}

				ushort lineHeight = (ushort)(frameYCoef * ((double)maxLineHeight));

			RecalcLineHeight:
				if (lineHeight < 4)
				{
					lineHeight = 4;
				}

				if (lineHeight > maxLineHeightClipping)
				{
					lineHeight = maxLineHeightClipping;
				}
				//lineHeight = maxLineHeightClipping;

				byte imageSizeR4 = (byte)(((ushort)maxLineHeight - (ushort)lineHeight) >> 3); // div 8
				byte cumulateLines = (byte)((ushort)maxLineHeight - ((ushort)(imageSizeR4) << 3) - (ushort)lineHeight);

				if (cumulateLines == 1)
				{
					lineHeight++;
					goto RecalcLineHeight;
				}

				int rasterTableOffset = (ushort)(276 - 3 - lineHeight - rasterOffset);
				if (rasterTableOffset < 0)
				{
					rasterTableOffset = 0;
				}

				content += prefix + "R12R13_" + (iFrame) + ":\t";

				content += "dw &" + rasterTableOffset.ToString("X4") + ":";
				content += "dw &" + (lineHeight - 2).ToString("X4") + ":";
				content += "db &" + imageSizeR4.ToString("X2") + ":";
				content += "db &" + cumulateLines.ToString("X2") + "\n";
			}

			frameYCoefDelta = Math.PI / ((float)NbFrames2);
			frameYCoefCalc = 0;

			for (int iFrame = 0; iFrame < NbFrames2; iFrame++)
			{
				var frameYCoef = 1.0f - Math.Abs(Math.Sin(frameYCoefCalc) * 0.40f);
				frameYCoefCalc += frameYCoefDelta;

				ushort lineHeight = (ushort)(frameYCoef * ((double)maxLineHeight));

			RecalcLineHeight2:
				if (lineHeight < 4)
				{
					lineHeight = 4;
				}

				if (lineHeight > maxLineHeightClipping)
				{
					lineHeight = maxLineHeightClipping;
				}

				byte imageSizeR4 = (byte)(((ushort)maxLineHeight - (ushort)lineHeight) >> 3); // div 8
				byte cumulateLines = (byte)((ushort)maxLineHeight - ((ushort)(imageSizeR4) << 3) - (ushort)lineHeight);

				if (cumulateLines == 1)
				{
					lineHeight++;
					goto RecalcLineHeight2;
				}

				ushort rasterTableOffset = (ushort)(276 - 3 - lineHeight - rasterOffset);
				if (rasterTableOffset < 0)
				{
					rasterTableOffset = 0;
				}

				/*if(imageSizeR4==0)
				{
					frameCount--;
					continue;
				}*/

				content += prefix + "R12R13_" + (NbFrames + iFrame) + ":\t";

				content += "dw &" + rasterTableOffset.ToString("X4") + ":";
				content += "dw &" + (lineHeight - 2).ToString("X4") + ":";
				content += "db &" + imageSizeR4.ToString("X2") + ":";
				content += "db &" + cumulateLines.ToString("X2") + "\n";
			}

			frameYCoefDelta = Math.PI / ((float)NbFrames3);
			frameYCoefCalc = 0;

			for (int iFrame = 0; iFrame < NbFrames3; iFrame++)
			{
				var frameYCoef = 1.0f - Math.Abs(Math.Sin(frameYCoefCalc) * 0.16f);
				frameYCoefCalc += frameYCoefDelta;

				ushort lineHeight = (ushort)(frameYCoef * ((double)maxLineHeight));

			RecalcLineHeight3:
				if (lineHeight < 4)
				{
					lineHeight = 4;
				}

				if (lineHeight > maxLineHeightClipping)
				{
					lineHeight = maxLineHeightClipping;
				}

				byte imageSizeR4 = (byte)(((ushort)maxLineHeight - (ushort)lineHeight) >> 3); // div 8
				byte cumulateLines = (byte)((ushort)maxLineHeight - ((ushort)(imageSizeR4) << 3) - (ushort)lineHeight);

				if (cumulateLines == 1)
				{
					lineHeight++;
					goto RecalcLineHeight3;
				}

				ushort rasterTableOffset = (ushort)(276 - 3 - lineHeight - rasterOffset);
				if (rasterTableOffset < 0)
				{
					rasterTableOffset = 0;
				}

				/*if(imageSizeR4==0)
				{
					frameCount--;
					continue;
				}*/

				content += prefix + "R12R13_" + (NbFrames + NbFrames2 + iFrame) + ":\t";

				content += "dw &" + rasterTableOffset.ToString("X4") + ":";
				content += "dw &" + (lineHeight - 2).ToString("X4") + ":";
				content += "db &" + imageSizeR4.ToString("X2") + ":";
				content += "db &" + cumulateLines.ToString("X2") + "\n";
			}

			content = ";\n"
							+ prefix + "R12R13_NBFRAMES equ " + (frameCount) + "\n"
							+ content;

			File.WriteAllText(compiledFolder + "R12R13_VertExit.z80", content);
		}

		public static void CreateVertR12R13Table(string prefix, int R2_Shift, string compiledFolder)
		{
			var baseOffsets = new List<ushort>();

			var shiftOffset = R2_Shift * 2;
			var leftShift = 4; // from 0 to 16
			byte R12 = 0x2C; // %00101100
			ushort offset = (ushort)(16 + leftShift);
			while (offset <= 1024 - 48 + shiftOffset)
			{
				byte _r12 = (byte)(R12 + (offset >> 8));
				byte _r13 = (byte)(offset & 255);

				ushort finalOffset = (ushort)((_r12 << 8) + _r13);
				finalOffset -= 0x2000;
				baseOffsets.Add(finalOffset);

				finalOffset += 0x2000;
				baseOffsets.Add(finalOffset);

				offset += (ushort)(48 + shiftOffset);
			}

			R12 = 0x3C; // %00111100
			offset = (ushort)(0 + leftShift);
			while (offset <= 1024 - 48 + shiftOffset)
			{
				byte _r12 = (byte)(R12 + (offset >> 8));
				byte _r13 = (byte)(offset & 255);

				ushort finalOffset = (ushort)((_r12 << 8) + _r13);
				finalOffset -= 0x2000;
				baseOffsets.Add(finalOffset);

				finalOffset += 0x2000;
				baseOffsets.Add(finalOffset);

				offset += (ushort)(48 + shiftOffset);
			}

			var content = ";\n";
			content += prefix + "Table:\n";

			foreach (var baseOffset in baseOffsets)
			{
				ushort _r12 = (ushort)((baseOffset >> 10) << 10);
				ushort curOffset = (ushort)(baseOffset & 1023);
				ushort final = (ushort)(_r12 + curOffset);

				content += "\t db &" + (final >> 8).ToString("X2") + ", &" + (final & 255).ToString("X2") + "\n";
			}

			content += "\n";

			File.WriteAllText(compiledFolder + "R12R13_VertTable.z80", content);
		}
	}
}

