﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _12_Tunnel
    {
        // using magic rasters, 8 shades are allowed. Color0=black
        const int NbFramesX = 128; // must be div 2
        const int NbSteps = 32;
        const int NbFramesY = 96; // must be div 2
        const int NbEdges = 7;
        const int Width = 192; // 192 pixels mode 0
        const int Height = 268 / 2; // screen height using R9=1
        const byte Vert_DummyOffsetInStackJPTable = (268 / 2) + 1;

        byte[] Magic0VLC1_PixelConvertedFromMagic0VLC0 = new byte[256];
        byte[] Magic1VLC0_PixelConvertedFromMagic0VLC0 = new byte[256];
        byte[] Magic1VLC1_PixelConvertedFromMagic0VLC0 = new byte[256];

        class YLineIndex
        {
            public int LineIndex = -1;
            public List<int> YPositions = new List<int>();
        }

        class LineEdge
        {
            public int iEdge;
            public int intLeftX;
            public int intRightX;
            public int y0;
            public int y1;
        }

        class Frame
        {
            public List<LineEdge> LineEdges;

            public Bitmap Magic0Vlc0Bitmap;
            public Bitmap Magic0Vlc1Bitmap;
            public Bitmap Magic1Vlc0Bitmap;
            public Bitmap Magic1Vlc1Bitmap;

            public Framework.Image Magic0Vlc0CPCImage;
            public Framework.Image Magic0Vlc1CPCImage;
            public Framework.Image Magic1Vlc0CPCImage;
            public Framework.Image Magic1Vlc1CPCImage;

            public List<byte> LineIndices = new List<byte>();
        }

        public _12_Tunnel()
        {
            //try
            {
                var palette = CreatePalette();
                List<Frame> XFrames = null;
                List<Frame> YFrames = null;
                CreateXYFrames(Program.ToZ80Folder("12_Tunnel/"), palette, out XFrames, out YFrames);

                int maxBytesChangedPerFrame = 0;
                int firstFrameLength = WriteCPCAnimation(ref XFrames, ref YFrames, Program.ToZ80Folder("12_Tunnel/"), out maxBytesChangedPerFrame);
                WriteConfig(ref XFrames, ref YFrames, firstFrameLength, maxBytesChangedPerFrame, Program.ToZ80Folder("12_Tunnel/"));
                WriteZ80MagicCases(Program.ToZ80Folder("12_Tunnel/"));
            }
            /*catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }*/
        }

        Palette CreatePalette()
        {
            var colors = new List<Color>();
            for (int i = 0; i < 16; i++) // 16 pens MODE 0.
            {
                colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[i] | 0xFF000000)));
            }

            var palette = new Palette();
            foreach (var color in colors)
            {
                palette.AddColor(color);
            }

            return palette;
        }

        void WriteConfig(ref List<Frame> XFrames, ref List<Frame> YFrames, int firstFrameLength, int maxBytesChangedPerFrame, string outFolder)
        {
            var z80 = "";

            z80 += "\n";
            z80 += ";\n";
            z80 += "TUNNEL_NB_FRAMES_Y\tequ " + YFrames.Count + "\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += "TUNNEL_NB_FRAMES_X\tequ " + XFrames.Count + "\n";
            z80 += "TUNNEL_MAX_BYTES_CHANGED_PER_FRAME\tequ " + maxBytesChangedPerFrame + "\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += "TUNNEL_LINE0_SCRPTR\tequ &8000\n";
            z80 += "TUNNEL_LINE1_SCRPTR\tequ &9000\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += "TUNNEL_FLIP0_SCRPTR\tequ &8000\n";
            z80 += "TUNNEL_FLIP1_SCRPTR\tequ &8060\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += "TUNNEL_NBSTEPS\tequ " + NbSteps + "\n";
            z80 += "\n";
            z80 += ";\n";
            z80 += "TUNNEL_FIRSTFRAME_LENGTH\tequ " + firstFrameLength + "\n";

            File.WriteAllText(outFolder + "_Tunnel_Config.z80", z80);
        }

        List<LineEdge> CreateFrameXYPositions(int iStep, bool isX, int iFrame)
        {
            var lineEdges = new List<LineEdge>();

            double nbFrames = (double)NbFramesX;
            if (!isX)
            {
                nbFrames = (double)NbFramesY;
            }

            double progress = (((double)iFrame) / ((double)nbFrames));
            //progress = 0; //////////
            double progressDeg = 360 * progress;
            double progressRad = (Math.PI * progressDeg) / 180; //Radians

            const double moveTunnelX = 1;// 50;
            const double moveTunnelY = 1;// 30;
            var frameSinX = Math.Sin(progressRad) * moveTunnelX;
            var frameSinY = Math.Cos(progressRad + 23) * moveTunnelY;

            double zoomSinDeltaX0 = progressRad + 2;
            double zoomSinDeltaX1 = progressRad - 45;
            double zoomSinDeltaY0 = progressRad;
            double zoomSinDeltaY1 = progressRad - 54;
            
            int iEdgeIndex = 0;
            //iStep = 0;
            for (int iEdge = 0; iEdge < NbEdges * NbSteps; iEdge++)
            {
                double x = ((double)Width) / 2;
                double y = ((double)Height) / 2;

                x += frameSinX;
                y += frameSinY;

                const double zoomTunnelX = 22;
                const double zoomTunnelY = 30;
                var zoomSinX = -Math.Cos(zoomSinDeltaX0) * zoomTunnelX;
                zoomSinX += Math.Sin(zoomSinDeltaX1) * zoomTunnelX;
                var zoomSinY = Math.Sin(zoomSinDeltaY0) * zoomTunnelY;
                zoomSinY += -Math.Cos(zoomSinDeltaY1) * zoomTunnelY;
                zoomSinDeltaX0 -= 0.18 / (double)(NbSteps);
                zoomSinDeltaX1 -= 0.260 / (double)(NbSteps);
                zoomSinDeltaY0 -= 0.381 / (double)(NbSteps);
                zoomSinDeltaY1 -= 0.266 / (double)(NbSteps);

                x += zoomSinX;
                y += zoomSinY;

                double scale = ((double)(iEdge)) / (double)(NbEdges * NbSteps); // scale from 0 to 1

                const double minEdgeX = 0;
                const double minEdgeY = 0;
                double edgeX = minEdgeX + 100 * scale;
                double edgeY = minEdgeY + 100 * scale;

                double x0 = x - edgeX;
                double x1 = x + edgeX;
                double y0 = y - edgeY;
                double y1 = y + edgeY;

                // clamp
                if (x0 < 0) x0 = 0;
                if (x0 > Width) x0 = Width;
                if (x1 < 0) x1 = 0;
                if (x1 > Width) x1 = Width;
                if (y0 < 0) y0 = 0;
                if (y0 > Height) y0 = Height;
                if (y1 < 0) y1 = 0;
                if (y1 > Height) y1 = Height;

                var y0int = (int)y0;
                var y1int = (int)y1;

                if (y0int > y1int)
                {
                    var tempY = y0int;
                    y0int = y1int;
                    y1int = tempY;
                }

                y0int = (y0int >> 1) << 1;
                y1int = (y1int >> 1) << 1;
                
                var lineEdge = new LineEdge
                {
                    intLeftX = (int)x0,
                    intRightX = (int)x1,
                    y0 = y0int,
                    y1 = y1int,
                    iEdge = iEdgeIndex
                };

                /*if(iColor==0)
                {
                    lineEdge.intLeftX = 0;
                    lineEdge.intRightX = Width;
                }*/
                
                iStep++;
                if (iStep == NbSteps)
                {
                    lineEdges.Add(lineEdge);
                    if(lineEdges.Count>NbEdges)
                    {
                        System.Diagnostics.Debugger.Break();
                    }
                    iEdgeIndex++;
                    iStep = 0;
                }
            }

            return lineEdges;
        }

        TextureBrush CreateTextureBrush(int edgeIndex, int vlcIndex, Color pen0Color, Color pen1Color)
        {
            var bitmap = new Bitmap(2, 1);

            switch (vlcIndex)
            {
                case 0:
                    bitmap.SetPixel(0, 0, pen1Color);
                    bitmap.SetPixel(1, 0, pen0Color);
                    break;
                case 1:
                    bitmap.SetPixel(0, 0, pen0Color);
                    bitmap.SetPixel(1, 0, pen1Color);
                    break;
                case 2:
                    bitmap.SetPixel(0, 0, pen1Color);
                    bitmap.SetPixel(1, 0, pen0Color);
                    break;
                case 3:
                    bitmap.SetPixel(0, 0, pen0Color);
                    bitmap.SetPixel(1, 0, pen1Color);
                    break;
            }

            /*if((edgeIndex & 1)==0)
            {
                switch (vlcIndex)
                {
                    case 0:
                        bitmap.SetPixel(0, 0, pen0Color);
                        bitmap.SetPixel(1, 0, pen1Color);
                        break;
                    case 1:
                        bitmap.SetPixel(0, 0, pen1Color);
                        bitmap.SetPixel(1, 0, pen0Color);
                        break;
                    case 2:
                        bitmap.SetPixel(0, 0, pen0Color);
                        bitmap.SetPixel(1, 0, pen1Color);
                        break;
                    case 3:
                        bitmap.SetPixel(0, 0, pen1Color);
                        bitmap.SetPixel(1, 0, pen0Color);
                        break;
                }
            }
            else
            {
                switch (vlcIndex)
                {
                    case 0:
                        bitmap.SetPixel(0, 0, pen1Color);
                        bitmap.SetPixel(1, 0, pen0Color);
                        break;
                    case 1:
                        bitmap.SetPixel(0, 0, pen0Color);
                        bitmap.SetPixel(1, 0, pen1Color);
                        break;
                    case 2:
                        bitmap.SetPixel(0, 0, pen1Color);
                        bitmap.SetPixel(1, 0, pen0Color);
                        break;
                    case 3:
                        bitmap.SetPixel(0, 0, pen0Color);
                        bitmap.SetPixel(1, 0, pen1Color);
                        break;
                }
            }*/

            return new TextureBrush(bitmap);
        }

        Bitmap CreateLineBitmap(int vlcIndex, Frame frame, List<TextureBrush> brushes)
        {
            var bitmap = new Bitmap(Width, 1);

            var g = Graphics.FromImage(bitmap);

            g.FillRectangle(brushes[0], 0, 0, Width, 1);

            for (int iEdge = 0; iEdge < NbEdges; iEdge++)
            {
                var lineEdge = frame.LineEdges[iEdge];

                var brush = brushes[iEdge];

                g.FillRectangle(brush, 0, 0, lineEdge.intLeftX, 1); // left
                g.FillRectangle(brush, lineEdge.intRightX, 0, Width - lineEdge.intRightX, 1); // right
            }

            g.Dispose();

            return bitmap;
        }

        void AddLineIndex(ref List<YLineIndex> YLineIndices, int lineIndex, int y)
        {
            foreach (var iLineIndex in YLineIndices)
            {
                if (iLineIndex.LineIndex == lineIndex)
                {
                    iLineIndex.YPositions.Add(y);
                    return;
                }
            }

            var yLineIndex = new YLineIndex();
            yLineIndex.LineIndex = lineIndex;
            yLineIndex.YPositions.Add(y);
            YLineIndices.Add(yLineIndex);
        }

        void CreateXYFrames(string outFolder, Palette palette, out List<Frame> OutXFrames, out List<Frame> OutYFrames)
        {
            var XFrames = new List<Frame>();
            var YFrames = new List<Frame>();
            int iStep = 0;

            // create brushes
            var magic0vlc0Brushes = new List<TextureBrush>();
            var magic0vlc1Brushes = new List<TextureBrush>();
            var magic1vlc0Brushes = new List<TextureBrush>();
            var magic1vlc1Brushes = new List<TextureBrush>();

            for (int iEdge = 0; iEdge < NbEdges; iEdge++)
            {
                int pen0 = 0;
                int pen1 = 0;

                switch (iEdge)
                {
                    case 6: pen0 = 6; pen1 = 7; break;
                    case 5: pen0 = 6; pen1 = 5; break;
                    case 4: pen0 = 4; pen1 = 5; break;
                    case 3: pen0 = 4; pen1 = 3; break;
                    case 2: pen0 = 2; pen1 = 3; break;
                    case 1: pen0 = 2; pen1 = 1; break;
                    case 0: pen0 = 0; pen1 = 1; break;
                    default:
                        System.Diagnostics.Debugger.Break();
                        break;
                }

                var magic0_pen0Color = palette.Pens[pen0].Color;
                var magic0_pen1Color = palette.Pens[pen1].Color;
                var magic1_pen0Color = palette.Pens[pen0 + 8].Color;
                var magic1_pen1Color = palette.Pens[pen1 + 8].Color;

                magic0vlc0Brushes.Add(CreateTextureBrush(iEdge, 0, magic0_pen0Color, magic0_pen1Color));
                magic0vlc1Brushes.Add(CreateTextureBrush(iEdge, 1, magic0_pen0Color, magic0_pen1Color));
                magic1vlc0Brushes.Add(CreateTextureBrush(iEdge, 2, magic1_pen0Color, magic1_pen1Color));
                magic1vlc1Brushes.Add(CreateTextureBrush(iEdge, 3, magic1_pen0Color, magic1_pen1Color));
            }

            // create lines (X-positions)
            iStep = 0;
            for (int iFrame = 0; iFrame < NbFramesX; iFrame++)
            {
                var frame = new Frame();

                frame.LineEdges = CreateFrameXYPositions(iStep, true, iFrame);
                iStep++;
                if (iStep == NbSteps)
                {
                    iStep = 0;
                }

                var magic0vlc0Bitmap = CreateLineBitmap(0, frame, magic0vlc0Brushes);
                var magic0vlc1Bitmap = CreateLineBitmap(1, frame, magic0vlc1Brushes);
                var magic1vlc0Bitmap = CreateLineBitmap(2, frame, magic1vlc0Brushes);
                var magic1vlc1Bitmap = CreateLineBitmap(3, frame, magic1vlc1Brushes);

                magic0vlc0Bitmap.Save(outFolder + "Frames\\magic0vlc0_" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                frame.Magic0Vlc0Bitmap = magic0vlc0Bitmap;
                frame.Magic0Vlc1Bitmap = magic0vlc1Bitmap;
                frame.Magic1Vlc0Bitmap = magic1vlc0Bitmap;
                frame.Magic1Vlc1Bitmap = magic1vlc1Bitmap;

                frame.Magic0Vlc0CPCImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = magic0vlc0Bitmap,
                    Mode = 0,
                    R1 = 48,
                    R4 = 21,
                    R9 = 0,
                    Palette = palette
                });
                frame.Magic0Vlc1CPCImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = magic0vlc1Bitmap,
                    Mode = 0,
                    R1 = 48,
                    R4 = 21,
                    R9 = 0,
                    Palette = palette
                });
                frame.Magic1Vlc0CPCImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = magic1vlc0Bitmap,
                    Mode = 0,
                    R1 = 48,
                    R4 = 21,
                    R9 = 0,
                    Palette = palette
                });
                frame.Magic1Vlc1CPCImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = magic1vlc1Bitmap,
                    Mode = 0,
                    R1 = 48,
                    R4 = 21,
                    R9 = 0,
                    Palette = palette
                });

                for (int x = 0; x < frame.Magic0Vlc0CPCImage.CPCRawBuffer_Width; x++)
                {
                    byte basePixel = frame.Magic0Vlc0CPCImage.CPCRawBuffer[x];

                    if(Magic0VLC1_PixelConvertedFromMagic0VLC0[basePixel]==0)
                    {
                        Magic0VLC1_PixelConvertedFromMagic0VLC0[basePixel] = frame.Magic0Vlc1CPCImage.CPCRawBuffer[x];
                    }
                    else
                    {
                        var oldValue = Magic0VLC1_PixelConvertedFromMagic0VLC0[basePixel];
                        var newValue = frame.Magic0Vlc1CPCImage.CPCRawBuffer[x];
                        if (oldValue!= newValue)
                        {
                            //System.Diagnostics.Debugger.Break();
                        }
                        Magic0VLC1_PixelConvertedFromMagic0VLC0[basePixel] = newValue;
                    }
                    Magic1VLC0_PixelConvertedFromMagic0VLC0[basePixel] = frame.Magic1Vlc0CPCImage.CPCRawBuffer[x];
                    Magic1VLC1_PixelConvertedFromMagic0VLC0[basePixel] = frame.Magic1Vlc1CPCImage.CPCRawBuffer[x];
                }

                XFrames.Add(frame);
            }

            // create Y positions
            iStep = 0;
            for (int iFrame = 0; iFrame < NbFramesY; iFrame++)
            {
                var frame = new Frame();

                frame.LineEdges = CreateFrameXYPositions(iStep, false, iFrame);
                iStep++;
                if (iStep == NbSteps)
                {
                    iStep = 0;
                }

                // create array of 268/2 values containing line index
                var verticalEdgeIndices = new int[Height];
                for (int i = 0; i < Height; i++)
                {
                    verticalEdgeIndices[i] = NbEdges - 1;
                }
                for (int iEdge = NbEdges - 1; iEdge >= 0; iEdge--)
                {
                    var lineEdge = frame.LineEdges[iEdge];
                    for (var y = lineEdge.y0; y < lineEdge.y1; y++)
                    {
                        verticalEdgeIndices[y] = iEdge;
                    }
                }

                // convert line indices change to bin format: 1 byte=yPos + 1 byte=line index
                var YLineIndices = new List<YLineIndex>();

                // 1st pass: set only line changes taking care of magic0/1
                var firstLineIndexMagic0 = verticalEdgeIndices[0];
                var firstLineIndexMagic1 = verticalEdgeIndices[1];

                // force lines 0/1 to set initial color values
                AddLineIndex(ref YLineIndices, firstLineIndexMagic1 + NbEdges /*+8*/, 0);
                AddLineIndex(ref YLineIndices, firstLineIndexMagic0, 1);

                var lineIndexMagic0 = firstLineIndexMagic0;
                var lineIndexMagic1 = firstLineIndexMagic1;

                for (int y = 0; y < Height; y++)
                {
                    bool isMagic0 = true;
                    if ((y & 1) == 0)
                    {
                        isMagic0 = false;
                    }

                    var iLineIndex = verticalEdgeIndices[y];
                    if (isMagic0)
                    {
                        if (iLineIndex != lineIndexMagic0)
                        {
                            AddLineIndex(ref YLineIndices, iLineIndex, y);
                            lineIndexMagic0 = iLineIndex;
                        }
                    }
                    else
                    {
                        if (iLineIndex != lineIndexMagic1)
                        {
                            AddLineIndex(ref YLineIndices, iLineIndex + NbEdges /*+8*/, y);
                            lineIndexMagic1 = iLineIndex;
                        }
                    }
                }

                // 2nd pass: write CPC format 1 byte=yPos + 1 byte=line index
                var frameLineIndices = new byte[NbEdges * 2 * 2]; // y positions for edges, can be used 2 times a frame and take care of magic0/1 (fixed size per frame for CPU-constant treatment)
                for (int i = 0; i < NbEdges * 2 * 2; i++)
                {
                    frameLineIndices[i] = Vert_DummyOffsetInStackJPTable;
                }
                foreach (var yLineIndex in YLineIndices)
                {
                    if (yLineIndex.YPositions.Count > 2)
                    {
                        System.Diagnostics.Debugger.Break();
                    }

                    var lineIndex = yLineIndex.LineIndex;

                    frameLineIndices[(lineIndex * 2)] = (byte)yLineIndex.YPositions[0];
                    if (yLineIndex.YPositions.Count > 1)
                    {
                        frameLineIndices[(lineIndex * 2) + 1] = (byte)yLineIndex.YPositions[1];
                    }
                }
                frame.LineIndices = new List<byte>(frameLineIndices);

                YFrames.Add(frame);
            }

            OutXFrames = XFrames;
            OutYFrames = YFrames;
        }

        int WriteCPCAnimation(ref List<Frame> XFrames, ref List<Frame> YFrames, string outFolder, out int outMaxBytesChangedPerFrame)
        {
            int firstFrameLength = 0;

            // write line indices
            var z80 = ";\n";
            z80 += "Vert_DummyOffsetInStackJPTable equ &" + Vert_DummyOffsetInStackJPTable.ToString("X2") + " ;; (268/2)+2\n";
            z80 += "\n";

            z80 += "TunnelYPosFrames:\n\n;";
            for (int iFrame = 0; iFrame < NbFramesY; iFrame++)
            {
                z80 += "\nTunnelYPos_Frame" + iFrame + ":";

                var frame = YFrames[iFrame];
                for (int lineIndex = 0; lineIndex < frame.LineIndices.Count; lineIndex += 2)
                {
                    //z80 += "\n\tdw Tunnel_JumpTable + &" + (frame.LineIndices[lineIndex] * 2).ToString("X4");
                    z80 += "\n\tdb &" + (frame.LineIndices[lineIndex]).ToString("X2") + ", &" + (frame.LineIndices[lineIndex + 1]).ToString("X2");
                }

                z80 += "\n";
            }
            File.WriteAllText(outFolder + "_TunnelYPosFrames.z80", z80);

            // write 1st frame (VLC CPC data)
            File.WriteAllBytes(outFolder + "_TunnelFirstFrame_magic0vlc0.bin", XFrames[0].Magic0Vlc0CPCImage.CPCRawBuffer);
            File.WriteAllBytes(outFolder + "_TunnelFirstFrame_magic0vlc1.bin", XFrames[0].Magic0Vlc1CPCImage.CPCRawBuffer);
            File.WriteAllBytes(outFolder + "_TunnelFirstFrame_magic1vlc0.bin", XFrames[0].Magic1Vlc0CPCImage.CPCRawBuffer);
            File.WriteAllBytes(outFolder + "_TunnelFirstFrame_magic1vlc1.bin", XFrames[0].Magic1Vlc1CPCImage.CPCRawBuffer);

            // write diff. Format: CPCScreenPtr_XPos/CPC Value
            var linePosValueStream = new List<byte>();

            outMaxBytesChangedPerFrame = 0;

            for (int iFrame = 0; iFrame < NbFramesX; iFrame++)
            {
                // get current / last frame
                var currentFrame = XFrames[iFrame];
                var iLastFrame = iFrame - 1;
                if (iLastFrame < 0)
                {
                    iLastFrame += NbFramesX;
                }
                var lastFrame = XFrames[iLastFrame];

                var currentCpcImage = currentFrame.Magic0Vlc0CPCImage;
                var lastCpcImage = lastFrame.Magic0Vlc0CPCImage;

                var linePosValueBin = new List<byte>();

                for (var x = 0; x < currentCpcImage.CPCRawBuffer_Width; x++)
                {
                    var currentByte = currentCpcImage.CPCRawBuffer[x];
                    var lastByte = lastCpcImage.CPCRawBuffer[x];

                    if (currentByte != lastByte)
                    {
                        linePosValueBin.Add((byte)x); // CPCScreenPtr offset
                        linePosValueBin.Add((byte)currentByte); // CPCScreen value
                    }
                }

                linePosValueStream.Add((byte)(linePosValueBin.Count / 2)); // line diff count
                linePosValueStream.AddRange(linePosValueBin.ToArray()); // add all CPCScreenPtr offset+values

                if (outMaxBytesChangedPerFrame < linePosValueBin.Count / 2)
                {
                    outMaxBytesChangedPerFrame = linePosValueBin.Count / 2;
                }

                Console.WriteLine("Frame " + iFrame + " : " + (linePosValueBin.Count / 2) + " bytes to change");
                if (linePosValueBin.Count > 255)
                {
                    System.Diagnostics.Debugger.Break();
                }

                if(iFrame==0)
                {
                    firstFrameLength = linePosValueBin.Count + 1;
                }
            }

            Console.WriteLine("Max Bytes changed per frame: " + outMaxBytesChangedPerFrame);

            File.WriteAllBytes(outFolder + "_Tunnel_XPosValues.bin", linePosValueStream.ToArray());

            File.WriteAllBytes(outFolder + "_Magic0VLC1_PixelConvertedFromMagic0VLC0.bin", Magic0VLC1_PixelConvertedFromMagic0VLC0);
            File.WriteAllBytes(outFolder + "_Magic1VLC0_PixelConvertedFromMagic0VLC0.bin", Magic1VLC0_PixelConvertedFromMagic0VLC0);
            File.WriteAllBytes(outFolder + "_Magic1VLC1_PixelConvertedFromMagic0VLC0.bin", Magic1VLC1_PixelConvertedFromMagic0VLC0);

            return firstFrameLength;
        }

        void WriteZ80MagicCases(string outFolder)
        {
            /* 
	            ld b, &7F
	            exx
	            ld b, COLOR1
	            ld c, COLOR2
	            ld d, COLOR3
	            ld e, COLOR4
	            ld h, COLOR5
	            ld l, COLOR6
	            exx	
	            ld a, COLOR7
             */

            /*
             6,7|6,5|4,5|4,3|2,3|2,1|0,1 = 7 edges using 8 colors
             */

            for (int iMagic = 0; iMagic < 2; iMagic++)
            {
                for (int iEdge = 0; iEdge < NbEdges; iEdge++)
                {
                    var z80 = "";

                    z80 += ";\n";

                    z80 += "Tunnel_Magic" + iMagic + "_Case" + iEdge + ":\n";

                    z80 += "\t_NOBRK\n";
                    z80 += "\n";

                    for (int iColor = 0; iColor < 8; iColor++)
                    {
                        if (iColor == 0)
                        {
                            z80 += "\tld c, " + ((iMagic * 8) + iColor) + "\n";
                            z80 += "\tout (c), c ;; Select PEN" + ((iMagic * 8) + iColor) + "\n";
                        }
                        else
                        {
                            z80 += "\tinc c\n";
                            z80 += "\tout (c), c ;; Select PEN" + ((iMagic * 8) + iColor) + "\n";
                        }

                        var color = iColor;
                        switch (iEdge)
                        {
                            case 6: // 6,7
                                switch (iColor)
                                {
                                    case 5: color = 7; break;
                                    case 4: color = 6; break;
                                    case 3: color = 7; break;
                                    case 2: color = 6; break;
                                    case 1: color = 7; break;
                                    case 0: color = 6; break;
                                }; break;

                            case 5: // 6,5
                                switch (iColor)
                                {
                                    case 4: color = 6; break;
                                    case 3: color = 5; break;
                                    case 2: color = 6; break;
                                    case 1: color = 5; break;
                                    case 0: color = 6; break;
                                }; break;

                            case 4: // 4,5
                                switch (iColor)
                                {
                                    case 3: color = 5; break;
                                    case 2: color = 4; break;
                                    case 1: color = 5; break;
                                    case 0: color = 4; break;
                                }; break;

                            case 3: // 4,3
                                switch (iColor)
                                {
                                    case 2: color = 4; break;
                                    case 1: color = 3; break;
                                    case 0: color = 4; break;
                                }; break;

                            case 2: // 2,3
                                switch (iColor)
                                {
                                    case 1: color = 3; break;
                                    case 0: color = 2; break;
                                }; break;

                            case 1: // 2,1
                                switch (iColor)
                                {
                                    case 0: color = 2; break;
                                }; break;

                            case 0: // 0,1
                                break;
                        }

                        if (color == 0)
                        {
                            z80 += "\tld a, IYL ;; COLOR0\n";
                            z80 += "\t_NOPS 1\n";
                            z80 += "\tout (c), a\n";
                        }
                        else if (color == 1)
                        {
                            z80 += "\tld a, IYH ;; COLOR1\n";
                            z80 += "\t_NOPS 1\n";
                            z80 += "\tout (c), a\n";
                        }
                        else
                        {
                            z80 += "\texx\n";
                            switch (color)
                            {
                                case 0: System.Diagnostics.Debugger.Break(); break;
                                case 1: System.Diagnostics.Debugger.Break(); break;
                                case 2: z80 += "\tld a, B ;; COLOR2\n"; break;
                                case 3: z80 += "\tld a, C ;; COLOR3\n"; break;
                                case 4: z80 += "\tld a, D ;; COLOR4\n"; break;
                                case 5: z80 += "\tld a, E ;; COLOR5\n"; break;
                                case 6: z80 += "\tld a, H ;; COLOR6\n"; break;
                                case 7: z80 += "\tld a, L ;; COLOR7\n"; break;
                                default:
                                    break;
                            }
                            z80 += "\texx\n";
                            z80 += "\tout (c), a\n";
                        }

                        z80 += "\n";
                    }

                    z80 += "\tjp (ix)\n";

                    z80 += "\n";

                    File.WriteAllText(outFolder + "_Magic" + iMagic + "_Case" + iEdge + ".z80", z80);
                }
            }
        }
    }
}
