﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Diagnostics;

namespace phX.Disc
{
    public class CreateDisc
    {
        public struct MemZone
        {
            public byte Bank;
            public ushort Start;
            public ushort End;
            public byte[] Data;
        }   

        public class BigFile
        {
            public byte Bank;
            public byte[] Data;

            public BigFile(byte bank)
            {
                Bank = bank;
                Data = new byte[65536];
                for(int i=0; i<65536; i++)
                {
                    Data[i] = 0xCC;
                }
            }
        }

        public CreateDisc()
        {
            try
            {
                var memZones = CreateMemZones(Program.ToMemZonesFolder("Part1\\"));
                foreach(var memZone in memZones)
                {
                    var memZoneMessage = String.Format("Bank: &{0:X} Start:&{1:X} End:&{2:X}", memZone.Bank, memZone.Start, memZone.End);
                }
                var bigFiles = CreateBigFiles(memZones);
                WriteBigFiles_Part1(bigFiles, Program.ToMemZonesFolder("\\"));

                memZones = CreateMemZones(Program.ToMemZonesFolder("Part2\\"));
                foreach (var memZone in memZones)
                {
                    var memZoneMessage = String.Format("Bank: &{0:X} Start:&{1:X} End:&{2:X}", memZone.Bank, memZone.Start, memZone.End);
                }
                bigFiles = CreateBigFiles(memZones);
                WriteBigFiles_Part2(bigFiles, Program.ToMemZonesFolder("\\"));

                memZones = CreateMemZones(Program.ToMemZonesFolder("Part3\\"));
                foreach (var memZone in memZones)
                {
                    var memZoneMessage = String.Format("Bank: &{0:X} Start:&{1:X} End:&{2:X}", memZone.Bank, memZone.Start, memZone.End);
                }
                bigFiles = CreateBigFiles(memZones);
                WriteBigFiles_Part3(bigFiles, Program.ToMemZonesFolder("\\"));

                memZones = CreateMemZones(Program.ToMemZonesFolder("Part4\\"));
                foreach (var memZone in memZones)
                {
                    var memZoneMessage = String.Format("Bank: &{0:X} Start:&{1:X} End:&{2:X}", memZone.Bank, memZone.Start, memZone.End);
                }
                bigFiles = CreateBigFiles(memZones);
                WriteBigFiles_Part4(bigFiles, Program.ToMemZonesFolder("\\"));

                CreateDSK(Program.GetBinDirectory()+"phX.dsk");
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public static bool StartProcess(string workingDirectory, string filename, string arguments)
        {
            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.CreateNoWindow = true;
            startInfo.UseShellExecute = false;
            startInfo.WorkingDirectory = workingDirectory;
            startInfo.FileName = filename;
            startInfo.Arguments = arguments;
            try
            {
                using (Process exeProcess = Process.Start(startInfo))
                {
                    exeProcess.WaitForExit();
                }
            }
            catch (Exception e)
            {
                throw e;
            }

            return true;
        }

        public static bool StartProcessNoWait(string workingDirectory, string filename, string arguments)
        {
            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.CreateNoWindow = true;
            startInfo.UseShellExecute = false;
            startInfo.WorkingDirectory = workingDirectory;
            startInfo.FileName = filename;
            startInfo.Arguments = arguments;
            try
            {
                Process.Start(startInfo);
            }
            catch (Exception e)
            {
                throw e;
            }

            return true;
        }

        public static bool ExecHFEConvert(string dskFilename, string hfeFilename)
        {
            string relativePath = "Emulation/HxCFloppyEmulator/Windows/";
            string toolName = "flopimage_convert";
            string workingDirectory = Program.GetRootDirectory() + "VS2015\\phx\\bin\\Debug\\";

            var tempFilename = workingDirectory + "image.dsk";
            File.Copy(dskFilename, tempFilename, true);

            var arguments = ". . -HFE";

            string binFilename = String.Empty;

            binFilename = Program.GetRootDirectory() + relativePath + toolName + ".exe";

            var result = StartProcess(workingDirectory, binFilename, arguments);
            if (!result)
            {
                return false;
            }

            var genHfeFilename = tempFilename.Replace(".dsk", "_dsk.hfe");
            File.Copy(genHfeFilename, dskFilename.Replace(".dsk", ".hfe"), true);
            File.Delete(tempFilename);
            File.Delete(genHfeFilename);

            return true;
        }

        private void ExecThirdParty(string execFilename, string arguments)
        {
            var thirdPartyFolder = Program.GetRootDirectory() + "VS2015\\phx\\bin\\Debug\\ThirdParty\\";
            var workingDirectory = thirdPartyFolder;
            StartProcess(workingDirectory, workingDirectory + execFilename, arguments);
        }

        public void CreateDSK(string dskFilename)
        {
            // kill WinAPE process
            var ps = Process.GetProcessesByName("WinAPE");
            foreach (Process p in ps)
            {
                p.Kill();
            }

            string workingDirectory = Program.GetRootDirectory() + "VS2015\\phx\\bin\\Debug\\ThirdParty\\";

            // Create DSK
            var arguments = "-f \"myd\" " + dskFilename;
            ExecThirdParty("discsys", arguments);

            var dskFilenameInfo = new FileInfo(dskFilename);

            var fat = new byte[256];
            var fatFilename = Program.GetRootDirectory() + "fat.bin";
            if (File.Exists(fatFilename))
            {
                File.Delete(fatFilename);
            }
            File.WriteAllBytes(fatFilename, fat);

            var fileEntries = new List<string>();
            fileEntries.Add(fatFilename);
            fileEntries.Add(Program.ToZ80Folder("Launcher\\_RatedM.png.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part1_BigFiles\\P1_C0_C000_FFFF.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part1_BigFiles\\P1_C0_0000_8000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part1_BigFiles\\P1_C2_0000_8000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part1_BigFiles\\P1_C2_8000_FA00.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part1_BigFiles\\P1_C0_8000_C000.bin"));

            fileEntries.Add(Program.ToMemZonesFolder("Part2_BigFiles\\P2_C2_8000_BD00.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part2_BigFiles\\P2_C0_0000_4000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part2_BigFiles\\P2_C0_4000_8000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part2_BigFiles\\P2_C2_4000_8000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part2_BigFiles\\P2_C2_C000_F800.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part2_BigFiles\\P2_C2_0100_4000.bin"));

            fileEntries.Add(Program.ToMemZonesFolder("Part3_BigFiles\\P3_C2_0000_4000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part3_BigFiles\\P3_C2_C000_F600.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part3_BigFiles\\P3_C2_4000_C000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part3_BigFiles\\P3_C0_0000_8000.bin"));

            fileEntries.Add(Program.ToMemZonesFolder("Part4_BigFiles\\P4_C2_0000_4000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part4_BigFiles\\P4_C2_4000_C000.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part4_BigFiles\\P4_C2_C000_F600.bin"));
            fileEntries.Add(Program.ToMemZonesFolder("Part4_BigFiles\\P4_C0_0000_8000.bin"));

            foreach (var fileEntry in fileEntries)
            {
                if(!File.Exists(fileEntry))
                {
                    // file does not exist!
                    Debugger.Break();
                }
            }

            var discsysContent = String.Empty;
            int itemCount = 1;
            foreach (var filename in fileEntries)
            {
                discsysContent += itemCount + " \"" + filename + "\"\n";
                itemCount++;
            }

            var discsysContentFilename = workingDirectory + "discsys.filelist";
            File.WriteAllText(discsysContentFilename, discsysContent);

            arguments = "-l \"" + discsysContentFilename + "\" \"" + dskFilename + "\"";
            ExecThirdParty("discsys", arguments);

            var dskAllBytes = File.ReadAllBytes(dskFilename);
            var offset = 0x3200 + 4;
            var z80Content = "";

            var fatOffset = 0;

            itemCount = 0;
            foreach (var filename in fileEntries)
            {
                byte track = dskAllBytes[offset];
                byte sector = dskAllBytes[offset + 1];
                ushort fileSize = (ushort)(dskAllBytes[offset + 2] + (dskAllBytes[offset + 3] << 8));
                offset += 4;

                if (itemCount != 0)
                {
                    fat[fatOffset++] = (byte)(fileSize & 255);
                    fat[fatOffset++] = (byte)(fileSize >> 8);
                    fat[fatOffset++] = sector;
                    fat[fatOffset++] = track;
                }

                var fileInfo = new FileInfo(filename);
                var id = fileInfo.Name.ToUpper().Replace(".", "_");

                if (itemCount == 0)
                {
                    z80Content += "\t" + id + "_FILESIZE equ &" + ((ushort)fileSize).ToString("X4") + "\n";
                    z80Content += "\t" + id + "_TRACK_SECTOR equ &" + ((ushort)((track << 8) + sector)).ToString("X4") + "\n\n";
                }
                else
                {
                    z80Content += "\t" + id + "_ID equ " + (itemCount - 1) + " ; TRACK &" + track.ToString("X2") + " --- SECTOR &" + sector.ToString("X2") + "\n";
                }

                itemCount++;
            }

            ExecThirdParty("discsys", "-f \"myd\" " + dskFilename);

            File.WriteAllBytes(fatFilename, fat);
            ExecThirdParty("discsys", arguments);
            File.Delete(fatFilename);

            var z80ContentFilename = Program.ToZ80Folder("_FileId.z80");
            File.WriteAllText(z80ContentFilename, z80Content);

            // Inject BASIC launcher
            var z80BinBasicFilename = Program.ToZ80Folder("Launcher/_Launcher.bas");
            var compiledBinBasicFilename = "\"" + "PHX".ToUpper() + "\"";
            arguments = "-a -t \"basic\" -s 368 -x 368 " + z80BinBasicFilename + " " + compiledBinBasicFilename;
            ExecThirdParty("addhead", arguments);
            arguments = "-o MLOD " + dskFilename + " -f -b -p " + compiledBinBasicFilename;
            ExecThirdParty("cpcxfs", arguments);
            //File.Delete(workingDirectory + Build.DemoName);

            ExecHFEConvert(dskFilenameInfo.FullName, dskFilenameInfo.FullName.Replace(".dsk", ".hfe"));

            // Launch WinAPE
            string winApeAbsolutePath = Program.GetRootDirectory() + "Emulation/WinAPE_AmsdosLauncher/WinAPE.exe";
            string winApeArguments = "\"" + dskFilenameInfo.FullName + "\" /A";
            StartProcessNoWait(Program.GetRootDirectory(), winApeAbsolutePath, winApeArguments);
        }

        public void WriteBigFiles_Part1(List<BigFile> bigFiles, string destFolder)
        {
            var bigFileC0 = bigFiles[0];
            var bigFileC2 = bigFiles[1];

            File.WriteAllBytes(destFolder + "Part1_BigFiles\\P1_C0_0000_8000.bin", bigFileC0.Data.Skip(0x0000).Take(0x8000).ToArray());
            File.WriteAllBytes(destFolder + "Part1_BigFiles\\P1_C0_8000_C000.bin", bigFileC0.Data.Skip(0x8000).Take(0x4000).ToArray());
            File.WriteAllBytes(destFolder + "Part1_BigFiles\\P1_C0_C000_FFFF.bin", bigFileC0.Data.Skip(0xC000).Take(0x4000).ToArray());

            File.WriteAllBytes(destFolder + "Part1_BigFiles\\P1_C2_0000_8000.bin", bigFileC2.Data.Skip(0x0000).Take(0x8000).ToArray());
            File.WriteAllBytes(destFolder + "Part1_BigFiles\\P1_C2_8000_FA00.bin", bigFileC2.Data.Skip(0x8000).Take(0x7A00).ToArray());
        }

        private byte[] CreateFakeData(byte fillData, int length)
        {
            var data = new byte[length];

            for(int i=0; i<length; i++)
            {
                data[i] = fillData;
            }

            return data;
        }

        public void WriteBigFiles_Part2(List<BigFile> bigFiles, string destFolder)
        {
            var bigFileC0 = bigFiles[0];
            var bigFileC2 = bigFiles[1];

            // music+player first (located at &8800)
            File.WriteAllBytes(destFolder + "Part2_BigFiles\\P2_C2_8000_BD00.bin", bigFileC2.Data.Skip(0x8000).Take(0x3D00).ToArray());

            File.WriteAllBytes(destFolder + "Part2_BigFiles\\P2_C2_4000_8000.bin", bigFileC2.Data.Skip(0x4000).Take(0x4000).ToArray());

            File.WriteAllBytes(destFolder + "Part2_BigFiles\\P2_C0_0000_4000.bin", bigFileC0.Data.Skip(0x0000).Take(0x4000).ToArray());
            File.WriteAllBytes(destFolder + "Part2_BigFiles\\P2_C0_4000_8000.bin", bigFileC0.Data.Skip(0x4000).Take(0x4000).ToArray());

            File.WriteAllBytes(destFolder + "Part2_BigFiles\\P2_C2_C000_F800.bin", bigFileC2.Data.Skip(0xC000).Take(0x3800).ToArray());

            File.WriteAllBytes(destFolder + "Part2_BigFiles\\P2_C2_0100_4000.bin", bigFileC2.Data.Skip(0x0100).Take(0x3F00).ToArray());
        }

        public void WriteBigFiles_Part3(List<BigFile> bigFiles, string destFolder)
        {
            var bigFileC0 = bigFiles[0];
            var bigFileC2 = bigFiles[1];

            File.WriteAllBytes(destFolder + "Part3_BigFiles\\P3_C2_0000_4000.bin", bigFileC2.Data.Skip(0x0000).Take(0x4000).ToArray());
            File.WriteAllBytes(destFolder + "Part3_BigFiles\\P3_C2_4000_C000.bin", bigFileC2.Data.Skip(0x4000).Take(0x8000).ToArray());
            File.WriteAllBytes(destFolder + "Part3_BigFiles\\P3_C2_C000_F600.bin", bigFileC2.Data.Skip(0xC000).Take(0x3600).ToArray());

            File.WriteAllBytes(destFolder + "Part3_BigFiles\\P3_C0_0000_8000.bin", bigFileC0.Data.Skip(0x0000).Take(0x8000).ToArray());
        }

        public void WriteBigFiles_Part4(List<BigFile> bigFiles, string destFolder)
        {
            var bigFileC0 = bigFiles[0];
            var bigFileC2 = bigFiles[1];

            File.WriteAllBytes(destFolder + "Part4_BigFiles\\P4_C2_0000_4000.bin", bigFileC2.Data.Skip(0x0000).Take(0x4000).ToArray());
            File.WriteAllBytes(destFolder + "Part4_BigFiles\\P4_C2_C000_F600.bin", bigFileC2.Data.Skip(0xC000).Take(0x3600).ToArray());
            File.WriteAllBytes(destFolder + "Part4_BigFiles\\P4_C2_4000_C000.bin", bigFileC2.Data.Skip(0x4000).Take(0x8000).ToArray());
            
            File.WriteAllBytes(destFolder + "Part4_BigFiles\\P4_C0_0000_8000.bin", bigFileC0.Data.Skip(0x0000).Take(0x8000).ToArray());
        }

        public List<BigFile> CreateBigFiles(List<MemZone> memZones)
        {
            var bigFiles = new List<BigFile>();

            var bigFileC0 = new BigFile(0xC0);
            bigFiles.Add(bigFileC0);
            var bigFileC2 = new BigFile(0xC2);
            bigFiles.Add(bigFileC2);

            foreach(var memZone in memZones)
            {
                BigFile bigFile;
                if(memZone.Bank==0xC0)
                {
                    bigFile = bigFileC0;
                }
                else
                {
                    bigFile = bigFileC2;
                }

                for(int i=0; i<memZone.Data.Length; i++)
                {
                    bigFile.Data[i+memZone.Start] = memZone.Data[i];
                }
            }

            return bigFiles;
        }

        public List<MemZone> CreateMemZones(string folderName)
        {
            var memZones = new List<MemZone>();

            var files = Directory.GetFiles(folderName);
            foreach (var file in files)
            {
                var fileInfo = new FileInfo(file);
                var fullname = fileInfo.Name;

                if(!fullname.Contains("&"))
                {
                    // not a mem zone file
                    continue; 
                }

                fullname = fullname.Replace("&", "");
                fullname = fullname.Replace(".bin", "");

                var splits = fullname.Split('_');

                var memZone = new MemZone();
                memZone.Bank = (byte)Convert.ToUInt16(splits[0], 16);
                memZone.Start = (ushort)Convert.ToUInt16(splits[1], 16);
                memZone.End = (ushort)Convert.ToUInt16(splits[2], 16);
                memZone.Data = File.ReadAllBytes(fileInfo.FullName);

                if((memZone.Bank==0xC3)&&(memZone.Start>=0x4000))
                {
                    memZone.Bank = 0xC0;
                    memZone.Start += 0x8000;
                    memZone.End += 0x8000;
                    if(memZone.End==0x0000)
                    {
                        memZone.End = 0xFFFF;
                    }
                }

                if ((memZone.Bank == 0xC1) && (memZone.Start >= 0x4000))
                {
                    memZone.Bank = 0xC0;
                }

                memZones.Add(memZone);
            }

            return memZones;
        }
    }
}
