using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace phX.Framework
{
	public class Image
	{
		public Palette Palette;
		
		public byte[] R12R13;
        public ushort[] ScreenPtr; // corresponds to VLC address conveted to CPC Screen addresses

        public class VLC
		{
			public int Index = 0;
			public byte[] CPCBytes = new byte[0x800];

			public VLC(int index)
			{
				Index = index;
			}
		}
		public class Page
		{
			public VLC[] VLCs;
			
			public int NbBytesWritten = 0;

			public Page(int R9)
			{
				VLCs = new VLC[R9 + 1];
				for(int VLC=0; VLC<=R9; VLC++)
				{
					VLCs[VLC] = new VLC(VLC);
				}
			}
		}
		public Page[] Pages;
		
		//--
		public class CreateImageStruct
		{
			public Bitmap Bitmap;
			public int Mode = 0;
			public int R1 = 0;
			public int R4 = 0;
			public int R9 = 0;
			public int StartR12 = 0; // 0x00, 0x40, 0x80, 0xC0 - Used to create R12R13 (line splitting)
			public int StartVLC = 0;
            public int ApplyOffsetAtNextPage = 0; // for fullscreen images when R1=54, cf. clown
            public int FirstPixelInPaletteX = 0;
            public int FirstPixelInPaletteY = 0;
			public Palette Palette;			
		}

        public byte[] CPCRawBuffer;
        public int CPCRawBuffer_Width;
        public int CPCRawBuffer_Height;

        public static Image CreateImage(CreateImageStruct s)
		{
			var image = new Image();			
			
            if (s.Palette == null)
            {
                s.Palette = new Palette();
            }

            int VCC = 0;
            int VLC = s.StartVLC;

            var pages = new List<Image.Page>();
            var page = new Image.Page(s.R9);
            pages.Add(page);

            int cpcBytesOffsetInVLC = 0;
            
			var R12R13 = new List<byte>();
            var ScreenPtr = new List<ushort>();

            var cpcRawBuffer = new List<byte>();
            image.CPCRawBuffer_Height = s.Bitmap.Height;

            Color p0 = s.Bitmap.GetPixel(s.FirstPixelInPaletteX, s.FirstPixelInPaletteY);
            s.Palette.AddColor(p0);

            for (int y = 0; y < s.Bitmap.Height; y++)
            {
                var line = Pixel.ToPixels(s.Bitmap, s.Mode, y, ref s.Palette);
                
                cpcRawBuffer.AddRange(line);
                image.CPCRawBuffer_Width = (int)line.Length;

                if ((cpcBytesOffsetInVLC+line.Length)>page.VLCs[VLC].CPCBytes.Length)
                {
                    VLC = 0;
                    VCC = 0;

                    page = new Image.Page(s.R9);
                    pages.Add(page);

                    cpcBytesOffsetInVLC = 0;

                    if ((pages.Count == 2)|| (pages.Count == 4))
                    {
                        cpcBytesOffsetInVLC = s.ApplyOffsetAtNextPage;
                    }
                }

                line.CopyTo(page.VLCs[VLC].CPCBytes, cpcBytesOffsetInVLC);
				
				page.NbBytesWritten += line.Length;
				
				byte R12 = (byte)((((byte)s.StartR12 / 0x40) * 0x10) + 0x0C);
                R12 += (byte)((pages.Count-1) * 0x10);
                //ushort cpcOffset = (ushort)(VCC * line.Length);
                ushort cpcOffset = (ushort)(VCC * s.R1*2);
                byte finalR12 = (byte)(R12 + ((cpcOffset / 2) >> 8));
				byte finalR13 = (byte)((cpcOffset / 2) & 255);
				R12R13.Add(finalR12);
				R12R13.Add(finalR13);

                ushort screenPtr = (ushort)(((ushort)s.StartR12)<<8);
                screenPtr = (ushort)(screenPtr + (pages.Count - 1) * 0x4000);
                screenPtr = (ushort)(screenPtr+cpcOffset);
                ScreenPtr.Add(screenPtr);

                //Console.WriteLine("R12=&" + finalR12.ToString("X2") + " R13=" + finalR13.ToString("X2"));

                VLC++;
                if(VLC==(s.R9+1))
                {
                    VLC = 0;
                    VCC++;

                    cpcBytesOffsetInVLC += s.R1*2;

                    if (VCC==(s.R4+1))
                    {
                        VCC = 0;

                        page = new Image.Page(s.R9);
                        pages.Add(page);

                        cpcBytesOffsetInVLC = 0;
                    }
                }
            }

			if(pages[pages.Count-1].NbBytesWritten==0)
			{
				pages.RemoveAt(pages.Count-1);
			}

            image.Pages = pages.ToArray();
			image.R12R13 = R12R13.ToArray();
            image.ScreenPtr = ScreenPtr.ToArray();
            image.Palette = s.Palette;
            image.CPCRawBuffer = cpcRawBuffer.ToArray();

            return image;
        }		
	}
}
