﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;

namespace phX.Framework
{
    public class Music
    {
        public byte[] Convert(string filename)
        {
            //var inBytes = File.ReadAllBytes(filename);
            var inTxt = File.ReadAllText(filename);

            var dump = DumpMusic(ref inTxt);
            //var cpcDump = DumpToCPCFrequency(ref dump);
            var frames = CreateFrames(ref dump);
            var outBytes = CreateBinary(ref frames);

            return outBytes;
        }

        const float FreqCPC = 1000000.0f;
        const float FreqSpeccy = 1773400.0f;
        const float FreqST = 2000000.0f;

        static float YMFreq = ((float)FreqCPC) / 1000000.0f;

        private class Frame
        {
            public byte UsedReg8To13 = 0; // if 0xFF, then repeat in UsedReg0To7
            public byte UsedReg0To7 = 0;
            public List<byte> RegValues = new List<byte>();
            public int Count = 1;
            public bool IsRegistersInBuffer256 = false;
        }

        private byte[] DumpMusic(ref string inTxt)
        {
            var outBytes = new List<byte>();

            var allLines = inTxt.Split('\n');
            foreach(var line in allLines)
            {
                if(line=="")
                {
                    continue;
                }
                int iReg = 0;

                var iLine = line.Replace(" \r", "");

                var allValues = iLine.Split(' ');
                foreach(var value in allValues)
                {
                    try
                    {

                        var regValue = (byte)System.Convert.ToUInt32(value, 16);
                        if (iReg < 14)
                        {
                            outBytes.Add(regValue);
                        }
                    }
                    catch(Exception e)
                    {
                        int a = 0;
                    }

                    iReg++;
                }
            }
            
            return outBytes.ToArray();
        }

        private byte[] DumpMusic(ref byte[] inBytes)
        {
            var outBytes = new List<byte>();

            var registers = new byte[14];

            int inOffset = 16 + 1; // skip PSG signature

            var done = false;
            while ((!done) && (inOffset < inBytes.Length))
            {
                var register = inBytes[inOffset];
                inOffset++;

                int repeatCount = 1;

                if (register == 0xFD)
                {
                    // end of music
                    done = true;
                }
                else if (register == 0xFE)
                {
                    // repeat times
                    repeatCount = (inBytes[inOffset] * 4);
                    inOffset++;
                }
                else if (register == 0xFF)
                {
                    // end of interrupt
                }
                else
                {
                    var value = inBytes[inOffset];
                    inOffset++;

                    registers[register] = value;
                }

                if (register >= 0xFD)
                {
                    for (int i = 0; i < repeatCount; i++)
                    {
                        outBytes.AddRange(registers);

                        registers[13] = 0xFF;
                    }
                }
            }

            return outBytes.ToArray();
        }

        private byte[] DumpToCPCFrequency(ref byte[] inBytes)
        {
            var outBytes = new List<byte>();

            var inOffset = 0;
            int nbFrames = inBytes.Length / 14;
            for (int iFrame = 0; iFrame < nbFrames; iFrame++)
            {
                var registers = new byte[14];
                for (int i = 0; i < 14; i++)
                {
                    registers[i] = inBytes[inOffset];
                    inOffset++;
                }

                registers = ToCPCFrequency(registers);
                outBytes.AddRange(registers);
            }

            return outBytes.ToArray();
        }

        private Frame[] CreateFrames(ref byte[] inBytes)
        {
            var outFrames = new List<Frame>();

            var prevRegisters = new byte[14];

            var inOffset = 0;
            int nbFrames = inBytes.Length / 14;
            for (int iFrame = 0; iFrame < nbFrames; iFrame++)
            {
                var frame = new Frame();

                var registers = new byte[14];
                for (int i = 0; i < 14; i++)
                {
                    registers[i] = inBytes[inOffset];
                    inOffset++;
                }

                for (int i = 0; i < 14; i++)
                {
                    var writeReg = true;
                    if ((iFrame > 0) && (registers[i] == prevRegisters[i]))
                    {
                        writeReg = false;
                    }
                    if ((iFrame > 0) && (i == 13) && (registers[i] == 0xFF))
                    {
                        writeReg = false;
                        registers[i] = prevRegisters[i];
                    }
                    if (writeReg)
                    {
                        byte register = (byte)i;
                        if (register < 8)
                        {
                            frame.UsedReg0To7 |= (byte)(1 << register);
                        }
                        else
                        {
                            frame.UsedReg8To13 |= (byte)(1 << (register - 8));
                        }

                        frame.RegValues.Add(registers[i]);
                    }
                }

                /*if ((frame.UsedReg0To7 == 0) && (frame.UsedReg8To13 == 0))
                {
                    outFrames[outFrames.Count - 1].Count++;
                }
                else*/
                {
                    outFrames.Add(frame);
                    prevRegisters = registers;
                }
            }

            return outFrames.ToArray();
        }

        private byte[] CreateBinary(ref Frame[] inFrames)
        {
            var outBytes = new List<byte>();

            foreach (var inFrame in inFrames)
            {
                if ((inFrame.UsedReg0To7 == 0) && (inFrame.Count == 1))
                {
                    outBytes.Add((byte)(128 + inFrame.UsedReg8To13));
                }
                else
                {
                    outBytes.Add((byte)(inFrame.UsedReg8To13));
                    outBytes.Add((byte)(inFrame.UsedReg0To7));
                }

                foreach (var regValue in inFrame.RegValues)
                {
                    outBytes.Add(regValue);
                }

                int totalFrameCount = inFrame.Count - 2;
                while (totalFrameCount > 0)
                {
                    if (totalFrameCount >= 63)
                    {
                        outBytes.Add((byte)(64 + 63));
                        totalFrameCount -= 63;
                    }
                    else
                    {
                        outBytes.Add((byte)(64 + totalFrameCount));
                        totalFrameCount = 0;
                    }
                }
            }

            return outBytes.ToArray();
        }

        static ushort ConvertShort(ushort reg16)
        {
            float fReg = (float)reg16;
            fReg /= YMFreq;
            return (ushort)fReg;
        }

        static byte ConvertByte(byte reg8)
        {
            float fReg = (float)reg8;
            fReg /= YMFreq;
            return (byte)fReg;
        }

        static byte[] ToCPCFrequency(byte[] inBytes)
        {
            var outBytes = new byte[14];
            for (int i = 0; i < 14; i++)
            {
                outBytes[i] = inBytes[i];
            }

            ushort toneChannelA = (ushort)(outBytes[0] + ((ushort)outBytes[1]) * 256);
            ushort toneChannelB = (ushort)(outBytes[2] + ((ushort)outBytes[3]) * 256);
            ushort toneChannelC = (ushort)(outBytes[4] + ((ushort)outBytes[5]) * 256);
            byte noiseFrequency = (byte)(outBytes[6] & 31);
            ushort volumeEnvelope = (ushort)(outBytes[11] + ((ushort)outBytes[12]) * 256);

            toneChannelA = ConvertShort(toneChannelA);
            toneChannelB = ConvertShort(toneChannelB);
            toneChannelC = ConvertShort(toneChannelC);
            noiseFrequency = ConvertByte(noiseFrequency);
            volumeEnvelope = ConvertShort(volumeEnvelope);

            outBytes[0] = (byte)(toneChannelA & 255);
            outBytes[1] = (byte)(toneChannelA >> 8);
            outBytes[2] = (byte)(toneChannelB & 255);
            outBytes[3] = (byte)(toneChannelB >> 8);
            outBytes[4] = (byte)(toneChannelC & 255);
            outBytes[5] = (byte)(toneChannelC >> 8);
            outBytes[6] = noiseFrequency;
            outBytes[11] = (byte)(volumeEnvelope & 255);
            outBytes[12] = (byte)(volumeEnvelope >> 8);

            return outBytes;
        }
    }
}
