﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace phX.Framework
{
    public enum OpcodeId
    {
        LDI, LDD, INC_D, DEC_D, INC_H, DEC_H, INC_L, DEC_L, DEC_E, DEC_DE, INC_HL, DEC_HL, LD_HL, LD_BC, LD_DE, LD_SP, LD_SP_HL, LD_C,
        ADD_HL_BC, LD_A, LD_H, ADD_A_H, LD_H_A, LD_L, EX_DE_HL, ADD_A_D, LD_D_A, LD_E, CALL, LD_A_HL, LD_E_HL, LD_D_HL, PUSH_DE, LD_DE_A,
        LD_A_PTR, LD_B_HL, LD_C_B, PUSH_BC
    }

    public class KnownOpcode
    {
        static private List<Opcode> Opcodes = new List<Opcode>()
        {
            new Opcode(OpcodeId.LDI, "LDI", 5),
            new Opcode(OpcodeId.LDD, "LDD", 5),
            new Opcode(OpcodeId.INC_D, "INC D", 1),
            new Opcode(OpcodeId.DEC_D, "DEC D", 1),
            new Opcode(OpcodeId.DEC_E, "DEC E", 1),
            new Opcode(OpcodeId.INC_H, "INC H", 1),
            new Opcode(OpcodeId.DEC_H, "DEC H", 1),
            new Opcode(OpcodeId.INC_L, "INC L", 1),
            new Opcode(OpcodeId.DEC_L, "DEC L", 1),
            new Opcode(OpcodeId.INC_HL, "INC HL", 2),
            new Opcode(OpcodeId.DEC_HL, "DEC HL", 2),
            new Opcode(OpcodeId.DEC_DE, "DEC DE", 2),
            new Opcode(OpcodeId.LD_BC, "LD BC, _num_", 3),
            new Opcode(OpcodeId.LD_DE, "LD DE, _num_", 3),
            new Opcode(OpcodeId.LD_HL, "LD HL, _num_", 3),
            new Opcode(OpcodeId.LD_SP, "LD SP, _num_", 3),
            new Opcode(OpcodeId.LD_SP_HL, "LD SP, HL", 2),
            new Opcode(OpcodeId.LD_C, "LD C, _num_", 2),
            new Opcode(OpcodeId.ADD_HL_BC, "ADD HL, BC", 3),
            new Opcode(OpcodeId.LD_A, "LD A, _num_", 2),
            new Opcode(OpcodeId.LD_H, "LD H, _num_", 2),
            new Opcode(OpcodeId.LD_A_PTR, "LD A, (_num_)", 4),
            new Opcode(OpcodeId.ADD_A_H, "ADD A, H", 1),
            new Opcode(OpcodeId.LD_H_A, "LD H, A", 1),
            new Opcode(OpcodeId.LD_L, "LD L, _num_", 2),
            new Opcode(OpcodeId.EX_DE_HL, "EX DE, HL", 1),
            new Opcode(OpcodeId.ADD_A_D, "ADD A, D", 1),
            new Opcode(OpcodeId.LD_D_A, "LD D, A", 1),
            new Opcode(OpcodeId.LD_E, "LD E, _num_", 2),
            new Opcode(OpcodeId.CALL, "CALL _str_", 5),
            new Opcode(OpcodeId.LD_A_HL, "LD A, (HL)", 2),
            new Opcode(OpcodeId.LD_E_HL, "LD E, (HL)", 2),
            new Opcode(OpcodeId.LD_D_HL, "LD D, (HL)", 2),
            new Opcode(OpcodeId.PUSH_DE, "PUSH DE", 4),
            new Opcode(OpcodeId.PUSH_BC, "PUSH BC", 4),
            new Opcode(OpcodeId.LD_B_HL, "LD B, (HL)", 2),
            new Opcode(OpcodeId.LD_C_B, "LD C, B", 1),
            new Opcode(OpcodeId.LD_DE_A, "LD (DE), A", 2),
        };

        static public Opcode GetKnownOpcode(OpcodeId Id)
        {
            foreach (var opcode in Opcodes)
            {
                if (opcode.Id == Id)
                {
                    return opcode;
                }
            }
            System.Diagnostics.Debugger.Break();
            return null;
        }
    }

    public class Opcode
    {
        public OpcodeId Id;
        private string _Z80;
        public int Nops;
        public string Comment = "";

        public bool HasValue = false;
        public ushort Value;
        public string StringValue = "";

        public Opcode(OpcodeId id, string z80, int nops, string comment="")
        {
            Id = id;
            _Z80 = z80;
            Nops = nops;
            Comment = comment;
        }

        public Opcode(OpcodeId id, string z80, int nops, ushort value, string comment="") : this(id, z80, nops, comment)
        {
            HasValue = true;
            Value = value;
        }

        public Opcode(OpcodeId id, string z80, int nops, string value, string comment="") : this(id, z80, nops, comment)
        {
            HasValue = true;
            StringValue = value;
        }

        public Opcode(Opcode opcode, string comment)
        {
            Id = opcode.Id;
            _Z80 = opcode._Z80;
            Nops = opcode.Nops;
            HasValue = opcode.HasValue;
            Value = opcode.Value;
            StringValue = opcode.StringValue;

            Comment = opcode.Comment;
            if(comment!="")
            {
                Comment = comment;
            }
        }
        
        public string GetZ80()
        {
            var z80 = _Z80;
            if (HasValue)
            {
                if (StringValue != "")
                {
                    z80 = _Z80.Replace("_str_", StringValue.ToString());
                }
                else
                {
                    z80 = _Z80.Replace("_num_", "&" + Value.ToString("X"));
                }
            }
            z80 += " ;;" + Nops;
            if(Comment.Length > 0)
            {
                z80 += " " + Comment;
            }
            return z80;
        }

        public void SetValue(ushort value)
        {
            HasValue = true;
            Value = value;
        }

        public void SetValue(string value)
        {
            HasValue = true;
            StringValue = value;
        }
    }

    public class Listing
    {
        public List<Opcode> Opcodes = new List<Opcode>();

        public Opcode AddOpcode(OpcodeId id, string comment = "")
        {
            var globalOpcode = KnownOpcode.GetKnownOpcode(id);
            var opcode = new Opcode(globalOpcode, comment);

            Opcodes.Add(opcode);
            return opcode;
        }

        public Opcode AddOpcode(OpcodeId id, byte value, string comment = "")
        {
            return AddOpcode(id, (ushort)value, comment);
        }

        public Opcode AddOpcode(OpcodeId id, ushort value, string comment = "")
        {
            var globalOpcode = KnownOpcode.GetKnownOpcode(id);
            var opcode = new Opcode(globalOpcode, comment);
            opcode.SetValue(value);

            Opcodes.Add(opcode);
            return opcode;
        }

        public Opcode AddOpcode(OpcodeId id, string value, string comment = "")
        {
            var globalOpcode = KnownOpcode.GetKnownOpcode(id);
            var opcode = new Opcode(globalOpcode, comment);
            opcode.SetValue(value);

            Opcodes.Add(opcode);
            return opcode;
        }

        public int GetNops()
        {
            var nops = 0;
            foreach (var opcode in Opcodes)
            {
                nops += opcode.Nops;
            }
            return nops;
        }

        public string GetZ80()
        {
            var z80 = "";
            foreach (var opcode in Opcodes)
            {
                z80 += opcode.GetZ80() + "\n";
            }
            return z80;
        }
    }
}
