﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;

namespace phX.Framework
{
    public class Packer
    {
        public byte[] Convert(ref byte[] inBytes)
        {
            InBytes = inBytes;
            Buffer = new List<byte>();
            IndexInBuffer = 0;

            var outBytes = Pack();

            // prevent corrupted output
            /*InBytes = outBytes;
            Buffer = new List<byte>();
            IndexInBuffer = 0;
            if (!Compare(inBytes, Unpack()))
            {
                Debugger.Break();
            }*/

            return outBytes;
        }

        private byte[] InBytes = null;
        private List<byte> Buffer = null;
        private byte IndexInBuffer = 0;

        private static bool Compare(byte[] a, byte[] b)
        {
            if (a.Length != b.Length)
            {
                return false;
            }

            for (int i = 0; i < a.Length; i++)
            {
                if (a[i] != b[i])
                {
                    return false;
                }
            }

            return true;
        }

        private void AddIntoBuffer(byte value)
        {
            if (Buffer.Count == 256)
            {
                Buffer[IndexInBuffer] = value;
            }
            else
            {
                Buffer.Add(value);
            }
            IndexInBuffer++;
        }

        private bool FindInBuffer(int inOffset, out byte index, out byte length)
        {
            index = 0;
            length = 0;

            byte iStartIndex = IndexInBuffer;
            do
            {
                int iInOffset = inOffset;
                byte iSrcIndex = iStartIndex;

                byte iDstIndex = IndexInBuffer;
                var srcIndices = new List<byte>();
                var dstIndices = new List<byte>();

                var isBufferOverlap = false;
                byte iLength = 0;
                while ((iInOffset < InBytes.Length)
                    && (iSrcIndex < Buffer.Count)
                    && (InBytes[iInOffset] == Buffer[iSrcIndex])
                    && (iLength != 255)
                    && !isBufferOverlap
                    )
                {
                    srcIndices.Add(iSrcIndex);
                    dstIndices.Add(iDstIndex);

                    if (iSrcIndex != iDstIndex)
                    {
                        foreach (var usedIndex in srcIndices)
                        {
                            if (dstIndices.Contains(usedIndex))
                            {
                                isBufferOverlap = true;
                            }
                        }
                    }

                    if (!isBufferOverlap)
                    {
                        iInOffset++;
                        iLength++;
                        iSrcIndex++;
                        iDstIndex++;
                    }
                }

                if (iLength > length)
                {
                    length = iLength;
                    index = iStartIndex;
                }

                iStartIndex++;
            } while (iStartIndex != IndexInBuffer);

            var found = (length > 1);
            return found;
        }

        private bool FindBufferRepeat(int srcOffset, byte index, byte length, out byte repeatLength)
        {
            repeatLength = 0;

            if (index != (byte)(IndexInBuffer - length))
            {
                return false;
            }

            var ended = false;
            while (!ended)
            {
                if ((repeatLength + length) >= (256 - length))
                {
                    ended = true;
                }

                srcOffset += length;
                repeatLength += length;

                if (((byte)(index + length) > Buffer.Count)
                    || ((srcOffset + length) > InBytes.Length))
                {
                    ended = true;
                }
                else
                {
                    for (int i = 0; i < length; i++)
                    {
                        if (Buffer[(byte)(index + i)] != InBytes[srcOffset + i])
                        {
                            ended = true;
                        }
                    }
                }
            }

            var found = (repeatLength > length);
            if (!found)
            {
                repeatLength = 0;
            }
            return found;
        }

        private byte[] Pack()
        {
            var outBytes = new List<byte>();

            int inOffset = 0;
            while (inOffset < InBytes.Length)
            {
                // find in Buffer
                byte indexBuffer = 0;
                byte lengthBuffer = 0;
                var isCopyFromBuffer = FindInBuffer(inOffset, out indexBuffer, out lengthBuffer);
                if (isCopyFromBuffer)
                {
                    byte repeatLength;
                    if (FindBufferRepeat(inOffset, indexBuffer, lengthBuffer, out repeatLength))
                    {
                        lengthBuffer = repeatLength;
                    }
                }

                if (!isCopyFromBuffer)
                {
                    var inByte = InBytes[inOffset];
                    inOffset++;
                    AddIntoBuffer(inByte);

                    outBytes.Add((byte)0);
                    outBytes.Add(inByte);
                }
                else
                {
                    outBytes.Add((byte)(lengthBuffer + 0));

                    for (int i = 0; i < lengthBuffer; i++)
                    {
                        var inByte = InBytes[inOffset];
                        inOffset++;
                        AddIntoBuffer(inByte);
                    }

                    outBytes.Add(indexBuffer);
                }
            }

            return outBytes.ToArray();
        }

        private byte[] Unpack()
        {
            var outBytes = new List<byte>();

            int inOffset = 0;

            while (inOffset < InBytes.Length)
            {
                var length = InBytes[inOffset];
                inOffset++;

                if (length == 0)
                {
                    var rawByte = InBytes[inOffset];
                    inOffset++;
                    AddIntoBuffer(rawByte);

                    outBytes.Add(rawByte);
                }
                else
                {
                    var unpackIndex = InBytes[inOffset];
                    inOffset++;

                    for (int i = 0; i < length; i++)
                    {
                        var packedByte = Buffer[unpackIndex];
                        unpackIndex++;
                        AddIntoBuffer(packedByte);

                        outBytes.Add(packedByte);
                    }
                }
            }

            return outBytes.ToArray();
        }
    }
}
