﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace phX.Framework
{
    public class Palette
    {
        public class Pen
        {
            public byte IndexInPalette = 0; // between 0 and 16
            public byte FirmwareColor = 255; // between 0 and 26
            public byte GateArrayColor = 255; // between 0 and 26

            public Color Color = Color.Black; // PC Color

            public bool IsDummyEntry = false;

            public Pen()
            {
            }
        }

        public List<Pen> Pens = new List<Pen>();

        public Palette(int startIndex = 0)
        {
            var color = Color.Black;
            var firmwareColor = Pixel.ToFirmwareColor(color);
            var gateArrayColor = Pixel.ToGateArrayColor(firmwareColor);

            for (int indexInPalette = 0; indexInPalette < startIndex; indexInPalette++)
            {
                var dummyPaletteEntry = new Pen()
                {
                    IsDummyEntry = true,
                    IndexInPalette = (byte)indexInPalette,
                    Color = color,
                    FirmwareColor = firmwareColor,
                    GateArrayColor = gateArrayColor
                };
                Pens.Add(dummyPaletteEntry);
            }
        }

        public Pen GetPen(Color color)
        {
            foreach (var paletteEntry in Pens)
            {
                if (paletteEntry.IsDummyEntry)
                {
                    continue;
                }

                if (paletteEntry.Color == color)
                {
                    return paletteEntry;
                }
            }

            return null;
        }

        public Pen AddColor(Color color)
        {
            Pen paletteEntry = null;

            var firmwareColor = Pixel.ToFirmwareColor(color);
            var gateArrayColor = Pixel.ToGateArrayColor(firmwareColor);

            paletteEntry = GetPen(color);
            if (paletteEntry != null)
            {
                return paletteEntry;
            }

            for (int iPal = 0; iPal < Pens.Count; iPal++)
            {
                if (Pens[iPal].IsDummyEntry)
                {
                    continue;
                }

                if (Pens[iPal].FirmwareColor == firmwareColor)
                {
                    return Pens[iPal];
                }
            }

            paletteEntry = new Pen()
            {
                IndexInPalette = (byte)Pens.Count,
                FirmwareColor = firmwareColor,
                GateArrayColor = gateArrayColor,
                Color = color
            };
            Pens.Add(paletteEntry);

            return paletteEntry;
        }

        public byte[] ToFirmwareByteArray()
        {
            var palette = new byte[16];

            for (int i = 0; i < 16; i++)
            {
                palette[i] = 0;
            }

            int offset = 0;
            foreach (var paletteEntry in Pens)
            {
                palette[offset] = paletteEntry.FirmwareColor;
                offset++;
            }

            return palette;
        }

        public byte[] ToGateArrayByteArray()
        {
            var palette = new byte[16];

            for (int i = 0; i < 16; i++)
            {
                palette[i] = Pixel.ToGateArrayColor(0);
            }

            int offset = 0;
            foreach (var paletteEntry in Pens)
            {
                palette[offset] = Pixel.ToGateArrayColor(paletteEntry.FirmwareColor);
                offset++;
            }

            return palette;
        }

        public string ToZ80TableString(String pointerName)
        {
            var content = ";\n";
            for(int i=0; i<Pens.Count; i++)
            {
                content += "_PAL_" + pointerName + "_PEN" + i + " equ COLOR" + Pens[i].FirmwareColor + "\n";
            }

            content += "\n";
            content += ";\n";
            content += "_PAL_" + pointerName + ":\n";

            int iPen = 0;
			
            int remaining = 16;
            foreach (var paletteEntry in Pens)
            {
                content += "\tdb _PAL_" + pointerName + "_PEN" + iPen + "\n";
                //content += "\tdb COLOR" + paletteEntry.FirmwareColor + "\n";
				iPen++;
                remaining--;
            }

            while (remaining > 0)
            {
                content += "\tdb COLOR0 ; unused\n";
				iPen++;
                remaining--;
            }

            return content;
        }

        public string ToZ80MacroString(String macroName)
        {
            var content = ";\n";
			content += "MACRO _MPAL_"+macroName+"\n";
			
			int remaining = 0;
			
			if (Pens.Count<2)
			{
				content += "\t_PAL_MODE2 ";
				remaining = 2;
			}
			else if (Pens.Count<4)
			{
				content += "\t_PAL_MODE1 ";
				remaining = 4;
			}
			else
			{
				content += "\t_PAL_MODE0 ";
				remaining = 16;
			}
			
			bool isFirst = true;
			foreach (var paletteEntry in Pens)
            {
				if(!isFirst)
				{
					content += ", ";
				}
				isFirst = false;
				
                content += "COLOR" + paletteEntry.FirmwareColor;
                remaining--;
            }

            while (remaining > 0)
            {
                content += ", COLOR_UNUSED";
                remaining--;
            }
			
			content += "\n";
			content += "MEND\n";

            return content;
        }
    }
}
