﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace phX.Framework
{
    public class Pixel
    {
        public static byte ToMode0Pixel(byte p0, byte p1)
        {
            return (byte)((((p0) & 1) << 7) + (((p1) & 1) << 6)
                        + (((p0 >> 2) & 1) << 5) + (((p1 >> 2) & 1) << 4)
                        + (((p0 >> 1) & 1) << 3) + (((p1 >> 1) & 1) << 2)
                        + (((p0 >> 3) & 1) << 1) + (((p1 >> 3) & 1)));
        }

        public static byte ToMode1Pixel(byte p0, byte p1, byte p2, byte p3)
        {
            return (byte)((((p0) & 1) << 7) + (((p1) & 1) << 6)
                        + (((p2) & 1) << 5) + (((p3) & 1) << 4)
                        + (((p0 >> 1) & 1) << 3) + (((p1 >> 1) & 1) << 2)
                        + (((p2 >> 1) & 1) << 1) + (((p3 >> 1) & 1)));
        }
        
        public static void FromMode1Pixel(byte mode1Pixel, out byte p0, out byte p1, out byte p2, out byte p3)
        {
            p0 = (byte)((((mode1Pixel) >> 7)&1) + (((mode1Pixel >> 3)&1) << 1));
            p1 = (byte)((((mode1Pixel) >> 6) & 1) + (((mode1Pixel >> 2) & 1) << 1));
            p2 = (byte)((((mode1Pixel) >> 5) & 1) + (((mode1Pixel >> 1) & 1) << 1));
            p3 = (byte)((((mode1Pixel) >> 4) & 1) + ((mode1Pixel & 1) << 1));
        }

        public static byte ToMode2Pixel(byte p0, byte p1, byte p2, byte p3, byte p4, byte p5, byte p6, byte p7)
        {
            return (byte)((((p0) & 1) << 7) + (((p1) & 1) << 6)
                        + (((p2) & 1) << 5) + (((p3) & 1) << 4)
                        + (((p4) & 1) << 3) + (((p5) & 1) << 2)
                        + (((p6) & 1) << 1) + (((p7) & 1)));
        }
		
		public static byte ToFirmwareColor(Color color)
        {
            var CPCcolor = new Color();
            int tolerance = 95;

            int colorR = color.R;
            int colorG = color.G;
            int colorB = color.B;

            int bestScore = 10000;
            byte index = 26;
            for (byte iCPCPal = 0; iCPCPal <= 26; iCPCPal++)
            {
                CPCcolor = Color.FromArgb(CPCAsicPalette[iCPCPal]);
                int CPCcolorR = CPCcolor.R;
                int CPCcolorG = CPCcolor.G;
                int CPCcolorB = CPCcolor.B;

                if ((colorR > (CPCcolorR - tolerance)) && (colorR <= (CPCcolorR + tolerance)))
                {
                    if ((colorG > (CPCcolorG - tolerance)) && (colorG <= (CPCcolorG + tolerance)))
                    {
                        if ((colorB > (CPCcolorB - tolerance)) && (colorB <= (CPCcolorB + tolerance)))
                        {
                            int diffR = CPCcolorR - colorR;
                            if (diffR < 0)
                            {
                                diffR = -colorR;
                            }

                            int diffG = CPCcolorG - colorG;
                            if (diffG < 0)
                            {
                                diffG = -colorG;
                            }

                            int diffB = CPCcolorB - colorB;
                            if (diffB < 0)
                            {
                                diffB = -colorB;
                            }

                            int score = diffR + diffG + diffB;
                            if (score <= bestScore)
                            {
                                bestScore = score;
                                index = iCPCPal;
                            }

                            break;
                        }
                    }
                }
            }

            return index;
        }

        public static byte ToGateArrayColor(Color color)
        {
            return ToGateArrayColor(ToFirmwareColor(color));
        }

        public static byte ToGateArrayColor(byte firmwareColor)
        {
            return CPCGateArrayPalette[firmwareColor];
        }

        public static byte[] CPCGateArrayPalette = new byte[]
        {
            0x54, 0x44, 0x55, 0x5C,
            0x58, 0x5D, 0x4C, 0x45,
            0x4D, 0x56, 0x46, 0x57,
            0x5E, 0x40, 0x5F, 0x4E,
            0x47, 0x4F, 0x52, 0x42,
            0x53, 0x5A, 0x59, 0x5B,
            0x4A, 0x43, 0x4B
        };

        public static int[] CPCAsicPalette = new int[]
        {
            0x020702, 0x050663, 0x0507f1, 0x670600,
            0x680764, 0x6807F1, 0xFD0704, 0xFF0764,
            0xFD07F2, 0x046703, 0x046764, 0x0567f1,
            0x686704, 0x686764, 0x6867f1, 0xfd6704,
            0xfd6763, 0xfd67f1, 0x04f502, 0x04f562,
            0x04f5f1, 0x68f500, 0x68f564, 0x68f5f1,
            0xfef504, 0xfdf563, 0xfdf5f0
        };
		
		public static byte[] ToPixels(Bitmap bitmap, int mode, int y, ref Palette palette)
        {
            var cpcBytes = new List<byte>();

            switch (mode)
            {
                case 0:
                    for (int x = 0; x < bitmap.Width; x += 2)
                    {
                        Color p0 = bitmap.GetPixel(x, y);
                        Color p1 = bitmap.GetPixel(x + 1, y);

                        var paletteEntry0 = palette.AddColor(p0);
                        var paletteEntry1 = palette.AddColor(p1);

                        cpcBytes.Add(
                            ToMode0Pixel(
                                paletteEntry0.IndexInPalette,
                                paletteEntry1.IndexInPalette
                                ));
                    }
                    break;

                case 1:
                    for (int x = 0; x < bitmap.Width; x += 4)
                    {
                        Color p0 = bitmap.GetPixel(x, y);
                        Color p1 = bitmap.GetPixel(x + 1, y);
                        Color p2 = bitmap.GetPixel(x + 2, y);
                        Color p3 = bitmap.GetPixel(x + 3, y);

						var paletteEntry0 = palette.AddColor(p0);
                        var paletteEntry1 = palette.AddColor(p1);
                        var paletteEntry2 = palette.AddColor(p2);
                        var paletteEntry3 = palette.AddColor(p3);

                        cpcBytes.Add(
                            ToMode1Pixel(
                                paletteEntry0.IndexInPalette,
                                paletteEntry1.IndexInPalette,
                                paletteEntry2.IndexInPalette,
                                paletteEntry3.IndexInPalette
                                ));
                    }
                    break;

                case 2:
                    for (int x = 0; x < bitmap.Width; x += 8)
                    {
                        Color p0 = bitmap.GetPixel(x, y);
                        Color p1 = bitmap.GetPixel(x + 1, y);
                        Color p2 = bitmap.GetPixel(x + 2, y);
                        Color p3 = bitmap.GetPixel(x + 3, y);
                        Color p4 = bitmap.GetPixel(x + 4, y);
                        Color p5 = bitmap.GetPixel(x + 5, y);
                        Color p6 = bitmap.GetPixel(x + 6, y);
                        Color p7 = bitmap.GetPixel(x + 7, y);

                        var paletteEntry0 = palette.AddColor(p0);
                        var paletteEntry1 = palette.AddColor(p1);
                        var paletteEntry2 = palette.AddColor(p2);
                        var paletteEntry3 = palette.AddColor(p3);
                        var paletteEntry4 = palette.AddColor(p4);
                        var paletteEntry5 = palette.AddColor(p5);
                        var paletteEntry6 = palette.AddColor(p6);
                        var paletteEntry7 = palette.AddColor(p7);

                        cpcBytes.Add(
                            ToMode2Pixel(
                                paletteEntry0.IndexInPalette,
                                paletteEntry1.IndexInPalette,
                                paletteEntry2.IndexInPalette,
                                paletteEntry3.IndexInPalette,
                                paletteEntry4.IndexInPalette,
                                paletteEntry5.IndexInPalette,
                                paletteEntry6.IndexInPalette,
                                paletteEntry7.IndexInPalette
                                ));
                    }
                    break;
            }

            return cpcBytes.ToArray();
        }
    }
}
