using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace phX.Framework
{
	public class SinCurve
	{
        static public byte[] CreateSinAbs(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
            var buffer = new byte[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - minValue)) - 0.00000001d);

            for (int i = 0; i < buffer.Length; i++)
            {
                var value = (byte)(Math.Abs((1.0 * ((double)(maxValue - minValue))) + (amplitude * Math.Sin((2 * Math.PI * ((double)i) / ((double)bufferSize))))));
                
                buffer[i] = (byte)(minValue + value);
            }

            return buffer;
        }

        static public byte[] CreateSin(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
			var buffer = new byte[bufferSize];
			
			double amplitude = 0.5 * (((double)(maxValue-minValue)) -0.00000001d);
			
			for(int i=0; i<buffer.Length; i++)
			{
				var value = (byte)( (0.5*((double)(maxValue-minValue))) + (amplitude * Math.Sin((2 * Math.PI * completeCount * ((double)i) / ((double)bufferSize)))));
				
				buffer[i] = (byte)(minValue+value);
			}
			
			return buffer;
        }

        static public double[] CreateSinDouble(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
            var buffer = new double[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - minValue)) - 0.00000001d);

            for (int i = 0; i < buffer.Length; i++)
            {
                var value = (double)((0.5 * ((double)(maxValue - minValue))) + (amplitude * completeCount * Math.Sin((2 * Math.PI * ((double)i) / ((double)bufferSize)))));

                buffer[i] = (double)(minValue + value);
            }

            return buffer;
        }

        static public double[] CreateCosDouble(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
            var buffer = new double[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - minValue)) - 0.00000001d);

            for (int i = 0; i < buffer.Length; i++)
            {
                var value = (double)((0.5 * ((double)(maxValue - minValue))) + (amplitude * completeCount * Math.Cos((2 * Math.PI * ((double)i) / ((double)bufferSize)))));

                buffer[i] = (double)(minValue + value);
            }

            return buffer;
        }

        static public byte[] CreateCos(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
            var buffer = new byte[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - minValue)) - 0.00000001d);

            for (int i = 0; i < buffer.Length; i++)
            {
                var value = (byte)((0.5 * ((double)(maxValue - minValue))) + (amplitude * Math.Cos((2 * Math.PI * completeCount *((double)i) / ((double)bufferSize)))));

                buffer[i] = (byte)(minValue + value);
            }

            return buffer;
        }

        // with minValue=0;maxValue=256  create index0=0;index128=255;index255=0
        static public byte[] CreateReverseCos(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
            var buffer = new byte[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - minValue)) - 0.00000001d);

            for (int i = 0; i < buffer.Length; i++)
            {
                var value = (byte)((0.5 * ((double)(maxValue - minValue))) + (amplitude * Math.Cos((2 * Math.PI * completeCount * ((double)i) / ((double)bufferSize)))));

                buffer[i] = (byte)((maxValue-1) - (minValue + value));
            }

            return buffer;
        }
        
        // with minValue=0;maxValue=256  create index0=0;index128=255;index255=0
        static public double[] CreateReverseCosDouble(int bufferSize, int minValue, int maxValue, double completeCount = 1)
        {
            var buffer = new double[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - minValue)) - 0.00000001d);

            for (int i = 0; i < buffer.Length; i++)
            {
                var value = (double)((0.5 * ((double)(maxValue - minValue))) + (amplitude * Math.Cos((2 * Math.PI * completeCount * ((double)i) / ((double)bufferSize)))));

                buffer[i] = (double)((maxValue - 1) - (minValue + value));
            }

            return buffer;
        }

        static public byte[] CreateSin16Bits(int bufferSize, int maxValue, double completeCount =1)
        {
            var buffer = new byte[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - 0)) - 0.00000001d);

            for (int i = 0; i < buffer.Length / 2; i++)
            {
                double value = ((0.5 * ((double)(maxValue - 0))) + (amplitude * Math.Sin((2 * Math.PI * completeCount * ((double)i) / ((double)bufferSize / 2)))));

                byte intPart = (byte)value;

                double oneScope = value - (double)((int)value);
                oneScope = oneScope * 256;
                byte floatPart = (byte)(oneScope);

                buffer[(i * 2)] = intPart;
                buffer[(i * 2) + 1] = floatPart;
            }

            return buffer;
        }

        static public byte[] CreateCos16Bits(int bufferSize, int maxValue, double completeCount = 1)
        {
            var buffer = new byte[bufferSize];

            double amplitude = 0.5 * (((double)(maxValue - 0)) - 0.00000001d);

            for (int i = 0; i < buffer.Length/2; i++)
            {
                double value = ((0.5 * ((double)(maxValue - 0))) + (amplitude * Math.Cos((2 * Math.PI * completeCount * ((double)i) / ((double)bufferSize/2)))));

                byte intPart = (byte)value;

                double oneScope = value - (double)((int)value);
                oneScope = oneScope * 256;
                byte floatPart = (byte)(oneScope);

                buffer[(i*2)] = intPart;
                buffer[(i*2)+1] = floatPart;
            }

            return buffer;
        }
    }
}
		
		