﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _01_Logo
    {
        public _01_Logo()
        {
			try
			{
                int height = CreateRaster(  Program.ToDataFolder("01_Logo/Rasters.png"),
                                            Program.ToZ80Folder("01_Logo/_Rasters.png.z80"));

                CreateSinCurves(            10, 50,
                                            Program.ToZ80Folder("01_Logo/_SinCurve1.bin"));

                CreateSplineFrames(         Program.ToZ80Folder(	"01_Logo/_SplineFrames.bin"));
										
				WriteLogo("01_Logo\\", "Logo.png");
			}
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}
        }

        private void CreateSinCurves(int minValue, int maxValue, string filename1)
        {
            var sinCurve1 = SinCurve.CreateSin(256, minValue, maxValue);
            File.WriteAllBytes(filename1, sinCurve1);
        }

        private int CreateRaster(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);

            return bmp.Height;
        }

        private byte[] firmwareColors = {
            13,7,15
        };

        const double NbBars = 2;
        double barHeight = 80;
        const int NbFrames = 5*4;
        const double Radian360 = 6.283185307179586;
        const double ScreenHeight = 268;

        private void PostEdit(ref List<byte> frame)
        {
            try
            {
                for (int i = 0; i < frame.Count; i++)
                {
                    frame[i] = Pixel.ToGateArrayColor(firmwareColors[frame[i]]);
                }
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        private void CreateBackgroundRasters(string filename)
        {
            var allFrames = new List<byte>();
            
            for (int iFrame=0; iFrame<NbFrames; iFrame++)
            {
                var frame = new List<byte>();
                for (int i = 0; i < 268; i++)
                {
                    frame.Add(0);
                }

                double fFrame = (double)iFrame;
                double fNbFrames = (double)NbFrames;
                double progress = (fFrame / fNbFrames); // 0 to 1

                double radAngle = ( progress * Radian360 ) / NbBars;
                double radDelta = Radian360 / NbBars;

                double maxPosY = (ScreenHeight-barHeight+1);
                double centerY = maxPosY / 2;
                
                for(double iBar=0; iBar<NbBars; iBar += 1)
                {
                    double y = Math.Sin(-radAngle) * (maxPosY / 2);
                    y += centerY;

                    for(int y2=0; y2<barHeight; y2++)
                    {
                        var fY = (int)(y + y2);
                        if(fY<ScreenHeight)
                        {
                            frame[fY]++;
                        }
                    }

                    radAngle += radDelta;
                }

                PostEdit(ref frame);
                
                allFrames.AddRange(frame);
            }

            File.WriteAllBytes(filename, allFrames.ToArray());
        }
		
		private void WriteLogo(string folder, string filename)
		{
			var bitmap = new Bitmap(Program.ToDataFolder(folder+filename));				
			var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
				{
					Bitmap = bitmap,
					Mode = 0,
					R1 = 32,
					R4 = 38+39,
					R9 = 3,
					StartVLC = 0,
                    FirstPixelInPaletteX = 0,
                    FirstPixelInPaletteY = 127
            }
			);
			foreach (var VLC in image.Pages[0].VLCs)
			{
				File.WriteAllBytes(Program.ToZ80Folder(folder + "_"+filename + ".vlc" + VLC.Index), VLC.CPCBytes);
			}				
			File.WriteAllText(
				Program.ToZ80Folder(folder + "_" + filename + ".pal.z80"), 
				image.Palette.ToZ80TableString(Program.FilenameToKeyWordName(filename))
			);
		}
		
		private void CreateSplineFrames(string filename)
		{
			var frames = new List<byte>();
			
			var keyPoints = new List<PointF>();
			
			float coef = 1.5f;

            keyPoints.Add(new PointF(0, 5*coef));
			keyPoints.Add(new PointF(0.2f, 5*coef));
		
			keyPoints.Add(new PointF(1, 10*coef));
			keyPoints.Add(new PointF(1.2f, 10*coef));
			
			keyPoints.Add(new PointF(2, 15*coef));
			keyPoints.Add(new PointF(2.2f, 15*coef));
			
			keyPoints.Add(new PointF(3, 20*coef));
			keyPoints.Add(new PointF(3.2f, 20*coef));

            keyPoints.Add(new PointF(4, 25*coef));
			keyPoints.Add(new PointF(4.2f, 25*coef));
		
			keyPoints.Add(new PointF(5, 30*coef));
			keyPoints.Add(new PointF(5.2f, 30*coef));
			
			keyPoints.Add(new PointF(6, 35*coef));
			keyPoints.Add(new PointF(6.2f, 35*coef));
			
			keyPoints.Add(new PointF(7, 40*coef));
			keyPoints.Add(new PointF(7.2f, 40*coef));
			
			keyPoints.Add(new PointF(8, 45*coef));
			keyPoints.Add(new PointF(8.2f, 45*coef));
			
			keyPoints.Add(new PointF(11, 5*coef));
			keyPoints.Add(new PointF(11.2f, 5*coef));
		
			keyPoints.Add(new PointF(12, 10*coef));
			keyPoints.Add(new PointF(12.2f, 10*coef));
			
			keyPoints.Add(new PointF(13, 15*coef));
			keyPoints.Add(new PointF(13.2f, 15*coef));
			
			keyPoints.Add(new PointF(16, 20*coef));
			keyPoints.Add(new PointF(16.2f, 20*coef));
			
			keyPoints.Add(new PointF(17, 25*coef));
			keyPoints.Add(new PointF(17.2f, 25*coef));
			
			keyPoints.Add(new PointF(18, 30*coef));
			keyPoints.Add(new PointF(18.2f, 30*coef));
			
			keyPoints.Add(new PointF(19, 35*coef));
			keyPoints.Add(new PointF(19.2f, 35*coef));
			
			keyPoints.Add(new PointF(22, 40*coef));
			keyPoints.Add(new PointF(22.2f, 40*coef));
			
			keyPoints.Add(new PointF(23, 45*coef));
			keyPoints.Add(new PointF(23.2f, 45*coef));
			
			keyPoints.Add(new PointF(24, 5*coef));
			keyPoints.Add(new PointF(24.2f, 5*coef));
			
			keyPoints.Add(new PointF(25, 10*coef));
			keyPoints.Add(new PointF(25.2f, 10*coef));
			
			keyPoints.Add(new PointF(26, 20*coef));
			keyPoints.Add(new PointF(26.2f, 20*coef));
			
			keyPoints.Add(new PointF(27, 20*coef));
			keyPoints.Add(new PointF(27.2f, 20*coef));
			
            var splinePoints = CatmullRomSpline.Generate(keyPoints.ToArray(), 10);
			
			foreach(var p in splinePoints)
			{
				frames.Add((byte)(p.Y-5));
			}

            /*
			
			var spline = new SplineInterpolator(nodes);			
            var start = nodes.First().Key;
            var end = nodes.Last().Key;
            var step = (end - start) / NbFrames;
			
			for (var x = start; x <= end; x += step)
			{
				var y = spline.GetValue(x);
				
				frames.Add((byte)y);
			}*/
			
			File.WriteAllBytes(filename, frames.ToArray());
		}
    }
}
