using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _02_SpectrumAnalyzer
    {
		public const int MaxSinValue = 128; // max 256
		
        public _02_SpectrumAnalyzer()
        {
			try
			{
				CreateHorizontalBars(	Program.ToDataFolder(	"02_SpectrumAnalyzer/HorizontalBars.png"), 
										Program.ToZ80Folder(	"02_SpectrumAnalyzer/_HorizontalBars.bin"));
										
				CreateAnimFrames(		1,
										Program.ToZ80Folder(	"02_SpectrumAnalyzer/_AnimFrames.bin"));
										
				CreateYSinMove(			Program.ToZ80Folder(	"02_SpectrumAnalyzer/_YSinMove.bin"));										
										
				WriteLogo(              "02_SpectrumAnalyzer\\",
                                        "BackgroundCRTC0.bmp",
										"BackgroundCRTC1.bmp",
										"BackgroundCRTC3.bmp",
                                        "BackgroundCRTC4.bmp"
                                        );
			}
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}
        }
		
		private void WriteLogo(string partFolder, string filename0, string filename1, string filename3, string filename4)
		{
			var image0 = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
				{
					Bitmap = new Bitmap(Program.ToDataFolder(partFolder+filename0)),
					Mode = 0,
					R1 = 48,
					R4 = 20,
					R9 = 0,
					StartVLC = 0,
					StartR12 = 0x40
				}
			);
			
			var image1 = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
				{
					Bitmap = new Bitmap(Program.ToDataFolder(partFolder+filename1)),
					Mode = 0,
					R1 = 48,
					R4 = 20,
					R9 = 0,
					StartVLC = 0,
					StartR12 = 0x40
				}
			);
			
			var image3 = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
				{
					Bitmap = new Bitmap(Program.ToDataFolder(partFolder+filename3)),
					Mode = 0,
					R1 = 48,
					R4 = 20,
					R9 = 0,
					StartVLC = 0,
					StartR12 = 0x40
				}
			);

            var image4 = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = new Bitmap(Program.ToDataFolder(partFolder + filename4)),
                Mode = 0,
                R1 = 48,
                R4 = 20,
                R9 = 0,
                StartVLC = 0,
                StartR12 = 0x40
            }
            );

            for (var iPage=0;iPage<image0.Pages.Length;iPage++)
			{
				foreach (var VLC in image0.Pages[iPage].VLCs)
				{
					File.WriteAllBytes(Program.ToZ80Folder(partFolder+"_"+filename0 + ".page" + iPage + ".vlc" + VLC.Index), VLC.CPCBytes);
				}	
			}
			
			File.WriteAllBytes(Program.ToZ80Folder(partFolder+"_"+filename0 + ".r12r13"), image0.R12R13);
							
			File.WriteAllText(
				Program.ToZ80Folder(partFolder+"_"+filename0 + ".pal.z80"), 
				image0.Palette.ToZ80TableString(Program.FilenameToKeyWordName(filename0))
			);
			
			WriteDiffBytes(image0, image1, Program.ToZ80Folder(partFolder + "_" + filename0 + ".diff.crtc1"));
			WriteDiffBytes(image0, image3, Program.ToZ80Folder(partFolder + "_" + filename0 + ".diff.crtc3"));
            WriteDiffBytes(image0, image4, Program.ToZ80Folder(partFolder + "_" + filename0 + ".diff.crtc4"));
        }
		
		private void WriteDiffBytes(Framework.Image imageA, Framework.Image imageB, string filename)
		{
			var diffBytes = new List<byte>();
			
			for(var iPage=0;iPage<imageA.Pages.Length;iPage++)
			{
				for(ushort offset=0; offset<imageA.Pages[iPage].VLCs[0].CPCBytes.Length; offset++)
				{
					var byteA = imageA.Pages[iPage].VLCs[0].CPCBytes[offset];
					var byteB = imageB.Pages[iPage].VLCs[0].CPCBytes[offset];
					
					if(byteA!=byteB)
					{	
						ushort CPCAddress = (ushort)(0x4000+offset); // Write into C3/&4000 (which is C0/&C000)
						
						diffBytes.Add((byte)(byteB));
						diffBytes.Add((byte)(CPCAddress&255));
						diffBytes.Add((byte)(CPCAddress>>8));
					}
				}
			}
			diffBytes.Insert(0, (byte)(1+(diffBytes.Count()/3)) );
			File.WriteAllBytes(filename, diffBytes.ToArray());
		}
		
		private void CreateYSinMove(string filename)
		{
			var bytes = SinCurve.CreateCos(256, 0, MaxSinValue);
			
			File.WriteAllBytes(filename, bytes);
		}
		
		private void CreateAnimFrames(int maxFrames, string baseFilename)
        {
            var inBar = new List<byte>();

            int inBarFrameSize = 0;

            for (int iFrame = 0; iFrame < maxFrames; iFrame++)
            {
                List<byte> frameBytes = CreateAnimFrame(iFrame, maxFrames);

                WriteAnimFrame(
                    frameBytes,
                    ref inBar);

                if (inBarFrameSize == 0)
                {
                    inBarFrameSize = inBar.Count;
                }
            }

            File.WriteAllBytes(baseFilename + ".inBar", inBar.ToArray());
        }
		
		private List<byte> CreateAnimFrame(int iFrame, int maxFrames)
        {
            var frameBytes = new List<byte>();

            double bitmapHeight = 63;

            double frameDelta = bitmapHeight/(double)(maxFrames);
            double yFrame = ((double)iFrame) * frameDelta;

            double progress = (double)iFrame / (double)(maxFrames);
            progress = -1 + (progress * 2); // from -1 to 1
            double sinProgress1 = 2 * Math.PI * progress / 180.0;

            for (int y=0; y<272+MaxSinValue; y++)
            {
                var v1 = Math.Sin(sinProgress1);

                sinProgress1 += 0.03f;

                double srcY = (bitmapHeight / 2.5) + (v1 * (bitmapHeight / 5));

                srcY += (double)(yFrame + y);

                while (srcY < 0)
                {
                    srcY += bitmapHeight;
                }
                while (srcY >= bitmapHeight)
                {
                    srcY -= bitmapHeight;
                }

                frameBytes.Add((byte)srcY);
            }

            return frameBytes;
        }
		
		private void WriteAnimFrame(
            List<byte> frameBytes,
            ref List<byte> inBar)
        {
            byte prevValue = frameBytes[0];
            foreach (var b in frameBytes)
            {
                byte _b = (byte)(b * 2); // R12/R13 takes 2 bytes, let's mul2 here instead of Z80

                inBar.Add(_b);
            }
        }
		
		private void CreateHorizontalBars(string baseBmpFilename, string outFilename)
        {
            Bitmap baseBmp = new Bitmap(System.Drawing.Image.FromFile(baseBmpFilename));

            var allBytes = new List<byte>(420);

            for(int iRasterBar=0; iRasterBar<5; iRasterBar++)
            {
                var verticalBar = new byte[84];

                for (int iY=0; iY<84;iY++)
                {
                    var pixel = baseBmp.GetPixel(0, (iRasterBar * 84) + iY);

                    var firmwareColor = Pixel.ToFirmwareColor(pixel);
                    var gateArrayColor = Pixel.CPCGateArrayPalette[firmwareColor];

                    verticalBar[iY] = gateArrayColor;
                }

                allBytes.AddRange(verticalBar);
            }

            File.WriteAllBytes(outFilename, allBytes.ToArray());
        }
    }
}
