using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _03_Bobs
    {
		public _03_Bobs()
        {
			try
			{
				CreateData(Program.ToZ80Folder(	"03_Bobs/Anim/"	));
				
				for(int i=1; i<=8; i++)
				{
					CreateRaster(	Program.ToDataFolder(	"03_Bobs/Raster"+i+".png"),
									Program.ToZ80Folder(	"03_Bobs/Rasters/_Raster" + i + ".png.z80")	);
				}
			}
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}
		}

        private void CreateRaster(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }
			
		public void CreateData(string compiledFolder)
        {
            var globalSettings = new BobsGlobalSettings()
            {
                TempFolder = compiledFolder + "Debug/",
				WinAPEFolder = compiledFolder,				
                BobWidth = 220 * 2, // must be multiple 4   max 256
                BobHeight = 220 + 1,
                CpcScreenHeight = 268+2, // was 274
                CpcScreenWidth = 48 * 2 * 8,
                CpcScreenByteWidth = 48 * 2,
                ClippingLeftRightWidth = 48 * 8,
                ClippingLeftRightByteWidth = 48
            };

            Console.WriteLine("[Step 1/4] Creating GIF images");
            var createImages = new CreateGifImage(globalSettings);
            createImages.Start();

            var allFiles = Directory.GetFiles(globalSettings.TempFolder + "");
            var bitmapFilenames = new List<string>();
            foreach (var file in allFiles)
            {
                if (file.Contains(".gif"))
                {
                    bitmapFilenames.Add(file);
                }
            }

            Console.WriteLine("[Step 2/4] Creating CPC images");
            var createCPCImages = new List<CreateCPCImage>();
            foreach (var bitmapFilename in bitmapFilenames)
            {
                var createCPCImage = new CreateCPCImage(globalSettings, bitmapFilename);
                createCPCImages.Add(createCPCImage);
                createCPCImage.Start();
            }

            Console.WriteLine("[Step 3/4] Creating CPC opcodes");
            int iCPCImage = 0;
            int iBitplan = 0;
            foreach (var bitmapFilename in bitmapFilenames)
            {
				//if((iCPCImage==0)||(iCPCImage==0))
				{
					var isCircle2 = (iCPCImage & 1) == 1;
					var createCPCOpcodes = new CreateCPCOpcodes(globalSettings,
						isCircle2,
						bitmapFilename,
						createCPCImages[iCPCImage].CPCImage,
						createCPCImages[iCPCImage].CPCImage_ByteWidth,
						createCPCImages[iCPCImage].CPCImage_Height);
					createCPCOpcodes.Start();

					if (isCircle2)
					{
						globalSettings.BitmapEdgeInfos[iBitplan].EdgeInfos = createCPCOpcodes.EdgeInfos;

						var lastEdgeInfo = new BobsEdgeInfo();
						lastEdgeInfo.LeftCpcByte = (byte)96;
						lastEdgeInfo.RightCpcByte = (byte)96;
						globalSettings.BitmapEdgeInfos[iBitplan].EdgeInfos.Add(lastEdgeInfo);

						iBitplan++;
					}
				}

                iCPCImage++;
            }

            Console.WriteLine("[Step 4/4] Creating CPC tables");
            var createCPCTables = new CreateCPCTables(globalSettings, createCPCImages[0].CPCImage_Height);
            createCPCTables.Start();
        }
    }
	
	public class BobsEdgeInfo
    {
        public byte LeftPtr;
        public byte LeftCpcByte;
        public byte RightPtr;
        public byte RightCpcByte;
    }
	
    public class BobsBitmapEdgeInfos
    {
        public List<BobsEdgeInfo> EdgeInfos = new List<BobsEdgeInfo>();
    }
	
    public class BobsGlobalSettings
    {
        public string TempFolder;
        public string WinAPEFolder;
        public int BobWidth;
        public int BobHeight;
        public int CpcScreenHeight;
        public int CpcScreenWidth;
        public int CpcScreenByteWidth;
        public int ClippingLeftRightWidth;
        public int ClippingLeftRightByteWidth;

        public string BackgroundMask = "10101010";

        public List<BobsBitmapEdgeInfos> BitmapEdgeInfos = new List<BobsBitmapEdgeInfos>();

        public BobsGlobalSettings()
        {
            BitmapEdgeInfos.Add(new BobsBitmapEdgeInfos());
            BitmapEdgeInfos.Add(new BobsBitmapEdgeInfos());
            BitmapEdgeInfos.Add(new BobsBitmapEdgeInfos());
            BitmapEdgeInfos.Add(new BobsBitmapEdgeInfos());
        }
    }
	
	public class CreateGifImage
    {
        private BobsGlobalSettings _globalSettings;

        public CreateGifImage(BobsGlobalSettings globalSettings)
        {
            _globalSettings = globalSettings;
        }

        private Brush GetBrush(int penIndex)
        {
            if (penIndex==0)
            {
                return Brushes.Black;
            }

            return Brushes.White;
        }
        
        private void ClearFrame(Graphics graphics, int penIndex)
        {
            var rect = new Rectangle(0, 0, _globalSettings.BobWidth+8, _globalSettings.BobHeight);
            graphics.FillRectangle(GetBrush(penIndex), rect);
        }
        
        private void DrawCircle(Graphics graphics, int penIndex, int offsetX)
        {
            graphics.FillEllipse(GetBrush(penIndex), new Rectangle(offsetX, 0, _globalSettings.BobWidth+1, _globalSettings.BobHeight-1 ));
        }

        private void CreateFrame(Graphics g, int circlePenIndex, int offsetX)
        {
            ClearFrame(g, 0);

            DrawCircle(g, circlePenIndex, offsetX);
        }

        public void Start()
		{
            var bitmap = new Bitmap(_globalSettings.BobWidth+8, _globalSettings.BobHeight);

            Graphics g = Graphics.FromImage(bitmap);
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighSpeed;
            g.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.HighQuality;

            for (int penIndex = 1; penIndex < 3; penIndex++)
            {
                for (int offsetX = 0; offsetX < 4; offsetX++)
                {
                    var filename = _globalSettings.TempFolder + "X" + offsetX + "P" + penIndex + ".gif";

                    Console.WriteLine(filename);

                    CreateFrame(g, penIndex, offsetX*2); // remove *2 for pixel precision
                    
					bitmap.Save(filename, System.Drawing.Imaging.ImageFormat.Gif);
                }
            }

            g.Dispose();
            bitmap.Dispose();
        }
    }
	
	public class CreateCPCImage
    {
        private BobsGlobalSettings _globalSettings;
        private string _bitmapFilename;

        public byte[] CPCImage = null;
        public int CPCImage_ByteWidth = 0;
        public int CPCImage_Height = 0;

        private int NbPixelsPerByte = 8;
        private List<int> _indices = new List<int>();
        
        public string _humanFilename = "";
        
        public CreateCPCImage(BobsGlobalSettings globalSettings, string bitmapFilename)
        {
            Console.WriteLine(bitmapFilename);

            _globalSettings = globalSettings;
            _bitmapFilename = bitmapFilename;
            _humanFilename = new FileInfo(_bitmapFilename).Name.Replace(".gif", "");
            
            _indices = new List<int>();
            //_indices.Add(251); // white = PEN0
            //_indices.Add(5);
            //_indices.Add(210);
            //_indices.Add(18);            
            _indices.Add(0);
            _indices.Add(0xFB);
            _indices.Add(0xD2);
            _indices.Add(0x05);
        }

#region CPC Image creation
        private Bitmap ConvertTo8bpp(System.Drawing.Image img)
        {
            var ms = new System.IO.MemoryStream();
            img.Save(ms, System.Drawing.Imaging.ImageFormat.Gif);
            ms.Position = 0;
            return new Bitmap(ms);
        }

        private byte ConvertPixelCPC_Mode2(byte p0, byte p1, byte p2, byte p3, byte p4, byte p5, byte p6, byte p7)
        {
            p0 = (byte)((p0 != 0) ? 0 : 1);
            p1 = (byte)((p1 != 0) ? 0 : 1);
            p2 = (byte)((p2 != 0) ? 0 : 1);
            p3 = (byte)((p3 != 0) ? 0 : 1);
            p4 = (byte)((p4 != 0) ? 0 : 1);
            p5 = (byte)((p5 != 0) ? 0 : 1);
            p6 = (byte)((p6 != 0) ? 0 : 1);
            p7 = (byte)((p7 != 0) ? 0 : 1);

            /*p0 = (byte)((p0 != 0) ? 1 : 0);
            p1 = (byte)((p1 != 0) ? 1 : 0);
            p2 = (byte)((p2 != 0) ? 1 : 0);
            p3 = (byte)((p3 != 0) ? 1 : 0);
            p4 = (byte)((p4 != 0) ? 1 : 0);
            p5 = (byte)((p5 != 0) ? 1 : 0);
            p6 = (byte)((p6 != 0) ? 1 : 0);
            p7 = (byte)((p7 != 0) ? 1 : 0);*/

            byte v;
            v = (byte)(
                (p0 << 7) +
                (p1 << 6) +
                (p2 << 5) +
                (p3 << 4) +
                (p4 << 3) +
                (p5 << 2) +
                (p6 << 1) +
                p7);

			v &= (byte)(Convert.ToByte(_globalSettings.BackgroundMask, 2));

            return v;
        }

        private byte GetIndex(Bitmap bitmap, Color pixel)
        {
            byte index = 0;

            foreach (var entry in bitmap.Palette.Entries)
            {
                if (entry.ToArgb() == pixel.ToArgb())
                {
                    if (!_indices.Contains(index))
                    {
                        System.Diagnostics.Debugger.Break();
                        return 0; // should not happen
                    }
                    return (byte)_indices.IndexOf(index);
                }

                index++;
            }

            return 0;
        }

        private byte GetIndex(Bitmap bitmap, int x, int y)
        {
            var pixel = bitmap.GetPixel(x, y);
            //var index = GetIndex(bitmap, pixel);

            var index = (byte)(pixel.ToArgb());

            return index;
        }

        private bool CreateRawCPCImage()
        {
            var image = System.Drawing.Image.FromFile(_bitmapFilename);
            var bitmap = ConvertTo8bpp(image);

            CPCImage_ByteWidth = bitmap.Width / NbPixelsPerByte;
            CPCImage_Height = bitmap.Height;
            CPCImage = new byte[CPCImage_ByteWidth * CPCImage_Height];
            
            /*if ((CPCImage_ByteWidth * 8) != bitmap.Width)
            {
                Console.WriteLine("Bad image width");
                return false;
            }*/

            for (int y = 0; y < bitmap.Height; y++)
            {
                for (int x = 0; x < bitmap.Width; x += 8)
                {
                    var p0 = GetIndex(bitmap, x, y);
                    var p1 = GetIndex(bitmap, x + 1, y);
                    var p2 = GetIndex(bitmap, x + 2, y);
                    var p3 = GetIndex(bitmap, x + 3, y);
                    var p4 = GetIndex(bitmap, x + 4, y);
                    var p5 = GetIndex(bitmap, x + 5, y);
                    var p6 = GetIndex(bitmap, x + 6, y);
                    var p7 = GetIndex(bitmap, x + 7, y);
                    var cpcByte = ConvertPixelCPC_Mode2(p0, p1, p2, p3, p4, p5, p6, p7);

                    CPCImage[(y * CPCImage_ByteWidth) + (x / NbPixelsPerByte)] = cpcByte;
                }
            }

            return true;
        }
#endregion

        public bool Start()
        {
            var rawTextFilename = new FileInfo(_bitmapFilename).Name.Replace(".gif", "_raw.txt");
            
            if (!CreateRawCPCImage())
            {
                Console.WriteLine("Failed to create CPC Image");
                return false;
            }

            var rawTxt = "";
            for (int iY = 0; iY < CPCImage_Height; iY++)
            {
                for (int iX = 0; iX < CPCImage_ByteWidth; iX++)
                {
                    rawTxt += CPCImage[(iY * CPCImage_ByteWidth) + iX].ToString("X2") + " ";
                }
                rawTxt += "\n";
            }
            File.WriteAllText(this._globalSettings.TempFolder + rawTextFilename, rawTxt);

            return true;
        }
    }
	
	public class CreateCPCOpcodes
    {
        private BobsGlobalSettings _globalSettings;
        private string _bitmapFilename;
        private byte[] _cpcImage;
        private int _cpcImageByteWidth;
        private int _cpcImageHeight;

        private byte[] _line;
        private byte _clearByte;
        private byte _fillByte;

        private bool _isCircle2 = false;

        public string _humanFilename = "";

        public List<BobsEdgeInfo> EdgeInfos = new List<BobsEdgeInfo>();

        public CreateCPCOpcodes(BobsGlobalSettings globalSettings, bool isCircle2, string bitmapFilename, byte[] cpcImage, int cpcImageByteWidth, int cpcImageHeight)
        {
            Console.WriteLine(bitmapFilename);

            _globalSettings = globalSettings;
            _bitmapFilename = bitmapFilename;
            _cpcImage = cpcImage;
            _cpcImageByteWidth = cpcImageByteWidth;
            _cpcImageHeight = cpcImageHeight;
            _humanFilename = new FileInfo(_bitmapFilename).Name.Replace(".gif", "");
            _isCircle2 = isCircle2;

            _clearByte = cpcImage[0];
            _fillByte = cpcImage[ (cpcImageByteWidth * (cpcImageHeight / 2)) + (cpcImageByteWidth / 2)];

            _line = new byte[_cpcImageByteWidth];
            for (int iX = 0; iX < _cpcImageByteWidth; iX++)
            {
                _line[iX] = _clearByte;
            }
        }

        public bool Start()
        {
            var debugTxt = "";

            var asmCode = "";
            
            int staticXLeft = ((_cpcImageByteWidth) / 2);
            int staticXRight = staticXLeft+1;

            int leftPtr = staticXLeft;
            int rightPtr = staticXRight;

            int minNOPS = 32;

            //string circle2Offset = "\n;\n";
            //var leftRightEdges = new List<string>();
            
            for (int iY = 0; iY < _cpcImageHeight; iY++)
            {
                if (_isCircle2)
                {
                    byte leftCpcByte = _cpcImage[(iY * _cpcImageByteWidth) + leftPtr];
                    byte rightCpcByte = _cpcImage[(iY * _cpcImageByteWidth) + rightPtr];

                    var edgeInfo = new BobsEdgeInfo();
                    edgeInfo.LeftPtr = (byte)leftPtr;
                    edgeInfo.LeftCpcByte = leftCpcByte;
                    edgeInfo.RightPtr = (byte)rightPtr;
                    edgeInfo.RightCpcByte = rightCpcByte;
                    EdgeInfos.Add(edgeInfo);

					if(iY==(_cpcImageHeight-1))
					{
						byte backgroundMask = (byte)(Convert.ToByte(_globalSettings.BackgroundMask, 2));

						edgeInfo = new BobsEdgeInfo();
						edgeInfo.LeftPtr = (byte)leftPtr;
						edgeInfo.LeftCpcByte = backgroundMask;
						edgeInfo.RightPtr = (byte)rightPtr;
						edgeInfo.RightCpcByte = backgroundMask;
						EdgeInfos.Add(edgeInfo);
					}
                }

                debugTxt += "[" + iY + "] ";
                int changeCount = 0;

                var NOPS = 32;

                var asmCodeLeftPtr = "";
                var asmCodeRightPtr = "";

                if (iY < (_cpcImageHeight / 2))
                {
                    for (int iX = staticXLeft; iX >= 0; iX--)
                    {
                        byte cpcByte = _cpcImage[(iY * _cpcImageByteWidth) + iX];
                        
                        if (_line[iX] != cpcByte)
                        {
                            _line[iX] = cpcByte;
                            debugTxt += iX.ToString("X2") + "=" + cpcByte.ToString("X2") + "; ";
                            changeCount++;

                            if (iX != leftPtr)
                            {
                                if (iX == (leftPtr - 1))
                                {
                                    asmCodeLeftPtr += "\tdec l ; 1\n"; NOPS -= 1;
                                }
                                else if (iX == (leftPtr + 1))
                                {
                                    asmCodeLeftPtr += "\tinc l ; 1\n"; NOPS -= 1;
                                }
                                else
                                {
                                    System.Diagnostics.Debugger.Break();
                                }

                                leftPtr = iX;
                            }

                            if (cpcByte == _clearByte)
                            {
                                asmCodeLeftPtr += "\tld (hl), c ; 2\n"; NOPS -= 2;
                            }
                            else if (cpcByte == _fillByte)
                            {
                                asmCodeLeftPtr += "\tld (hl), h ; 2\n"; NOPS -= 2;
                            }
                            else
                            {
                                asmCodeLeftPtr += "\tld (hl), &" + cpcByte.ToString("X2") + " ; 3\n"; NOPS -= 3;
                            }
                        }
                    }

                    for (int iX = staticXRight; iX < _cpcImageByteWidth; iX++)
                    {
                        byte cpcByte = _cpcImage[(iY * _cpcImageByteWidth) + iX];

                        if (_line[iX] != cpcByte)
                        {
                            _line[iX] = cpcByte;
                            debugTxt += iX.ToString("X2") + "=" + cpcByte.ToString("X2") + "; ";
                            changeCount++;

                            if (iX != rightPtr)
                            {
                                if (iX == (rightPtr - 1))
                                {
                                    asmCodeRightPtr += "\tdec l ; 1\n"; NOPS -= 1;
                                }
                                else if (iX == (rightPtr + 1))
                                {
                                    asmCodeRightPtr += "\tinc l ; 1\n"; NOPS -= 1;
                                }
                                else
                                {
                                    System.Diagnostics.Debugger.Break();
                                }

                                rightPtr = iX;
                            }

                            if (cpcByte == _clearByte)
                            {
                                asmCodeRightPtr += "\tld (hl), c ; 2\n"; NOPS -= 2;
                            }
                            else if (cpcByte == _fillByte)
                            {
                                asmCodeRightPtr += "\tld (hl), h ; 2\n"; NOPS -= 2;
                            }
                            else
                            {
                                asmCodeRightPtr += "\tld (hl), &" + cpcByte.ToString("X2") + " ; 3\n"; NOPS -= 3;
                            }
                        }
                    }
                }
                else
                {
                    for (int iX = 0; iX <= staticXLeft; iX++)
                    {
                        byte cpcByte = _cpcImage[(iY * _cpcImageByteWidth) + iX];

                        if (_line[iX] != cpcByte)
                        {
                            _line[iX] = cpcByte;
                            debugTxt += iX.ToString("X2") + "=" + cpcByte.ToString("X2") + "; ";
                            changeCount++;

                            if (iX != leftPtr)
                            {
                                if (iX == (leftPtr - 1))
                                {
                                    asmCodeLeftPtr += "\tdec l ; 1\n"; NOPS -= 1;
                                }
                                else if (iX == (leftPtr + 1))
                                {
                                    asmCodeLeftPtr += "\tinc l ; 1\n"; NOPS -= 1;
                                }
                                else
                                {
                                    System.Diagnostics.Debugger.Break();
                                }

                                leftPtr = iX;
                            }

                            if (cpcByte == _clearByte)
                            {
                                asmCodeLeftPtr += "\tld (hl), c ; 2\n"; NOPS -= 2;
                            }
                            else if (cpcByte == _fillByte)
                            {
                                asmCodeLeftPtr += "\tld (hl), h ; 2\n"; NOPS -= 2;
                            }
                            else
                            {
                                asmCodeLeftPtr += "\tld (hl), &" + cpcByte.ToString("X2") + " ; 3\n"; NOPS -= 3;
                            }
                        }
                    }

                    for (int iX = _cpcImageByteWidth - 1; iX >= staticXRight; iX--)
                    {
                        byte cpcByte = _cpcImage[(iY * _cpcImageByteWidth) + iX];

                        if (_line[iX] != cpcByte)
                        {
                            _line[iX] = cpcByte;
                            debugTxt += iX.ToString("X2") + "=" + cpcByte.ToString("X2") + "; ";
                            changeCount++;

                            if (iX != rightPtr)
                            {
                                if (iX == (rightPtr - 1))
                                {
                                    asmCodeRightPtr += "\tdec l ; 1\n"; NOPS -= 1;
                                }
                                else if (iX == (rightPtr + 1))
                                {
                                    asmCodeRightPtr += "\tinc l ; 1\n"; NOPS -= 1;
                                }
                                else
                                {
                                    System.Diagnostics.Debugger.Break();
                                }

                                rightPtr = iX;
                            }

                            if (cpcByte == _clearByte)
                            {
                                asmCodeRightPtr += "\tld (hl), c ; 2\n"; NOPS -= 2;
                            }
                            else if (cpcByte == _fillByte)
                            {
                                asmCodeRightPtr += "\tld (hl), h ; 2\n"; NOPS -= 2;
                            }
                            else
                            {
                                asmCodeRightPtr += "\tld (hl), &" + cpcByte.ToString("X2") + " ; 3\n"; NOPS -= 3;
                            }
                        }
                    }
                }

                asmCode += "; LINE " + iY + "\n";
                //asmCode += "\tORG SCANLINES + &" + iY.ToString("X2") + "00 + SCANLINES_ALIGN\n";
                asmCode += this._humanFilename + "_" + iY + ":\n";

                if (!_isCircle2)
                {
                    asmCode += "\tout (c), a ; 4\n"; NOPS -= 4;
                }

                asmCode += "\texx ; 1\n"; NOPS -= 1;

                if ((iY & 1) == 0)
                {
                    asmCode += asmCodeLeftPtr;
                    asmCode += "\tex de, hl ; 1\n"; NOPS -= 1;
                    asmCode += asmCodeRightPtr;
                }
                else
                {
                    asmCode += asmCodeRightPtr;
                    asmCode += "\tex de, hl ; 1\n"; NOPS -= 1;
                    asmCode += asmCodeLeftPtr;
                }

                if (!_isCircle2)
                {
                    if (iY < (_cpcImageHeight-1))
                    {
                        asmCode += "\tld ix, " + this._humanFilename + "_" + iY + "_END ; 4\n"; NOPS -= 4;
                    }
                    else
                    {
                        asmCode += "\tld ix, SCANLINE_IDLE1 ; 4\n"; NOPS -= 4;
                    }
                }

				var asmCodeIYManagement = "";
                if (_isCircle2)
                {                    
					asmCodeIYManagement += "\tld a, (iy) ; 5\n"; NOPS -= 5;
                    asmCodeIYManagement += "\tinc iy ; 3\n"; NOPS -= 3;
                }

                var asmCodeEnd = "";
                int lastInstructionNops = 0;
                int minNopsForJump = 0;
                if (_isCircle2)
                {
                    minNopsForJump = 5; // jp JPIXWaitNops
					
					asmCodeEnd += asmCodeIYManagement;
                    asmCodeEnd += "\tjp (ix) ; 2\n"; NOPS -= 2; lastInstructionNops = 2;        
					asmCodeEnd += this._humanFilename + "_" + iY + "_END:\n";   
                }
                else
                {
                    minNopsForJump = 6; // jp JPRETWaitNops
                    asmCodeEnd += "\tret ; 3\n"; NOPS -= 3; lastInstructionNops = 3;
                }

                if (minNOPS > NOPS)
                {
                    minNOPS = NOPS;
                }

                if (NOPS >= 0)
                {
                    if (NOPS >= (minNopsForJump - lastInstructionNops)) // jp (IX)/jp (IY)  or  RET
                    {
                        NOPS += lastInstructionNops;
                        if (_isCircle2)
                        {
                            asmCode += "\tjp Bobs_JPIXWaitNops" + NOPS + " ; " + NOPS + "\n"; NOPS -= NOPS;														
							asmCode += this._humanFilename + "_" + iY + "_END:\n";         
                        }
                        else
                        {
                            asmCode += "\tjp WaitNops_GLOBAL_End - " + (NOPS-6) + " ; " + NOPS + "\n"; NOPS -= NOPS;
                        }
                    }
                    else if (NOPS>0)
                    {
                        asmCode += "\tds " + NOPS + ", 0 ; " + NOPS + "\n"; NOPS -= NOPS;
                        asmCode += asmCodeEnd;
                    }
                    else if (NOPS==0)
                    {
                        asmCode += asmCodeEnd;
                    }
                    else
                    {
                        System.Diagnostics.Debugger.Break();
                    }
                }
                else
                {
                    System.Diagnostics.Debugger.Break();
                }

                if (!_isCircle2)
                {
                    asmCode += this._humanFilename + "_" + iY + "_END:\n";
                }
                asmCode += "\n";
                
                debugTxt += "[" + changeCount + "]\n";
            }

            var finalAsmCode = ";\n";
            finalAsmCode += "; Scanline max. " + (32 - minNOPS) + "\n\n";

            byte fillByteCPC = (byte)this._cpcImage[0];// (Convert.ToByte("10111010", 2));
            //byte fillByteCPC = this._fillByte;

            finalAsmCode += "LET " + this._humanFilename + "_REGB\t= &7F\t\t; GA port\n";
            finalAsmCode += "LET " + this._humanFilename + "_REGC\t= &" + fillByteCPC.ToString("X2") + "\t\t; fill byte\n";
            finalAsmCode += "LET " + this._humanFilename + "_REGBC\t= &7F" + fillByteCPC.ToString("X2") + "\t; GA port + fill byte\n";
            finalAsmCode += "LET " + this._humanFilename + "_REGHL\t= &" + staticXLeft.ToString("X2") + "\t\t; left offset, H=0\n";
            finalAsmCode += "LET " + this._humanFilename + "_REGDE\t= &" + staticXRight.ToString("X2") + "\t\t; right offset, D=0\n";
            finalAsmCode += "\n";

            finalAsmCode += asmCode;

            /*if (_isCircle2)
            {
                circle2Offset += this._humanFilename + "_CLIPLIST_TOP_EDGES:\n";
                for (int i = 0; i < _globalSettings.CpcScreenHeight; i++)
                {
                    circle2Offset += leftRightEdges[0];
                }
                circle2Offset += this._humanFilename + "_EDGEVALUES:\n";
                foreach (var s in leftRightEdges)
                {
                    circle2Offset += s;
                }
                circle2Offset += this._humanFilename + "_CLIPLIST_BOTTOM_EDGES:\n";
                for (int i = 0; i < _globalSettings.CpcScreenHeight; i++)
                {
                    circle2Offset += leftRightEdges[0];
                }
                
                circle2Offset += "\n";
                finalAsmCode += circle2Offset;
            }*/

            if(_humanFilename.Contains("X0")||_humanFilename.Contains("X2"))
			{
				File.WriteAllText(this._globalSettings.TempFolder + _humanFilename + "_diff.txt", debugTxt);
				File.WriteAllText(this._globalSettings.WinAPEFolder + "_" + _humanFilename + ".z80", finalAsmCode);	
			}

            return true;
        }
    }
	
	public class CreateCPCTables
    {
        private BobsGlobalSettings _globalSettings;
        private int _height;

        public CreateCPCTables(BobsGlobalSettings globalSettings, int height)
        {
            _globalSettings = globalSettings;
            _height = height;
        }

        private void WriteClippingFile()
        {
            for (int iX = 0; iX < 4; iX++)
            {
				if((iX!=0)&&(iX!=2))
				{
					continue;
				}
				
				var content = "";

				/*content += "; " + _globalSettings.CpcScreenHeight + " lines\n";
                content += "X" + iX + "P2_CLIPLIST_TOP:\n";
                for (int iY = 0; iY < _globalSettings.CpcScreenHeight; iY++)
                {
                    content += "\tdw SCANLINE_IDLE2\n";
                }*/
				
				content += "\n;; auto-generated here: 270 lines of SCANLINE_IDLE2 (used for clipping)\n";
				content += ";; warning: also another one declared end of this source !\n";

                content += "\n; " + _height + " lines\n";
                content += "X" + iX + "P2_FRAMELIST_INTERMEDIATE:\n";
                for (int iY = 0; iY < _height; iY++)
                {
                    content += "\tdb X" + iX + "P2_" + iY + "_END - X" + iX + "P2_" + iY + "\n";
                }
				content += "\tdb 0 ; terminator\n";

                /*content += "\n; " + _globalSettings.CpcScreenHeight + " lines\n";
                content += "X" + iX + "P2_CLIPLIST_BOTTOM:\n";
                for (int iY = 0; iY < _globalSettings.CpcScreenHeight; iY++)
                {
                    content += "\tdw SCANLINE_IDLE2\n";
                }*/
				content += "\n;; auto-generated here: 270 lines of SCANLINE_IDLE2 (used for clipping)\n";

				content += "\n";
				
				File.WriteAllText(this._globalSettings.WinAPEFolder + "_X" + iX + "P2_RET.z80", content);
            }
        }

        private void WriteBobPosFile()
        {
            var content = ";\n";

            content += "\n";

			content += "BACKGROUND_MASK\tequ %"+_globalSettings.BackgroundMask+"\n";
			content += "\n";
            
            //Random random = new Random();

            //int nbFrames = 50*13; // in seconds
            int nbFrames = 320; // in seconds

            bool isFirstLine = true;

            var center = 114.0d;
            var distance = 135.0d;
            bool hasTouchedCenter = false;
            var defaultIncCoef = 0.095d;
            var incrementCoef = defaultIncCoef;
            var d = 0.0d;
            bool isTurningReverse = false;
            int iWaitFrame = 65555;
            var distanceInc = 0.0d;

            int iBeatCount = 0;

            for (int i = 0; i < nbFrames; i++)
            {
                const int beatInterval = 5 * 4;
                //float bounce = 0;
                int bounceIndex = i;
                while(bounceIndex >= beatInterval)
                {
                    bounceIndex -= beatInterval;
                }
                if(bounceIndex==0)
                {
                    //bounce += 0.07f*(float)Math.Cos((double)(bounceIndex)*0.05f);
                    iBeatCount++;
                }

                d += incrementCoef;

                if ((iBeatCount & 1) == 1)
                {
                 //   d += 0.1f;
                 //.. to be implemented
                }

                var iDistance = center + distance;
                distance -= 1;
                if(distance<0)
                {
                    distance = 0;
                    hasTouchedCenter = true;
                }
                if(hasTouchedCenter)
                {
                    incrementCoef -= 0.0005f; // anim length for 320 frames


                    if(incrementCoef<-defaultIncCoef)
                    {
                        incrementCoef = -defaultIncCoef;
                        isTurningReverse = true;
                        if (iWaitFrame==65555)
                        {
                            iWaitFrame = i + 200;
                        }
                    }
                }
                if(isTurningReverse && (i>iWaitFrame))
                {
                    distance += 1; // reverse minus
                    distanceInc += 0.10f;
                    distance += distanceInc;
                    incrementCoef += 0.0004f;
                }

                //iDistance *= bounce;
                //d += bounce;

                var x1 = Math.Sin(d) * iDistance;
                var y1 = Math.Cos(d) * iDistance;
                var x2 = -Math.Sin(d) * iDistance;
                var y2 = -Math.Cos(d) * iDistance;

                // center on screen
                x1 += (_globalSettings.CpcScreenWidth / 4);
                x2 += (_globalSettings.CpcScreenWidth / 4);
                y1 += (_globalSettings.CpcScreenHeight / 2);
                y2 += (_globalSettings.CpcScreenHeight / 2);

                // center ball
                x1 -= (_globalSettings.BobWidth)/4;
                y1 -= (_globalSettings.BobHeight) / 2;
                x2 -= (_globalSettings.BobWidth)/4;
                y2 -= (_globalSettings.BobHeight) / 2;

                if(x2>(x1-16)) // -16 to center artefact effect in raster scanline
                {
                    var t = x1;
                    x1 = x2;
                    x2 = t; 
                    
                    t = y1;
                    y1 = y2;
                    y2 = t;
                }

                var y1Old = y1;
				
				// left/right clamp
				if (x1 < -_globalSettings.ClippingLeftRightWidth/2)
                {
					x1 = -_globalSettings.ClippingLeftRightWidth/2;
                }
                else if (x1 > _globalSettings.CpcScreenWidth/2)
                {
                    x1 = _globalSettings.CpcScreenWidth/2;
                }
                if (x2 < -_globalSettings.ClippingLeftRightWidth/2)
                {
					x2 = -_globalSettings.ClippingLeftRightWidth/2;
                }
				else if (x2 > (_globalSettings.CpcScreenWidth/2))
                {
					x2 = (_globalSettings.CpcScreenWidth/2);
                }

                if (y1 <= -_globalSettings.BobHeight)
                {
                    y1 = _globalSettings.CpcScreenHeight; // not an error
                }
                if (y2 < -_globalSettings.BobHeight)
                {
                    y2 = -_globalSettings.BobHeight;
                }
                if (y1 >= _globalSettings.CpcScreenHeight)
                {
                    y1 = _globalSettings.CpcScreenHeight;
                }
                if (y2 >= _globalSettings.CpcScreenHeight)
                {
                    y2 = _globalSettings.CpcScreenHeight;
                }
                
                byte Y1_CLIPSTART = (byte)(0);
                if(y1 <= 1)
                {
                	Y1_CLIPSTART = (byte)(-y1);
                }
                
				byte Y2_CLIPSTART = (byte)(0);
                if (y2 <= 1)
                {
                	Y2_CLIPSTART = (byte)(-y2);
                }

                byte BITPLAN_X1 = (byte) ((((ushort)(x1))) & 3);
                byte BITPLAN_X2 = (byte) ((((ushort)(x2))) & 3);

                if(BITPLAN_X1==1)
                {
                    BITPLAN_X1 = 0;
                }
                if (BITPLAN_X1 == 3)
                {
                    BITPLAN_X1 = 2;
                } 
                if (BITPLAN_X2 == 1)
                {
                    BITPLAN_X2 = 0;
                }
                if (BITPLAN_X2 == 3)
                {
                    BITPLAN_X2 = 2;
                }

				byte CPC_VISIBLEPTR = (byte)(0x72);

				var edgeInfos1 = _globalSettings.BitmapEdgeInfos[BITPLAN_X1].EdgeInfos[Y1_CLIPSTART];
				byte X1_LEFT_VIDEO_PTR = (byte)((byte)(((ushort)(x1)) / 4) + (byte)(edgeInfos1.LeftPtr + CPC_VISIBLEPTR));
				byte X1_RIGHT_VIDEO_PTR = (byte)((byte)(((ushort)(x1)) / 4) + (byte)(edgeInfos1.RightPtr + CPC_VISIBLEPTR));

                var edgeInfos2 = _globalSettings.BitmapEdgeInfos[BITPLAN_X2].EdgeInfos[Y2_CLIPSTART];
				byte X2_LEFT_VIDEO_PTR = (byte)((byte)(((ushort)(x2)) / 4) + (byte)(edgeInfos2.LeftPtr + CPC_VISIBLEPTR));
				byte X2_RIGHT_VIDEO_PTR = (byte)((byte)(((ushort)(x2)) / 4) + (byte)(edgeInfos2.RightPtr + CPC_VISIBLEPTR));

                byte TOP_CLIP_LEFT_POS = (byte)96;
                byte TOP_CLIP_LEFT_VALUE = (byte)0;
                byte TOP_CLIP_RIGHT_POS = (byte)96;
                byte TOP_CLIP_RIGHT_VALUE = (byte)0;

                if(y1<0)
                {
					TOP_CLIP_LEFT_POS = (byte)(edgeInfos1.LeftPtr);
					TOP_CLIP_LEFT_VALUE = (byte)(edgeInfos1.LeftCpcByte);
					TOP_CLIP_RIGHT_POS = (byte)(edgeInfos1.RightPtr);
					TOP_CLIP_RIGHT_VALUE = (byte)(edgeInfos1.RightCpcByte);

					byte bX = (byte)(((ushort)(x1)) / 4);
					TOP_CLIP_LEFT_POS += bX;
					TOP_CLIP_RIGHT_POS += bX;

					if((TOP_CLIP_LEFT_POS>=96)&&(TOP_CLIP_RIGHT_POS>=96))
					{
						TOP_CLIP_LEFT_POS = (byte)96;
                        TOP_CLIP_RIGHT_POS = (byte)96;
					}
					else if(((TOP_CLIP_LEFT_POS)>=(_globalSettings.CpcScreenWidth/8))&&(TOP_CLIP_RIGHT_POS>=(_globalSettings.CpcScreenWidth/8)))
					{
						TOP_CLIP_LEFT_POS = (byte)96;
						TOP_CLIP_RIGHT_POS = (byte)96;
					}
					else
					{
						if((TOP_CLIP_LEFT_POS>=(_globalSettings.CpcScreenWidth/8))&&(TOP_CLIP_RIGHT_POS<(_globalSettings.CpcScreenWidth/8)))
						{
							TOP_CLIP_LEFT_POS = (byte)0;
							TOP_CLIP_LEFT_VALUE = (byte)0; //edgeInfos1.FillCpcByte;
						}
						if((TOP_CLIP_LEFT_POS<(_globalSettings.CpcScreenWidth/8))&&(TOP_CLIP_RIGHT_POS>=(_globalSettings.CpcScreenWidth/8)))
						{
							TOP_CLIP_RIGHT_POS = (byte)((_globalSettings.CpcScreenWidth/8));
							TOP_CLIP_RIGHT_VALUE = (byte)0; //edgeInfos1.FillCpcByte;
						}
					}
                }
                else if(y2<0)
				{
					TOP_CLIP_LEFT_POS = (byte)(edgeInfos2.LeftPtr);
					TOP_CLIP_LEFT_VALUE = (byte)(edgeInfos2.LeftCpcByte);
					TOP_CLIP_RIGHT_POS = (byte)(edgeInfos2.RightPtr);
					TOP_CLIP_RIGHT_VALUE = (byte)(edgeInfos2.RightCpcByte);

					byte bX = (byte)(((ushort)(x2)) / 4);
					TOP_CLIP_LEFT_POS += bX;
					TOP_CLIP_RIGHT_POS += bX;

					if(Y2_CLIPSTART==_globalSettings.BobHeight)
					{
						TOP_CLIP_LEFT_POS = (byte)96;
						TOP_CLIP_RIGHT_POS = (byte)96;
					}
					else if((TOP_CLIP_LEFT_POS>=96)&&(TOP_CLIP_RIGHT_POS>=96))
					{
						TOP_CLIP_LEFT_POS = (byte)96;
						TOP_CLIP_RIGHT_POS = (byte)96;
					}
					else if(((TOP_CLIP_LEFT_POS)>=(_globalSettings.CpcScreenWidth/8))&&(TOP_CLIP_RIGHT_POS>=(_globalSettings.CpcScreenWidth/8)))
					{
						TOP_CLIP_LEFT_POS = (byte)96;
						TOP_CLIP_RIGHT_POS = (byte)96;
					}
					else
					{
						if((TOP_CLIP_LEFT_POS>=(_globalSettings.CpcScreenWidth/8))&&(TOP_CLIP_RIGHT_POS<(_globalSettings.CpcScreenWidth/8)))
						{
							TOP_CLIP_LEFT_POS = (byte)0;
							TOP_CLIP_LEFT_VALUE = (byte)0; //edgeInfos2.FillCpcByte;
						}
						if((TOP_CLIP_LEFT_POS<(_globalSettings.CpcScreenWidth/8))&&(TOP_CLIP_RIGHT_POS>=(_globalSettings.CpcScreenWidth/8)))
						{
							TOP_CLIP_RIGHT_POS = (byte)((_globalSettings.CpcScreenWidth/8));
							TOP_CLIP_RIGHT_VALUE = (byte)0; //edgeInfos2.FillCpcByte;
						}
					}
                }

				//TOP_CLIP_LEFT_POS = TOP_CLIP_RIGHT_POS = (byte)96;

				var Y1_JPLABEL = "";
				var Y1_JPLABEL_TOPIDLE = "";
                ushort Y1_WAITLINES = 0;
				if(y1==0)
				{
					Y1_JPLABEL = "X" + BITPLAN_X1 + "P1_0";
					Y1_JPLABEL_TOPIDLE = "&0000";
					Y1_WAITLINES = 0;
				}
				else if (y1 >= _globalSettings.CpcScreenHeight)
                {
					Y1_JPLABEL = "SCANLINE_IDLE1";
					Y1_JPLABEL_TOPIDLE = "&0000";
					Y1_WAITLINES = 0;
				}
				else if (y1 >= 1)
				{
					Y1_JPLABEL = "SCANLINE_TOPIDLE1";
					Y1_WAITLINES = (ushort)(y1);
					Y1_JPLABEL_TOPIDLE = "X" + BITPLAN_X1 + "P1_0";
				}
				else if (y1 < 1)
				{
					Y1_JPLABEL = "X" + BITPLAN_X1 + "P1_" + (ushort)(-y1);
					Y1_WAITLINES = 0;
					Y1_JPLABEL_TOPIDLE = "&0000";
				}

                var Y2_FRAMELISTSTART = "X" + BITPLAN_X2 + "P2_FRAMELIST";
                if (y2 < 0)
                {
                    Y2_FRAMELISTSTART += "+" + (((ushort)-y2)*2);
                }
				else if (y2>0)
                {
                    Y2_FRAMELISTSTART += "-" + (((ushort)y2)*2);
                }

                if(isFirstLine)
                {
                    isFirstLine = false;

                    content += ";\n";
                    content += "ANIM_FRAME_COUNT\t\tequ " + nbFrames + " ; " + (nbFrames / 50) + " seconds\n";
                    content += "ANIM_FRAME_CHUNKSIZE\tequ " + (1+1+1+1+2+2+2+2+1+1+1+1) +" ; MUST BE POWER 2 (INC L/INC HL)\n";

                    content += "\n";
                    content += ";\n";
                    content += "\tALIGN 2\n";
                    content += "ANIM_FRAMES:\n";
                }

                content += "ANIM_FRAME_" + i + ":\n";

				// apply reverse EX DE,HL if odd Y line (z80 optimization)
				if(y1<=0)
				{
					if((((ushort)y1)&1)==1)
					{
						byte b = 0;
						b=X1_LEFT_VIDEO_PTR; X1_LEFT_VIDEO_PTR=X1_RIGHT_VIDEO_PTR; X1_RIGHT_VIDEO_PTR=b;
					}
				}
				else if(y2<=0)
				{
					if((((ushort)y2)&1)==1)
					{
						byte b = 0;
						b=X2_RIGHT_VIDEO_PTR; X2_RIGHT_VIDEO_PTR=X2_LEFT_VIDEO_PTR; X2_LEFT_VIDEO_PTR=b;
					}
				}

                content += "\t" + "db &" + X1_LEFT_VIDEO_PTR.ToString("X2")
                                + ":db &" + X1_RIGHT_VIDEO_PTR.ToString("X2")
                                + ":db &" + X2_LEFT_VIDEO_PTR.ToString("X2")
                                + ":db &" + X2_RIGHT_VIDEO_PTR.ToString("X2");

                if (Y1_JPLABEL == "SCANLINE_TOPIDLE1")
                {
                    content += ":dw " + Y1_JPLABEL // SCANLINE_TOPIDLE1 or X2P1_34, etc.
                                + ":dw " + Y1_JPLABEL_TOPIDLE
                                + ":dw &" + Y1_WAITLINES.ToString("X4"); // to set to SCANLINE_TOPIDLE1_WaitLines
                }
                else
                {
                    content += ":dw " + Y1_JPLABEL; // SCANLINE_TOPIDLE1
                }

                content += ":dw " + Y2_FRAMELISTSTART; // Stack-based (RET) jumps..

                if ((TOP_CLIP_LEFT_POS == (byte)96) && (TOP_CLIP_RIGHT_POS == (byte)96))
                {
                    content += ":db &FF";
                }
                else
                {
                    content += ":db &" + TOP_CLIP_LEFT_POS.ToString("X2")
                                + ":db &" + TOP_CLIP_LEFT_VALUE.ToString("X2")
                                + ":db &" + TOP_CLIP_RIGHT_POS.ToString("X2")
                                + ":db &" + TOP_CLIP_RIGHT_VALUE.ToString("X2");
                }
                content += "\n";

                y1 = y1Old;
            }

            content += "\n";
            content += ";\n";
            content += "ANIM_FRAMES_END:\n";
            content += "\n";

            content += "ANIM_FRAMES_PTR:\n";
            for (int i = 0; i < nbFrames; i++)
            {
                content += "\t dw ANIM_FRAME_" + i + "\n";
            }

            File.WriteAllText(this._globalSettings.WinAPEFolder + "_AnimFrames.z80", content);
        }

        public void Start()
        {
            WriteClippingFile();
            WriteBobPosFile();
        }
    }
}
