using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _05_Chessboard
    {
		public static int StaticMaxIterations = 12; // -1 to ignore hardcoded max iterations. All 3 logos should get same max iterations.
        public const int NbAvailableNopsPerScanline = 35; // complete scanline including RET
        public const int LogoPosX = 6;
        public const byte CPCPixelON = 0xAF;
		
		public const int ZoomScrollWidth = 32;
        public const int ZoomScreenWidth = 96 + ZoomScrollWidth;
        public const int ZoomScrollHeight = 0;
        public const int ZoomScreenHeight = 272;
        public static int MoveYHeight = 63;
        public static int DamierZoomNbSteps = 40; // number of possible zoom iterations
        public static int DamierMinBlockSize = 32; // block size in pixels
        public static int DamierBlockSizeInc = 1; // how fast zoom increases
        public static int AnimNbFrames = 128;

        public _05_Chessboard()
        {
			try
			{
				CreateDamierRaster(
					Program.ToDataFolder("05_Chessboard/DamierRasters1.png"),
					Program.ToDataFolder("05_Chessboard/DamierRasters2.png"),
					Program.ToDataFolder("05_Chessboard/DamierLogoRasters1.png"),
					Program.ToZ80Folder("05_Chessboard/_DamierRasters1.png.z80") );
				
				CreateZ80Opcodes(
					Program.ToDataFolder("05_Chessboard/Image1.bmp"),				
					Program.ToZ80Folder("05_Chessboard/_ThisOpcodes.z80"),
					"THIS");
					
				CreateZ80Opcodes(
					Program.ToDataFolder("05_Chessboard/Image2.bmp"),				
					Program.ToZ80Folder("05_Chessboard/_IsOpcodes.z80"),
					"IS");
					            
				CreateZ80Opcodes(
					Program.ToDataFolder("05_Chessboard/Image3.bmp"),				
					Program.ToZ80Folder("05_Chessboard/_PhxOpcodes.z80"),
					"PHX");
					
				CreateSinPosXCurves(20 + 14 + 40, 20 - 14 - 40, 1, false, 1, false, false, 
					Program.ToZ80Folder("05_Chessboard/_LogoPosX.z80"));
				CreateSinCurves(255, 1, 2, false, 1, true, true, 
					Program.ToZ80Folder("05_Chessboard/_LogoPosY.z80"));
				CreateSinCurves(1, 80, 1, false, 2, false, false, 
					Program.ToZ80Folder("05_Chessboard/_LogoPosY_Centered.z80")); // when logo is sticked at center of screen

				var damierXSinOffset = CreateSinCurves(ZoomScrollWidth*8, (ZoomScrollWidth*12)-1, 1, false, 2, false, true, 
					Program.ToZ80Folder("05_Chessboard/_DamierXPos.z80"));
				var damierYSinOffset = CreateSinCurvesDamierY(0, MoveYHeight, 1, false, 2, false, true, 
					Program.ToZ80Folder("05_Chessboard/_DamierYPos.z80"));
				var damierXZoomValues = CreateSinCurves(0, DamierZoomNbSteps - 1, 1, false, 2, false, false, 
					Program.ToZ80Folder("05_Chessboard/_DamierXZoom.z80"));

				CreateZoom(DamierZoomNbSteps, DamierMinBlockSize, DamierBlockSizeInc, damierXZoomValues, damierXSinOffset, damierYSinOffset,
					Program.ToZ80Folder("05_Chessboard/_HorizontalSteps.z80"),
					Program.ToZ80Folder("05_Chessboard/_VerticalSteps.z80"),
					Program.ToZ80Folder("05_Chessboard/_HorizontalSteps.bin")
					);
			}
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}					
        }
		
		private void CreateDamierRaster(string imageFilename1, string imageFilename2, string logoRasterFilename1, string destZ80)
        {
            Bitmap bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            Bitmap bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));
            Bitmap bmp3 = new Bitmap(System.Drawing.Image.FromFile(logoRasterFilename1));

            var z80Content = "";

            //z80Content += "\tds 10, COLOR0\n";

            bool isFlipped = false;

            for (int y = 0; y < bmp1.Height; y++)
            {
                var pixel1 = bmp1.GetPixel(0, y);
                var palette1 = new Palette();
                palette1.AddColor(pixel1);

                var pixel2 = bmp2.GetPixel(0, y);
                var palette2 = new Palette();
                palette2.AddColor(pixel2);

                //var pixel3 = bmp3.GetPixel(0, y);
                var palette3 = new Palette();
                //palette3.AddColor(pixel3);

                var gateArrayColor1 = palette1.ToFirmwareByteArray()[0];
                var gateArrayColor2 = palette2.ToFirmwareByteArray()[0];
                //var gateArrayColor3 = palette3.ToFirmwareByteArray()[0];

                if(!isFlipped)
                {
                    z80Content += "\tdb COLOR" + gateArrayColor1 + ", COLOR" + gateArrayColor2 + "\n";
                    isFlipped = true;
                }
                else
                {
                    z80Content += "\tdb COLOR" + gateArrayColor2 + ", COLOR" + gateArrayColor1 + "\n";
                    isFlipped = false;
                }
            }

            File.WriteAllText(destZ80, z80Content);
        }
		
		private enum BlockOperation
        {
            WritePixelON, // pixel switched to ON (cf. write PEN2 ScreenPtr0/1)
            WritePixelOFF // pixel switched back to OFF (cf. read from ScreenPtr2, write PEN2 ScreenPtr0/1)
        }
        private class Block
        {
            public int BlockIndex = 0;
            public Dictionary<int, BlockOperation> PixelWrites = new Dictionary<int, BlockOperation>();
        }

        // parse logo
        // compute nb pixels changed for each lines in bitmap (= blocks)
        private List<Block> CreateBlocks(string bmpFilename)
        {
            var blocks = new List<Block>();

            var bitmap = new Bitmap(bmpFilename);

            var previousLine = new int[bitmap.Width];
            for (int i = 0; i < previousLine.Length; i++)
            {
                previousLine[i] = 0; // all pixel off
            }

            for (int y = 0; y < bitmap.Height; y++)
            {
                var block = new Block();
                block.BlockIndex = blocks.Count;

                var line = new int[bitmap.Width];
                for (int x = 0; x < bitmap.Width; x++)
                {
                    int pixelON = (bitmap.GetPixel(x, y).R == 255) ? 0 : 1;
                    line[x] = pixelON;

                    if ((line[x] == 1) && (previousLine[x] == 0))
                    {
                        block.PixelWrites.Add(x, BlockOperation.WritePixelON);
                    }
                    else if ((line[x] == 0) && (previousLine[x] == 1))
                    {
                        block.PixelWrites.Add(x, BlockOperation.WritePixelOFF);
                    }
                }

                previousLine = line;
                blocks.Add(block);
            }

            return blocks;
        }

        public string CommandPrefix = "";

        private string CreateBlockFunctionName(int iBlock, int iScreen, int iIteration)
        {
            return "CMD_" + CommandPrefix + "_DrawBlock" + iBlock + "_Screen" + iScreen + "_Iteration" + iIteration;
        }

        //private int Nops = 0;
        private int LRegister = -10;
        private int ERegister = -10;

        private void InitRegisters()
        {
            LRegister = -10;
            ERegister = -10;
        }

        public class TimeOpcode
        {
            public string Z80 = "";
            public int Nops = 0;
            public int Iteration = -1;

            public TimeOpcode(string z80, int _Nops)
            {
                Z80 = "\t" + z80 + " ; " + _Nops + "\n";
                Nops = _Nops;
            }
        }

        public class Iteration
        {
            public int iBlock = -1;
            public int iScreen = -1;
            public int iIteration = -1;

            public List<TimeOpcode> TimeOpcodes = new List<TimeOpcode>();

            public Iteration(int block, int screen, int iteration)
            {
                iBlock = block;
                iScreen = screen;
                iIteration = iteration;
            }
        }

        // generate all cumulated opcodes necessary to draw 1 block
        private List<TimeOpcode> CreateRawOpcodes(int posX, BlockOperation blockOperation)
        {
            var opcodes = new List<TimeOpcode>();

            if (LRegister == (LogoPosX + posX - 1))
            {
                opcodes.Add(new TimeOpcode("inc l", 1));
                LRegister = LogoPosX + posX;
            }
            else
            {
                opcodes.Add(new TimeOpcode("ld a, c", 1));
                opcodes.Add(new TimeOpcode("add a, " + (LogoPosX + posX).ToString(), 2));
                opcodes.Add(new TimeOpcode("ld l, a", 1));
                LRegister = LogoPosX + posX;
            }

            if (blockOperation == BlockOperation.WritePixelON)
            {
                if (ERegister != (int)CPCPixelON)
                {
                    opcodes.Add(new TimeOpcode("ld e, &" + CPCPixelON.ToString("X2"), 2));
                    ERegister = (int)CPCPixelON;
                }

                opcodes.Add(new TimeOpcode("ld (hl), e", 2));
            }
            else if (blockOperation == BlockOperation.WritePixelOFF)
            {
                opcodes.Add(new TimeOpcode("ld e, l", 1));
                ERegister = LogoPosX + posX;

                opcodes.Add(new TimeOpcode("ld a, (de)", 2));
                opcodes.Add(new TimeOpcode("ld (hl), a", 2));
            }

            return opcodes;
        }

        // generate all cumulated opcodes necessary to draw 1 block
        private List<TimeOpcode> CreateRawBlockOpcodes(Block block)
        {
            var blockOpcodes = new List<TimeOpcode>();
            foreach (var pixelWrite in block.PixelWrites)
            {
                int posX = pixelWrite.Key;
                var blockOperation = pixelWrite.Value;

                var pixelOpcodes = CreateRawOpcodes(posX, blockOperation);
                blockOpcodes.AddRange(pixelOpcodes);
            }

            return blockOpcodes;
        }

        private int GetIterationCount(Block block)
        {
            TimeOpcode timeOpcode = null;
            int remainingNops = 0;
            int iIteration = 0;

            for (int iScreen = 0; iScreen < 2; iScreen++)
            {
                InitRegisters();

                var blockOpcodes = CreateRawBlockOpcodes(block);

                // same pattern for both screen 0 and 1
                blockOpcodes.Add(new TimeOpcode("ld ix, 0", 4));
                blockOpcodes.Add(new TimeOpcode("ld iy, 0", 4));
                blockOpcodes.Add(new TimeOpcode("ld l, 0", 2));
                blockOpcodes.Add(new TimeOpcode("ret", 3));

                // iterates from end to start
                // inject opcodes to split scanlines and pad Nops
                int iTimeOpcode = blockOpcodes.Count - 1;
                int Nops = 0;
                while (iTimeOpcode >= 0)
                {
                    timeOpcode = blockOpcodes[iTimeOpcode];

                    if ((Nops+timeOpcode.Nops) > NbAvailableNopsPerScanline)
                    {
                        blockOpcodes.Insert(iTimeOpcode+1, new TimeOpcode("ld ix, 0", 4));
                        blockOpcodes.Insert(iTimeOpcode+2, new TimeOpcode("ret", 3));

                        remainingNops = NbAvailableNopsPerScanline - Nops;
                        if (remainingNops > 0)
                        {
                            blockOpcodes.Insert(iTimeOpcode + 3, new TimeOpcode("_NOPS " + remainingNops, remainingNops));
                        }

                        iTimeOpcode += 2;

                        iIteration++;
                        Nops = 0;
                    }
                    else
                    {
                        Nops += timeOpcode.Nops;
                        iTimeOpcode--;
                    }
                }

                if(Nops>0)
                {
                    remainingNops = NbAvailableNopsPerScanline - Nops;
                    blockOpcodes.Insert(0, new TimeOpcode("_NOPS " + remainingNops, remainingNops));
                }
            }

            iIteration += 2; // for CMD_WaitY_ThenSet_DrawScreen1_ShowScreen0 / CMD_WaitY_ThenSet_DrawScreen0_ShowScreen1

            return iIteration;
        }

        // first pass: create non-optimized iterations
        // cf. scanline is 26 NOPS (13 Nops for draw / 4+4+2+3=13 Nops for IX+IY+L+RET)
        private List<Iteration> CreateIterations(Block block, bool isLastBlock, int maxIterations)
        {
            int NbIterations = GetIterationCount(block);

            var iterations = new List<Iteration>();

            TimeOpcode timeOpcode = null;
            int remainingNops = 0;
            int iIteration = NbIterations;

            for (int iScreen = 1; iScreen >= 0; iScreen--)
            {
                InitRegisters();

                var blockOpcodes = CreateRawBlockOpcodes(block);

                if(iScreen==0)
                {
                    blockOpcodes.Add(new TimeOpcode("ld ix, CMD_WaitY_ThenSet_DrawScreen1_ShowScreen0", 4));
                    blockOpcodes.Add(new TimeOpcode("ld iy, " + CreateBlockFunctionName(block.BlockIndex, 1, iIteration), 4));
                    blockOpcodes.Add(new TimeOpcode("ld l, 1", 2));
                    blockOpcodes.Add(new TimeOpcode("ret", 3));
                }
                else if (iScreen == 1)
                {
                    blockOpcodes.Add(new TimeOpcode("ld ix, CMD_WaitY_ThenSet_DrawScreen0_ShowScreen1", 4));
                    if(isLastBlock)
                    {
                        blockOpcodes.Add(new TimeOpcode("ld iy, CMD_ShowScreen2_Then_Idle", 4));
                    }
                    else
                    {
                        blockOpcodes.Add(new TimeOpcode("ld iy, " + CreateBlockFunctionName(block.BlockIndex+1, 0, 0), 4));
                    }
                    int waitLines = (maxIterations - NbIterations);
                    if(waitLines<0)
                    {
                        System.Diagnostics.Debugger.Break();
                    }
                    blockOpcodes.Add(new TimeOpcode("ld l, " + waitLines, 2));
                    blockOpcodes.Add(new TimeOpcode("ret", 3));
                }

                // iterates from end to start
                // inject opcodes to split scanlines and pad Nops
                int iTimeOpcode = blockOpcodes.Count - 1;
                int Nops = 0;
                while (iTimeOpcode >= 0)
                {
                    timeOpcode = blockOpcodes[iTimeOpcode];

                    if ((Nops + timeOpcode.Nops) > NbAvailableNopsPerScanline)
                    {
                        blockOpcodes.Insert(iTimeOpcode+1, new TimeOpcode("ld ix, " + CreateBlockFunctionName(block.BlockIndex, iScreen, iIteration-1), 4));
                        blockOpcodes.Insert(iTimeOpcode+2, new TimeOpcode("ret", 3));

                        remainingNops = NbAvailableNopsPerScanline - Nops;
                        if (remainingNops > 0)
                        {
                            blockOpcodes.Insert(iTimeOpcode + 3, new TimeOpcode("_NOPS " + remainingNops, remainingNops));
                        }

                        var iteration = new Iteration(block.BlockIndex, iScreen, iIteration - 1);
                        for(int iOpcodes=iTimeOpcode+3; iOpcodes < blockOpcodes.Count; iOpcodes++)
                        {
                            var timeOpcodeInIteration = blockOpcodes[iOpcodes];
                            iteration.TimeOpcodes.Add(timeOpcodeInIteration);
                        }
                        blockOpcodes.RemoveRange(iTimeOpcode + 3, blockOpcodes.Count - iTimeOpcode - 3);
                        iterations.Insert(0, iteration);
                        iIteration--;

                        iTimeOpcode+=2;

                        Nops = 0;
                    }
                    else
                    {
                        Nops += timeOpcode.Nops;
                        iTimeOpcode--;
                    }
                }

                if (Nops > 0)
                {
                    remainingNops = NbAvailableNopsPerScanline - Nops;
                    blockOpcodes.Insert(0, new TimeOpcode("_NOPS " + remainingNops, remainingNops));

                    var iteration = new Iteration(block.BlockIndex, iScreen, iIteration - 1);
                    for (int iOpcodes = 0; iOpcodes < blockOpcodes.Count; iOpcodes++)
                    {
                        var timeOpcodeInIteration = blockOpcodes[iOpcodes];
                        iteration.TimeOpcodes.Add(timeOpcodeInIteration);
                    }
                    iterations.Insert(0, iteration);
                    iIteration--;
                }
            }

            return iterations;
        }

        private void CreateZ80Opcodes(string bmpFilename, string cpcFilename, string prefix)
        {
            CommandPrefix = prefix;

            var blocks = CreateBlocks(bmpFilename);

            int maxIterations = -1;
            foreach (var block in blocks)
            {
                var nbIterations = GetIterationCount(block);
                Console.WriteLine("Iteration " + block.BlockIndex + ": " + nbIterations);
                if(maxIterations < nbIterations)
                {
                    maxIterations = nbIterations;
                }
            }
            if(StaticMaxIterations!=-1)
            {
                maxIterations = StaticMaxIterations;
            }
            maxIterations++;

            var z80 = "";

            foreach (var block in blocks)
            {
                var isLastBlock = false;
                if(block.BlockIndex==(blocks.Count-1))
                {
                    isLastBlock = true;
                }

                var iterations = CreateIterations(block, isLastBlock, maxIterations);

                foreach(var iteration in iterations)
                {
                    z80 += ";\n";
                    z80 += CreateBlockFunctionName(iteration.iBlock, iteration.iScreen, iteration.iIteration) + ":\n";
                    foreach(var timeOpcode in iteration.TimeOpcodes)
                    {
                        z80 += timeOpcode.Z80;
                    }
                    z80 += "\n";
                }
            }

            File.WriteAllText(cpcFilename, z80);
        }
		
		private class VerticalStepInfo
        {
            public byte FirstValue = 0;
            public List<int> YColorSwitch = new List<int>();
        }

        private void CreateReversedPixel(string cpcFilename)
        {
            string z80 = "";

            z80 += ";\n";
            // z80 += "ReversedPixels:\n";
            for (int i = 0; i < 256; i++)
            {
                byte v = (byte)(i);

                /*byte p0 = (byte)(v >> 7);
                byte p2 = (byte)(v >> 6);
                byte p4 = (byte)(v >> 5);
                byte p6 = (byte)(v >> 4);*/

                // read reversed order
                byte p6 = (byte)((v >> 7) & 1);
                byte p4 = (byte)((v >> 6) & 1);
                byte p2 = (byte)((v >> 5) & 1);
                byte p0 = (byte)((v >> 4) & 1);

                // also swap pixel on/off
                p0 = (byte)(1 - p0);
                p2 = (byte)(1 - p2);
                p4 = (byte)(1 - p4);
                p6 = (byte)(1 - p6);

                byte cpcPixel = CreatePixelCPC_Mode1(p0, p2, p4, p6);

                z80 += "\tdb &" + cpcPixel.ToString("X2") + " ; &" + v.ToString("X2") + "\n";
            }

            File.WriteAllText(cpcFilename, z80);
        }

        public class StepPixel
        {
            public int StepIndex = -1;
            public int IsPointingTo = -1; // if !=-1, this is index in StepPixels. Which means exact same content
            public List<byte> HorizontalLineZoom = new List<byte>();
            public List<byte> CPCPixels = new List<byte>();
        }

        private void CreateZoom(int nbSteps, int minBlockSize, int blockSizeInc, List<float> sinZoomX, List<float> sinOffsetX, List<float> sinOffsetY, string cpcHorizontalFilename, string cpcVerticalFilename, string cpcHorizontalFilenameBin)
        {
            var verticalStepInfos = new List<VerticalStepInfo>();

            var screenWidth = ZoomScreenWidth * 8; // must be power of 2
            
            var blockSize = minBlockSize;
            int maxColorSwitchCount = 0;

            var stepPixels = new List<StepPixel>();
            var horizontalStepPixels = new List<StepPixel>();

            int iStep;

            int iY;
            for (iStep = 0; iStep < nbSteps; iStep++)
            {
                var zoomPattern = new byte[screenWidth / 2];

                int count = 0;
                byte pixel = 1;
                for (int i = 0; i < screenWidth / 2; i++)
                {
                    zoomPattern[i] = pixel;

                    count++;
                    if (count == blockSize)
                    {
                        count = 0;

                        if (pixel == 0)
                        {
                            pixel = 1;
                        }
                        else
                        {
                            pixel = 0;
                        }
                    }
                }

                var horizontalLineZoom = new byte[screenWidth];
                for (int iX = 0; iX < screenWidth / 2; iX++)
                {
                    pixel = zoomPattern[iX];
                    byte reversedPixel = 0;
                    if (pixel == 0)
                    {
                        reversedPixel = 1;
                    }

                    horizontalLineZoom[(screenWidth / 2) + iX] = pixel;
                    horizontalLineZoom[(screenWidth / 2) - iX - 1] = reversedPixel;
                }

#region Vertical Zoom
                int YOverzone = 220;
                var heightZoom = new byte[ZoomScreenHeight + YOverzone];
                int yOffset = (screenWidth - ZoomScreenHeight) / 2;
                for (iY = 0; iY < ZoomScreenHeight + YOverzone; iY++)
                {
                    byte p = horizontalLineZoom[iY + yOffset];
                    heightZoom[iY] = p;
                }

                byte value = heightZoom[0];
                var stepInfo = new VerticalStepInfo();
                stepInfo.FirstValue = value;
                int colorSwitchCount = 0;
                for (iY = 0; iY < ZoomScreenHeight + YOverzone; iY++)
                {
                    byte iValue = heightZoom[iY];
                    if (iValue != value)
                    {
                        stepInfo.YColorSwitch.Add(iY);
                        value = iValue;
                    }

                    colorSwitchCount++;
                }
                verticalStepInfos.Add(stepInfo);
                if (maxColorSwitchCount < colorSwitchCount)
                {
                    maxColorSwitchCount = colorSwitchCount;
                }
                #endregion

                var stepPixel = new StepPixel();
                stepPixel.HorizontalLineZoom = new List<byte>(horizontalLineZoom);
                stepPixels.Add(stepPixel);

                blockSize += blockSizeInc;                
            }

            var z80_uniqueHorizSteps = ";\n";
            for (iStep = 0; iStep < AnimNbFrames; iStep++)
            {
                byte zoomValue = (byte)(sinZoomX[iStep]);
                var stepPixel = stepPixels[zoomValue];
                //int offsetX = ZoomScrollWidth*10;
                int offsetX = (int)sinOffsetX[iStep];

                //offsetX = (offsetX >> 2) << 2;

                var horizStepPixel = new StepPixel();
                horizStepPixel.StepIndex = iStep;

                for (int iX = 0; iX < 96; iX++)
                {
                    var p0 = stepPixel.HorizontalLineZoom[(iX * 4) + offsetX + 0];
                    var p1 = stepPixel.HorizontalLineZoom[(iX * 4) + offsetX + 1];
                    var p2 = stepPixel.HorizontalLineZoom[(iX * 4) + offsetX + 2];
                    var p3 = stepPixel.HorizontalLineZoom[(iX * 4) + offsetX + 3];

                    byte cpcPixel = CreatePixelCPC_Mode1(p0, p1, p2, p3);
                    horizStepPixel.CPCPixels.Add(cpcPixel);
                }

                horizontalStepPixels.Add(horizStepPixel);
            }
			var allBytes = new List<byte>();
			
            for (iStep = 0; iStep < AnimNbFrames; iStep++)
            {
                var horizStepPixel = horizontalStepPixels[iStep];

                for(var i=0; i<iStep; i++)
                {
                    var iHorizStepPixel = horizontalStepPixels[i];
                    if (iHorizStepPixel.CPCPixels.SequenceEqual(horizStepPixel.CPCPixels))
                    {
                        horizStepPixel.IsPointingTo = iHorizStepPixel.StepIndex;
                        break;
                    }
                }

               //if (horizStepPixel.IsPointingTo == -1)
                {
                    z80_uniqueHorizSteps += ";\n";
                    z80_uniqueHorizSteps += "HorizontalStep_" + horizStepPixel.StepIndex + ":\n";
                    foreach (var cpcPixel in horizStepPixel.CPCPixels)
                    {
                        z80_uniqueHorizSteps += "\tdb &" + cpcPixel.ToString("X2") + "\n";
						allBytes.Add(cpcPixel);
                    }
					z80_uniqueHorizSteps += "HorizontalStep_" + horizStepPixel.StepIndex + "_End:\n";
                    z80_uniqueHorizSteps += "\n";
                }
            }
            /*var z80_horizontal = "HorizontalSteps:\n";
            for (iStep = 0; iStep < AnimNbFrames; iStep++)
            {
                var horizStepPixel = horizontalStepPixels[iStep];
                /*if (horizStepPixel.IsPointingTo != -1)
                {
                    z80_horizontal += "\tdw HorizontalStep_" + (horizStepPixel.IsPointingTo) + " ; reference (saved mem!)\n";
                }
                else
                {
                    z80_horizontal += "\tdw HorizontalStep_" + (horizStepPixel.StepIndex) + "\n";					
                }*/
				//z80_horizontal += "\tdw HorizontalStep_" + (horizStepPixel.StepIndex) + "\n";
            //}
            //z80_horizontal += "\n";
			var z80_horizontal = "";
			
            z80_horizontal += z80_uniqueHorizSteps;
            File.WriteAllText(cpcHorizontalFilename, z80_horizontal);
			
			var unpackedBytes = allBytes.ToArray();
			File.WriteAllBytes(cpcHorizontalFilenameBin, unpackedBytes);
			var packer = new Packer();
			var packedBytes = packer.Convert(ref unpackedBytes);
			File.WriteAllBytes(cpcHorizontalFilenameBin+".packed", packedBytes);
			Console.Write(cpcHorizontalFilename+" = "+unpackedBytes.Length+ " bytes\n");
			Console.Write(cpcHorizontalFilename+".packed = "+packedBytes.Length+ " bytes\n");	

            var z80_vertical = "VerticalSteps:\n";

            for (int iVerticalStep = 0; iVerticalStep < AnimNbFrames; iVerticalStep++)
            {
                byte zoomValue = (byte)(sinZoomX[iVerticalStep]);
                var stepInfo = verticalStepInfos[zoomValue];
                int offsetY = (int)(sinOffsetY[iVerticalStep]);
                var YColorSwitch = new List<int>(stepInfo.YColorSwitch);
                byte currentPen = stepInfo.FirstValue; // 0 or 1

                int ColorSwitchCounterIndex = 0;
                var prevOffset = 0;

                while (offsetY>0)
                {
                    prevOffset++;

                    if (prevOffset == YColorSwitch[ColorSwitchCounterIndex])
                    {
                        if(currentPen==0)
                        {
                            currentPen = 1;
                        }
                        else
                        {
                            currentPen = 0;
                        }

                        ColorSwitchCounterIndex++;
                    }

                    offsetY--;
                }

                var bytes = new List<byte>();
                for (int i = ColorSwitchCounterIndex; i < maxColorSwitchCount; i++)
                {
                    if (i < YColorSwitch.Count)
                    {
                        int iOffset = YColorSwitch[i];

                        byte cpcByte = (byte)(iOffset - prevOffset);
                        bytes.Add(cpcByte);

                        prevOffset = iOffset;
                    }
                }

                byte currentPenAndCurrentOffset = (byte)((currentPen << 7) + bytes[0]);

                z80_vertical += "\tdb &" + bytes[1].ToString("X2")+" ; block size\n";
                z80_vertical += "\tdb &" + currentPenAndCurrentOffset.ToString("X2") + " ; step=" + iStep + " pen=" + currentPen + " firstOffset=" + bytes[0] + "\n";
            }
            File.WriteAllText(cpcVerticalFilename, z80_vertical);
        }

        private byte CreatePixelCPC_Mode2(byte p0, byte p1, byte p2, byte p3, byte p4, byte p5, byte p6, byte p7)
        {
            byte v;
            v = (byte)(
                    (((p0) & 1) << 7) + (((p1) & 1) << 6)
                + (((p2) & 1) << 5) + (((p3) & 1) << 4)
                + (((p4) & 1) << 3) + (((p5) & 1) << 2)
                + (((p6) & 1) << 1) + (((p7) & 1)));

            return v;
        }
        private byte CreatePixelCPC_Mode1(byte p0, byte p1, byte p2, byte p3)
        {
            byte v;
            v = (byte)(
                    (((p0) & 1) << 7) + (((p1) & 1) << 6)
                + (((p2) & 1) << 5) + (((p3) & 1) << 4)
                + (((p0 >> 1) & 1) << 3) + (((p1 >> 1) & 1) << 2)
                + (((p2 >> 1) & 1) << 1) + (((p3 >> 1) & 1)));

            return v;
        }

        private void CreateBackgroundPixels(byte pen1, byte pen2, string cpcFilename)
        {
            string z80 = "";

            var pixels = new byte[4]; // MODE 1
            
            for (int i = 0 - 1; i < 256 - 1; i++) // -1 because of first erase, draw pixel, second erase
            //for (int i = 0; i < 256; i++)
            {
                string s = "";

                byte value = 0;

                if (((i) & 15) < 8)
                {
                    value = CreatePixelCPC_Mode1(pen1, pen1, pen1, pen1);
                }
                else
                {
                    value = CreatePixelCPC_Mode1(pen2, pen2, pen2, pen2);
                }

                s += "\tdb &" + value.ToString("X2") + "\n";

                z80 += s;
            }

            File.WriteAllText(cpcFilename, z80);
        }

        private void CreateMovingPixels(string cpcFilename)
        {
            string z80 = "";

            var pixels = new byte[4]; // MODE 1
            int iPixel;

            for (int i = 0; i < 256; i++)
            {
                string s = "";

                int v = i & 3;

                for (iPixel = 0; iPixel < v; iPixel++)
                {
                    if (((i) & 63) < 32)
                    {
                        pixels[iPixel] = 0;
                    }
                    else
                    {
                        pixels[iPixel] = 1;
                    }
                }
                for (; iPixel < 4; iPixel++)
                {
                    if (((i) & 63) < 32)
                    {
                        pixels[iPixel] = 1;
                    }
                    else
                    {
                        pixels[iPixel] = 0;
                    }
                }

                var value = CreatePixelCPC_Mode1(pixels[0], pixels[1], pixels[2], pixels[3]);

                s += "\tdb &" + value.ToString("X2") + "\n";

                z80 += s;
            }

            File.WriteAllText(cpcFilename, z80);
        }

        private List<float> CreateSinCurves(double minValue, double maxValue, double coef, bool is512Buffer, int nbCurves, bool isAbs, bool isSin, string cpcFilename)
        {
            var allBytes = new List<byte>();
            var allFloats= new List<float>();

            string z80 = "";

            string z80Low = ";\n";
            string z80High = "\n;\n";

            const int bufferSize = 256;

            //maxValue += 1;

            double amplitude = 0.5 * (double)((maxValue - minValue));

            for (int iFrame = 0; iFrame < bufferSize; iFrame++)
            {
                var cos = Math.Cos((2 * Math.PI * ((double)(iFrame * nbCurves)) / bufferSize));
                if (isSin)
                {
                    cos = Math.Sin((2 * Math.PI * ((double)(iFrame * nbCurves)) / bufferSize));
                }
                if (isAbs)
                {
                    cos = Math.Abs(cos);
                }

                double value = (double)(amplitude * cos);

                value += (double)(amplitude * 1);
                value += (double)minValue;

                value *= coef;

                z80Low += "\tdb " + (((ushort)value) & 255) + "\n";
                z80High += "\tdb " + ((((ushort)value) >> 8) & 255) + "\n";

                var b = (byte)((((ushort)value) & 255));
                allBytes.Add(b);
                allFloats.Add((float)value);
            }

            z80 = z80Low;
            if (is512Buffer)
            {
                z80 += z80High;
            }

            File.WriteAllText(cpcFilename, z80);

            return allFloats;
        }

        private List<float> CreateSinCurvesDamierY(double minValue, double maxValue, double coef, bool is512Buffer, int nbCurves, bool isAbs, bool isSin, string cpcFilename)
        {
            var allBytes = new List<byte>();
            var allFloats = new List<float>();

            string z80 = ";\n";

            const int bufferSize = 256;
            
            double amplitude = 0.5 * (double)((maxValue - minValue));

            for (int iFrame = 0; iFrame < bufferSize; iFrame++)
            {
                var cos = Math.Cos((2 * Math.PI * ((double)(iFrame * nbCurves)) / bufferSize));
                if (isSin)
                {
                    cos = Math.Sin((2 * Math.PI * ((double)(iFrame * nbCurves)) / bufferSize));
                }
                if (isAbs)
                {
                    cos = Math.Abs(cos);
                }

                double value = (double)(amplitude * cos);
                value += (double)(amplitude * 1);
                value += (double)minValue;

                value *= coef;

                var b = (byte)((((ushort)value) & 255));
                allBytes.Add(b);
                allFloats.Add((float)value);

                z80 += "\tdb " + (2*(MoveYHeight - b)) + "\n";
            }

            File.WriteAllText(cpcFilename, z80);

            return allFloats;
        }

        private void CreateSinPosXCurves(double minValue, double maxValue, double coef, bool is512Buffer, int nbCurves, bool isAbs, bool isSin, string cpcFilename)
        {
            string z80 = "";

            string z80Low = ";\n";
            string z80High = "\n;\n";

            const int bufferSize = 256;

            //maxValue += 1;

            double amplitude = 0.5 * (double)((maxValue - minValue));

            for (int iFrame = 0; iFrame < bufferSize; iFrame++)
            {
                int it = iFrame;
                if(iFrame< bufferSize/2)
                {
                    it += bufferSize / 2;
                }
               /* else
                {
                    it = 3 * (bufferSize / 4);
                }*/

                var cos = Math.Cos((2 * Math.PI * ((double)(it * nbCurves)) / bufferSize));
                if (isSin)
                {
                    cos = Math.Sin((2 * Math.PI * ((double)(it * nbCurves)) / bufferSize));
                }
                if (isAbs)
                {
                    cos = Math.Abs(cos);
                }

                double value = (double)(amplitude * cos);

                value += (double)(amplitude * 1);
                value += (double)minValue;

                value *= coef;

                z80Low += "\tdb " + (((ushort)value) & 255) + "\n";
                z80High += "\tdb " + ((((ushort)value) >> 8) & 255) + "\n";
            }

            z80 = z80Low;
            if (is512Buffer)
            {
                z80 += z80High;
            }

            File.WriteAllText(cpcFilename, z80);
        }
    }
}
