using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
	public class _06_Clown
    {	
        public _06_Clown()
        {
			try
			{
                var nbBlocks = CreateR12R13Table(Program.ToZ80Folder("06_Clown/_Clown_R12R13.z80"));

                CreateR12R13Frames(nbBlocks, Program.ToZ80Folder("06_Clown/_R12R13_Vertical.bin"));

                CreateFades(Program.ToZ80Folder("06_Clown/_ClownFadePalettes.bin"));

                PackR12R13Vertical(
					Program.ToZ80Folder("06_Clown/_R12R13_Vertical.bin"),
					Program.ToZ80Folder("06_Clown/_R12R13_Vertical.bin.packed")
				);

                WriteImage("06_Clown/FreakyClownShhh.bmp", "06_Clown/_FreakyClownShhh.bmp");
			}
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}					
        }
		
		public void PackR12R13Vertical(string srcFilename, string packedFilename)
        {
			var unpackedBytes = File.ReadAllBytes(srcFilename);
			var packer = new Packer();
			var packedBytes = packer.Convert(ref unpackedBytes);
			File.WriteAllBytes(packedFilename, packedBytes);
			Console.Write(srcFilename+" = "+unpackedBytes.Length+ " bytes\n");
			Console.Write(packedFilename+" = "+packedBytes.Length+ " bytes\n");
        }
        
        public int CreateR12R13Table(string filename)
        {
            var baseOffsets = new List<ushort>();
            byte R1 = 54;

            byte R12 = 0x2C; // %00101100 8000
            ushort offset = (ushort)(0);
            while (offset <= 1024)
            {
                byte _r12 = (byte)(R12 + (offset >> 8));
                byte _r13 = (byte)(offset & 255);

                ushort finalOffset = (ushort)((_r12 << 8) + _r13);
                baseOffsets.Add(finalOffset);

                finalOffset = (ushort)(finalOffset - 0x1000);
                baseOffsets.Add(finalOffset);

                offset += (ushort)(R1);
            }

            R12 = 0x3C; // %00111100 4000
            offset = (ushort)(2);
            while (offset <= 1024)
            {
                byte _r12 = (byte)(R12 + (offset >> 8));
                byte _r13 = (byte)(offset & 255);

                ushort finalOffset = (ushort)((_r12 << 8) + _r13);
                baseOffsets.Add(finalOffset);

                finalOffset = (ushort)(finalOffset - 0x3000);
                baseOffsets.Add(finalOffset);

                offset += (ushort)(R1);
            }

            // discard last blocks
            for (int i=0; i<9; i++)
            {
                baseOffsets.RemoveAt(baseOffsets.Count - 1);
            }

            var content = ";\n";
            content += "Clown_R12R13_Table:\n";

            foreach (var baseOffset in baseOffsets)
            {
                ushort _r12 = (ushort)((baseOffset >> 10) << 10);
                ushort curOffset = (ushort)(baseOffset & 1023);
                ushort final = (ushort)(_r12 + curOffset);

                content += "\t db &" + (final >> 8).ToString("X2") + ", &" + (final & 255).ToString("X2") + "\n";
            }

            content += "\n";

            File.WriteAllText(filename, content);

            return baseOffsets.Count;
        }

        public static void CreateR12R13Frames(int nbBlocks, string binFilename)
        {
            var allBytes = new List<byte>();

            // dont forget to update FrameManager, cf:
            // dw THERIPPER_VBLFrame_Vertical, 160
            int NbFrames = 132;

            float frameSin1 = 0.0f;
            float frameSin2 = 0.0f;

            for (int iFrame = 0; iFrame < NbFrames; iFrame++)
            {
                var nbBlockLines = nbBlocks;

                var values = new byte[nbBlockLines];
                for (int i = 0; i < nbBlockLines; i++)
                {
                    values[i] = 0;
                }

                float frameCoef = 1.0f;
                if (iFrame != 0)
                {
                    var delta = 0.999f / ((float)NbFrames);
                    frameCoef = ((float)iFrame) * delta;
                    frameCoef = 0.999f - frameCoef;

                    /*frameCoef *= 0.8f;
					if (frameCoef > 1.0f)
					{
						frameCoef = 1.0f;
					}*/
                }

                ushort currentFlatBlock = 0;

                frameSin1 += 0.06f;
                frameSin2 += 0.043f;

                var sin1 = frameSin1;
                var sin2 = frameSin2;

                for (int iBlock = 0; iBlock < nbBlockLines; iBlock++)
                {
                    float yBlockCenter = (float)(nbBlockLines / 2);

                    float distance = 50000;
                    if (iFrame != 0)
                    {
                        distance = (float)(NbFrames) / (float)(iFrame);
                    }

                    float fY = -(yBlockCenter - 0.1f - (float)iBlock) * distance;
                    fY += yBlockCenter;// * (0.5f + (0.5f*(float) Math.Abs(Math.Sin(distance*0.5f))));

                    var plasmaDelta = (float)Math.Sin(sin1) + (float)Math.Sin(sin2);
                    plasmaDelta *= 16;
                    plasmaDelta *= frameCoef;

                    fY += plasmaDelta;

                    int y = (int)fY;
                    if ((y >= 0) && (y < nbBlockLines))
                    {
                        values[y] = (byte)iBlock;

                        if (currentFlatBlock == 0)
                        {
                            currentFlatBlock = (byte)iBlock;
                        }
                    }
                }

                for (int y = 0; y < nbBlockLines; y++)
                {
                    if (values[y] == 0)
                    {
                        values[y] = (byte)currentFlatBlock;
                    }
                    else
                    {
                        currentFlatBlock = values[y];
                    }
                }

                // set first frame to repeat 1st blockline instead of block 0 
                // since block 0 only contains 1 scanline
                /*if (iFrame == 0)
                {
                    for (int y = 0; y < nbBlockLines; y++)
                    {
                        values[y] = (byte)1;
                    }
                }*/

                // temp hack in data
                for (int y = 0; y < nbBlockLines; y++)
                {
                    //values[y] = (byte)iFrame;
                }

                foreach (byte final in values)
                {
                    // *2 because of right offset on CPC - ready to read R12 then R13 (2*8bits)
                    allBytes.Add((byte)(final * 2));
                }
            }

            File.WriteAllBytes(binFilename, allBytes.ToArray());
        }

        private void WriteImage(string srcFilename, string dstFilename)
		{
			var bitmap = new Bitmap(Program.ToDataFolder(srcFilename));
            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 0,
                R1 = 54, // 54, not 48, to get linear data
                R4 = 38,
                R9 = 7, // for static pages
                StartVLC = 0,
                ApplyOffsetAtNextPage = 4
            });
            for(int iPage=0; iPage<image.Pages.Length;iPage++)
            {
                foreach (var VLC in image.Pages[iPage].VLCs)
                {
                    // clear out invisible right border with pen having COLOR0
                    byte penColor0 = 0;
                    foreach(var p in image.Palette.Pens)
                    {
                        if (p.FirmwareColor==0)
                        {
                            penColor0 = p.IndexInPalette;
                            break;
                        }
                    }
                    for (int i = 0; i < 0x800; i += 54*2)
                    {
                        for (int j = iPage*4+i + 48*2; j < iPage * 4 + i + 54*2; j++)
                        {
                            var pixelToWrite = Pixel.ToMode0Pixel(penColor0, penColor0);
                            if(j<0x800)
                            {
                                VLC.CPCBytes[j] = pixelToWrite;
                            }
                        }
                    }
                    File.WriteAllBytes(Program.ToZ80Folder(dstFilename + ".page" + iPage + ".vlc" + VLC.Index), VLC.CPCBytes);
                }
            }
			File.WriteAllText(
				Program.ToZ80Folder(dstFilename + ".pal.z80"), 
				image.Palette.ToZ80TableString(Program.FilenameToKeyWordName(srcFilename))
			);
		}

        public void CreateFades(string outFilename)
        {
            var fadeFilenames = new List<string>();
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn0.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn1.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn2.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn3.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn4.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn5.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeIn6.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/FreakyClownShhh.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut1.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut2.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut3.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut4.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut5.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut6.bmp"));
            fadeFilenames.Add(Program.ToDataFolder("06_Clown/ClownFadeOut7.bmp"));

            var fadePalettes = new List<byte>();

            foreach (var fadeFilename in fadeFilenames)
            {
                var allBytes = CreateFadeMode0Raw(
                                    Program.ToDataFolder("06_Clown/FreakyClownShhh.bmp"),
                                    fadeFilename);

                fadePalettes.AddRange(allBytes);
            }

            File.WriteAllBytes(outFilename, fadePalettes.ToArray());
        }

        public byte[] CreateFadeMode0Raw(
            string baseBmpFilename,
            string fadeBmpFilename,
            int skipPensInPalette = 0)
        {
            var baseBitmap = new Bitmap(baseBmpFilename);
            var fadeBitmap = new Bitmap(fadeBmpFilename);

            var basePalette = new Palette(skipPensInPalette);
            var fadePalette = new Palette(skipPensInPalette);
            var outPalette = new Palette(skipPensInPalette);

            var cpcRaw = new byte[((baseBitmap.Width + 1) / 2) * baseBitmap.Height];

            var basePaletteIndices = new List<byte>();
            var fadePaletteIndices = new List<byte>();

            for (int y = 0; y < baseBitmap.Height; y++)
            {
                for (int x = 0; x < baseBitmap.Width; x += 2)
                {
                    Color baseP0 = baseBitmap.GetPixel(x, y);
                    Color baseP1 = baseP0;
                    if ((x + 1) < baseBitmap.Width)
                    {
                        baseP1 = baseBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = basePalette.AddColor(baseP0);
                    var paletteEntry1 = basePalette.AddColor(baseP1);
                    basePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    basePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x += 2)
                {
                    Color fadeP0 = fadeBitmap.GetPixel(x, y);
                    Color fadeP1 = fadeP0;
                    if ((x + 1) < fadeBitmap.Width)
                    {
                        fadeP1 = fadeBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = fadePalette.AddColor(fadeP0);
                    var paletteEntry1 = fadePalette.AddColor(fadeP1);
                    fadePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    fadePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x++)
                {
                    int offset = (y * fadeBitmap.Width) + x;

                    int baseIndex = basePaletteIndices[offset];
                    int fadeIndex = fadePaletteIndices[offset];

                    var basePaletteEntry = basePalette.Pens[baseIndex];
                    var fadePaletteEntry = fadePalette.Pens[fadeIndex];

                    basePaletteEntry.Color = fadePaletteEntry.Color;
                    basePaletteEntry.FirmwareColor = fadePaletteEntry.FirmwareColor;
                    basePaletteEntry.GateArrayColor = fadePaletteEntry.GateArrayColor;
                }
            }

            return basePalette.ToGateArrayByteArray();
        }
    }
}
