﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _07_Epsilon
    {
        public int NbFrames = 21+21;//59
        public int BarSize = 32;
        public int NbBarsPerBranches = 8; // must be div 2
        public int NbBranches = 4;
        public int BAR_TYPE = 15; // 3 6 8 14
        public double Scale = 4.32*0.78*1.0;

        public _07_Epsilon()
        {
            try
            {
                //TestSmoothFloatValue();
                //TestSmoothIntegerValue();

                CreateSinMoves(
                    Program.ToZ80Folder("07_Epsilon/_SinXDisto1.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinXDisto2.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinYDisto1.bin"),
                    Program.ToZ80Folder("07_Epsilon/_SinYDisto2.bin")
                    );

                CreateScaleValues(
                    Program.ToZ80Folder("07_Epsilon/_ScaleValues.z80")
                    );

                CreateJumpTable(   
                    Program.ToDataFolder("07_Epsilon/JumpTable1_PEN0.png"), 
                    Program.ToDataFolder("07_Epsilon/JumpTable1_PEN1.png"), 
                    Program.ToZ80Folder("07_Epsilon/_JumpTable1.z80"));

                CreateVerticalBars(
                    96*4, 
                    NbFrames*2, // *2 because R9=1
                    Program.ToDataFolder("07_Epsilon/Bars.png"),
                    Program.ToZ80Folder("07_Epsilon/_Epsilon.bmp")
                    );
            }
			catch(Exception e)
			{
				Log.Fatal(e.ToString());
			}
        }

        private void TestSmoothFloatValue()
        {
            float value = 4;
            float targetValue = 50;

            for(int i=0; i<100; i++)
            {
                Console.WriteLine("" + value);
                value = (float)(value + (targetValue - value)*0.5f);
            }
        }

        private void TestSmoothIntegerValue()
        {
            int value = 4;
            int targetValue = 50;

            for (int i = 0; i < 20; i++)
            {
                Console.WriteLine("" + value);
                int delta = (targetValue - value)>>2;
                value += delta;
            }
        }

        private void CreateJumpTable(string imageFilename1, string imageFilename2, string destZ80)
        {
            Bitmap bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            Bitmap bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));

            var colors = new List<byte>();

            for (int y = 0; y < 268; y++)
            {
                var palette1 = new Palette();
                palette1.AddColor(bmp1.GetPixel(0, y));
                var gateArrayColor1 = palette1.ToFirmwareByteArray()[0];
                colors.Add(gateArrayColor1);

                var palette2 = new Palette();
                palette2.AddColor(bmp2.GetPixel(0, y));
                var gateArrayColor2 = palette2.ToFirmwareByteArray()[0];
                colors.Add(gateArrayColor2);
            }

            var z80Content = "";
            
            var iColor = 0;
            for (int i=0; i<267/4; i++)
            {
                z80Content += "\tdw Epsilon_LineFX_Pair\n";                
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;

                z80Content += "\tdw Epsilon_LineFX_Impair\n";
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
                z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;
            }

            z80Content += "\tdw Epsilon_LineFX_Pair\n";
            z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
            z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;
            z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
            z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;

            z80Content += "\tdw Epsilon_LineFX_LAST_Impair ;; LAST line has special treatment for BORDER\n";
            z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
            z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;
            //z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 0\n"; iColor++;
            //z80Content += "\tdb COLOR" + colors[iColor] + " ;; PEN 1\n"; iColor++;
            z80Content += "\tdw Epsilon_EndLineFX\n";

            File.WriteAllText(destZ80, z80Content);
        }

        private Framework.Math3D.Vector3D[] CreateGeometry(double scale, out double minZ, out double maxZ)
        {
            var vertices = new Framework.Math3D.Vector3D[NbBarsPerBranches * NbBranches];
            var iVert = 0;

            int iSkipFirst = 1;

            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D((iSkipFirst + iNbBarPerBranch) * scale, 0, 0);
                iVert++;
            }
            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D((-iSkipFirst - iNbBarPerBranch) * scale, 0, 0);
                iVert++;
            }
            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D(0, 0, (iSkipFirst + iNbBarPerBranch) * scale);
                iVert++;
            }
            for (int iNbBarPerBranch = 0; iNbBarPerBranch < NbBarsPerBranches; iNbBarPerBranch++)
            {
                vertices[iVert] = new Framework.Math3D.Vector3D(0, 0, (-iSkipFirst - iNbBarPerBranch) * scale);
                iVert++;
            }

            minZ = (double)(-(iSkipFirst + (NbBarsPerBranches - 1)) * scale);
            maxZ = (double)((iSkipFirst + (NbBarsPerBranches - 1)) * scale);

            return vertices;
        }

        // zoomScale = 0 mininum  1 maximum
        private void DrawBar(Bitmap barsBitmap, Bitmap bitmap, int x, int y, double zoomScale) 
        {
            var max = 255 / NbBarsPerBranches;

            if(zoomScale<0)
            {
                zoomScale = 0;
            }
            else if (zoomScale > 1)
            {
                zoomScale = 1;
            }

            zoomScale = 1 - zoomScale;

            int iBar = (int)((zoomScale) * (16.9999));
            if (iBar<0)
            {
                iBar = 0;
            }
            else if (iBar>16)
            {
                iBar = 16;
            }
            var edgePixelsToRemove = iBar;
            iBar *= 36;

            iBar += 2 * BAR_TYPE;

            for(int i=0; i<BarSize-(edgePixelsToRemove*2); i++)
            {
                var iX = x + i + edgePixelsToRemove;
                if ((iX>=0)&&(iX<bitmap.Width))
                {
                    bitmap.SetPixel(iX, y, barsBitmap.GetPixel(i, iBar));
                    bitmap.SetPixel(iX, y+1, barsBitmap.GetPixel(i, iBar+1));
                }
            }
        }

        public void CreateVerticalBars(int width, int height, string barsFilename, string filename)
        {
            var barsBitmap = new Bitmap(Bitmap.FromFile(barsFilename));
            var bitmap = new Bitmap(width, height);

            for (int iFrame = 0; iFrame < NbFrames; iFrame++)
            {
                double yRot = (360.0 / NbFrames / 4) * iFrame;

                double minZ; // -7.0 ..
                double maxZ; // 7.0 ...
                var vertices = CreateGeometry(Scale, out minZ, out maxZ);

                vertices = Math3D.RotateY(vertices, yRot);

                vertices = Math3D.DepthSort(vertices);
                
                for (int i = 0; i < vertices.Length; i++)
                {
                    var vec = vertices[i];

                    var x2D = (double)(-vec.X/(0.1));

                    x2D += (width / 2)-(barsBitmap.Width/2);

                    var ratio = vec.Z - minZ;
                    var length = maxZ - minZ;

                    double zoomScale = 0;
                    double limit = 3.4;// 3.4;// 5.3;
                    if (ratio>(length/ limit))
                    {
                        ratio -= length / limit;
                        ratio *= limit;
                        zoomScale = (double)(ratio / length);
                    }

                    //zoomScale = (double)((vec.Z - minZ) /(maxZ-minZ));

                    DrawBar(barsBitmap, bitmap, (int)(x2D), iFrame*2, zoomScale);
                }
            }

            var tempImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 48,
                R4 = 21,
                R9 = 1,
                StartVLC = 0,
                StartR12 = 0x00
            });

            var blackColor = tempImage.Palette.Pens[0].Color;
            var noiseColor = Color.FromArgb(0, tempImage.Palette.Pens[1].Color);
            for (float iY = 1; iY < bitmap.Height;)
            {
                int y = (int)iY;

                int offset = y & 1;

                for (int x = 0; x < bitmap.Width; x += 2)
                {
                    int _iX = x + offset;
                    int _iY = bitmap.Height - y;

                    //var pixel = Color.FromArgb(255, bitmap.GetPixel(_iX, _iY));
                    //if (pixel != blackColor)
                    {
                        //bitmap.SetPixel(_iX, _iY, noiseColor);
                    }
                }

                iY += 0.8f;
                iY *= 1.32f;
            }

            bitmap.Save(filename, System.Drawing.Imaging.ImageFormat.Bmp);

            CreateVerticalBars_CPC(bitmap, filename);
        }

        private void CreateVerticalBars_CPC(Bitmap bitmap, string bmpFilename)
        {
            var packer = new Packer();
            var filename = "";
            
            var palette = new Palette();
            //palette.AddColor(Color.Black);
            //var gateArrayColor1 = palette.ToFirmwareByteArray()[0];

            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 48,
                R4 = 21,
                R9 = 1,
                StartVLC = 0,
                StartR12 = 0x00,
                Palette = palette
            });

            for (var iPage = 0; iPage < image.Pages.Length; iPage++)
            {
                foreach (var VLC in image.Pages[iPage].VLCs)
                {
                    var cpcBytes = new List<byte>(VLC.CPCBytes).Take(image.Pages[iPage].NbBytesWritten / image.Pages[iPage].VLCs.Length).ToArray();

                    //var bbPackedBytes = BBPacker.Pack(ref cpcBytes);
                    //File.WriteAllBytes(filename + ".bbPacked", bbPackedBytes);

                    /*if((iPage==1)&&(VLC.Index==1))
                    {
                        filename = bmpFilename + ".page" + iPage + ".vlc" + VLC.Index;
                        File.WriteAllBytes(filename, cpcBytes);

                        var remainingSplitSize = 0x100;
                        var part1 = new List<byte>(cpcBytes).Take(cpcBytes.Length - remainingSplitSize);
                        var part2 = new List<byte>(cpcBytes).Skip(cpcBytes.Length - remainingSplitSize);

                        File.WriteAllBytes(filename+".part1", part1.ToArray());
                        File.WriteAllBytes(filename + ".part2", part2.ToArray());
                    }
                    else*/
                    {
                        filename = bmpFilename + ".page" + iPage + ".vlc" + VLC.Index;
                        File.WriteAllBytes(filename, cpcBytes);
                    }

                    var vlcBBPackedBytes = BBPacker.Pack(ref cpcBytes);
                    File.WriteAllBytes(filename + ".bbPacked", vlcBBPackedBytes);

                    //var vlcPackedBytes = packer.Convert(ref cpcBytes);
                    //File.WriteAllBytes(filename + ".packed", vlcPackedBytes);
                }
            }

            var r12r13 = new List<byte>();
            for(int i=0; i<image.R12R13.Length; i+=4)
            {
                var r12 = image.R12R13[i];

                // force page 0 (not 1) to be in &8000 instead of &0000
                // force page 1 (not 0) to be in &C000 instead of &4000
                r12 += 0x20;
                /*if (r12 >= 0x1C)
                {
                    r12 += 0x10;
                }*/
                                
                var r13 = image.R12R13[i + 1];
                r12r13.Add(r12);
                r12r13.Add(r13);
            }

            filename = bmpFilename + ".r12r13";
            var r12r13CpcBytes = r12r13.ToArray();
            File.WriteAllBytes(filename, r12r13CpcBytes);

            filename = bmpFilename + ".r12r13.takePart";
            r12r13CpcBytes = r12r13.Take(r12r13.Count/4).ToArray();
            File.WriteAllBytes(filename, r12r13CpcBytes);

            File.WriteAllText(
                bmpFilename + ".pal.z80",
                image.Palette.ToZ80TableString(Program.FilenameToKeyWordName(bmpFilename))
            );
        }

        public void CreateSinMoves( string sinXDisto1Filename, 
                                    string sinXDisto2Filename,
                                    string sinYDisto1Filename,
                                    string sinYDisto2Filename)
        {
            File.WriteAllBytes(sinXDisto1Filename, SinCurve.CreateCos(256, 0, 40, 2));
            File.WriteAllBytes(sinXDisto2Filename, SinCurve.CreateSin(256, 0, 12, 2));
        }

        public void CreateScaleValues(string scaleValuesFilename)
        {
            var z80TablePtr = "";
            var z80TableValues = "";

            z80TablePtr += ";\n";
            z80TablePtr += "Epsilon_ScaleValue_Ptr:\n";

            const int MaxVumetreValue = 54;
            const int NbValueScales = 16;

            for (int iVumetre=0; iVumetre < NbValueScales; iVumetre++)
            {
                z80TablePtr += "\tdw Epsilon_ScaleValue_Vumetre" + iVumetre + "\n";
                
                double maxScaleValue = ((double)(iVumetre)/(double)(NbValueScales))*(double)MaxVumetreValue;

                maxScaleValue *= 0.5;

                z80TableValues += ";; MaxScaleValue=" + (int)maxScaleValue + "\n";
                z80TableValues += "\tALIGN 32\n";
                z80TableValues += "Epsilon_ScaleValue_Vumetre" + iVumetre + ":\n";
                for (double v = 0; v < (double)MaxVumetreValue; v += 1)
                {
                    int intValue = (int)((v / (double)MaxVumetreValue) * maxScaleValue);
                    z80TableValues += "\tdb " + intValue + "\n";
                }
                z80TableValues += "\n";
            }

            var z80 = z80TablePtr + "\n";
            z80 += z80TableValues;
            File.WriteAllText(scaleValuesFilename, z80);
        }
    }
}
