﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using phX.Framework;

namespace phX.Parts
{
    public class _08_CubeRotX
    {
        private const int NbFrames90Deg = 90;
        private const int BmpWidth = 48 * 4 * 2; // MODE 1
        private const int BmpHeight = 268;
        private const double CubeSize = 1.128;
        private const double CubeZDiv = 0.5;
        private const double CubeDistance = 3.1;

        private int MinX = 9999;
        private int MaxX = 0;
        
        public class FrameX
        {
            public Bitmap TopBitmap;
            public List<Math3D.Vector2D> VerticallySortedDots;
            public List<Math3D.Vector2D> HorizontallySortedDots;

            public byte Z80TileSide0 = 0;
            public byte Z80YLengthSide0 = 0;
            public ushort Z80StartLeftSide0 = 0;
            public ushort Z80StartRightSide0 = 0;
            public ushort Z80DeltaLeftSide0 = 0;
            public ushort Z80DeltaRightSide0 = 0;

            public byte Z80TileSide1 = 0;
            public byte Z80YLengthSide1 = 0;
            public ushort Z80StartLeftSide1 = 0;
            public ushort Z80StartRightSide1 = 0;
            public ushort Z80DeltaLeftSide1 = 0;
            public ushort Z80DeltaRightSide1 = 0;

            public byte Z80_Y0 = 0;
            public byte Z80_Y1 = 0;
            public byte Z80_Y2 = 0;
        }

        public _08_CubeRotX()
        {
            CreateCube(Program.ToZ80Folder("08_Cube/RotX/"));
        }

        private void CreateCube(string outFolder)
        {
            CreateRotX(outFolder);
        }

        private Math3D.Vector3D[] CreateGeometry()
        {
            var vertices = new Math3D.Vector3D[4 * 2];
            vertices[0] = new Math3D.Vector3D(-CubeSize, -CubeSize, -CubeSize);
            vertices[1] = new Math3D.Vector3D(CubeSize, -CubeSize, -CubeSize);
            vertices[2] = new Math3D.Vector3D(CubeSize, -CubeSize, CubeSize);
            vertices[3] = new Math3D.Vector3D(-CubeSize, -CubeSize, CubeSize);
            vertices[4] = new Math3D.Vector3D(-CubeSize, CubeSize, -CubeSize);
            vertices[5] = new Math3D.Vector3D(CubeSize, CubeSize, -CubeSize);
            vertices[6] = new Math3D.Vector3D(CubeSize, CubeSize, CubeSize);
            vertices[7] = new Math3D.Vector3D(-CubeSize, CubeSize, CubeSize);
            return vertices;
        }

        private List<Math3D.Vector2D> CreateSequentialFrameX(int iFrame)
        {
            var vertices = CreateGeometry();

            double rotDeg = 90 * (((double)iFrame) / (double)(NbFrames90Deg));
            var rotatedVertices = Math3D.RotateX(vertices, rotDeg);

            var translatedVertices = Math3D.Translate(rotatedVertices, new Math3D.Vector3D(0, 0, -CubeSize - CubeDistance));

            var projectedVertices = new List<Math3D.Vector2D>();
            foreach (var v in translatedVertices)
            {
                const int X2D_OffsetX = -4;
                const double ScaleX = 0.96;
                const double ScaleY = 1.185;
                const double ScaleZ = 1;

                v.Z *= CubeZDiv;
                var x2D = BmpWidth / 2 + (int)((double)(((v.X* ScaleX) / (ScaleZ*v.Z)) * (BmpWidth / 2)));
                var y2D = BmpHeight / 2 + (int)((double)(((v.Y* ScaleY) / (ScaleZ*v.Z)) * (BmpHeight / 2)));

                x2D += X2D_OffsetX;
                
                if (x2D > BmpWidth / 2)
                {
                    x2D = x2D + 1;
                }
                else
                {
                    x2D = x2D - 1;
                }

                y2D >>= 1;
                y2D <<= 1;

                projectedVertices.Add(new Math3D.Vector2D(x2D, y2D));
            }

            var verticallySortedDots = projectedVertices.OrderBy(o => o.Y).ToList();
            
            var minY = (int)verticallySortedDots[0].Y;
            var maxY = (int)verticallySortedDots[verticallySortedDots.Count - 1].Y;
            var height = maxY - minY;
            var centeredYStart = (BmpHeight / 2) - (height / 2);

            var offsetYStart = centeredYStart - minY;
            offsetYStart >>= 1;
            offsetYStart <<= 1;

            if (offsetYStart > 0)
            {
                foreach (var v in verticallySortedDots)
                {
                  //  v.Y += offsetYStart;
                }
            }

            /*foreach (var v in verticallySortedDots)
            {
                var clamp = 14;
                if (v.Y < clamp)
                {
                    v.Y = clamp;
                }
                if (v.Y >= (BmpHeight - clamp))
                {
                    v.Y = BmpHeight - clamp;
                }
            }*/

            if (iFrame < 16)
            {
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
            }
            else if (iFrame < 45)
            {
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
            }
            else if (iFrame < 75)
            {
                verticallySortedDots.RemoveAt(4);
                verticallySortedDots.RemoveAt(4);
            }
            else
            {
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);
                verticallySortedDots.RemoveAt(2);

                var t = verticallySortedDots[3];
                verticallySortedDots[3] = verticallySortedDots[2];
                verticallySortedDots[2] = t;
            }

            return verticallySortedDots;
        }

        private Bitmap CreateTextureBitmap(Color color0, Color color1)
        {
            var bitmap = new Bitmap(2, 2);
            bitmap.SetPixel(0, 0, color1); bitmap.SetPixel(1, 0, color0);
            bitmap.SetPixel(0, 1, color0); bitmap.SetPixel(1, 1, color1);

            return bitmap;
        }

        private void MinMaxUpdate(int x)
        {
            if (MinX > x)
            {
                MinX = x;
            }
            if (MaxX < x)
            {
                if (x == 356)
                {
                    x = 357;
                }
                MaxX = x;
            }
        }

        private void CreateRotX(string outFolder)
        {
            var z80 = "";

            #region Colors
            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000)); // background (raster)
            var color1 = Color.FromArgb((int)(Pixel.CPCAsicPalette[11] | 0xFF000000)); // side A
            var color2 = Color.FromArgb((int)(Pixel.CPCAsicPalette[14] | 0xFF000000)); // side B
            var color3 = Color.FromArgb((int)(Pixel.CPCAsicPalette[26] | 0xFF000000)); // both sides

            var brushColor0 = new SolidBrush(color0);
            var brushColor1 = new SolidBrush(color1);
            var brushColor2 = new SolidBrush(color2);
            var brushColor3 = new SolidBrush(color3);
            
            var textureBitmap0 = CreateTextureBitmap(color1, color3);
            var textureBitmap1 = CreateTextureBitmap(color2, color3);
            var textureBrush0 = new TextureBrush(textureBitmap0);
            var textureBrush1 = new TextureBrush(textureBitmap1);

            var penColor0 = new Pen(color0);
            var penColor1 = new Pen(color1);
            var penColor2 = new Pen(color2);
            var penColor3 = new Pen(color3);
            #endregion

            #region Compute palette
            var palette = new Palette();
            palette.AddColor(color0);
            palette.AddColor(color1);
            palette.AddColor(color2);
            palette.AddColor(color3);

            // create palette
            File.WriteAllText(outFolder + "_CubeXPalette.z80", palette.ToZ80TableString(Program.FilenameToKeyWordName("Cube")));
            #endregion

            var frames = new List<FrameX>();

            int TopY1 = BmpHeight;
            int TopY2 = 0;
            MinX = 9999;
            MaxX = 0;

            //byte Reg1 = 43;

            #region create frames and compute TOP/BOTTOM
            for (int iFrame = 0; iFrame < NbFrames90Deg + 1; iFrame++)
            {
                var frameY = new FrameX();
                frameY.VerticallySortedDots = CreateSequentialFrameX(iFrame);
                frameY.HorizontallySortedDots = frameY.VerticallySortedDots.OrderBy(o => o.X).ToList();
                frames.Add(frameY);

                if (frameY.VerticallySortedDots.Count == 6)
                {
                    if (frameY.VerticallySortedDots[0].Y < TopY1)
                    {
                        TopY1 = (int)frameY.VerticallySortedDots[0].Y;
                    }
                    if (frameY.VerticallySortedDots[2].Y > TopY2)
                    {
                        TopY2 = (int)frameY.VerticallySortedDots[2].Y;
                    }
                }

                if (frameY.HorizontallySortedDots.Count == 4)
                {
                    var x1 = (int)frameY.HorizontallySortedDots[0].X;
                    var x2 = (int)frameY.HorizontallySortedDots[1].X;

                    MinMaxUpdate(x1);
                    MinMaxUpdate(x2);
                }
                else if (frameY.HorizontallySortedDots.Count == 6)
                {
                    var x1 = (int)frameY.HorizontallySortedDots[0].X;
                    var x2 = (int)frameY.HorizontallySortedDots[1].X;
                    var x3 = (int)frameY.HorizontallySortedDots[2].X;

                    MinMaxUpdate(x1);
                    MinMaxUpdate(x2);
                    MinMaxUpdate(x3);
                }
                else
                {
                    // can't happen
                    MaxX = 12;
                }
            }
            #endregion

            Console.WriteLine("MinX=" + MinX);
            Console.WriteLine("MaxX=" + MaxX);

            var nbLines = MaxX - MinX;
            Console.WriteLine("NbLines=" + nbLines);

            #region draw flat triangles (top only) + write png frame
            for (int iFrame = 0; iFrame < NbFrames90Deg+1; iFrame++)
            {
                var frameY = frames[iFrame];

                frameY.TopBitmap = new Bitmap(BmpWidth, BmpHeight);
                var topGraphics = Graphics.FromImage(frameY.TopBitmap);
                topGraphics.FillRectangle(brushColor0, 0, 0, BmpWidth, BmpHeight);

                // draw flat triangles
                if (frameY.VerticallySortedDots.Count == 4)
                {
                    var b = textureBrush0;
                    if (frameY.VerticallySortedDots[1].X > BmpWidth / 2)
                    {
                        b = textureBrush1;
                    }
                    if (frameY.VerticallySortedDots[0].Y == frameY.VerticallySortedDots[1].Y)
                    {
                        b = textureBrush0;
                    }

                    if (iFrame == NbFrames90Deg)
                    {
                        b = textureBrush1;
                    }

                    var points = new List<Point>();

                    if (iFrame < 45)
                    {
                        frameY.Z80TileSide0 = 0;

                        points.Add(new Point((int)frameY.VerticallySortedDots[0].X, (int)frameY.VerticallySortedDots[0].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[2].X, (int)frameY.VerticallySortedDots[2].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[3].X, (int)frameY.VerticallySortedDots[3].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[1].X, (int)frameY.VerticallySortedDots[1].Y));
                    }
                    else if (iFrame<75)
                    {
                        frameY.Z80TileSide0 = 1;

                        b = textureBrush1;
                        points.Add(new Point((int)frameY.VerticallySortedDots[0].X, (int)frameY.VerticallySortedDots[0].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[1].X, (int)frameY.VerticallySortedDots[1].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[2].X, (int)frameY.VerticallySortedDots[2].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[3].X, (int)frameY.VerticallySortedDots[3].Y));
                    }
                    else
                    {
                        frameY.Z80TileSide0 = 1;

                        b = textureBrush1;
                        points.Add(new Point((int)frameY.VerticallySortedDots[0].X, (int)frameY.VerticallySortedDots[0].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[1].X, (int)frameY.VerticallySortedDots[1].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[3].X, (int)frameY.VerticallySortedDots[3].Y));
                        points.Add(new Point((int)frameY.VerticallySortedDots[2].X, (int)frameY.VerticallySortedDots[2].Y));
                    }

                    frameY.Z80YLengthSide0 = (byte)(((int)frameY.VerticallySortedDots[2].Y) - ((int)frameY.VerticallySortedDots[0].Y));

                    double deltaLeft = (frameY.VerticallySortedDots[2].X - frameY.VerticallySortedDots[0].X)
                                        / (frameY.VerticallySortedDots[2].Y - frameY.VerticallySortedDots[0].Y);
                    frameY.Z80StartLeftSide0 = DoubleToUShort(frameY.VerticallySortedDots[0].X/4);
                    frameY.Z80DeltaLeftSide0 = DoubleToUShort256(deltaLeft/4);

                    double deltaRight = (frameY.VerticallySortedDots[3].X - frameY.VerticallySortedDots[1].X)
                                        / (frameY.VerticallySortedDots[3].Y - frameY.VerticallySortedDots[1].Y);
                    frameY.Z80StartRightSide0 = DoubleToUShort(frameY.VerticallySortedDots[1].X/4);
                    frameY.Z80DeltaRightSide0 = DoubleToUShort256(deltaRight/4);

                    topGraphics.FillPolygon(b, points.ToArray());

                    frameY.Z80YLengthSide1 = 0;

                    if (iFrame < 75)
                    {
                        frameY.Z80_Y0 = (byte)(((int)frameY.VerticallySortedDots[0].Y) >> 1);
                        frameY.Z80_Y1 = (byte)(((int)frameY.VerticallySortedDots[2].Y) >> 1);
                        frameY.Z80_Y2 = frameY.Z80_Y1;
                    }
                    else
                    {
                        frameY.Z80_Y0 = (byte)(((int)frameY.VerticallySortedDots[0].Y) >> 1);
                        frameY.Z80_Y1 = frameY.Z80_Y0;
                        frameY.Z80_Y2 = (byte)(((int)frameY.VerticallySortedDots[2].Y) >> 1);
                    }
                }
                else if (frameY.VerticallySortedDots.Count == 6)
                {
                    var b0 = textureBrush0;
                    var b1 = textureBrush1;

                    // top
                    var points = new List<Point>();
                    points.Add(new Point((int)frameY.VerticallySortedDots[0].X, (int)frameY.VerticallySortedDots[0].Y));
                    points.Add(new Point((int)frameY.VerticallySortedDots[2].X, (int)frameY.VerticallySortedDots[2].Y));
                    points.Add(new Point((int)frameY.VerticallySortedDots[3].X, (int)frameY.VerticallySortedDots[3].Y));
                    points.Add(new Point((int)frameY.VerticallySortedDots[1].X, (int)frameY.VerticallySortedDots[1].Y));
                    topGraphics.FillPolygon(b0, points.ToArray());

                    frameY.Z80TileSide0 = 0;

                    frameY.Z80YLengthSide0 = (byte)(((int)frameY.VerticallySortedDots[2].Y) - ((int)frameY.VerticallySortedDots[0].Y));

                    double deltaLeft = (frameY.VerticallySortedDots[2].X - frameY.VerticallySortedDots[0].X)
                                        / (frameY.VerticallySortedDots[2].Y - frameY.VerticallySortedDots[0].Y);
                    frameY.Z80StartLeftSide0 = DoubleToUShort(frameY.VerticallySortedDots[0].X/4);
                    frameY.Z80DeltaLeftSide0 = DoubleToUShort256(deltaLeft/4);

                    double deltaRight = (frameY.VerticallySortedDots[3].X - frameY.VerticallySortedDots[1].X)
                                        / (frameY.VerticallySortedDots[3].Y - frameY.VerticallySortedDots[1].Y);
                    frameY.Z80StartRightSide0 = DoubleToUShort(frameY.VerticallySortedDots[1].X/4);
                    frameY.Z80DeltaRightSide0 = DoubleToUShort256(deltaRight/4);
                        

                    // bottom
                    points = new List<Point>();
                    points.Add(new Point((int)frameY.VerticallySortedDots[2 + 3].X, (int)frameY.VerticallySortedDots[2 + 3].Y));
                    points.Add(new Point((int)frameY.VerticallySortedDots[2 + 0].X, (int)frameY.VerticallySortedDots[2 + 0].Y));
                    points.Add(new Point((int)frameY.VerticallySortedDots[2 + 1].X, (int)frameY.VerticallySortedDots[2 + 1].Y));
                    points.Add(new Point((int)frameY.VerticallySortedDots[2 + 2].X, (int)frameY.VerticallySortedDots[2 + 2].Y));
                    topGraphics.FillPolygon(b1, points.ToArray());
                    
                    frameY.Z80TileSide1 = 1;

                    frameY.Z80YLengthSide1 = (byte)(((int)frameY.VerticallySortedDots[4].Y) - ((int)frameY.VerticallySortedDots[2].Y));

                    // X are reversed for 4,5 (bottom only!)

                    deltaLeft = (frameY.VerticallySortedDots[5].X - frameY.VerticallySortedDots[2].X)
                                        / (frameY.VerticallySortedDots[4].Y - frameY.VerticallySortedDots[2].Y);
                    frameY.Z80StartLeftSide1 = DoubleToUShort(frameY.VerticallySortedDots[2].X / 4);
                    frameY.Z80DeltaLeftSide1 = DoubleToUShort256(deltaLeft/4);

                    deltaRight = (frameY.VerticallySortedDots[4].X - frameY.VerticallySortedDots[3].X)
                                        / (frameY.VerticallySortedDots[5].Y - frameY.VerticallySortedDots[3].Y);
                    frameY.Z80StartRightSide1 = DoubleToUShort(frameY.VerticallySortedDots[3].X / 4);
                    frameY.Z80DeltaRightSide1 = DoubleToUShort256(deltaRight/4);

                    frameY.Z80_Y0 = (byte)(((int)frameY.VerticallySortedDots[0].Y) >> 1);

                    // -1 because set R13 is active for next VCC
                    frameY.Z80_Y1 = (byte)(-1+(int)(frameY.VerticallySortedDots[2].Y) >> 1);
                    if(frameY.Z80_Y0==frameY.Z80_Y1)
                    {
                        frameY.Z80_Y1++;
                    }

                    // -1 because set R13 is active for next VCC
                    frameY.Z80_Y2 = (byte)(-1+((int)frameY.VerticallySortedDots[4].Y) >> 1);
                    if (frameY.Z80_Y1 == frameY.Z80_Y2)
                    {
                        frameY.Z80_Y2++;
                    }
                }

                topGraphics.Dispose();
                frameY.TopBitmap.Save(outFolder + "AnimSequence\\" + iFrame.ToString("00") + ".png", ImageFormat.Png);
            }
            #endregion

            #region Create Edge Pixels
            // color0 = background (raster)
            // color1 = side A
            // color2 = side B
            // color3 = both sides
            byte[] _SideA_VLC0_Left_Bytes = null;
            byte[] _SideA_VLC1_Left_Bytes = null;
            byte[] _SideB_VLC0_Left_Bytes = null;
            byte[] _SideB_VLC1_Left_Bytes = null;
            byte[] _SideA_VLC0_Right_Bytes = null;
            byte[] _SideA_VLC1_Right_Bytes = null;
            byte[] _SideB_VLC0_Right_Bytes = null;
            byte[] _SideB_VLC1_Right_Bytes = null;
            var plain_SideA_VLC0 = 
                WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color1, color3), "_SideA_VLC0_Left.bin", ref _SideA_VLC0_Left_Bytes, true);
            var plain_SideA_VLC1 = 
                WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color3, color1), "_SideA_VLC1_Left.bin", ref _SideA_VLC1_Left_Bytes, true);
            var plain_SideB_VLC0 = 
                WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color2, color3), "_SideB_VLC0_Left.bin", ref _SideB_VLC0_Left_Bytes, true);
            var plain_SideB_VLC1 =
                WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color3, color2), "_SideB_VLC1_Left.bin", ref _SideB_VLC1_Left_Bytes, true);

            WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color3, color1), "_SideA_VLC0_Right.bin", ref _SideA_VLC0_Right_Bytes, false);
            WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color1, color3), "_SideA_VLC1_Right.bin", ref _SideA_VLC1_Right_Bytes, false);
            WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color3, color2), "_SideB_VLC0_Right.bin", ref _SideB_VLC0_Right_Bytes, false);
            WriteEdgePixels(outFolder, brushColor0, palette, CreateTileBitmap(color2, color3), "_SideB_VLC1_Right.bin", ref _SideB_VLC1_Right_Bytes, false);
            #endregion

            #region Create Z80 Frames
            z80 = "";
            for (int iFrame = 0; iFrame < NbFrames90Deg + 1; iFrame++)
            {
                var frameY = frames[iFrame];

                z80 += ";\n";
                z80 += "CubeRotX_Frame" + iFrame + ":\n";

                // special case for frame > 75
              /*  if (frameY.Z80_Y0 == frameY.Z80_Y1)
                {
                    z80 += "\tdb &" + frameY.Z80_Y0.ToString("X2") + " ;; Z80_Y0\n";
                    z80 += "\tdb &" + frameY.Z80_Y2.ToString("X2") + " ;; Z80_Y1\n";
                    z80 += "\tdb &" + frameY.Z80_Y2.ToString("X2") + " ;; Z80_Y2\n";
                }
                else*/
                {
                    z80 += "\tdb &" + frameY.Z80_Y0.ToString("X2") + " ;; Z80_Y0\n";
                    z80 += "\tdb &" + frameY.Z80_Y1.ToString("X2") + " ;; Z80_Y1\n";
                    z80 += "\tdb &" + frameY.Z80_Y2.ToString("X2") + " ;; Z80_Y2\n";
                }

                // TOP
                if (iFrame == 71)
                {
                    z80 += "\tdw &" + frameY.Z80StartLeftSide0.ToString("X4") + " ;; Z80StartLeftSide0\n";
                    z80 += "\tdw &FEC0 ;; Z80DeltaLeftSide0  (hard-coded)\n";
                }
                else if (iFrame == 72)
                {
                    z80 += "\tdw &" + frameY.Z80StartLeftSide0.ToString("X4") + " ;; Z80StartLeftSide0\n";
                    z80 += "\tdw &FEAD ;; Z80DeltaLeftSide0  (hard-coded)\n";
                }
                else if ((iFrame == 73) || (iFrame == 74))
                {
                    z80 += "\tdw &1400 ;; Z80StartLeftSide0  (Special case dummy)\n";
                    z80 += "\tdw &0000 ;; Z80DeltaLeftSide0  (Special case dummy)\n";
                }
                else
                {
                    z80 += "\tdw &" + frameY.Z80StartLeftSide0.ToString("X4") + " ;; Z80StartLeftSide0\n";
                    z80 += "\tdw &" + frameY.Z80DeltaLeftSide0.ToString("X4") + " ;; Z80DeltaLeftSide0\n";
                }

                // BOTTOM
                if ((iFrame == 16)|| (iFrame == 17)|| (iFrame == 18))
                {
                    z80 += "\tdw &1400 ;; Z80StartLeftSide1  (Special case dummy)\n";
                    z80 += "\tdw &0000 ;; Z80DeltaLeftSide1  (Special case dummy)\n";
                }
                else
                if (frameY.Z80YLengthSide1 > 0)
                {
                    z80 += "\tdw &" + frameY.Z80StartLeftSide1.ToString("X4") + " ;; Z80StartLeftSide1\n";
                    z80 += "\tdw &" + frameY.Z80DeltaLeftSide1.ToString("X4") + " ;; Z80DeltaLeftSide1\n";
                }

                // special case for frame > 75
                if(frameY.Z80_Y0 == frameY.Z80_Y1)
                {
                    z80 += "\tdw &" + frameY.Z80StartLeftSide0.ToString("X4") + " ;; Z80StartLeftSide0\n";
                    z80 += "\tdw &" + frameY.Z80DeltaLeftSide0.ToString("X4") + " ;; Z80DeltaLeftSide0\n";
                }

                z80 += "\n";
            }
            File.WriteAllText(outFolder + "_Frames.z80", z80);
            #endregion

            #region Create Z80 Frames index
            z80 = "";
            z80 += ";\n";
            z80 += "CubeRotX_FramesIndex:\n";
            for (int iFrame = 0; iFrame < NbFrames90Deg + 1; iFrame++)
            {
                var frameY = frames[iFrame];

                z80 += "\tdw CubeRotX_Frame" + iFrame + "\n";
            }
            File.WriteAllText(outFolder + "_Frames_Index.z80", z80);
            #endregion

            #region Create GFX of SPECIAL CASEs
            var specialBitmap = new Bitmap(frames[0].TopBitmap.Width, 12 * 2);
            int iSpecialY = 0;
            var specialYR12R13Start = new List<int>();
            var specialYFrameIndex = new List<int>();

            var bottomFrameIndices = new List<int>();
            bottomFrameIndices.Add(16);
            bottomFrameIndices.Add(17);
            bottomFrameIndices.Add(18);
            foreach (var iFrame in bottomFrameIndices)
            {
                specialYR12R13Start.Add(iSpecialY / 2);
                specialYFrameIndex.Add(iFrame);

                var frameY = frames[iFrame];

                var y0 = frameY.Z80_Y1 + 1;
                var y1 = frameY.Z80_Y2;

                var height = ((y1 - y0) + 1) * 2;
                for (var iY = 0; iY < height; iY++)
                {
                    for (var x = 0; x < specialBitmap.Width; x++)
                    {
                        var color = frameY.TopBitmap.GetPixel(x, (y0 * 2) + iY);
                        specialBitmap.SetPixel(x, iSpecialY+iY, color);
                    }
                }

                iSpecialY += height;
            }

            var topFrameIndices = new List<int>();
            topFrameIndices.Add(73);
            topFrameIndices.Add(74);
            foreach (var iFrame in topFrameIndices)
            {
                specialYR12R13Start.Add(iSpecialY / 2);
                specialYFrameIndex.Add(iFrame);

                var frameY = frames[iFrame];

                var y0 = frameY.Z80_Y0;
                var y1 = frameY.Z80_Y1;

                var height = ((y1 - y0)+1)*2;
                if (iFrame == 74)
                {
                    height -= 2;
                }
                for (var iY=0; iY< height; iY++)
                {
                    for (var x = 0; x < specialBitmap.Width; x++)
                    {
                        var color = frameY.TopBitmap.GetPixel(x, (y0 * 2)+iY);
                        specialBitmap.SetPixel(x, iSpecialY+iY, color);
                    }
                }
                iSpecialY += height;
            }
            specialYR12R13Start.Add(iSpecialY / 2);

            specialBitmap.Save(outFolder + "SpecialCases\\SpecialCases.png", ImageFormat.Png);
            
            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = specialBitmap,
                    Mode = 1,
                    R1 = 48,
                    R4 = 21,
                    R9 = 1,
                    StartR12 = 0x00,
                    Palette = palette
                });
            var vlcLength = (specialBitmap.Height / 2) * 96;

            int offset = 0;
            int iIndex = 0;
            for(int i=0; i<12; i++)
            {
                var filename = outFolder + "SpecialCases\\_SpecialCases_VLC0_Index" + iIndex.ToString("00") + ".bin";
                var rawBytes = cpcImage.Pages[0].VLCs[0].CPCBytes.Skip(offset).Take(96).ToArray();

                iIndex++;
                offset += 96;

                File.WriteAllBytes(filename, rawBytes);

                var vlcBBPackedBytes = BBPacker.Pack(ref rawBytes);
                File.WriteAllBytes(filename + ".bbPacked", vlcBBPackedBytes);
            }

            offset = 0;
            iIndex = 0;
            for (int i = 0; i < 12; i++)
            {
                var filename = outFolder + "SpecialCases\\_SpecialCases_VLC1_Index" + iIndex.ToString("00") + ".bin";
                var rawBytes = cpcImage.Pages[0].VLCs[1].CPCBytes.Skip(offset).Take(96).ToArray();

                iIndex++;
                offset += 96;

                File.WriteAllBytes(filename, rawBytes);

                var vlcBBPackedBytes = BBPacker.Pack(ref rawBytes);
                File.WriteAllBytes(filename + ".bbPacked", vlcBBPackedBytes);
            }

            //File.WriteAllBytes(outFolder + "_SpecialCases_VLC0.bin", cpcImage.Pages[0].VLCs[0].CPCBytes.Take(vlcLength).ToArray());
            //File.WriteAllBytes(outFolder + "_SpecialCases_VLC1.bin", cpcImage.Pages[0].VLCs[1].CPCBytes.Take(vlcLength).ToArray());

            z80 = ";\n";
            for(int i=0; i<12; i++)
            {
                z80 += "SPECIALCASE_VLC0_INDEX" + i.ToString("00") + ":\n";
                var filename = "08_Cube\\RotX\\SpecialCases\\_SpecialCases_VLC0_Index" + i.ToString("00") + ".bin.bbPacked";
                z80 += "\t_INCBIN \"" + filename + "\"\n";
                z80 += "\n";

                z80 += "SPECIALCASE_VLC1_INDEX" + i.ToString("00") + ":\n";
                filename = "08_Cube\\RotX\\SpecialCases\\_SpecialCases_VLC1_Index" + i.ToString("00") + ".bin.bbPacked";
                z80 += "\t_INCBIN \"" + filename + "\"\n";
                z80 += "\n";
            }
            File.WriteAllText(outFolder + "_IncBin_SpecialCases.z80", z80);

            var R12R13SpecialCases = new List<short>();
            R12R13SpecialCases.Add(0x560);
            R12R13SpecialCases.Add(0x6C0);
            R12R13SpecialCases.Add(0x720);
            R12R13SpecialCases.Add(0x780);
            R12R13SpecialCases.Add(0x44C0);
            R12R13SpecialCases.Add(0x4520);
            R12R13SpecialCases.Add(0x4580);
            R12R13SpecialCases.Add(0x45E0);
            R12R13SpecialCases.Add(0x4640);
            R12R13SpecialCases.Add(0x46A0);
            R12R13SpecialCases.Add(0x4700);
            R12R13SpecialCases.Add(0x4760);

            z80 = ";\n";
            z80 += "SPECIAL_CASES_UNPACKPTR:\n";
            for (int i = 0; i < 12; i++)
            {
                z80 += "\tdw SPECIALCASE_VLC0_INDEX" + i.ToString("00") + ", &" + R12R13SpecialCases[i].ToString("X4") + "\n";
                z80 += "\tdw SPECIALCASE_VLC1_INDEX" + i.ToString("00") + ", &" + R12R13SpecialCases[i].ToString("X4") + "+&800\n";
            }
            File.WriteAllText(outFolder + "_SpecialCases_UnpackPtr.z80", z80);

            var cubeRotXEdgeBytes = new List<byte>();
            cubeRotXEdgeBytes.AddRange(_SideA_VLC0_Left_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideA_VLC0_Right_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideA_VLC1_Left_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideA_VLC1_Right_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideB_VLC0_Left_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideB_VLC0_Right_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideB_VLC1_Left_Bytes);
            cubeRotXEdgeBytes.AddRange(_SideB_VLC1_Right_Bytes);

            var unpackedBytes = cubeRotXEdgeBytes.ToArray();
            var unpackedFilename = outFolder + "_CubeRotX_Data_VLC.bin";
            File.WriteAllBytes(unpackedFilename, unpackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref unpackedBytes);
            var packedFilename = unpackedFilename + ".packed";
            File.WriteAllBytes(packedFilename, packedBytes);
            
            z80 = "";
            int R12 = 0x20;
            int R13 = 0xF0; // (0x100+96+96+32)/2
            int ScreenPtr = R12R13SpecialCases[0];

            int iR12R13SpecialCaseIndex = 0;

            z80 += "\n";
            z80 += "SpecialCase_StartPtr\t\t\tequ &" + ScreenPtr.ToString("X4") + "\n";
            for (int i = 0; i < specialYFrameIndex.Count; i++)
            {
                var r12R13Start = specialYR12R13Start[i];
                var r12R13End = specialYR12R13Start[specialYR12R13Start.Count - 1];
                if ((i + 1) != specialYR12R13Start.Count)
                {
                    r12R13End = specialYR12R13Start[i + 1];
                }
                var nbVCC = r12R13End - r12R13Start;

                z80 += "\n";

                int vcc = 0;
                for (var iR12R13 = r12R13Start; iR12R13 < r12R13End; iR12R13++)
                {
                    var screenPtr = R12R13SpecialCases[iR12R13SpecialCaseIndex];
                    iR12R13SpecialCaseIndex++;

                    if (vcc == 0)
                    {
                        //z80 += "SpecialCase_Frame" + specialYFrameIndex[i] + "_ScreenPtr\tequ &" + (ScreenPtr).ToString("X4") + "\n";
                        z80 += "SpecialCase_Frame" + specialYFrameIndex[i] + "_ScreenPtr\tequ &" + (screenPtr).ToString("X4") + "\n";
                    }

                    byte r12 = 0;
                    byte r13 = 0;

                    bool isSecond = false;
                    if (screenPtr >= 0x4000)
                    {
                        screenPtr -= 0x4000;
                        isSecond = true;
                    }

                    screenPtr /= 2;
                    r12 = (byte)(screenPtr >> 8);
                    if(isSecond)
                    {
                        r12 += 0x10; // + %00010000
                    }
                    r13 = (byte)(screenPtr & 255);

                    z80 += "SpecialCase_Frame" + specialYFrameIndex[i] + "_VCC" + vcc + "_R12R13\tequ &" + r13.ToString("X2") + r12.ToString("X2") + "\n";
                    z80 += "SpecialCase_Frame" + specialYFrameIndex[i] + "_VCC" + vcc + "_R12\tequ &" + r12.ToString("X2") + "\n";
                    z80 += "SpecialCase_Frame" + specialYFrameIndex[i] + "_VCC" + vcc + "_R13\tequ &" + r13.ToString("X2") + "\n";

                    vcc++;

                    ScreenPtr += 96;
                    R13 += 96 / 2;
                    if (R13 > 255)
                    {
                        R13 -= 256;
                        R12++;
                    }
                }
            }
            File.WriteAllText(outFolder + "_SpecialCases.z80", z80);
            #endregion

            #region Create X reverse (for right)
            z80 = "";
            z80 += ";\n";
            z80 += "XLeftToRight:\n";
            for (int i = 0; i < 96; i++)
            {
                var left = i;
                var right = (0x50 + 0x0F) - left;
                z80 += "\t db &" + right.ToString("X2") + " ;; LEFT=&" + left.ToString("X2") + "\n";
            }
            File.WriteAllText(outFolder + "_XLeftToRight.z80", z80);
            #endregion

            #region Config
            z80 = "";
            z80 += ";\n";
            z80 += "BYTE_Plain_SideA_VLC0\tequ &" + plain_SideA_VLC0.ToString("X2") + "\n";
            z80 += "BYTE_Plain_SideA_VLC1\tequ &" + plain_SideA_VLC1.ToString("X2") + "\n";
            z80 += "BYTE_Plain_SideB_VLC0\tequ &" + plain_SideB_VLC0.ToString("X2") + "\n";
            z80 += "BYTE_Plain_SideB_VLC1\tequ &" + plain_SideB_VLC1.ToString("X2") + "\n";

            File.WriteAllText(outFolder + "_Config.z80", z80);
            #endregion 
        }

        private ushort DoubleToUShort(double vDouble)
        {
            ushort outValue = (ushort)(vDouble * 256);

            byte b = (byte)(outValue >> 8);

            outValue = (ushort)((b << 8) + (outValue & 255));

            return outValue;
        }

        private ushort DoubleToUShort256(double vDouble)
        {
            ushort outValue = (ushort)(vDouble * 256);
            return outValue;
        }

        private Bitmap CreateTileBitmap(Color colorA, Color colorB)
        {
            var bitmap = new Bitmap(4 /* mode 1 */, 1);
            bitmap.SetPixel(0, 0, colorA);
            bitmap.SetPixel(1, 0, colorB);
            bitmap.SetPixel(2, 0, colorA);
            bitmap.SetPixel(3, 0, colorB);
            return bitmap;
        }

        private byte WriteEdgePixels(string outFolder, Brush backgroundBrush, Palette palette, Bitmap tileBitmap, string filename, ref byte[] outBytes, bool isLeft)
        {
            var edgeBitmap = new Bitmap(256 * 4 /* mode 1 */, 1);
            var edgeGraphics = Graphics.FromImage(edgeBitmap);
            edgeGraphics.FillRectangle(backgroundBrush, 0, 0, 256*4, 1);
            edgeGraphics.Dispose();

            int iMaxPen = 1;
            for (int iByte = 0; iByte < 256; iByte++)
            {
                //if ((iByte == 51) || (iByte == 102) || (iByte == 153) || (iByte == 204)) // 256/5=51.2
                if ((iByte == 64) || (iByte == 128) || (iByte == 192))
                {
                    iMaxPen++;
                }

                var edgeBitmapX = iByte * 4;

                for (int i = 0; i < iMaxPen; i++)
                {
                    var pixel = tileBitmap.GetPixel(i, 0);
                    edgeBitmap.SetPixel(edgeBitmapX + i, 0, pixel);
                }
            }

            if (isLeft)
            {
                var pixels = new List<Color>();
                for(int x=0; x<256*4; x++)
                {
                    pixels.Insert(0, edgeBitmap.GetPixel(x, 0));
                }
                for (int x = 0; x < 256 * 4; x++)
                {
                    edgeBitmap.SetPixel(x, 0, pixels[x]);
                }
            }

            edgeBitmap.Save(outFolder + "EdgePixels\\" + filename + ".png", ImageFormat.Png);

            // convert to CPC
            var bytes = new byte[256];
            for (int i = 0; i < 256; i++)
            {
                var edgeBitmapX = i * 4;

                Color p0 = edgeBitmap.GetPixel(edgeBitmapX + 0, 0);
                Color p1 = edgeBitmap.GetPixel(edgeBitmapX + 1, 0);
                Color p2 = edgeBitmap.GetPixel(edgeBitmapX + 2, 0);
                Color p3 = edgeBitmap.GetPixel(edgeBitmapX + 3, 0);

                var paletteEntry0 = palette.AddColor(p0);
                var paletteEntry1 = palette.AddColor(p1);
                var paletteEntry2 = palette.AddColor(p2);
                var paletteEntry3 = palette.AddColor(p3);

                var cpcByte = Pixel.ToMode1Pixel(
                            paletteEntry0.IndexInPalette,
                            paletteEntry1.IndexInPalette,
                            paletteEntry2.IndexInPalette,
                            paletteEntry3.IndexInPalette
                            );

                bytes[i] = cpcByte;
            }

            if (isLeft == false)
            {
                var bytes2 = new byte[256];
                for (int i = 0; i < 256; i++)
                {
                    bytes2[255 - i] = bytes[i];
                }
                bytes = bytes2;
            }

            //File.WriteAllBytes(outFolder + filename, bytes);

            outBytes = bytes;

            //var packer = new Packer();
            //var packedBytes = packer.Convert(ref bytes);
            //File.WriteAllBytes(outFolder + filename + ".packed", packedBytes);

            var firstPixel = bytes[0];
            return firstPixel;
        }
    }
}
