﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _09_SineScroll
    {
        private const int CharHeight = 21;

        public _09_SineScroll()
        {
            try
            {
                CreateBackgroundRasters(
                    Program.ToDataFolder("09_SineScroll/BackgroundRasters.png"),
                    Program.ToDataFolder("09_SineScroll/BackgroundRasters.png"),
                    Program.ToZ80Folder("09_SineScroll/_BackgroundRasters.png.z80"));

                CreateBackgroundRastersBeat(
                    Program.ToDataFolder("09_SineScroll/BackgroundRasters.png"),
                    Program.ToDataFolder("09_SineScroll/BackgroundRasters.png"),
                    Program.ToZ80Folder("09_SineScroll/_BackgroundRastersBeat.png.z80"));

                CreateForegroundRasters(
                    Program.ToDataFolder("09_SineScroll/ForegroundRasters.png"),
                    Program.ToDataFolder("09_SineScroll/ForegroundRasters.png"),
                    Program.ToZ80Folder("09_SineScroll/_ForegroundRasters.png.z80"));

                CreateBackgroundZoom(
                    Program.ToZ80Folder("09_SineScroll/Debug/"),
                    Program.ToZ80Folder("09_SineScroll/_BackgroundXSine.bin"),
                    Program.ToZ80Folder("09_SineScroll/_BackgroundZoomLevels.z80"));

                CreateFont(     Program.ToDataFolder("09_SineScroll/ProportionalFont_x21_Fixed.png"),
                                Program.ToZ80Folder("09_SineScroll/_ScrollText.z80"),
                                Program.ToZ80Folder("09_SineScroll/_VertSlices.bin"),
                                " ABCDEFGHIJKLMNOPQRSTUVWXYZ!:.?()1234567890\\/@",
                                  //use rock the party instead?
                                  //"                           SO YOU WANT MORE AMSTRAD DEMOS    THEN HERE WE ARE    SINUS SCROLLERS STILL BEAT THE PARTY                                                     "
                                    "                           SO YOU WANT MORE AMSTRAD DEMOS    THEN HERE WE ARE    NOW LETS REFRESH THE OLD PHAT TUNNEL                                                   "
                                //     "                            HUMAN NOT HUMAN   FUTURE NO FUTURE   CHANGE NO CHANGE   MEMORY NO MEMORY   R E V O L U T I O N                                                "
                                //"                           GREETINGS TO   LOGON SYSTEM   FUTURS   BENEDICTION   VANITY   BATMAN GROUP   GPA   IMPACT   ARKOS                                              "
                                //"                           EVERYTHING CHANGES TO A POINT THAT IT STOPS AND IT TURNS AROUND ... THIS IS WHAT HAPPENS RIGHT NOW                                          "
                                //"                               WHO NEEDS GREETINGS IN SCROLLERS  ??? ?    BLABLA BLAH THIS IS UNREADABLE ! ! .. ...  : )                                               "
                                //"                               CREATING THIS DEMO TOOK US FOUR YEARS  !!!  HELLO  REVISION  2 0 1 8  !!!                                              "
                                //"                               BLUE  IS  GREENER  THAN  PURPLE  FOR  SURE                                                                 "
                                //"                               THE  OLD SCHOOL  ONE PIXEL  SINE  SCROLLER  STILL  RULES  FOR  SURE . .... P H A N T A S T I C                                          "
                                //"                               PRIVATE JOKE ... HI  A L K A M A  AND  S O U N D Y  !!! .. HOPE YOU DID NOT WAIT TOO LONG !!!                                           "
                                );

                CreateSinMoves(
                    Program.ToZ80Folder("09_SineScroll/_BackgroundYSine.bin"),
                    Program.ToZ80Folder("09_SineScroll/_ForegroundYSine.bin"),
                    Program.ToZ80Folder("09_SineScroll/_ScrollYSin1.bin"),
                    Program.ToZ80Folder("09_SineScroll/_ScrollYSin2.bin"),
                    Program.ToZ80Folder("09_SineScroll/_ScrollCharHeight.bin"),
                    Program.ToZ80Folder("09_SineScroll/_PrecaSine.bin")
                    );
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        private void CreateBackgroundZoom(string outFolder, string backgroundXSineFilename, string destZ80)
        {
            const int zoomLevels = 1;

            var cpcBytes = SinCurve.CreateSin(256, 0, zoomLevels, 4);
            for(int i=0; i<256; i++)
            {
                cpcBytes[i] = (byte)(cpcBytes[i] * 2);
            }
            //File.WriteAllBytes(backgroundXSineFilename, cpcBytes);
            
            var z80 = "";

            z80 += ";\n";
            z80 += "CharHeightMul2   equ 21 + 21\n";
            z80 += "Empty            equ CharHeightMul2 * 0\n";
            z80 += "FlipBackground   equ CharHeightMul2 * 1\n";
            z80 += "FlipForeground   equ CharHeightMul2 * 2\n";
            z80 += "\n";

            var baseBitmap = new Bitmap(96 * 4, 1);
            for (int x = 0; x < baseBitmap.Width; x++)
            {
                var on = false;
                var iX = x;
                while (iX>=32)
                {
                    iX -= 32;
                }
                if(iX<16)
                {
                    on = true;
                }

                var color = Color.Black;
                if(on)
                {
                    color = Color.White;
                }

                baseBitmap.SetPixel(x, 0, color);
            }

            z80 += ";\n";
            z80 += "ZoomLevels:\n";
            for (int iZoomLevel = 0; iZoomLevel < zoomLevels; iZoomLevel++)
            {
                z80 += "\tdw ZoomLevel" + iZoomLevel + "\n";
            }
            z80 += "\n";

            const int lineWidth = 96 + 32;

            for (int iZoomLevel=0; iZoomLevel<zoomLevels; iZoomLevel++)
            {
                z80 += ";\n";
                z80 += "ZoomLevel" + iZoomLevel + ":\n";

                var newWidth = baseBitmap.Width - (iZoomLevel*19);
                var newSize = new Size(newWidth, 1);
                var bmp = new Bitmap(baseBitmap, newSize);

                var xStart = newWidth / 2;
                xStart -= lineWidth / 2;
                var xEnd = xStart + lineWidth;
                for(int iX=xStart; iX<xEnd; iX+=2)
                {
                    Color pixel;

                    if ((iX + 1) < xEnd)
                    {
                        pixel = bmp.GetPixel(iX + 1, 0);
                        if (pixel.G > 0)
                        {
                            z80 += "\tdb FlipBackground\n";
                        }
                        else
                        {
                            z80 += "\tdb Empty\n";
                        }
                    }

                    pixel = bmp.GetPixel(iX, 0);
                    if (pixel.G > 0)
                    {
                        z80 += "\tdb FlipBackground\n";
                    }
                    else
                    {
                        z80 += "\tdb Empty\n";
                    }
                }

                bmp.Save(outFolder + "_ZoomLevel" + iZoomLevel + ".png", ImageFormat.Png);
            }
            
            //File.WriteAllText(destZ80, z80);
        }

        Bitmap bmp1;
        Bitmap bmp2;
        int bmpY;
        private void GetNextColors(out string cpcColorLabel1, out string cpcColorLabel2)
        {
            var y = bmpY;

            var pixel1 = bmp1.GetPixel(0, y);
            var palette1 = new Palette();
            palette1.AddColor(pixel1);

            var pixel2 = bmp2.GetPixel(0, y);
            var palette2 = new Palette();
            palette2.AddColor(pixel2);

            var gateArrayColor1 = palette1.ToFirmwareByteArray()[0];
            var gateArrayColor2 = palette2.ToFirmwareByteArray()[0];

            cpcColorLabel1 = "COLOR" + gateArrayColor1;
            cpcColorLabel2 = "COLOR" + gateArrayColor2;

            bmpY++;
        }

        private void CreateBackgroundRasters(string imageFilename1, string imageFilename2, string destZ80)
        {
            bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));
            bmpY = 50;

            string c0 = "";
            string c1 = "";
            var z80 = "";

            for (var iRepeat1 = 0; iRepeat1 < 2 * 2 * 4 * 2; iRepeat1++)
            {
                /*for (var iRepeat2 = 0; iRepeat2 < 2; iRepeat2++)
                {
                    c0 = "COLOR1";
                    c1 = "COLOR2";

                    z80 += "\tdb BACKGROUND_PEN0, " + c0 + ", BACKGROUND_PEN1, " + c1 + "\n";
                    z80 += "\tdb BACKGROUND_PEN1, " + c1 + ", BACKGROUND_PEN0, " + c0 + "\n";
                }*/
                for (var iRepeat3 = 0; iRepeat3 < 1; iRepeat3++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb BACKGROUND_PEN0, " + c1 + ", BACKGROUND_PEN1, " + c0 + "\n";
                    z80 += "\tdb BACKGROUND_PEN1, " + c0 + ", BACKGROUND_PEN0, " + c1 + "\n";
                }
            }

            File.WriteAllText(destZ80, z80);
        }

        private void CreateBackgroundRastersBeat(string imageFilename1, string imageFilename2, string destZ80)
        {
            bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));
            bmpY = 80;

            string c0 = "";
            string c1 = "";
            var z80 = "";

            for (var iRepeat1 = 0; iRepeat1 < 2 * 2 * 4 * 2; iRepeat1++)
            {
                /*for (var iRepeat2 = 0; iRepeat2 < 2; iRepeat2++)
                {
                    c0 = "COLOR1";
                    c1 = "COLOR2";

                    z80 += "\tdb BACKGROUND_PEN0, " + c0 + ", BACKGROUND_PEN1, " + c1 + "\n";
                    z80 += "\tdb BACKGROUND_PEN1, " + c1 + ", BACKGROUND_PEN0, " + c0 + "\n";
                }*/
                for (var iRepeat3 = 0; iRepeat3 < 1; iRepeat3++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb BACKGROUND_PEN0, " + c1 + ", BACKGROUND_PEN1, " + c0 + "\n";
                    z80 += "\tdb BACKGROUND_PEN1, " + c0 + ", BACKGROUND_PEN0, " + c1 + "\n";
                }
            }

            File.WriteAllText(destZ80, z80);
        }

        private void CreateForegroundRasters(string imageFilename1, string imageFilename2, string destZ80)
        {
            bmp1 = new Bitmap(System.Drawing.Image.FromFile(imageFilename1));
            bmp2 = new Bitmap(System.Drawing.Image.FromFile(imageFilename2));
            bmpY = 35;

            string c0 = "";
            string c1 = "";
            var z80 = "";

            for (var iRepeat1 = 0; iRepeat1 < 4; iRepeat1++)
            {
                for (var iRepeat2 = 0; iRepeat2 < 8; iRepeat2++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN0, " + c0 + ", FOREGROUND_PEN1, " + c1 + "\n";
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN1, " + c1 + ", FOREGROUND_PEN0, " + c0 + "\n";
                }
                for (var iRepeat3 = 0; iRepeat3 < 8; iRepeat3++)
                {
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN0, " + c1 + ", FOREGROUND_PEN1, " + c0 + "\n";
                    GetNextColors(out c0, out c0);
                    GetNextColors(out c1, out c1);
                    z80 += "\tdb FOREGROUND_PEN1, " + c0 + ", FOREGROUND_PEN0, " + c1 + "\n";
                }
            }

            File.WriteAllText(destZ80, z80);
        }

        private byte[] DoubleValues(byte[] src)
        {
            for (int i = 0; i < src.Length; i++)
            {
                src[i] = (byte)(src[i] * 4);
            }

            return src;
        }

        public void CreateSinMoves( string backgroundYSineFilename,
                                    string foregroundYSineFilename,
                                    string scrollYSin1Filename,
                                    string scrollYSin2Filename,
                                    string scrollCharHeightTableFilename,
                                    string precalculatedSineFilename)
        {
            File.WriteAllBytes(backgroundYSineFilename, DoubleValues(SinCurve.CreateSinAbs(256, 0, 96, 1)));
            File.WriteAllBytes(foregroundYSineFilename, DoubleValues(SinCurve.CreateCos(256, 0, 50, 1)));

            //File.WriteAllBytes(scrollYSin1Filename, SinCurve.CreateSin(256, 0, 56, 2));
            //File.WriteAllBytes(scrollYSin2Filename, SinCurve.CreateCos(256, 0, 24, 4));
            //File.WriteAllBytes(scrollYSin1Filename, SinCurve.CreateSin(256, 0, 66, 2));
            //File.WriteAllBytes(scrollYSin2Filename, SinCurve.CreateCos(256, 0, 8, 4));
            var charHeightList = new List<byte>();
            byte iCharHeight = 0;
            for(int i=0; i<256; i++)
            {
                charHeightList.Add(iCharHeight);
                iCharHeight++;
                if (iCharHeight == CharHeight)
                {
                    iCharHeight = 0;
                }
            }
            //File.WriteAllBytes(scrollCharHeightTableFilename, charHeightList.ToArray());

            // scroll anim
            var bgNbFrames = 64;
            var cpcNbFrames = 31; // cpcNbFrames is /2 -1 because it's a sin (same anim)

            var frameBytes = new List<byte>();
            for (int iFrame = 0; iFrame < cpcNbFrames; iFrame++)
            {
                double animProgress = (double)iFrame/(double)bgNbFrames;

                //var iFrameSine1 = Math.Sin(2 * Math.PI * animProgress*1.8)*1.2;
                //var iFrameSine2 = Math.Cos(2 * Math.PI * animProgress)*0.9;
                var iFrameSine1 = (2 * Math.PI * animProgress)*2;
                var iFrameSine2 = (2 * Math.PI * animProgress)*2;

                var iFrameSine12 = (2 * Math.PI * animProgress) * 1;
                var iFrameSine22 = (2 * Math.PI * animProgress) * 1;

                var frameYSine = 0 *((Math.Cos(2 * Math.PI * animProgress)));

                for (int iX = 0; iX < 96; iX++)
                {
                    double floatValue = (((8.0f * Math.Cos(iFrameSine1)) - (8.0f * (Math.Sin(iFrameSine2)))));
                    floatValue += frameYSine;

                    iFrameSine1 += 0.024f;
                    iFrameSine2 += 0.012f;

                    int b = (int)floatValue;
                    while (b >= CharHeight)
                    {
                        b -= CharHeight;
                    }
                    while (b<0)
                    {
                        b += CharHeight;
                    }

                    // flipBackground and shadeBackground are mutually EXCLUSIVE
                    //var flipBackground = false;
                    var shadeBackground = false;

                    var flipForeground = false;

                    double floatValue2 = Math.Abs(((14.0f * Math.Cos(iFrameSine12)) - (14.0f * (Math.Sin(iFrameSine22)))));
                    floatValue2 += frameYSine;
                    iFrameSine12 += 0.074f;
                    iFrameSine22 += 0.054f;

                    var v = 16;
                    var w = 0;
                    var f = animProgress * 67; // ((double)iFrame*0.5) * 1;
                    var ix = -animProgress+floatValue2;
                    var ix2 = animProgress + floatValue2;

                    if ((Math.Abs((w+ix - f) % (v-1)) < v/4))
                    {
                        //flipBackground = true;
                    }

                    if ((Math.Abs((w + ix - f) % (v-1)) < v/2))
                    {
                        shadeBackground = true;
                    }
                    
                    if ((Math.Abs((w - iX+ animProgress*96) % (v*2-1)) < v/4))
                    {
                        flipForeground = true;
                    }

                    if (shadeBackground)
                    {
                        b += (CharHeight + CharHeight) * 1;
                    }
                    if (flipForeground)
                    {
                        b += (CharHeight + CharHeight) *2;
                    }
                    frameBytes.Add((byte)b);
                }
            }

            var unpackedBytes = frameBytes.ToArray();
            File.WriteAllBytes(precalculatedSineFilename, unpackedBytes);
            
            var packer = new Packer();
            var packedBytes = packer.Convert(ref unpackedBytes);
            //File.WriteAllBytes(precalculatedSineFilename + ".packed", packedBytes);
        }

        private class CharBitmap
        {
            public byte CPCCharWidthInBytes;
            public char Letter;
            public byte[] Data;
        }

        private void CreateFont(string srcImage, string dstZ80, string dstSliceFilename, string charset, string scrollText)
        {
            var bitmap = new Bitmap(srcImage);

            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 1000,
                R4 = 38,
                R9 = 7,
                StartVLC = 0,
                FirstPixelInPaletteX = 0,
                FirstPixelInPaletteY = 0
            }
            );

            var charBitmaps = new List<CharBitmap>();
            int iCharsetIndex = 0;

            //var cpcCharWidth = image.CPCRawBuffer_Width / 8;
            int baseY = 0;
            var isFirst = true;
            int cpcCharWidth = 0;
            int baseX = 0;
            for (int iChar=0; iChar<charset.Length; iChar++)
            {
                baseX += cpcCharWidth;

                char letter = charset[iCharsetIndex];

                if (isFirst)
                {
                    isFirst = false;
                    cpcCharWidth = 16 / 4; // 1st character (space) is 16 pixels width
                }
                else
                {
                    // search for complete empty vertical line to determine width of a character
                    byte p0, p1, p2, p3;
                    var emptyPixel = image.CPCRawBuffer[0];                    
                    Pixel.FromMode1Pixel(emptyPixel, out p0, out p1, out p2, out p3);
                    byte emptyBitPixelP3 = p3;

                    cpcCharWidth = 0;
                    var vertLineFound = false;
                    while(!vertLineFound)
                    {
                        int wrongPixelsCount = 0;
                        for (int y=0; y<image.CPCRawBuffer_Height; y++)
                        {
                            var currentPixel = image.CPCRawBuffer[(y*image.CPCRawBuffer_Width)+baseX+cpcCharWidth];
                            Pixel.FromMode1Pixel(currentPixel, out p0, out p1, out p2, out p3);

                            if(p3 != emptyBitPixelP3)
                            {
                                wrongPixelsCount++;
                            }
                        }

                        if(wrongPixelsCount>0)
                        {
                            cpcCharWidth++;
                        }
                        else
                        {
                            vertLineFound = true;
                        }
                    }
                    cpcCharWidth++;
                }

                var data = new List<byte>();

                // WARNING!!! Vertical slice of cpc-pixel-data!
                for (int x = 0; x < cpcCharWidth; x++)
                {
                    #region CASE 1: BACKGROUND NORMAL - FOREGROUND NORMAL
                    for (var y=0; y<CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];
                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];
                        data.Add(b);
                    }
                    #endregion

                    #region CASE 3: FLIP BACKGROUND - FLIP FOREGROUND
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN0 to PEN2 SHADE
                        if (p0 == 0) { p0 = 2; };
                        if (p1 == 0) { p1 = 2; };
                        if (p2 == 0) { p2 = 2; };
                        if (p3 == 0) { p3 = 2; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN0 to PEN2 SHADE
                        if (p0 == 0) { p0 = 2; };
                        if (p1 == 0) { p1 = 2; };
                        if (p2 == 0) { p2 = 2; };
                        if (p3 == 0) { p3 = 2; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    #endregion

                    #region CASE 4: BACKGROUND NORMAL - FOREGROUND REVERSE
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);

                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    #endregion

                    #region CASE 6: FLIP BACKGROUND - FLIP FOREGROUND
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);


                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        // PEN0 to PEN2 SHADE
                        if (p0 == 0) { p0 = 2; };
                        if (p1 == 0) { p1 = 2; };
                        if (p2 == 0) { p2 = 2; };
                        if (p3 == 0) { p3 = 2; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    for (var y = 0; y < CharHeight; y++)
                    {
                        var b = image.CPCRawBuffer[((baseY + y) * image.CPCRawBuffer_Width) + (baseX + x)];

                        byte p0, p1, p2, p3;
                        Pixel.FromMode1Pixel(b, out p0, out p1, out p2, out p3);


                        // PEN1 to PEN3
                        if (p0 == 1) { p0 = 3; };
                        if (p1 == 1) { p1 = 3; };
                        if (p2 == 1) { p2 = 3; };
                        if (p3 == 1) { p3 = 3; };

                        // PEN0 to PEN2 SHADE
                        if (p0 == 0) { p0 = 2; };
                        if (p1 == 0) { p1 = 2; };
                        if (p2 == 0) { p2 = 2; };
                        if (p3 == 0) { p3 = 2; };

                        b = Pixel.ToMode1Pixel(p0, p1, p2, p3);

                        data.Add(b);
                    }
                    #endregion
                }

                var charBitmap = new CharBitmap
                {
                    CPCCharWidthInBytes = (byte)cpcCharWidth,
                    Letter = letter,
                    Data = data.ToArray()
                };
                charBitmaps.Add(charBitmap);

                iCharsetIndex++;
            }

            // parse scroll text and generate bitmap data
            var scrollText_ScrollTextSlicesData = new List<int>();
            var vertSlices_Data = new List<byte[]>();
            var sliceByteSize = (CharHeight + CharHeight) * 4;

            foreach (var letter in scrollText)
            {
                int foundIndex = -1;
                for (int i = 0; i < charBitmaps.Count; i++)
                {
                    var charBitmap = charBitmaps[i];

                    if (charBitmap.Letter == letter)
                    {
                        foundIndex = i;
                        break;
                    }
                }
                if (foundIndex == -1)
                {
                    // letter not found in charset
                    System.Diagnostics.Debugger.Break();
                }

                var foundCharBitmap = charBitmaps[foundIndex];

                cpcCharWidth = foundCharBitmap.CPCCharWidthInBytes;

                int iOffset = 0;
                for (int iSlice = 0; iSlice < cpcCharWidth; iSlice++)
                {
                    var slice = new List<byte>();
                    for (int i = 0; i < sliceByteSize; i++)
                    {
                        slice.Add(foundCharBitmap.Data[iOffset]);
                        iOffset++;
                    }

                    var vertSliceFoundIndex = -1;
                    for (int iVertSliceIndex = 0; iVertSliceIndex < vertSlices_Data.Count(); iVertSliceIndex++)
                    {
                        var found = true;
                        var iVertSlice = vertSlices_Data[iVertSliceIndex];
                        for (int i = 0; i < sliceByteSize; i++)
                        {
                            if (iVertSlice[i] != slice[i])
                            {
                                found = false;
                            }
                        }
                        if (found)
                        {
                            vertSliceFoundIndex = iVertSliceIndex;
                            break;
                        }
                    }

                    if (vertSliceFoundIndex != -1)
                    {
                        scrollText_ScrollTextSlicesData.Add(vertSliceFoundIndex);
                    }
                    else
                    {
                        var newIndex = vertSlices_Data.Count();
                        vertSlices_Data.Add(slice.ToArray());

                        scrollText_ScrollTextSlicesData.Add(newIndex);
                    }
                }
            }

            var vertSlices_ByteData = new List<byte>();
            foreach (var sliceData in vertSlices_Data)
            {
                foreach (var b in sliceData)
                {
                    vertSlices_ByteData.Add(b);
                }
            }
            var unpackedBytes = vertSlices_ByteData.ToArray();
            File.WriteAllBytes(dstSliceFilename, unpackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref unpackedBytes);
            File.WriteAllBytes(dstSliceFilename + ".packed", packedBytes);
            
            var keyword = Program.FilenameToKeyWordName(dstZ80);

            var z80 = ";\n";
            foreach (var sliceIndex in scrollText_ScrollTextSlicesData)
            {
                var offset = sliceIndex * sliceByteSize;

                z80 += "\tdw SineScroll_VertSlices - " + (0x7C * 256) + " + " + offset + " ;; " + sliceIndex + "\n";
            }
            
            File.WriteAllText(dstZ80, z80);
        }
    }
}
