﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _10_Circle
    {
        const int BarSize = 140;

        const int LogoMaxRepeatLines = 8;
        const int LogoMaxSkipLines = 0;

        public _10_Circle()
        {
            try
            {
                var logoPalette = new Palette();
                int logoHeight = 0;
                var logoRawBytes = CreateLogo( Program.ToDataFolder("10_Circle/LogoPhX_192x16_1.png"),
                            Program.ToZ80Folder("10_Circle/"),
                            ref logoHeight,
                            ref logoPalette);

                CreateSinMoves(Program.ToZ80Folder("10_Circle/"));

                CreateLogoOpcodes(Program.ToZ80Folder("10_Circle/"));
                CreateLogoSinMoves(Program.ToZ80Folder("10_Circle/"));

                var r12r13 = new List<byte>();
                CreateCircle(Program.ToZ80Folder("10_Circle/"), ref r12r13, logoHeight, logoRawBytes);

                //var flashBeat = new int[] { 0, 1, 4, 5, 11, 14, 21 };
                //var flashBeat = new int[] { 0, 5, 11, 14 };
                var flashBeat = new int[] { 0, 1, 2, 5 };
                //var flashBeat = new int[] { 0, 1, 3, 4 };

                CreateRasterAnim(Program.ToDataFolder("10_Circle/Bar420_1.png"),
                                flashBeat[0], 
                                Program.ToZ80Folder("10_Circle/_BeatRasterBars0.bin"));
                CreateRasterAnim(Program.ToDataFolder("10_Circle/Bar420_1.png"),
                                flashBeat[1],
                                Program.ToZ80Folder("10_Circle/_BeatRasterBars1.bin"));
                CreateRasterAnim(Program.ToDataFolder("10_Circle/Bar420_1.png"),
                                flashBeat[2],
                                Program.ToZ80Folder("10_Circle/_BeatRasterBars2.bin"));
                CreateRasterAnim(Program.ToDataFolder("10_Circle/Bar420_1.png"),
                                flashBeat[3],
                                Program.ToZ80Folder("10_Circle/_BeatRasterBars3.bin"));

                //CreateR12R13Anim(Program.ToZ80Folder("10_Circle/"), ref r12r13);
                CreateR12R13_BC00BD00_SinTable(Program.ToZ80Folder("10_Circle/"), ref r12r13, logoHeight, logoRawBytes);
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public class LogoFrame
        {
            public byte PosY = 0;
            public byte Height = 0;
            public byte DeltaMinus1 = 0;
        }

        public void CreateLogoSinMoves(string outFolder)
        {
            int maxHeightRatio1 = 268 - 16+2;

            var frames = new List<LogoFrame>();

            var sinCurve = SinCurve.CreateCos(256, 0, maxHeightRatio1+1, 1).Take(128).ToArray();
            var sinHeight = SinCurve.CreateCos(256, 16, 64+1, 2).Skip(64).Take(128).ToArray();
            for(int i=0; i<128; i++)
            {
                var frame = new LogoFrame
                {
                    PosY = sinCurve[i],
                    Height = sinHeight[i]
                };
                frames.Add(frame);
            }

            for (int i = 127; i >=0; i--)
            {
                var frame = new LogoFrame
                {
                    PosY = sinCurve[i],
                    Height = sinHeight[i]
                };
                frames.Add(frame);
            }

            foreach (var frame in frames)
            {
                frame.PosY = (byte)((frame.PosY / 2) * 2);
                frame.Height = (byte)((frame.Height / 2) * 2);
            }

            var b0 = new List<byte>();
            var b1 = new List<byte>();
            var b2 = new List<byte>();

            foreach (var frame in frames)
            {
                double posY = frame.PosY;
                double height = frame.Height;
                
                double ratio = 256 / ((height / 16) * 256);
                ratio *= 256;

                frame.DeltaMinus1 = (byte)(ratio - 1);

                b0.Add(frame.PosY);
                b1.Add(frame.Height);
                b2.Add(frame.DeltaMinus1);
            }

            var b = new List<byte>();
            b.AddRange(b0); // posy
            b.AddRange(b1); // height
            b.AddRange(b2); // delta zoom

            var bytes = b.ToArray();
            File.WriteAllBytes(outFolder + "_SinLogoMove.bin", bytes);
        }

        private void AddPushLabel(string label, ref bool isPair, ref string z80, ref int nops, ref int pushCount)
        {
            var pairImpairName = "Pair";
            if(!isPair)
            {
                pairImpairName = "Impair";
            }

            label = label.Replace("XXX", pairImpairName);

            z80 += "\tld (hl), " + label + " AND 255 ;3\n"; nops += 3;
            z80 += "\tinc l ;1\n"; nops += 1;
            z80 += "\tld (hl), " + label + " / 256 ;3\n"; nops += 3;
            z80 += "\tinc hl ;2\n"; nops += 2;

            if (isPair)
            {
                isPair = false;
            }
            else
            {
                isPair = true;
            }

            pushCount++;
        }

        public void CreateLogoOpcodes(int repeatCount, int skipCount, ref string z80, ref int nops, int maxNops)
        {
            bool isPair = true;
            z80 = "";
            nops = 0;

            int pushCount = 0;

            bool lastIsBlackLine = false;

            if (skipCount > 2)
            {
                z80 += "\tld bc, " + ((skipCount - 2) * 2) + " ;3\n"; nops += 3;
            }

            AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);
            AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);

            for (int iLine = 0; iLine < 14; iLine++)
            {
                lastIsBlackLine = false;

                for (int iRepeat=0; iRepeat<repeatCount; iRepeat++)
                {
                    AddPushLabel("CircleXXXShowLine"+iLine, ref isPair, ref z80, ref nops, ref pushCount);
                }

                if(skipCount==0)
                {
                    // do nothing
                }
                else if (skipCount == 1)
                {
                    // one single black line
                    AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);

                    lastIsBlackLine = true;
                }
                else if (skipCount == 2)
                {
                    // two single black line
                    AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);
                    AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);

                    lastIsBlackLine = true;
                }
                else if (skipCount > 2)
                {
                    // this will display 2 black lines (cf. +1 DoFX below)
                    AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);

                    var skipRepeat = (skipCount - 2)+1+1;

                    z80 += "\tadd hl, bc ;3\n"; nops += 3;

                    for (int i=0; i<skipRepeat; i++)
                    {
                        //AddPushLabel("Circle_XXX_DoFX", ref isPair, ref z80, ref nops, ref pushCount);
                        if (isPair)
                        {
                            isPair = false;
                        }
                        else
                        {
                            isPair = true;
                        }
                    }

                    // 2 black lines
                    AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);
                    AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);

                    lastIsBlackLine = true;
                }
            }

            // ends with one (not 2, because it's automatically repeated twice)
            if(!lastIsBlackLine)
            {
                AddPushLabel("CircleXXXShowBlackLine", ref isPair, ref z80, ref nops, ref pushCount);
            }

            // pad on Pair lines
            if(!isPair)
            {
                AddPushLabel("Circle_XXX_DoFX", ref isPair, ref z80, ref nops, ref pushCount);
                isPair = true;
            }

            if (maxNops > 0)
            {
                var remainingNops = maxNops - nops;
                if(remainingNops>0)
                {
                    var nbLines = 0;
                    while(remainingNops>=64)
                    {
                        nbLines++;
                        remainingNops -= 64;
                    }

                    if(nbLines>0)
                    {
                        z80 += "\t_LINES " + nbLines + "\n";
                    }
                    if(remainingNops>0)
                    {
                        z80 += "\t_NOPS " + remainingNops + "\n";
                    }
                }
            }

            //var relocateStackZ80 = "";
            //relocateStackZ80 += "\tld hl, " + (pushCount + 2) + "\n";

            z80 += "\tjp Circle_WriteRepeatSkip_End\n";
        }

        public string CreateLabel(int iRepeat, int iSkip)
        {
            if(iSkip==3)
            {
                //return "CircleLogo_Repeat" + iRepeat + "_Skip3AndBeyond";
            }

            return "CircleLogo_Repeat" + iRepeat + "_Skip" + iSkip;
        }

        public void CreateLogoOpcodes(string outFolder)
        {
            var maxNops = 0;

            // 1st pass: find max nops
            var z80Labels = "";
            for(int iRepeat=1; iRepeat<LogoMaxRepeatLines; iRepeat++)
            {
                int iSkip = 0;

                string z80 = "";
                int nops = 0;
                CreateLogoOpcodes(iRepeat, iSkip, ref z80, ref nops, 0);
                if (maxNops < nops)
                {
                    maxNops = nops;
                }
                z80Labels += ";; " + CreateLabel(iRepeat, iSkip) + " ;; " + nops + " nops\n";

                if (iRepeat == (LogoMaxRepeatLines - 1))
                {
                    for (iSkip = 1; iSkip < LogoMaxSkipLines; iSkip++)
                    {
                        z80 = "";
                        nops = 0;
                        CreateLogoOpcodes(iRepeat, iSkip, ref z80, ref nops, 0);
                        if (maxNops < nops)
                        {
                            maxNops = nops;
                        }
                        z80Labels += ";; " + CreateLabel(iRepeat, iSkip) + " ;; " + nops + " nops\n";
                    }
                }
            }

            // 2nd pass: pad to max nops + write full z80
            var nbLines = 0;
            var remainingNops = maxNops;
            while (remainingNops >= 64)
            {
                nbLines++;
                remainingNops -= 64;
            }
            var fullZ80 = "";
            fullZ80 += ";; MaxRepeatLines=" + LogoMaxRepeatLines + "  MaxSkipLines=" + LogoMaxSkipLines + "\n";
            fullZ80 += ";; MaxMacroLines=" + nbLines + "  MaxMacroNops=" + remainingNops + "  MaxNops=" + maxNops + "\n";
            fullZ80 += "\n" + z80Labels;

            for (int iRepeat = 1; iRepeat < LogoMaxRepeatLines; iRepeat++)
            {
                int iSkip = 0;

                string z80 = "";
                int nops = 0;
                CreateLogoOpcodes(iRepeat, iSkip, ref z80, ref nops, maxNops);                
                fullZ80 += "\n";
                fullZ80 += ";" + nops + "\n";
                fullZ80 += CreateLabel(iRepeat, iSkip) + ":\n";
                fullZ80 += z80;

                if (iRepeat == (LogoMaxRepeatLines - 1))
                {
                    for (iSkip = 1; iSkip < LogoMaxSkipLines; iSkip++)
                    {
                        z80 = "";
                        nops = 0;
                        CreateLogoOpcodes(iRepeat, iSkip, ref z80, ref nops, maxNops);
                        fullZ80 += "\n";
                        fullZ80 += ";" + nops + "\n";
                        fullZ80 += CreateLabel(iRepeat, iSkip) + ":\n";
                        fullZ80 += z80;
                    }
                }
            }

            File.WriteAllText(outFolder + "_CircleLogoRepeatSkip.z80", fullZ80);
        }

        public byte[] CreateLogo(string bitmapFilename, string outFolder, ref int logoHeight, ref Palette logoPalette)
        {
            var bitmap = new Bitmap(bitmapFilename);

            logoHeight = bitmap.Height; // +1 because of black line

            var palette = new Palette(3);
            palette.AddColor(Color.Black);
            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 0,
                R1 = 48,
                R4 = 0,
                R9 = 0,
                Palette = palette
            }
            );

            logoPalette = palette;

            File.WriteAllText(outFolder + "_CircleLogoPalette.z80", palette.ToZ80TableString(Program.FilenameToKeyWordName("Circle_Logo")));

            var bytes = new List<byte>();
            /*for(int i=0; i<96; i++)
            {
                bytes.Add((byte)0xFF); // write black line, MODE1 will draw PEN3, MODE0 will draw PEN15, both pens are set to black color
            }
            bytes.AddRange(image.CPCRawBuffer.Take(96 * 15).ToArray());*/
            bytes.AddRange(image.CPCRawBuffer.Take(96 * 16).ToArray());

            var rawBytes = bytes.ToArray();
            //File.WriteAllBytes(outFolder + "_Logo.bin", rawBytes);

            return rawBytes;
        }

        public void CreateSinMoves(string outFolder)
        {
            File.WriteAllBytes(outFolder + "_SinRasterBarsMove.bin", SinCurve.CreateSin(256, 10, 268-BarSize, 2)); // vitesse haut en bas
            //File.WriteAllBytes(outFolder + "_SinDisto.bin", SinCurve.CreateSin(256, 0, 0, 2)); // vitesse haut en bas

            var height = 68;// 21 * 4;
            var curve0 = SinCurve.CreateCos(256, 1, height/2, 1);
            File.WriteAllBytes(outFolder + "_SinDistoMove0.bin", curve0);
            var curve1 = SinCurve.CreateCos(268 * 2, 0, height/2, 1).Take(286).ToArray();
            File.WriteAllBytes(outFolder + "_SinDistoMove1.bin", curve1);
        }

        public void CreateCircle(string outFolder, ref List<byte> r12r13, int logoHeight, byte[] logoRawBytes)
        {
            var color0 = Color.FromArgb((int)(Pixel.CPCAsicPalette[1] | 0xFF000000));
            var color1 = Color.FromArgb((int)(Pixel.CPCAsicPalette[3] | 0xFF000000));
            var color2 = Color.FromArgb((int)(Pixel.CPCAsicPalette[4] | 0xFF000000));
            var color3 = Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000));

            var palette = new Palette();
            palette.AddColor(color0);
            palette.AddColor(color1);
            palette.AddColor(color2);
            palette.AddColor(color3);

            var pen0 = new SolidBrush(color0);
            var pen1 = new SolidBrush(color1);
            var pen2 = new SolidBrush(color2);
            var pen3 = new SolidBrush(color3);

            var pens = new List<SolidBrush>();
            pens.Add(pen0);
            pens.Add(pen1);
            pens.Add(pen2);
            pens.Add(pen1);

            var penIndex = 0;

            var bitmap = new Bitmap(48 * 8, 21*4);

            var graphics = Graphics.FromImage(bitmap);

            double zoomStep = 1.9;
            double zoom = 1;
            for (double iStep = 300; iStep >= 2; iStep -= (double)zoomStep)
            {
                zoom -= 0.01321;

                double x = (double)((bitmap.Width / 2)) - (iStep*zoom);
                double y = (double)((bitmap.Height)) - (iStep * zoom);
                double width = (iStep * zoom) * 4;
                double height = (iStep * zoom) * 2;

                penIndex++;
                if(penIndex>=pens.Count)
                {
                    penIndex = 0;
                }
                var pen = pens[penIndex];

                if(width<0)
                {
                    x += 20;
                }

                graphics.FillEllipse(pen, (float)x, (float)y, (float)width, (float)height);
            }

            // left circle a.k.a. Overflow's circle
            penIndex = 0;
            zoomStep = 1.9f;
            zoom = 1;
            for (double iStep = 300; iStep >= 2; iStep -= (double)zoomStep)
            {
                zoom -= 0.01321f;

                double x = (double)((bitmap.Width / 2)) - (iStep * zoom);
                double y = (double)((bitmap.Height)) - (iStep * zoom);
                double width = (iStep * zoom) * 4;
                double height = (iStep * zoom) * 2;

                penIndex++;
                if (penIndex >= pens.Count)
                {
                    penIndex = 0;
                }
                var pen = pens[penIndex];

                if (width < 0)
                {
                    height *= 1.0;
                    width *= 1.0;

                    x -= 90.0;
                    y -= 90.0;

                    graphics.FillEllipse(pen, (float)x, (float)y, (float)width, (float)height);
                }
            }

            graphics.Dispose();

            for(float iY=0; iY<bitmap.Height;)
            {
                int y = (int)iY;

                int offset = y & 1;

                for (int x = 0; x < bitmap.Width; x+=2)
                {
                    //bitmap.SetPixel(x + offset, y, Color.Black);
                }

                iY += 1;
                iY *= 1.32f;
            }

            var bmpFilename = outFolder + "_Circle.png";
            bitmap.Save(bmpFilename, ImageFormat.Png);
            bitmap = new Bitmap(System.Drawing.Image.FromFile(bmpFilename));

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 1,
                R1 = 48,
                R4 = 21,
                R9 = 0,
                Palette = palette
            }
            );

            int logoByteOffsetInPage3 = 0;
            for (int i = 0; i < cpcImage.R12R13.Length; i += 2)
            {
                if(i==(cpcImage.R12R13.Length-logoHeight*2))
                {
                    byte r12 = (byte)(cpcImage.R12R13[i] & 3);
                    byte r13 = cpcImage.R12R13[i+1];
                    logoByteOffsetInPage3 = (int)(((r12<<8) + r13)*2);
                }

                r12r13.Add(cpcImage.R12R13[i]);
                r12r13.Add(cpcImage.R12R13[i + 1]);
            }

            for (int iPage = 0; iPage < cpcImage.Pages.Length; iPage++)
            {
                var page = cpcImage.Pages[iPage];

                foreach (var VLC in cpcImage.Pages[iPage].VLCs)
                {
                    var filename = outFolder + "_Circle.page" + iPage + ".vlc" + VLC.Index;

                    var cpcBytes = VLC.CPCBytes;

                    if((iPage==0)&&(VLC.Index==0))
                    {
                        for(int i=0; i<96; i++)
                        {
                            cpcBytes[i] = 0xFF;
                        }
                    }

                    // inject Logo in VLC Page 3
                    if (iPage == 3)
                    {
                        foreach (var b in logoRawBytes)
                        {
                            cpcBytes[logoByteOffsetInPage3] = b;
                            logoByteOffsetInPage3++;
                        }
                    }

                    File.WriteAllBytes(filename, cpcBytes);

                    var bbPackedBytes = BBPacker.Pack(ref cpcBytes);
                    File.WriteAllBytes(filename + ".bbPacked", bbPackedBytes);

                    var packer = new Packer();
                    var packedBytes = packer.Convert(ref cpcBytes);
                    File.WriteAllBytes(filename + ".packed", packedBytes);
                }
            }
        }

        private void CreateRasterAnim(string imageFilename, int blackColorToReplace, string outFilename)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var raster = new List<byte>();

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);
                var gateArrayColor = palette.ToGateArrayByteArray()[0];
                raster.Add(gateArrayColor);
            }

            var blackColor = Pixel.CPCGateArrayPalette[0];
            var newBlackColor = Pixel.CPCGateArrayPalette[blackColorToReplace];

            var frame = new List<byte>();

            var delta = 5;
            var p0Index = delta*0;
            var p1Index = delta*1;
            var p2Index = delta*2;

            for (int i = 0; i < BarSize*3; i++)
            {
                byte p0 = Pixel.CPCGateArrayPalette[0];
                byte p1 = Pixel.CPCGateArrayPalette[0];
                byte p2 = Pixel.CPCGateArrayPalette[0];
                    
                {
                    if (p0Index < raster.Count)
                    {
                        p0 = raster[p0Index];
                    }
                    if (p1Index < raster.Count)
                    {
                        p1 = raster[p1Index];
                    }
                    if (p2Index < raster.Count)
                    {
                        p2 = raster[p2Index];
                    }

                    p0Index++;
                    p1Index++;
                    p2Index++;
                }

                if(p0==blackColor)
                {
                    p0 = newBlackColor;
                }
                if (p1 == blackColor)
                {
                    p1 = newBlackColor;
                }
                if (p2 == blackColor)
                {
                    p2 = newBlackColor;
                }

                frame.Add(p0);
                frame.Add(p1);
                frame.Add(p2);
            }

            File.WriteAllBytes(outFilename, frame.ToArray());
        }
        
        private void CreateR12R13_BC00BD00_SinTable(string outFolder, ref List<byte> r12r13, int logoHeight, byte[] logoRawBytes)
        {
            var byteArray = r12r13.ToArray();

            var z80 = "";
            z80 += "\n";
            z80 += ";; Must be compiled at &BC00\n";
            z80 += "Circle_R12_Table:\n";

            for(int i=0; i<byteArray.Length; i+=2)
            {
                if(i==(byteArray.Length-logoHeight*2))
                {
                    z80 += "Logo_R12_Table:\n";
                }

                var r12 = byteArray[i];
                z80 += "\tdb &" + r12.ToString("X2") + "\n";
            }

            z80 += "\n";
            z80 += ";; Must be compiled at &BD00\n";
            z80 += "\tALIGN 256\n";
            z80 += "Circle_R13_Table:\n";
            for (int i = 1; i < byteArray.Length; i += 2)
            {
                if ((i-1) == (byteArray.Length - logoHeight * 2))
                {
                    z80 += "Logo_R13_Table:\n";
                }

                var r13 = byteArray[i];
                z80 += "\tdb &" + r13.ToString("X2") + "\n";
            }

            File.WriteAllText(outFolder + "_R12R13_BC00_BD00.z80", z80);
        }
    }
}
