﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _11_DistoGrid
    {
        const int R1 = 64; // due to offset
        const int EdgeWidthInPixels = 14;//14;//8;
        const int TileWidthInPixels = 8 * 4;
        const int NbPixelsMove = 16 * 4; // 16=64-48  4=4 lines in block

        public _11_DistoGrid()
        {
            try
            {
                var r12r13 = new List<byte>();
                CreateGridLines(Program.ToZ80Folder("11_DistoGrid/"), ref r12r13);

                CreateGridAnim(Program.ToZ80Folder("11_DistoGrid/_DistoAnim.bin"));

                CreatePalettePtr(   Program.ToZ80Folder("11_DistoGrid/_DistoGrid_0To7_PalettePtr_High"),
                                    Program.ToZ80Folder("11_DistoGrid/_DistoGrid_0To7_PalettePtr_Low"));
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        private Bitmap CreateTexturePattern(ref List<Color> colors, bool is8_15Range)
        {
            var bitmap = new Bitmap(4, 4);

            var pattern = new byte[]
            {
                0,2,0,5,
                1,0,1,0,
                0,4,0,3,
                1,0,1,0
            };

            int indexShift = 0;
            if (is8_15Range)
            {
                indexShift += 8;
            }

            for (int y = 0; y < 4; y++)
            {
                for (int x = 0; x < 4; x++)
                {
                    var colorIndex = pattern[(y * 4) + x];
                    bitmap.SetPixel(x, y, colors[indexShift + colorIndex]);
                }
            }

            return bitmap;
        }

        private Framework.Image CreateCPCBitmap(ref List<Color> colors, bool is8_15Range, string baseName)
        {
            var texturePattern = CreateTexturePattern(ref colors, is8_15Range);
            var textureBrush = new TextureBrush(texturePattern);

            var baseBitmap = new Bitmap(R1 * 4, 4 * 4); // 1 CRTC block is 4 pixels-wide (mode 0)    4*4=pattern has height of 4, and we have 4 horizontal shifts (mode 0)
            var reorderedBitmap = new Bitmap(R1 * 4, 4 * 4); // 1 CRTC block is 4 pixels-wide (mode 0)    4*4=pattern has height of 4, and we have 4 horizontal shifts (mode 0)

            var baseGraphics = Graphics.FromImage(baseBitmap);
            baseGraphics.FillRectangle(new SolidBrush(colors[0]), new Rectangle(0, 0, baseBitmap.Width, baseBitmap.Height));
            for (int iShift = 0; iShift < 4; iShift++)
            {
                int iX = 0;
                for (int i = 0; i < 80; i++)
                {
                    baseGraphics.FillRectangle(textureBrush, new Rectangle(iX - iShift, iShift * 4, TileWidthInPixels - EdgeWidthInPixels, 4));

                    iX += TileWidthInPixels;
                }
            }
            baseGraphics.Dispose();

            int dstY = 0;
            for (int y1 = 0; y1 < 4; y1++)
            {
                for (int y2 = 0; y2 < 4; y2++)
                {
                    int srcY = (y2 * 4) + y1;

                    for (int x = 0; x < baseBitmap.Width; x++)
                    {
                        var pixel = baseBitmap.GetPixel(x, srcY);
                        reorderedBitmap.SetPixel(x, dstY, pixel);
                    }

                    dstY++;
                }
            }

            var bitmap = reorderedBitmap;
            bitmap.Save(baseName + ".png", ImageFormat.Png);

            var palette = new Palette();
            foreach (var c in colors)
            {
                palette.AddColor(c);
            }

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 0,
                R1 = R1,
                R4 = 21,
                R9 = 0,
                Palette = palette
            }
            );

            for (int iPage = 0; iPage < cpcImage.Pages.Length; iPage++)
            {
                var page = cpcImage.Pages[iPage];

                foreach (var VLC in cpcImage.Pages[iPage].VLCs)
                {
                    File.WriteAllBytes(baseName + ".bin", new List<byte>(VLC.CPCBytes).Take(0x0800).ToArray());
                }
            }

            return cpcImage;
        }

        public void CreateGridLines(string outFolder, ref List<byte> r12r13)
        {
            var colors = new List<Color>();
            for (int i = 0; i < 16; i++)
            {
                colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[i] | 0xFF000000)));
            }

            var firstCPCImage = CreateCPCBitmap(ref colors, false, outFolder + "_DistoGrid_0to7");
            CreateCPCBitmap(ref colors, true, outFolder + "_DistoGrid_8to15"); ;

            for (int i = 0; i < firstCPCImage.R12R13.Length; i += 2)
            {
                r12r13.Add(firstCPCImage.R12R13[i]);
                r12r13.Add(firstCPCImage.R12R13[i + 1]);
            }

            for (int iRange = 0; iRange < 2; iRange++)
            {
                for (int iLineInBlock = 0; iLineInBlock < 4; iLineInBlock++)
                {
                    var r12_256 = new List<byte>();
                    var r13_256 = new List<byte>();
                    byte iXOffset = 0;

                    int iBlock = 0;
                    // const int NbPixelsMove = 16 * 4; // 16=64-48  4=4 lines in block
                    for (int i = 0; i < NbPixelsMove; i++)
                    {
                        var r12r13Offset = (iLineInBlock * 4) + iBlock;
                        var r12 = r12r13[r12r13Offset * 2];
                        var r13 = r12r13[r12r13Offset * 2 + 1];

                        r13 += iXOffset;

                        if (iRange == 0)
                        {
                            r12_256.Add((byte)(0x20 + r12)); // &8000
                        }
                        else
                        {
                            r12_256.Add((byte)(0x30 + r12)); // &C000
                        }
                        r13_256.Add(r13);

                        iBlock++;
                        if (iBlock == 4)
                        {
                            iBlock = 0;
                            iXOffset++;
                        }
                    }

                    r12_256.Reverse();
                    r13_256.Reverse();

                    for (int i = 0; i < 3; i++) // 16=64-48  4=4 lines
                    {
                        for (int j = 0; j < NbPixelsMove; j++)
                        {
                            r12_256.Add(r12_256[j]);
                            r13_256.Add(r13_256[j]);
                        }
                    }

                    if (iRange == 0)
                    {
                        File.WriteAllBytes(outFolder + "_DistoGrid_0To7_Line" + iLineInBlock + ".r12", r12_256.ToArray());
                        File.WriteAllBytes(outFolder + "_DistoGrid_0To7_Line" + iLineInBlock + ".r13", r13_256.ToArray());
                    }
                    else
                    {
                        File.WriteAllBytes(outFolder + "_DistoGrid_8To15_Line" + iLineInBlock + ".r12", r12_256.ToArray());
                        File.WriteAllBytes(outFolder + "_DistoGrid_8To15_Line" + iLineInBlock + ".r13", r13_256.ToArray());
                    }
                }
            }
        }

        public byte[] DoubleToCpcBytes(double v)
        {
            var bytes = new byte[2];

            var intPartString = v.ToString().Replace(",", ".").Split('.')[0];
            var intPart = double.Parse(intPartString);
            var floatPart = v - intPart;
            var floatPart255 = floatPart * 255.0f;

            if(v>=0)
            {
                bytes[0] = (byte)(intPart);
                bytes[1] = (byte)(floatPart255);
            }
            else
            {
                //bytes[0] = (byte)(255-intPart);
                bytes[0] = (byte)(255);

                if (((byte)floatPart255)==0)
                {
                    bytes[1] = (byte)(255);
                }
                else
                {
                    bytes[1] = (byte)((byte)floatPart255);
                }
            }
            
            return bytes;
        }

        public void CreateGridAnim(string animName)
        {
            //int screenHeight = 268;
            int NbFrames = 100;

            double maxX = 26;// 31;
            double centerX = 64 / 2;

            //maxX /= 2;
            //centerX /= 2;

            var allBytes = new List<byte>();

            for (int iFrame = 0; iFrame < NbFrames; iFrame++)
            {
                double progress = (((double)iFrame) / ((double)NbFrames));
                double progressDeg = 360 * progress;
                double progressRad = (Math.PI * progressDeg) / 180; //Radians

                var sin0 = -Math.Sin(progressRad);
                var sin1 = Math.Cos(progressRad + 123);

                var xFrameSin = Math.Cos(progressRad) * 16;

                var yList = new List<double>();
                int y = 0;
                for (; y < 268 + 100; y++)
                {
                    var s0 = Math.Cos(sin0) * (maxX / 2);
                    var s1 = Math.Sin(sin1) * (maxX / 2);
                    var value = centerX + s0 + s1;

                    //value += progress * 64;
                    value += xFrameSin;

                    sin0 += 0.019;
                    sin1 += 0.011;

                    yList.Add(value);
                }

                var frameBytes = new List<byte>();
                y = 0;
                int lastY = 0;
                byte[] lastDeltaCpcBytes = null;
                bool firstYinFrame = true;
                int foundLen = 0;
                for (; y < 268; y += foundLen)
                {
                    lastY = y;
                    var xPosTop = yList[y];

                    // find next where distance is acceptable
                    var x0 = xPosTop;
                    var y0 = (double)y;

                    int y1Iter = y + 1;
                    var y1 = (double)y1Iter;
                    var x1 = yList[y1Iter];
                    //var found = false;
                    /*while(!found)
                    {
                        var scaledX0 = x0 * 5;
                        var scaledX1 = x1 * 2;

                        var distance = Math.Sqrt((scaledX0 - scaledX1) * (scaledX0 - scaledX1) + (y0 - y1) * (y0 - y1));
                        if(distance < 30)
                        {
                            y1Iter++;
                            y1 = (double)y1Iter;
                            x1 = yList[y1Iter];
                        }
                        else
                        {
                            found = true;
                            foundLen = y1Iter - y;
                        }
                    }*/
                    //foundLen = 40;
                    
                    var s0 = Math.Cos(sin0) * 5;
                    var s1 = Math.Sin(sin1) * 5;
                    var value = 16 + s0 + s1;
                    sin0 += 0.22;
                    sin1 += 0.24;

                    foundLen = (int)value;
                    if(foundLen<6)
                    {
                        foundLen = 6;
                    }
                    if (foundLen >23)
                    {
                        foundLen = 23;
                    }

                    var xPosBottom = yList[y + foundLen];

                    if (firstYinFrame)
                    {
                        var xPosCpcBytes = DoubleToCpcBytes(xPosTop);
                        frameBytes.Add(xPosCpcBytes[0]); // H xpos int part
                        frameBytes.Add(xPosCpcBytes[1]); // L xpos float part

                        firstYinFrame = false;
                    }

                    var delta = (xPosBottom - xPosTop) / ((double)foundLen);
                    var deltaCpcBytes = DoubleToCpcBytes(delta);
                    lastDeltaCpcBytes = deltaCpcBytes;

                    if((delta<-1)||(delta>=1))
                    {
                        // delta not supported
                        System.Diagnostics.Debugger.Break();
                    }

                    // possible opti:
                    // use bit 7 of len to set toLeft or toRight
                    // then remove deltaCpcBytes[0]

                    frameBytes.Add((byte)(foundLen));
                    frameBytes.Add(deltaCpcBytes[0]); // delta int
                    frameBytes.Add(deltaCpcBytes[1]); // delta float part
                }

                if (lastY != 268)
                {
                    frameBytes.RemoveAt(frameBytes.Count - 1); // len
                    frameBytes.RemoveAt(frameBytes.Count - 1); // delta int
                    frameBytes.RemoveAt(frameBytes.Count - 1); // delta float part

                    frameBytes.Add((byte)(268 - lastY));
                    frameBytes.Add(lastDeltaCpcBytes[0]); // delta int
                    frameBytes.Add(lastDeltaCpcBytes[1]); // delta float part
                }

                frameBytes.Add((byte)0); // write len=0 -> end of frame

                // if (iFrame==0) // TEMP - 1st frame only for now
               // if (iFrame > 8) // TEMP - 1st frame only for now
                {
                    allBytes.AddRange(frameBytes);
                }
            }

            File.WriteAllBytes(animName, allBytes.ToArray());
        }
        
        private void CreatePalettePtr(string highFilename, string lowFilename)
        {
            for (int iPalette = 0; iPalette < 4; iPalette++)
            {
                var z80_High = "";
                var z80_Low = "";

                int nbLines = 0;

                // const int NbPixelsMove = 16 * 4; // 16=64-48  4=4 lines in block
                var z80Loop_High = ";\n";
                var z80Loop_Low = ";\n";
                for (int i = 0; i < NbPixelsMove; i++)
                {
                    for (int j = 0; j < 4; j++)
                    {
                        z80Loop_High += "\tLET ptr = DistoGrid_Palettes_" + iPalette + " + " + i * 5 + "\n";
                        z80Loop_Low += "\tLET ptr = DistoGrid_Palettes_" + iPalette + " + " + i * 5 + "\n";

                        z80Loop_High += "\tdb ptr/256\n\n";
                        z80Loop_Low += "\tdb ptr AND 255\n\n";

                        nbLines++;
                    }
                }

                z80_High = z80Loop_High;
                z80_Low = z80Loop_Low;
                /*for (int i = 0; i < 4; i++)
                {
                    z80_High += z80Loop_High;
                    z80_Low += z80Loop_Low;
                }*/

                //nbLines *= 4;

                if (nbLines != 256)
                {
                    System.Diagnostics.Debugger.Break();
                }

                File.WriteAllText(highFilename + "_" + iPalette + ".z80", z80_High);
                File.WriteAllText(lowFilename + "_" + iPalette + ".z80", z80_Low);
            }
        }
    }
}
