﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _15_Condense
    {
        private const int ImagePage0_TakeBytes = 0x980;

        public _15_Condense()
        {
            try
            {
                WriteConfig("15_Condense/_CondenseSince1994.config.z80");
                CreateFades(Program.ToZ80Folder("15_Condense/_CondenseSince1994FadePalettes.bin"));
                WriteImage("15_Condense/CondenseSince1994.png", "15_Condense/_CondenseSince1994.png");
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        private void WriteConfig(string dstFilename)
        {
            var z80 = ";\n";
            z80 += "\tCondenseSince1994_TakeBytes\tequ &" + ImagePage0_TakeBytes.ToString("X4") + "\n";
            File.WriteAllText(Program.ToZ80Folder(dstFilename), z80);
        }

        public void CreateFades(string outFilename)
        {
            var fadeFilenames = new List<string>();
            /*fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeIn0.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeIn1.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeIn2.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeIn3.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeIn4.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeIn5.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/CondenseSince1994.png"));*/
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut1.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut2.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut3.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut4.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut5.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut6.png"));
            fadeFilenames.Add(Program.ToDataFolder("15_Condense/FadeOut7.png"));

            var fadePalettes = new List<byte>();

            foreach (var fadeFilename in fadeFilenames)
            {
                var allBytes = CreateFadeMode0Raw(
                                    Program.ToDataFolder("15_Condense/CondenseSince1994.png"),
                                    fadeFilename);

                fadePalettes.AddRange(allBytes);
            }

            File.WriteAllBytes(outFilename, fadePalettes.ToArray());
        }

        public byte[] CreateFadeMode0Raw(
            string baseBmpFilename,
            string fadeBmpFilename,
            int skipPensInPalette = 0)
        {
            var baseBitmap = new Bitmap(baseBmpFilename);
            var fadeBitmap = new Bitmap(fadeBmpFilename);

            var basePalette = new Palette(skipPensInPalette);
            var fadePalette = new Palette(skipPensInPalette);
            var outPalette = new Palette(skipPensInPalette);

            var cpcRaw = new byte[((baseBitmap.Width + 1) / 2) * baseBitmap.Height];

            var basePaletteIndices = new List<byte>();
            var fadePaletteIndices = new List<byte>();

            for (int y = 0; y < baseBitmap.Height; y++)
            {
                for (int x = 0; x < baseBitmap.Width; x += 2)
                {
                    Color baseP0 = baseBitmap.GetPixel(x, y);
                    Color baseP1 = baseP0;
                    if ((x + 1) < baseBitmap.Width)
                    {
                        baseP1 = baseBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = basePalette.AddColor(baseP0);
                    var paletteEntry1 = basePalette.AddColor(baseP1);
                    basePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    basePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x += 2)
                {
                    Color fadeP0 = fadeBitmap.GetPixel(x, y);
                    Color fadeP1 = fadeP0;
                    if ((x + 1) < fadeBitmap.Width)
                    {
                        fadeP1 = fadeBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = fadePalette.AddColor(fadeP0);
                    var paletteEntry1 = fadePalette.AddColor(fadeP1);
                    fadePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    fadePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x++)
                {
                    int offset = (y * fadeBitmap.Width) + x;

                    int baseIndex = basePaletteIndices[offset];
                    int fadeIndex = fadePaletteIndices[offset];

                    var basePaletteEntry = basePalette.Pens[baseIndex];
                    var fadePaletteEntry = fadePalette.Pens[fadeIndex];

                    basePaletteEntry.Color = fadePaletteEntry.Color;
                    basePaletteEntry.FirmwareColor = fadePaletteEntry.FirmwareColor;
                    basePaletteEntry.GateArrayColor = fadePaletteEntry.GateArrayColor;
                }
            }

            return basePalette.ToGateArrayByteArray();
        }

        private void WriteImage(string srcFilename, string dstFilename)
        {
            var bitmap = new Bitmap(Program.ToDataFolder(srcFilename));
            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 0,
                R1 = 54, // 54, not 48, to get linear data
                R4 = 38,
                R9 = 7, // for static pages
                StartVLC = 0,
                ApplyOffsetAtNextPage = 4
            });

            for (int iPage = 0; iPage < image.Pages.Length; iPage++)
            {
                var bytes = new List<byte>();
                foreach (var VLC in image.Pages[iPage].VLCs)
                {
                    // clear out invisible right border with pen having COLOR0
                    byte penColor0 = 0;
                    foreach (var p in image.Palette.Pens)
                    {
                        if (p.FirmwareColor == 0)
                        {
                            penColor0 = p.IndexInPalette;
                            break;
                        }
                    }
                    for (int i = 0; i < 0x800; i += 54 * 2)
                    {
                        for (int j = iPage * 4 + i + 48 * 2; j < iPage * 4 + i + 54 * 2; j++)
                        {
                            var pixelToWrite = Pixel.ToMode0Pixel(penColor0, penColor0);
                            if (j < 0x800)
                            {
                                VLC.CPCBytes[j] = pixelToWrite;
                            }
                        }
                    }

                    File.WriteAllBytes(Program.ToZ80Folder(dstFilename + ".page" + iPage + ".vlc" + VLC.Index), VLC.CPCBytes);
                    bytes.AddRange(VLC.CPCBytes);
                }
                var raw = bytes.ToArray();
                var filename = Program.ToZ80Folder(dstFilename + ".page" + iPage);
                File.WriteAllBytes(filename, raw);

                var BBPackedBytes = BBPacker.Pack(ref raw);
                File.WriteAllBytes(filename + ".bbPacked", BBPackedBytes);

                if(iPage==0)
                {
                    var BBPackedBytes_Part0 = BBPackedBytes.Take(ImagePage0_TakeBytes);
                    var BBPackedBytes_Part1 = BBPackedBytes.Skip(ImagePage0_TakeBytes);
                    
                    File.WriteAllBytes(filename + ".bbPacked.0", BBPackedBytes_Part0.ToArray());
                    File.WriteAllBytes(filename + ".bbPacked.1", BBPackedBytes_Part1.ToArray());
                }

                var packer = new Packer();
                var packedBytes = packer.Convert(ref raw);
                File.WriteAllBytes(filename + ".packed", packedBytes);
            }
            File.WriteAllText(
                Program.ToZ80Folder(dstFilename + ".pal.z80"),
                image.Palette.ToZ80TableString(Program.FilenameToKeyWordName(srcFilename))
            );
        }
    }
}
