﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public class _17_ThirdPartEndScreen
    {
        public _17_ThirdPartEndScreen()
        {
            try
            {
                CreateFades(Program.ToZ80Folder("17_ThirdPartEndScreen/_EndScreenFadeOutPalettes.bin"));

                var linePtr = CalcYLinePtr(
                    Program.ToZ80Folder("17_ThirdPartEndScreen/_ScreenPtr.z80"));

                var endScreenFilename = "EndScreen";
                //var endScreenFilename = "NoRecess_EvilFire";

                //var globalPalette = WriteImage("17_ThirdPartEndScreen/NoRecess_Evil.png", "17_ThirdPartEndScreen/Reference/_NoRecess_Evil.png");
                var globalPalette = WriteImage("17_ThirdPartEndScreen/"+ endScreenFilename+".png", "17_ThirdPartEndScreen/Reference/_NoRecess_Evil.png");

                WriteSplitImage(
                    linePtr,
                    Program.ToDataFolder("17_ThirdPartEndScreen/Cloud.png"),
                    globalPalette,
                    "17_ThirdPartEndScreen/"+ endScreenFilename+".png",
                    Program.ToZ80Folder("17_ThirdPartEndScreen/Slices/"));
                
                File.WriteAllText(
                    Program.ToZ80Folder("17_ThirdPartEndScreen/_GlobalPalette.pal.z80"),
                    globalPalette.ToZ80TableString(Program.FilenameToKeyWordName("ThirdPart_EndScreen"))
                );
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public void CreateFades(string outFilename)
        {
            var fadeFilenames = new List<string>();
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut0.png"));
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut1.png"));
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut2.png"));
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut3.png"));
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut4.png"));
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut5.png"));
            fadeFilenames.Add(Program.ToDataFolder("17_ThirdPartEndScreen/FadeOut6.png"));

            var fadePalettes = new List<byte>();

            foreach (var fadeFilename in fadeFilenames)
            {
                var allBytes = CreateFadeMode0Raw(
                                    Program.ToDataFolder("17_ThirdPartEndScreen/EndScreen.png"),
                                    fadeFilename);

                fadePalettes.AddRange(allBytes);
            }

            File.WriteAllBytes(outFilename, fadePalettes.ToArray());
        }

        public byte[] CreateFadeMode0Raw(
            string baseBmpFilename,
            string fadeBmpFilename,
            int skipPensInPalette = 0)
        {
            var baseBitmap = new Bitmap(baseBmpFilename);
            var fadeBitmap = new Bitmap(fadeBmpFilename);

            var basePalette = new Palette(skipPensInPalette);
            var fadePalette = new Palette(skipPensInPalette);
            var outPalette = new Palette(skipPensInPalette);

            var cpcRaw = new byte[((baseBitmap.Width + 1) / 2) * baseBitmap.Height];

            var basePaletteIndices = new List<byte>();
            var fadePaletteIndices = new List<byte>();

            for (int y = 0; y < baseBitmap.Height; y++)
            {
                for (int x = 0; x < baseBitmap.Width; x += 2)
                {
                    Color baseP0 = baseBitmap.GetPixel(x, y);
                    Color baseP1 = baseP0;
                    if ((x + 1) < baseBitmap.Width)
                    {
                        baseP1 = baseBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = basePalette.AddColor(baseP0);
                    var paletteEntry1 = basePalette.AddColor(baseP1);
                    basePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    basePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x += 2)
                {
                    Color fadeP0 = fadeBitmap.GetPixel(x, y);
                    Color fadeP1 = fadeP0;
                    if ((x + 1) < fadeBitmap.Width)
                    {
                        fadeP1 = fadeBitmap.GetPixel(x + 1, y);
                    }

                    var paletteEntry0 = fadePalette.AddColor(fadeP0);
                    var paletteEntry1 = fadePalette.AddColor(fadeP1);
                    fadePaletteIndices.Add(paletteEntry0.IndexInPalette);
                    fadePaletteIndices.Add(paletteEntry1.IndexInPalette);
                }
            }

            for (int y = 0; y < fadeBitmap.Height; y++)
            {
                for (int x = 0; x < fadeBitmap.Width; x++)
                {
                    int offset = (y * fadeBitmap.Width) + x;

                    int baseIndex = basePaletteIndices[offset];
                    int fadeIndex = fadePaletteIndices[offset];

                    var basePaletteEntry = basePalette.Pens[baseIndex];
                    var fadePaletteEntry = fadePalette.Pens[fadeIndex];

                    basePaletteEntry.Color = fadePaletteEntry.Color;
                    basePaletteEntry.FirmwareColor = fadePaletteEntry.FirmwareColor;
                    basePaletteEntry.GateArrayColor = fadePaletteEntry.GateArrayColor;
                }
            }

            return basePalette.ToGateArrayByteArray();
        }

        private List<ushort> CalcYLinePtr(string z80Filename)
        {
            /* 
	        ;;R9=7 R1=54
	        ;;page2:
	        ;;start at &8000
	        ;;19 carac (de 0 a 18)  cf 18*108
	        ;;last VCC= &B800 (normal)
	        ;;page3:
	        ;;start at &C000+4
	        ;;15 carac (de 0 a 14)  cf 14*108 pour dernier VCC affiche
	        ;;last VCC= &E000 est le dernier VLC affiche (s'arrete a VLC=5 au lieu de 7)
            */

            var lines = new List<ushort>();

            var z80 = ";\n";
            var bin = new List<byte>();

            // page 2
            ushort ptr = (ushort) 0x8000;
            for(int iVCC=0; iVCC<19; iVCC++)
            {
                for(int iVLC=0; iVLC<8; iVLC++)
                {
                    lines.Add(ptr);
                    z80 += "\tdw &" + ptr.ToString("X4") + "\n";
                    bin.Add((byte)(ptr & 255));
                    bin.Add((byte)(ptr>>8));
                    ptr = (ushort)(ptr + 0x800);
                }
                ptr = (ushort)(ptr - (0x800 * 8));
                ptr = (ushort)(ptr + 54 + 54);
            }
            // page 3
            ptr = (ushort)0xC000+4;
            int maxVCC = 19;// 15;  calc lines even in not visible part
            for (int iVCC = 0; iVCC < maxVCC; iVCC++)
            {
                for (int iVLC = 0; iVLC < 8; iVLC++)
                {
                    /*if((iVCC==14)&&(iVLC>=4))
                    {
                    }
                    else*/
                    {
                        lines.Add(ptr);
                        z80 += "\tdw &" + ptr.ToString("X4") + "\n";
                        bin.Add((byte)(ptr & 255));
                        bin.Add((byte)(ptr >> 8));
                    }
                    ptr = (ushort)(ptr + 0x800);
                }
                ptr = (ushort)(ptr - (0x800 * 8));
                ptr = (ushort)(ptr + 54+54);
            }

            File.WriteAllText(z80Filename, z80);
            File.WriteAllBytes(z80Filename+".bin", bin.ToArray());

            var raw = bin.ToArray();
            var BBPackedBytes = BBPacker.Pack(ref raw);
            File.WriteAllBytes(z80Filename + ".bin.bbPacked", BBPackedBytes);

            return lines;
        }

        private Palette WriteImage(string srcFilename, string dstFilename)
        {
            var palette = new Palette();
            palette.AddColor(Color.Black);
            var bitmap = new Bitmap(Program.ToDataFolder(srcFilename));
            var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 0,
                R1 = 54, // 54, not 48, to get linear data
                R4 = 38,
                R9 = 7, // for static pages
                StartVLC = 0,
                ApplyOffsetAtNextPage = 4,
                Palette = palette
            });

            int cumulatedBBPackedSize = 0;
            int cumulatedPackedSize = 0;

            for (int iPage = 0; iPage < image.Pages.Length; iPage++)
            {
                var bytes = new List<byte>();
                foreach (var VLC in image.Pages[iPage].VLCs)
                {
                    // clear out invisible right border with pen having COLOR0
                    byte penColor0 = 0;
                    foreach (var p in image.Palette.Pens)
                    {
                        if (p.FirmwareColor == 0)
                        {
                            penColor0 = p.IndexInPalette;
                            break;
                        }
                    }
                    for (int i = 0; i < 0x800; i += 54 * 2)
                    {
                        for (int j = iPage * 4 + i + 48 * 2; j < iPage * 4 + i + 54 * 2; j++)
                        {
                            var pixelToWrite = Pixel.ToMode0Pixel(penColor0, penColor0);
                            if (j < 0x800)
                            {
                                VLC.CPCBytes[j] = pixelToWrite;
                            }
                        }
                    }

                    File.WriteAllBytes(Program.ToZ80Folder(dstFilename + ".page" + iPage + ".vlc" + VLC.Index), VLC.CPCBytes);
                    bytes.AddRange(VLC.CPCBytes);
                }
                var raw = bytes.ToArray();
                var filename = Program.ToZ80Folder(dstFilename + ".page" + iPage);
                File.WriteAllBytes(filename, raw);

                var BBPackedBytes = BBPacker.Pack(ref raw);
                File.WriteAllBytes(filename + ".bbPacked", BBPackedBytes);
                cumulatedBBPackedSize += BBPackedBytes.Length;

                var packer = new Packer();
                var packedBytes = packer.Convert(ref raw);
                File.WriteAllBytes(filename + ".packed", packedBytes);
                cumulatedPackedSize += packedBytes.Length;
            }
            
            Console.WriteLine("Complete bitmap - BBPack Cumulated Size="+ cumulatedBBPackedSize + " bytes");
            Console.WriteLine("Complete bitmap - Pack Cumulated Size=" + cumulatedPackedSize + " bytes");

            return image.Palette;
        }

        private void WriteSplitImage(List<ushort> linesPtr, 
            string cloudFilename,
            Palette globalPalette, 
            string fullscreenImage, 
            string outFolder)
        {
            var globalBitmap = new Bitmap(Program.ToDataFolder(fullscreenImage));

            int cumulatedBBPackedSize = 0;
            int cumulatedPackedSize = 0;

            const int boxWidth = 12;// must be multiple of 4 (because of R1)
            const int boxHeight = 18; // 268

            int gridWidth = 0;
            int gridHeight = 0;

            var binFile = new List<byte[]>();
            var allPackedTiles = new List<byte>();
            int binFileCumulatedSize = 0;
            int tileOffset = 0;
            int iTile = 0;
            var z80TileConfig = ";\n";
            var gridDwLabels = new List<string>();
            var gridDbBanks = new List<string>();

            for (int y = 0; y < globalBitmap.Height; y += boxHeight)
            {
                for (int x = 0; x < globalBitmap.Width; x += boxWidth)
                {
                    int iBoxX = x / boxWidth;
                    int iBoxY = y / boxHeight;

                    if(gridWidth<iBoxX)
                    {
                        gridWidth = iBoxX;
                    }
                    if (gridHeight < iBoxY)
                    {
                        gridHeight = iBoxY;
                    }

                    var sliceBitmap = new Bitmap(boxWidth, boxHeight);
                    for (int iX = 0; iX < boxWidth; iX++)
                    {
                        for (int iY = 0; iY < boxHeight; iY++)
                        {
                            var pixel = Color.Black;// globalBitmap.GetPixel(0,0);
                            if (iY + y < globalBitmap.Height)
                            {
                                pixel = globalBitmap.GetPixel(iX + x, iY + y);
                            }
                            sliceBitmap.SetPixel(iX, iY, pixel);
                        }
                    }

                    var bmpFilename = outFolder + "_Slice_Y" + iBoxY + "_X" + iBoxX + ".png";
                    sliceBitmap.Save(bmpFilename, ImageFormat.Png);

                    var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                    {
                        Bitmap = sliceBitmap,
                        Mode = 0,
                        R1 = boxWidth / 4, // 54, not 48, to get linear data
                        R4 = sliceBitmap.Height / 8,
                        R9 = 7, // for static pages
                        StartVLC = 0,
                        Palette = globalPalette
                    });

                    var rawBuffer = image.CPCRawBuffer;

                    var binFilename = bmpFilename + ".bin";

                    File.WriteAllBytes(binFilename, rawBuffer);

                    var BBPackedBytes = BBPacker.Pack(ref rawBuffer);
                    File.WriteAllBytes(binFilename + ".bbPacked", BBPackedBytes);
                    cumulatedBBPackedSize += BBPackedBytes.Length;

                    binFile.Add(BBPackedBytes);
                    binFileCumulatedSize += BBPackedBytes.Length;
                    if ((binFileCumulatedSize>12500)&&(tileOffset==0))
                    {
                        allPackedTiles = new List<byte>();
                        foreach(var binArray in binFile)
                        {
                            allPackedTiles.AddRange(binArray);
                            // -&4000 because EndPackedTiles1==&C6 offset=-&4000
                            z80TileConfig += "END_TILE_" + iTile + " equ EndPackedTiles0 -&4000 + &" + tileOffset.ToString("X4") + "\n";
                            z80TileConfig += "END_TILE_" + iTile + "_BANK equ &C6\n";
                            z80TileConfig += "\n";
                            gridDwLabels.Add("END_TILE_" + iTile);
                            gridDbBanks.Add("END_TILE_" + iTile+"_BANK");
                            iTile++;
                            tileOffset += binArray.Length;

                        }
                        File.WriteAllBytes(outFolder + "../_EndScreen_PackedTiles_0.bin", allPackedTiles.ToArray());
                        
                        binFile = new List<byte[]>();
                        binFileCumulatedSize = 0;
                    }

                    var packer = new Packer();
                    var packedBytes = packer.Convert(ref rawBuffer);
                    File.WriteAllBytes(binFilename + ".packed", packedBytes);
                    cumulatedPackedSize += packedBytes.Length;
                }
            }

            z80TileConfig += "\n;\n";

            allPackedTiles = new List<byte>();
            tileOffset = 0;
            foreach (var binArray in binFile)
            {
                allPackedTiles.AddRange(binArray);
                // + &4000 because EndPackedTiles1==&C4 offset=+&4000
                z80TileConfig += "END_TILE_" + iTile + " equ EndPackedTiles1 +&4000 + &" + tileOffset.ToString("X4") + "\n";
                z80TileConfig += "END_TILE_" + iTile + "_BANK equ &C4\n";
                z80TileConfig += "\n";
                gridDwLabels.Add("END_TILE_" + iTile);
                gridDbBanks.Add("END_TILE_" + iTile + "_BANK");
                iTile++;
                tileOffset += binArray.Length;
            }
            File.WriteAllBytes(outFolder + "../_EndScreen_PackedTiles_1.bin", allPackedTiles.ToArray());
            File.WriteAllText(outFolder + "../_EndTileConfig.z80", z80TileConfig);

            Console.WriteLine("Slice bitmap - BBPack Cumulated Size=" + cumulatedBBPackedSize + " bytes");
            Console.WriteLine("Slice bitmap - Pack Cumulated Size=" + cumulatedPackedSize + " bytes");

            var z80TilesPtr = ";\n"; // sort by cloud order
            int tileCount = 0;
            var cloudBitmap = new Bitmap(System.Drawing.Image.FromFile(cloudFilename));
            for(int iColor=255; iColor>=0; iColor--)
            {
                for(int y=0; y<cloudBitmap.Height; y++)
                {
                    for(int x=0; x<cloudBitmap.Width; x++)
                    {
                        var pixel = cloudBitmap.GetPixel(x, y);
                        if(pixel.R==(byte)iColor)
                        {
                            z80TilesPtr += "\t;; X=" + x + " Y=" + y + "\n";

                            int scrY = y * (boxHeight);
                            if(scrY>=268)
                            {
                                scrY = 268-1;
                            }
                            int scrX = x * (boxWidth);
                            ushort screenPtr = (ushort)(linesPtr[scrY]+(scrX/2));
                            var offset = (y * (gridWidth + 1)) + x;
                            z80TilesPtr += "\tdb " + gridDbBanks[offset] + " ;; X=" + x + " Y=" + y + "\n";
                            z80TilesPtr += "\tdw " + gridDwLabels[offset] + " ;; X=" + x + " Y=" + y + "\n";
                            z80TilesPtr += "\tdw " + (scrY*2) + " ;; ScrY*2\n";
                            z80TilesPtr += "\tdb " + (scrX/2) + " ;; ScrX/2\n";
                            z80TilesPtr += "\n";
                            tileCount++;
                        }
                    }
                }
            }
            var bitmapCloudSize = cloudBitmap.Width * cloudBitmap.Height;
            File.WriteAllText(outFolder + "../_PtrTiles.z80", z80TilesPtr);

            var z80 = "";
            z80 += ";\n";
            z80 += "ENDSCREEN_BOXWIDTH equ " + boxWidth + "\n";
            z80 += "ENDSCREEN_BOXHEIGHT equ " + boxHeight + "\n";
            z80 += "\n";
            z80 += "ENDSCREEN_GRIDWIDTH equ " + (gridWidth + 1) + "\n";
            z80 += "ENDSCREEN_GRIDHEIGHT equ " + (gridHeight + 1) + "\n";
            z80 += "ENDSCREEN_GRIDSIZE equ " + (gridWidth * gridHeight) + "\n";
            z80 += "\n";
            z80 += "ENDSCREEN_TILECOUNT equ " + tileCount + "\n";
            z80 += "\n";
            z80 += "ENDSCREEN_BBPACKED_CUMLATED_SIZE equ " + cumulatedPackedSize + "\n";
            z80 += "\n";
            z80 += "ENDSCREEN_UNPACKED_TILEDATA equ &1000\n";
            z80 += "ENDSCREEN_SCREENPTR_TABLE equ &2000\n";
            File.WriteAllText(outFolder + "../_SliceConfig.z80", z80);
        }
    }
}
