﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public partial class _18_MultiFX
    {
        public const int NbStaticFrames = 16 * 6;
        public const int NbScrollFrames = 16 * 6;

        public _18_MultiFX()
        {
            try
            {
                // Write Global Config
                CreateGlobalConfig(Program.ToZ80Folder("18_MultiFX/"));

                // Create Scroll Vertical Moves
                CreateScrollVerticalMoves_Up(Program.ToZ80Folder("18_MultiFX/"));
                CreateScrollVerticalMoves_Down(Program.ToZ80Folder("18_MultiFX/"));

                var globalPalette = new Palette();
                var chessColors = new List<Color>();
                chessColors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000))); // background
                chessColors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[14] | 0xFF000000))); // alternate color for multiscroll
                chessColors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[4] | 0xFF000000))); // Chess color2
                chessColors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[5] | 0xFF000000))); // Chess color3
                
                var colors = new List<Color>();
                colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000))); // background

                // color condense multiscroll
                colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[14] | 0xFF000000))); // alternate color for multiscroll

                //colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[4] | 0xFF000000))); // Chess color2
                //colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[5] | 0xFF000000))); // Chess color3
                foreach(var color in colors)
                {
                    globalPalette.AddColor(color);
                }
                
                // Plasma
                CreatePlasmaSinMoves(Program.ToZ80Folder("18_MultiFX/Plasma/"));
                CreatePlasmaR13(Program.ToZ80Folder("18_MultiFX/Plasma/"));
                var plasmaPalette = new Palette(2);
                
                // chess pens 2-3
                plasmaPalette.AddColor(Color.FromArgb((int)(Pixel.CPCAsicPalette[1] | 0xFF000000))); // background
                plasmaPalette.AddColor(Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000))); // background

                plasmaPalette = CreatePlasmaImage(
                    plasmaPalette,
                    Program.ToDataFolder("18_MultiFX/Plasma/plasma256_14.png"),
                    Program.ToZ80Folder("18_MultiFX/"),
                    Program.ToZ80Folder("18_MultiFX/Plasma/"));

                plasmaPalette.Pens[0] = globalPalette.Pens[0];
                plasmaPalette.Pens[1] = globalPalette.Pens[1];
                globalPalette = plasmaPalette;
                      /*
                for (int i = 2; i < 16; i++)
                {
                    globalPalette.AddColor(plasmaPalette.Pens[i].Color);
                }*/

                // MultiScroll
                CreateMultiScrollRasters(
                    Program.ToDataFolder("18_MultiFX/MultiScroll/Rasters256_2.png"),
                    Program.ToZ80Folder("18_MultiFX/MultiScroll/_Rasters256_2.png.z80"));
                CreateMultiScrollRasters(
                    Program.ToDataFolder("18_MultiFX/MultiScroll/Rasters256_4.png"),
                    Program.ToZ80Folder("18_MultiFX/MultiScroll/_Rasters256_3.png.z80"));
                CreateMultiScrollSinMoves(Program.ToZ80Folder("18_MultiFX/MultiScroll/"));
                var r12r13 = CreateMultiScrollImage(
                    Program.ToDataFolder("18_MultiFX/MultiScroll/CONDENSE_256x8.png"),
                    Program.ToZ80Folder("18_MultiFX/MultiScroll/"));
                CreateMultiScrollR12R13(Program.ToZ80Folder("18_MultiFX/MultiScroll/"), r12r13);
                CreateMultiScrollScreenConfig(Program.ToZ80Folder("18_MultiFX/"));

                // Chess
                Bitmap notFlippedDebugBitmap = null;
                Bitmap flippedDebugBitmap = null;
                CreateChessBitmap(
                    ref chessColors, Program.ToZ80Folder("18_MultiFX/Chess/"), out notFlippedDebugBitmap, out flippedDebugBitmap);
                CreateChessVerticalData(Program.ToZ80Folder("18_MultiFX/Chess/"), notFlippedDebugBitmap, flippedDebugBitmap);

                // Aurore
                CreateAuroreSinMoves(
                    Program.ToZ80Folder("18_MultiFX/Aurore/"));
                var aurorePalette = new Palette();
                aurorePalette.AddColor(Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000)));
                aurorePalette.AddColor(Color.FromArgb((int)(Pixel.CPCAsicPalette[18] | 0xFF000000))); // unused
                aurorePalette = CreateAuroreBars(
                    ref aurorePalette,
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar0_Left.png"),
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar0_Right.png"),
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar0_Left.png"), //used
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar0_Right.png"),//used  vert=2  classic=0
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar2_Left.png"),
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar2_Right.png"),
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar3_Left.png"),//used
                    Program.ToDataFolder("18_MultiFX/Aurore/Bar3_Right.png"),//used
                    Program.ToZ80Folder("18_MultiFX/Aurore/"));

                // Aurore Palette
                File.WriteAllText(
                    Program.ToZ80Folder("18_MultiFX/") + "_AurorePalette.pal.z80",
                    aurorePalette.ToZ80TableString("AURORE")
                );
                
                // Global Palette
                File.WriteAllText(
                    Program.ToZ80Folder("18_MultiFX/") + "_GlobalPalette.pal.z80",
                    globalPalette.ToZ80TableString("PLASMA_GLOBAL")
                );

                MergeAllMultiFXVLCs(Program.ToZ80Folder("18_MultiFX/"));
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        public void MergeAllMultiFXVLCs(string outFolder)
        {
            var page0 = File.ReadAllBytes(outFolder + "MultiScroll\\_ShiftImage.png.page0.vlc0");
            var plasmaScanline0 = File.ReadAllBytes(outFolder + "Plasma\\_PlasmaScanlines\\_PlasmaScanline_0.png.bin");
            plasmaScanline0.CopyTo(page0, 0x380);
            var plasmaScanline1 = File.ReadAllBytes(outFolder + "Plasma\\_PlasmaScanlines\\_PlasmaScanline_2.png.bin");
            plasmaScanline1.CopyTo(page0, 0x780);

            var page1 = File.ReadAllBytes(outFolder + "Chess\\_ChessFrames\\_ChessFrames.page0.vlc0");
            var page2 = File.ReadAllBytes(outFolder + "Chess\\_ChessFrames\\_ChessFrames.page1.vlc0");
            var page3 = File.ReadAllBytes(outFolder + "Chess\\_ChessFrames\\_ChessFrames.page2.vlc0");

            File.WriteAllBytes(outFolder + "_MultiFX.page0.vlc0", page0);
            //File.WriteAllBytes(outFolder + "_MultiFX.page0.vlc0.bbPacked", BBPacker.Pack(ref page0));
            File.WriteAllBytes(outFolder + "_MultiFX.page1.vlc0", page1);
            //File.WriteAllBytes(outFolder + "_MultiFX.page1.vlc0.bbPacked", BBPacker.Pack(ref page1));
            File.WriteAllBytes(outFolder + "_MultiFX.page2.vlc0.bbPacked", BBPacker.Pack(ref page2));
            File.WriteAllBytes(outFolder + "_MultiFX.page3.vlc0.bbPacked", BBPacker.Pack(ref page3));
        }

        public void CreateGlobalConfig(string outFolder)
        {
            var z80 = "";
            z80 += ";\n";
            z80 += "MULTIFX_1FX_NB_FRAMES\tequ " + (NbStaticFrames+ NbScrollFrames) + "\n";
            z80 += "MULTIFX_NB_STATIC_FRAMES\tequ " + NbStaticFrames + "\n";
            z80 += "MULTIFX_NB_SCROLL_FRAMES\tequ " + NbScrollFrames + "\n";

            File.WriteAllText(outFolder + "_MultiFX_Config.z80", z80);
        }

        double BounceMaxValue = 268 / 2;
        double BounceNbItems = NbScrollFrames;

        private List<byte> CreateBounce_Byte(double percentItemLength, double percentBounce)
        {
            var bytes = SinCurve.CreateReverseCos((int)(percentItemLength* BounceNbItems), 0, (int)(BounceMaxValue*percentBounce), 1);

            for(int i=0; i<bytes.Length; i++)
            {
                bytes[i] = (byte)(BounceMaxValue - bytes[i]);
            }
            return new List<byte>(bytes);
        }

        private List<short> CreateBounce_Short(double percentItemLength, double percentBounce)
        {
            var doubles = SinCurve.CreateReverseCosDouble((int)(percentItemLength * BounceNbItems), 0, (int)(3 * BounceMaxValue * percentBounce), 1);

            var s = new List<short>();
            foreach(var d in doubles)
            {
                s.Add((short)d);
            }
            
            return new List<short>(s);
        }

        public void CreateScrollVerticalMoves_Up(string outFolder)
        {
            var moveBytes = new List<byte>();

            double startItems = 0.9;
            double startHeight = 1;

            var original = CreateBounce_Byte(startItems, startHeight);
            moveBytes.AddRange(original.Skip(original.Count / 2));

            for (int i=0; i<2; i++)
            {
                startItems *= 0.6;
                startHeight *= 0.2;
                moveBytes.AddRange(CreateBounce_Byte(startItems, startHeight));
            }

            var finalBytes = new List<byte>();
            for (int i = 0; i < NbScrollFrames+NbStaticFrames-moveBytes.Count; i++)
            {
                finalBytes.Add(0);
            }
            finalBytes.AddRange(moveBytes);

            if (finalBytes.Count != (NbStaticFrames + NbScrollFrames))
            {
                System.Diagnostics.Debugger.Break();
            }

            /*for(int i=0; i<finalBytes.Count; i++)
            {
                finalBytes[i] = (byte)(134-finalBytes[i]);
            }*/

            File.WriteAllBytes(outFolder + "_GlobalVerticalScrollOffset_Up.bin", finalBytes.ToArray());
        }

        public void CreateScrollVerticalMoves_Down(string outFolder)
        {
            var moveShorts = new List<short>();

            double startItems = 2.3;
            double startHeight = 1;

            var original = CreateBounce_Short(startItems, startHeight);
            moveShorts.AddRange(original.Take(original.Count / 2));

            var finalBytes = new List<byte>();
            foreach (var s in moveShorts)
            {
                finalBytes.Add((byte)(s & 255));
                finalBytes.Add((byte)(s >> 8));
            }
            var count = NbScrollFrames - moveShorts.Count + 20 + 25;
            for (int i = 0; i < count; i++)
            {
                finalBytes.Add(0x92);
                finalBytes.Add(0x01);
            }

            File.WriteAllBytes(outFolder + "_GlobalVerticalScrollOffset_Down.bin", finalBytes.ToArray());
        }
    }
}
