﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public partial class _18_MultiFX
    {
        public void CreateAuroreSinMoves(string outFolder)
        {
            File.WriteAllBytes(outFolder + "_AuroreSinDisto.bin", SinCurve.CreateSin(256, 0, 47*2, 1));
        }

        public Palette CreateAuroreBars(
            ref Palette aurorePalette,
            string bar0LeftFilename,
            string bar0RightFilename,
            string bar1LeftFilename,
            string bar1RightFilename,
            string bar2LeftFilename,
            string bar2RightFilename,
            string bar3LeftFilename,
            string bar3RightFilename,
            string outFolder)
        {
            
            var leftBitmaps = new List<Bitmap>();
            leftBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar0LeftFilename)));
            leftBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar1LeftFilename)));
            leftBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar2LeftFilename)));
            leftBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar3LeftFilename)));

            var rightBitmaps = new List<Bitmap>();
            rightBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar0RightFilename)));
            rightBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar1RightFilename)));
            rightBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar2RightFilename)));
            rightBitmaps.Add(new Bitmap(System.Drawing.Image.FromFile(bar3RightFilename)));

            var z80 = "";
            int iBar = 0;
            foreach (var bitmap in leftBitmaps)
            {
                z80 += ";\n";
                z80 += "MACRO MFXAurore_LeftBar" + iBar + "\n";

                var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = bitmap,
                    Palette = aurorePalette,
                    Mode = 0,
                    R1 = 64,
                    R4 = 8 * 4, // 8 VCC * 4 Shifts
                    R9 = 0,
                    StartR12 = 0x00 // VRAM at 0x0000
                });
                aurorePalette = cpcImage.Palette;

                z80 += "\tld (hl), &" + cpcImage.CPCRawBuffer[0].ToString("X2") + "\n";
                z80 += "\tinc l\n";
                z80 += "\tld (hl), &" + cpcImage.CPCRawBuffer[1].ToString("X2") + "\n";
                z80 += "\tinc l\n";
                z80 += "\tld a, (hl)\n";
                z80 += "\tand d ;; &55 ;; mise a 0 du pixel a gauche\n";
                z80 += "\tor &" + cpcImage.CPCRawBuffer[2].ToString("X2") + "\n";
                z80 += "\tld (hl), a\n";
                
                z80 += "MEND\n";

                iBar++;
            }

            File.WriteAllText(outFolder + "_BarLeftMacros.z80", z80);


            z80 = "";
            iBar = 0;
            foreach (var bitmap in rightBitmaps)
            {
                z80 += ";\n";
                z80 += "MACRO MFXAurore_RightBar" + iBar + "\n";

                var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = bitmap,
                    Palette = aurorePalette,
                    Mode = 0,
                    R1 = 64,
                    R4 = 8 * 4, // 8 VCC * 4 Shifts
                    R9 = 0,
                    StartR12 = 0x00 // VRAM at 0x0000
                });
                aurorePalette = cpcImage.Palette;

                z80 += "\tld a, (hl)\n";
                z80 += "\tand e ;; &AA ;; mise a 0 du pixel a droite\n";
                z80 += "\tor &" + cpcImage.CPCRawBuffer[0].ToString("X2") + "\n";
                z80 += "\tld (hl), a\n";
                z80 += "\tinc l\n";
                z80 += "\tld (hl), &" + cpcImage.CPCRawBuffer[1].ToString("X2") + "\n";
                z80 += "\tinc l\n";
                z80 += "\tld (hl), &" + cpcImage.CPCRawBuffer[2].ToString("X2") + "\n";
                
                z80 += "MEND\n";

                iBar++;
            }

            File.WriteAllText(outFolder + "_BarRightMacros.z80", z80);

            return aurorePalette;
        }
    }
}
