﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public partial class _18_MultiFX
    {
        const int ChessNbFrames = 16;
        const int ChessWidth = 64;
        const int ChessHeight = 64;
        const int CPCScreenWidthMode1 = 96 * 2 * 2;
        const int HorizFrameSkip = 1;
        const int VertFrameSkip = 1;

        // swap color 0 and 1
        // 8 frames * 4 cases = 32 CRTC lines

        private void CopyLine(ref Bitmap srcBitmap, ref Bitmap dstBitmap, int srcY, int dstY)
        {
            for (int x = 0; x < srcBitmap.Width; x++)
            {
                var pixel = srcBitmap.GetPixel(x, srcY);
                dstBitmap.SetPixel(x, dstY, pixel);
            }
        }

        private class Line
        {
            public int LineIndexInR12R13 = -1;

            // 1st pass: large cubes
            public enum E1stPass
            {
                Interval_01,
                Interval_23
            }
            public E1stPass LargeCubePass;

            // 2nd pass: mid cubes
            public int LineIndexInSrcBitmap = -1; // 0->3

            // 3rd pass: small cubes (IsSwapColors)            
            public bool IsSwapColors = false;
        }

        private List<Line> CreateChessVerticalData(string outFolder, Bitmap notFlippedDebugBitmap, Bitmap flippedDebugBitmap, int iFrame)
        {
            var bitmap = new Bitmap(notFlippedDebugBitmap.Width, ChessHeight);

            var lines = new Line[ChessHeight];
            for (int y = 0; y < ChessHeight; y++)
            {
                lines[y] = new Line();
            }

            // large cubes
            int index = -1;
            int offset = iFrame * 4;
            for (int y = -offset; y < ChessHeight; y++)
            {
                index = (index + 1) & 63;
                if (y < 0)
                {
                    continue;
                }

                var isFlipped = (index >= 32);
                if (!isFlipped)
                {
                    lines[y].LargeCubePass = Line.E1stPass.Interval_01;
                }
                else
                {
                    lines[y].LargeCubePass = Line.E1stPass.Interval_23;
                }
            }

            // mid cubes
            offset = iFrame * 2;
            index = -1;
            for (int y = -offset; y < ChessHeight; y++)
            {
                index = (index + 1) & 31;
                if (y < 0)
                {
                    continue;
                }

                var isFlipped = (index >= 16);
                if (lines[y].LargeCubePass == Line.E1stPass.Interval_01)
                {
                    if (!isFlipped)
                    {
                        lines[y].LineIndexInSrcBitmap = 0;
                    }
                    else
                    {
                        lines[y].LineIndexInSrcBitmap = 1;
                    }
                }
                else
                {
                    if (!isFlipped)
                    {
                        lines[y].LineIndexInSrcBitmap = 2;
                    }
                    else
                    {
                        lines[y].LineIndexInSrcBitmap = 3;
                    }
                }
            }

            // small cubes
            offset = iFrame;
            index = -1;
            for (int y = -offset; y < ChessHeight; y++)
            {
                index = (index + 1) & 15;
                if (y < 0)
                {
                    continue;
                }

                var isFlipped = (index >= 8);
                lines[y].IsSwapColors = isFlipped;
            }

            // create bitmap
            for (int y = 0; y < ChessHeight; y++)
            {
                var srcY = lines[y].LineIndexInSrcBitmap; // 0->3
                var dstY = y;

                var srcBitmap = notFlippedDebugBitmap;
                if (lines[y].IsSwapColors)
                {
                    srcBitmap = flippedDebugBitmap;
                }

                CopyLine(ref srcBitmap, ref bitmap, srcY, dstY);
            }

            // LineIndexInR12R13
            for (int y = 0; y < ChessHeight; y++)
            {
                lines[y].LineIndexInR12R13 = (lines[y].LineIndexInSrcBitmap);

                // next line is for horiz offset
                // commented out because done on CPC!
                //lines[y].LineIndexInR12R13 += ((iFrame/ HorizFrameSkip) * 4); // (0->7)*4

                lines[y].LineIndexInR12R13 *= 2; // *2=r12+r13
            }

            bitmap.Save(outFolder + "Debug_VerticalFrames/_Frame_" + (iFrame / VertFrameSkip).ToString("00") + ".png", ImageFormat.Png);

            return new List<Line>(lines);
        }

        private string VerticalDataToZ80(int iFrame, List<Line> lines)
        {
            var z80 = "";

            z80 += "\n";
            z80 += ";; format: db (LineIndexInR12R13<<1)+IsSwapColors (0 or 1)\n";
            z80 += "\n";

            z80 += ";\n";
            z80 += "ChessVerticalData_" + (iFrame / VertFrameSkip).ToString("00") + ":\n";

            foreach (var line in lines)
            {
                byte isSwapColors = 0;
                if (line.IsSwapColors)
                {
                    isSwapColors = 1;
                }
                z80 += "\tdb " + (byte)((line.LineIndexInR12R13 << 1) + isSwapColors) + "\n";
            }

            return z80;
        }

        public void CreateChessVerticalData(string outFolder, Bitmap notFlippedDebugBitmap, Bitmap flippedDebugBitmap)
        {
            for (int iFrame = 0; iFrame < ChessNbFrames; iFrame += VertFrameSkip)
            {
                var lines = CreateChessVerticalData(outFolder, notFlippedDebugBitmap, flippedDebugBitmap, iFrame);
                var z80 = VerticalDataToZ80(iFrame, lines);
                File.WriteAllText(outFolder + "/_VerticalFrameData/_FXChess_VerticalFrameData_" + (iFrame / VertFrameSkip).ToString("00") + ".z80", z80);
            }
        }

        public void CreateChessBitmap(ref List<Color> colors, string outFolder, out Bitmap notFlippedDebugBitmap, out Bitmap flippedDebugBitmap)
        {
            notFlippedDebugBitmap = null;
            flippedDebugBitmap = null;

            /*var colors = new List<Color>();
            colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[0] | 0xFF000000)));
            colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[7] | 0xFF000000)));
            colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[15] | 0xFF000000)));
            colors.Add(Color.FromArgb((int)(Pixel.CPCAsicPalette[3] | 0xFF000000)));*/

            var cpcBitmaps = new List<Bitmap>();

            for (int iFrame = 0; iFrame < ChessNbFrames; iFrame += HorizFrameSkip)
            {
                // Debug
                Bitmap bitmap = new Bitmap(ChessWidth, ChessHeight);
                DrawLayers(ref colors, ref bitmap, iFrame, false, false);
                bitmap.Save(outFolder + "Debug_HorizontalFrames/_Frame_" + (iFrame / HorizFrameSkip).ToString("00") + ".png", ImageFormat.Png);

                bitmap = new Bitmap(CPCScreenWidthMode1, 4);
                DrawLayers(ref colors, ref bitmap, iFrame, true, false);
                bitmap.Save(outFolder + "CPCFrames/_Frame_" + (iFrame / HorizFrameSkip).ToString("00") + ".png", ImageFormat.Png);
                cpcBitmaps.Add(bitmap);

                if (iFrame == 0)
                {
                    notFlippedDebugBitmap = bitmap;

                    flippedDebugBitmap = new Bitmap(notFlippedDebugBitmap.Width, notFlippedDebugBitmap.Height);
                    DrawLayers(ref colors, ref flippedDebugBitmap, iFrame, true, true);
                }
            }

            var finalCpcBitmap = new Bitmap(CPCScreenWidthMode1, cpcBitmaps.Count * 4);
            int iY = 0;
            foreach (var bitmap in cpcBitmaps)
            {
                for (int y = 0; y < 4; y++)
                {
                    for (int x = 0; x < CPCScreenWidthMode1; x++)
                    {
                        var pixel = bitmap.GetPixel(x, y);
                        finalCpcBitmap.SetPixel(x, iY + y, pixel);
                    }
                }

                iY += 4;
            }
            finalCpcBitmap.Save(outFolder + "_AllFrames.png", ImageFormat.Png);

            var palette = new Palette();
            palette.AddColor(colors[0]);
            palette.AddColor(colors[1]);
            palette.AddColor(colors[2]);
            palette.AddColor(colors[3]);

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = finalCpcBitmap,
                Mode = 1,
                R1 = 48,
                R4 = 21,
                R9 = 0,
                StartVLC = 0,
                StartR12 = 0x40,
                Palette = palette
            });

            for (var iPage = 0; iPage < cpcImage.Pages.Length; iPage++)
            {
                foreach (var VLC in cpcImage.Pages[iPage].VLCs)
                {
                    int nbWrittenBytes = cpcImage.Pages[iPage].NbBytesWritten;
                    var cutArray = new List<byte>(VLC.CPCBytes).Take(nbWrittenBytes).ToArray();

                    var filename = outFolder + "_ChessFrames/_ChessFrames.page" + iPage + ".vlc" + VLC.Index;
                    File.WriteAllBytes(filename, cutArray);

                    var vlcBBPackedBytes = BBPacker.Pack(ref cutArray);
                    File.WriteAllBytes(filename + ".bbPacked", vlcBBPackedBytes);

                    var packer = new Packer();
                    var packedBytes = packer.Convert(ref cutArray);
                    File.WriteAllBytes(filename + ".packed", packedBytes);

                    /*if (iPage == 0)
                    {
                        var listBytes = new List<byte>(cutArray);

                        var first0040Bytes = listBytes.Take(0x0040).ToArray();
                        var after0040Bytes = listBytes.Skip(0x0040).ToArray();

                        var filename_first0040 = filename + ".first0040bytes";
                        File.WriteAllBytes(filename_first0040, first0040Bytes.ToArray());

                        var bbPackedBytes = BBPacker.Pack(ref after0040Bytes);
                        var filename_after0040 = filename + ".after0040bytes";
                        File.WriteAllBytes(filename + ".bbPacked", bbPackedBytes);
                    }*/
                }
            }

            /*
            FX_MULTISCROLL_VCC_15_SCREENPTR equ &0780
            FX_MULTISCROLL_VCC_15_R12 equ &03
            FX_MULTISCROLL_VCC_15_R13 equ &C0

            FX_MULTISCROLL_VCC_08_SCREENPTR	equ &0400
            FX_MULTISCROLL_VCC_08_R12	equ &02
            FX_MULTISCROLL_VCC_08_R13	equ &00
            */
            cpcImage.R12R13[126] = (byte)(0x02);
            cpcImage.R12R13[127] = (byte)(0x00);
            File.WriteAllBytes(outFolder + "_ChessFrames.r12r13", cpcImage.R12R13);
        }

        private int GetY(bool isCPCFrame, int y)
        {
            if (isCPCFrame == false)
            {
                return y;
            }

            const int delta = ChessHeight / 4;

            switch (y)
            {
                case 0 * delta: return 0;
                case 1 * delta: return 1;
                case 2 * delta: return 2;
                case 3 * delta: return 3;
                default:
                    return -1;
            }
        }

        private void DrawLayers(ref List<Color> colors, ref Bitmap bitmap, int iFrame, bool isCPCFrame, bool isReverseFlipForSmallCubes)
        {
            int width = bitmap.Width;
            bool isFlipped;

            // empty background
            for (int y = 0; y < bitmap.Height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    bitmap.SetPixel(x, y, colors[2]);
                }
            }

            // small cubes
            for (int y = 0; y < ChessHeight; y++)
            {
                int iY = y & 15;
                isFlipped = (iY < 8);

                if (isReverseFlipForSmallCubes)
                {
                    if (isFlipped)
                    {
                        isFlipped = false;
                    }
                    else
                    {
                        isFlipped = true;
                    }
                }

                int drawY = GetY(isCPCFrame, y);
                if (drawY == -1)
                {
                    continue;
                }

                int iX = (iFrame) & 15;
                for (int x = 0; x < width; x++)
                {
                    if (((!isFlipped) && (iX < 8)) || ((isFlipped) && (iX >= 8)))
                    {
                        bitmap.SetPixel(x, drawY, colors[3]);
                    }

                    iX++;
                    if (iX == 16)
                    {
                        iX = 0;
                    }
                }
            }

            // mid cubes
            for (int y = 0; y < ChessHeight; y++)
            {
                int iY = y & 31;
                isFlipped = (iY < 16);

                int drawY = GetY(isCPCFrame, y);
                if (drawY == -1)
                {
                    continue;
                }

                int c = (iFrame * 2) & 31;
                for (int x = 0; x < width; x++)
                {
                    if (((!isFlipped) && (c < 16)) || ((isFlipped) && (c >= 16)))
                    {
                        bitmap.SetPixel(x, drawY, colors[1]);
                    }

                    c++;
                    if (c == 32)
                    {
                        c = 0;
                    }
                }
            }

            // large cubes
            for (int y = 0; y < ChessHeight; y++)
            {
                int iY = y & 63;
                isFlipped = (iY < 32);

                int drawY = GetY(isCPCFrame, y);
                if (drawY == -1)
                {
                    continue;
                }

                int c = (iFrame * 4) & 63;
                for (int x = 0; x < width; x++)
                {
                    if (((!isFlipped) && (c < 32)) || ((isFlipped) && (c >= 32)))
                    {
                        bitmap.SetPixel(x, drawY, colors[0]);
                    }

                    c++;
                    if (c == 64)
                    {
                        c = 0;
                    }
                }
            }
        }
    }
}
