﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public partial class _18_MultiFX
    {
        private void CreateMultiScrollRasters(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";
                //z80Content += "\tdb COLOR24\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }

        public void CreateMultiScrollSinMoves(string outFolder)
        {
            File.WriteAllBytes(outFolder + "_MultiScrollSinDisto.bin", SinCurve.CreateCos(256, 0, 60, 1));
            File.WriteAllBytes(outFolder + "_MultiScrollSinDistoVertical.bin", SinCurve.CreateSin(256, 0, 240, 2));
            //File.WriteAllBytes(outFolder + "_MultiScrollSinDisto.bin", SinCurve.CreateCos(256, 0, 0, 1));
            //File.WriteAllBytes(outFolder + "_MultiScrollSinDistoVertical.bin", SinCurve.CreateSin(256, 0, 0, 2));
        }

        const int Mode2_ScreenWidthInPixels = 192 * 2 * 2;

        private void CreateMultiScrollR12R13(string outFolder, byte[] r12r13)
        {
            // CONDENSE is made of 8 lines (8 VCC with R9=0)
            for (int iVCC = 0; iVCC < 8; iVCC++)
            {
                var r12_bytes = new List<byte>();
                var r13_bytes = new List<byte>();

                // CONDENSE can scroll 16 CRTC blocks horizontally
                for (int iCharX = 0; iCharX < 16; iCharX++)
                {
                    // 2 shifts per CRTC blocks
                    for (int iShift = 0; iShift < 2; iShift++)
                    {
                        int r12r13_index = iVCC+(iShift*8);

                        // redirect empty lines to 1st empty line
                        // so remaining empty lines can be used for other FX
                        if((r12r13_index==7)||(r12r13_index == 8)||(r12r13_index == 15))
                        {
                            r12r13_index = 0;
                        }

                        var r12 = r12r13[r12r13_index*2];
                        var r13 = (byte)(r12r13[r12r13_index*2 + 1] + iCharX);

                        r12_bytes.Add(r12);
                        r13_bytes.Add(r13);
                    }
                }

                File.WriteAllBytes(outFolder + "_R12R13\\_r12_vcc" + iVCC + ".bin", r12_bytes.ToArray());
                File.WriteAllBytes(outFolder + "_R12R13\\_r13_vcc" + iVCC + ".bin", r13_bytes.ToArray());
            }
        }

        private void CreateMultiScrollScreenConfig(string outFolder)
        {
            var z80 = "";
            z80 += ";\n";

            int offset = 0;
            for(int iVCC=0; iVCC<16; iVCC++)
            {
                ushort screenPtr = (ushort)(offset * 2);
                z80 += "FX_MULTISCROLL_VCC_" + iVCC.ToString("00") + "_SCREENPTR\tequ &" + screenPtr.ToString("X4") + "\n";

                ushort R12R13 = (ushort)(offset);
                byte R12 = (byte)(R12R13 >> 8);
                byte R13 = (byte)(R12R13 & 255);

                z80 += "FX_MULTISCROLL_VCC_" + iVCC.ToString("00") + "_R12\tequ &" + R12.ToString("X2") + "\n";
                z80 += "FX_MULTISCROLL_VCC_" + iVCC.ToString("00") + "_R13\tequ &" + R13.ToString("X2") + "\n";

                z80 += "\n";

                offset += 64;
            }

            File.WriteAllText(outFolder+ "_MultiScroll_ScreenConfig.z80", z80);
        }

        private byte[] CreateMultiScrollImage(string pngFilename, string outFolder)
        {
            var srcBitmap = new Bitmap(System.Drawing.Image.FromFile(pngFilename));

            const int nbShift = 2;
            const int horizDuplicateCount = 5; // number of times we repeat CONDENSE words through the horizontal line

            int screenWidth_WithScrollWidth = Mode2_ScreenWidthInPixels + srcBitmap.Width;
            // cpcScreenWidth=768 w=256 w*2=512 w*3=768 w*4=1024

            // 1024 width=&400

            var giantBitmap = new Bitmap(screenWidth_WithScrollWidth, srcBitmap.Height * nbShift);
            var bmpFilename = "";

            var shiftedBitmaps = new List<Bitmap>();
            for (int iShift = 0; iShift < nbShift; iShift++)
            {
                var shiftedBitmap = new Bitmap(screenWidth_WithScrollWidth, srcBitmap.Height);
                for (int iDuplicate = 0; iDuplicate < horizDuplicateCount; iDuplicate++)
                {
                    for (int y = 0; y < srcBitmap.Height; y++)
                    {
                        for (int x = 0; x < srcBitmap.Width; x++)
                        {
                            int iSrcX = x + iShift * 8; // *8=move 8 pixels per shift .. using mode 2
                            if (iSrcX >= srcBitmap.Width)
                            {
                                iSrcX -= srcBitmap.Width;
                            }

                            var dstX = x + srcBitmap.Width * iDuplicate;
                            if ((dstX < 0) || (dstX >= screenWidth_WithScrollWidth))
                            {
                                continue;
                            }

                            var pixel = srcBitmap.GetPixel(iSrcX, y);
                            shiftedBitmap.SetPixel(dstX, y, pixel);

                            giantBitmap.SetPixel(dstX, y + iShift * srcBitmap.Height, pixel);
                        }
                    }
                }
                shiftedBitmaps.Add(shiftedBitmap);

                bmpFilename = outFolder + "_ShiftImage_" + iShift + ".png";
                //shiftedBitmap.Save(bmpFilename, ImageFormat.Png);
            }

            bmpFilename = outFolder + "_ShiftImage.png";
            giantBitmap.Save(bmpFilename, ImageFormat.Png);

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = giantBitmap,
                Mode = 2,
                R1 = 64,
                R4 = 8 * 2, // 8 VCC * 2 Shifts
                R9 = 0,
                StartR12 = 0x00 // VRAM at 0x0000
            });

            var imageArray = cpcImage.CPCRawBuffer.ToArray();
            File.WriteAllBytes(bmpFilename + ".page0.vlc0", imageArray);

            var BBPackedBytes = BBPacker.Pack(ref imageArray);
            File.WriteAllBytes(bmpFilename + ".page0.vlc0.bbPacked", BBPackedBytes);
            
            var packer = new Packer();
            var packedBytes = packer.Convert(ref imageArray);
            File.WriteAllBytes(bmpFilename + ".page0.vlc0.packed", packedBytes);

            return cpcImage.R12R13;
        }
    }
}
