﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;

namespace phX.Parts
{
    public partial class _18_MultiFX
    {
        private Palette WritePlasmaImage(Palette palette, string pngFilename, int imageIndex, string outFolder)
        {
            var srcBitmap = new Bitmap(System.Drawing.Image.FromFile(pngFilename));

            var bitmap = new Bitmap(srcBitmap.Width, 1);
            for (int x = 0; x < srcBitmap.Width; x++)
            {
                int iX = x + imageIndex;
                if (iX >= srcBitmap.Width)
                {
                    iX -= srcBitmap.Width;
                }
                bitmap.SetPixel(x, 0, srcBitmap.GetPixel(iX, 0));
            }

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = bitmap,
                Mode = 0,
                R1 = 128,
                R4 = 0,
                R9 = 0,
                Palette = palette
            });

            var filename = outFolder + "_PlasmaScanlines\\_PlasmaScanline_" + imageIndex + ".png.bin";

            var cpcBytes = cpcImage.CPCRawBuffer;
            File.WriteAllBytes(filename, cpcBytes);

            return cpcImage.Palette;
        }

        private Palette CreatePlasmaImage(Palette globalPalette, string pngFilename, string outRootFolder, string outFolder)
        {
            globalPalette = WritePlasmaImage(globalPalette, pngFilename, 0, outFolder);
            //globalPalette = WritePlasmaImage(globalPalette, pngFilename, 1, outFolder);
            globalPalette = WritePlasmaImage(globalPalette, pngFilename, 2, outFolder);
            //globalPalette = WritePlasmaImage(globalPalette, pngFilename, 3, outFolder);

            return globalPalette;
        }

        public void CreatePlasmaR13(string outFolder)
        {
            var bytes = new List<byte>();

            for (int i = 0; i < 32; i++)
            {
                if ((i&1)==0)
                {
                    bytes.Add(1); // FX_MULTISCROLL_VCC_07_R12
                }
                else
                {
                    bytes.Add(3); // FX_MULTISCROLL_VCC_15_R12
                }
                
                //int offset = ((i & 3) * 64) + (i / 4);
                byte offset = (byte)(i / 2);
                bytes.Add((byte)(0xC0+offset)); // 0xC0 = FX_MULTISCROLL_VCC_07_R13  0xC0 = FX_MULTISCROLL_VCC_15_R13
            }

            File.WriteAllBytes(outFolder + "_PlasmaR13.bin", bytes.ToArray());
        }

        public void CreatePlasmaSinMoves(string outFolder)
        {
            File.WriteAllBytes(outFolder + "_PlasmaSinDisto.bin", SinCurve.CreateSin(256, 0, 16, 1));
        }
    }
}
