﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;

namespace phX.Parts
{
    public partial class _19_Zoom
    {
        public const int ScreenWidth = 96-FXLogoWidth;
        public const int ScreenHeight = 67; // 42*6=252 displayed lines  (6 = R9=5)   42=21*2 (R12/R13)
        public const int ZoomDeltaX = 1;//4;
        public const int ZoomDeltaY = 1;//2;
        public const int FXLogoWidth = 42;//16;
        public const int FXWidth = ScreenWidth - ZoomDeltaX*2;
        public const int FXHeight = ScreenHeight - ZoomDeltaY*1; // last line is displayed
        public const int FXReduceIncX = 1;
        public const int FXReduceIncY = 1;
        public const int NbSimulationFrames = 64;
        public const float RotateDegrees = 0.9F;
        
        [XmlRoot("SrcPoint")]
        public class SrcPoint
        {
            public int X;
            public int Y;
        }
        public class MappingTable
        {
            public int ScreenWidth = 0;
            public int ScreenHeight = 0;
            public int FXWidth = 0;
            public int FXHeight = 0;
            public float RotateDegrees = 0;

            [XmlArray("SrcPoints"), XmlArrayItem(typeof(SrcPoint), ElementName = "SrcPoint")]
            public List<SrcPoint> SrcPoints { get; set; }
        }

        public class CPCLine
        {
            public int Page;
            public byte LineIndex;
            public byte R12;
            public byte R13;
            public ushort CPCScreenPtr;
        }
        private List<CPCLine> CPCLines_Screen0;

        public _19_Zoom()
        {
            try
            {
                // MODE 0 pens
                CreateMode0Pens(Program.ToZ80Folder("19_Zoom/_Mode0Pens.z80"));
                
                // Text (GFX+SinMoves)
                int textWidth = 0;
                int textHeight = 0;
                CreateText(Program.ToZ80Folder("19_Zoom/"), out textWidth, out textHeight);
                CreateSinCurves(Program.ToZ80Folder("19_Zoom/_SinCurveX.bin"), Program.ToZ80Folder("19_Zoom/_SinCurveY.bin"), textWidth, textHeight);

                // Screen R12/R13/Ptr
                CreateR12R13Screens(Program.ToZ80Folder("19_Zoom/"));
                WriteR12R13Screen(Program.ToZ80Folder("19_Zoom/_R12R13_Screen0.z80"), "Zoom_Screen0_R12R13", CPCLines_Screen0);
                WriteScreenPtr(Program.ToZ80Folder("19_Zoom/_ScreenPtr_Screen0.z80"), "Zoom_Screen0_ScreenPtr", CPCLines_Screen0);
                
                // Zoom MappingTable/Simulations
                var mappingTable = CreateMappingTable(Program.ToZ80Folder("19_Zoom/"));
                CreateZoomSimulation(Program.ToZ80Folder("19_Zoom/"));
                CreateZoomSimulationWithMappingTable(Program.ToZ80Folder("19_Zoom/"), ref mappingTable);
                CreateZoomSimulationWithMappingTable_UpdateSameBitmap(Program.ToZ80Folder("19_Zoom/"), ref mappingTable);

                // CPC Zoom
                var walkZoom = ComputeWalkZoom(Program.ToZ80Folder("19_Zoom/_WalkStep.z80"), "WalkStep", ref mappingTable);
                WriteZoomFirstCPCImage(Program.ToDataFolder("19_Zoom/"), Program.ToZ80Folder("19_Zoom/"));
            }
            catch (Exception e)
            {
                // try deleting _MappingTable.xml
                Log.Fatal(e.ToString());
            }
        }

        private void CreateSinCurves(string sinXFilename, string sinYFilename, int textWidth, int textHeight)
        {
            int borderWidth = 10;
            int borderHeight = 5;

            int minX = borderWidth;
            int maxX = ScreenWidth - borderWidth - textWidth;
            int minY = borderHeight;
            int maxY = ScreenHeight - borderHeight - textHeight;

            File.WriteAllBytes(sinXFilename, SinCurve.CreateCos(256, minX, maxX, 1));
            File.WriteAllBytes(sinYFilename, SinCurve.CreateSin(256, minY, maxY, 1));
        }

        private void CreateMode0Pens(string destZ80)
        {
            var z80 = "";
            z80 += ";\n";
            z80 += "Zoom_Pens:\n";

            for(int iPen=0; iPen<16; iPen++)
            {
                var pixel = Pixel.ToMode0Pixel((byte)iPen, (byte)iPen);
                z80 += "\tdb &" + pixel.ToString("X2") + "\n";
            }

            File.WriteAllText(destZ80, z80);
        }
    }
}
