﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;

namespace phX.Parts
{
    public partial class _19_Zoom
    {
        private bool HasVerticalPixels(Bitmap bitmap, int x)
        {
            for (int iY = 0; iY < bitmap.Height; iY++)
            {
                var pixel = bitmap.GetPixel(x, iY);
                if ((pixel.R != Color.White.R) && (pixel.G != Color.White.G) && (pixel.B != Color.White.B))
                {
                    return true;
                }
            }

            return false;
        }

        private bool HasHorizontalPixels(Bitmap bitmap, int y)
        {
            for (int iX = 0; iX < bitmap.Width; iX++)
            {
                var pixel = bitmap.GetPixel(iX, y);
                if ((pixel.R != Color.White.R) && (pixel.G != Color.White.G) && (pixel.B != Color.White.B))
                {
                    return true;
                }
            }

            return false;
        }

        private Bitmap CreateSubBitmap(Bitmap bitmap, int left, int right, int top, int bottom)
        {
            var newBitmap = new Bitmap(right - left, bottom - top);

            for (int iY = top; iY < bottom; iY++)
            {
                for (int iX = left; iX < right; iX++)
                {
                    var pixel = bitmap.GetPixel(iX, iY);
                    newBitmap.SetPixel(iX - left, iY - top, pixel);
                }
            }

            return newBitmap;
        }

        private Bitmap CreateBitmap(string text)
        {
            var bitmap = new Bitmap(ScreenWidth, ScreenHeight);
            for (int y = 0; y < bitmap.Height; y++)
            {
                for (int x = 0; x < bitmap.Width; x++)
                {
                    bitmap.SetPixel(x, y, Color.White);
                }
            }

            using (Graphics g = Graphics.FromImage(bitmap))
            {
                using (Font arialFont = new Font("Verdana", 10, FontStyle.Bold))
                {
                    g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.SingleBitPerPixelGridFit;
                    g.DrawString(text, arialFont, Brushes.Black, new Point(0, 0));
                }
            }

            int left = 0;
            while (!HasVerticalPixels(bitmap, left))
            {
                left++;
            }
            int right = bitmap.Width - 1;
            while (!HasVerticalPixels(bitmap, right))
            {
                right--;
            }
            right++;
            int top = 0;
            while (!HasHorizontalPixels(bitmap, top))
            {
                top++;
            }
            int bottom = bitmap.Height - 1;
            while (!HasHorizontalPixels(bitmap, bottom))
            {
                bottom--;
            }
            bottom++;

            bitmap = CreateSubBitmap(bitmap, left, right, top, bottom);

            return bitmap;
        }

        private string CreateZ80Opcodes(Bitmap bitmap)
        {
            var z80 = "";

            for (int iY = 0; iY < bitmap.Height; iY++)
            {
                z80 += ";; line " + iY + "\n";

                var lineZ80 = "";

                lineZ80 += "\tpop hl\n";
                lineZ80 += "\tld c, e\n";
                lineZ80 += "\tadd hl, bc\n";

                int RegC = 0;

                int X = 0;

                for (int iX = 0; iX < bitmap.Width; iX++)
                {
                    var pixel = bitmap.GetPixel(iX, iY);
                    if ((pixel.R != Color.White.R) && (pixel.G != Color.White.G) && (pixel.B != Color.White.B))
                    {
                        if (iX == 0)
                        {
                            // don't update HL
                        }
                        else if ((X + 1) == iX)
                        {
                            lineZ80 += "\tinc hl\n";
                        }
                        else if ((X + 2) == iX)
                        {
                            lineZ80 += "\tinc hl\n";
                            lineZ80 += "\tinc hl\n";
                        }
                        else if ((X + 3) == iX)
                        {
                            lineZ80 += "\tinc hl\n";
                            lineZ80 += "\tinc hl\n";
                            lineZ80 += "\tinc hl\n";
                        }
                        else
                        {
                            //if(RegC!=(iX-X))
                            {
                                RegC = iX - X;
                                lineZ80 += "\tld c, " + RegC + "\n";
                            }
                            lineZ80 += "\tadd hl, bc\n";
                        }
                        lineZ80 += "\tld (hl), d\n";

                        X = iX;
                    }
                }

                z80 += lineZ80;
                z80 += "\n";
            }

            return z80;
        }

        private void CreateText(string outFolder, out int textWidth, out int textHeight)
        {
            // Madness is not over yet.. phX!

            var bitmap = CreateBitmap("phX");
            bitmap.Save(outFolder + "Simulation_Text\\Text_Hello.png", ImageFormat.Png);

            var z80 = CreateZ80Opcodes(bitmap);
            File.WriteAllText(outFolder + "_Text.z80", z80);

            textWidth = bitmap.Width;
            textHeight = bitmap.Height;
        }
    }
}
