﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;

namespace phX.Parts
{
    public partial class _19_Zoom
    {
        private Bitmap CreateReducedBitmap(Bitmap currentBitmap)
        {
            var newBitmap = new Bitmap(FXWidth, FXHeight);
            for (int y = 0; y < FXHeight; y++)
            {
                for (int x = 0; x < FXWidth; x++)
                {
                    //var pixel = currentBitmap.GetPixel(x, y);
                    //var pixel = currentBitmap.GetPixel(0, 0);
                    var pixel = currentBitmap.GetPixel(10, 0);
                    newBitmap.SetPixel(x, y, pixel);
                }
            }

            using (Graphics g = Graphics.FromImage((System.Drawing.Image)newBitmap))
            {
                g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
                g.RotateTransform(RotateDegrees);
                g.DrawImage(currentBitmap, FXReduceIncX, FXReduceIncY, FXWidth, FXHeight);
            }

            return newBitmap;
        }

        private Bitmap CreateReducedBitmapWithMappingTable(Bitmap currentBitmap, ref MappingTable mappingTable)
        {
            var newBitmap = new Bitmap(FXWidth, FXHeight);
            for (int y = 0; y < FXHeight; y++)
            {
                for (int x = 0; x < FXWidth; x++)
                {
                    //var pixel = currentBitmap.GetPixel(x, y);
                    //var pixel = currentBitmap.GetPixel(0, 0);
                    var pixel = Color.White; // not used to created mapping table, so no need to use 0,0
                    newBitmap.SetPixel(x, y, pixel);
                }
            }

            int mappingTableIndex = 0;

            for (int y = 0; y < FXHeight; y++)
            {
                for (int x = 0; x < FXWidth; x++)
                {
                    var srcPoint = mappingTable.SrcPoints[mappingTableIndex];
                    mappingTableIndex++;

                    var srcPixel = currentBitmap.GetPixel(srcPoint.X, srcPoint.Y);
                    newBitmap.SetPixel(x, y, srcPixel);
                }
            }

            return newBitmap;
        }

        private void CreateReducedBitmapWithMappingTable_UpdateSameBitmap(ref Bitmap currentBitmap, ref MappingTable mappingTable)
        {
            // left
            for (int y = FXHeight - 1; y >= 0; y--)
            {
                for (int x = FXWidth - 1; x >= 0; x--)
                {
                    var srcPoint = mappingTable.SrcPoints[(y * FXWidth) + x];
                    var srcPixel = currentBitmap.GetPixel(srcPoint.X, srcPoint.Y);
                    currentBitmap.SetPixel(x + ZoomDeltaX, y + ZoomDeltaY, srcPixel);
                }
            }

            /*// top left
            for (int y = (FXHeight / 2) - 1; y >= 0; y--)
            {
                for (int x = (FXWidth / 2) - 1; x >= 0; x--)
                {
                    var srcPoint = mappingTable.SrcPoints[(y * FXWidth) + x];
                    var srcPixel = currentBitmap.GetPixel(srcPoint.X, srcPoint.Y);
                    currentBitmap.SetPixel(x+ZoomDeltaX, y + ZoomDeltaY, srcPixel);
                }
            }

            // top right
            for (int y = (FXHeight / 2) - 1; y >= 0; y--)
            {
                for (int x = (FXWidth / 2); x < FXWidth; x++)
                {
                    var srcPoint = mappingTable.SrcPoints[(y * FXWidth) + x];
                    var srcPixel = currentBitmap.GetPixel(srcPoint.X, srcPoint.Y);
                    currentBitmap.SetPixel(x + ZoomDeltaX, y + ZoomDeltaY, srcPixel);
                }
            }

            // bottom left
            for (int y = (FXHeight / 2); y < FXHeight; y++)
            {
                for (int x = (FXWidth / 2) - 1; x >= 0; x--)
                {
                    var srcPoint = mappingTable.SrcPoints[(y * FXWidth) + x];
                    var srcPixel = currentBitmap.GetPixel(srcPoint.X, srcPoint.Y);
                    currentBitmap.SetPixel(x + ZoomDeltaX, y + ZoomDeltaY, srcPixel);
                }
            }

            // bottom right
            for (int y = (FXHeight / 2); y < FXHeight; y++)
            {
                for (int x = (FXWidth / 2); x < FXWidth; x++)
                {
                    var srcPoint = mappingTable.SrcPoints[(y * FXWidth) + x];
                    var srcPixel = currentBitmap.GetPixel(srcPoint.X, srcPoint.Y);
                    currentBitmap.SetPixel(x + ZoomDeltaX, y + ZoomDeltaY, srcPixel);
                }
            }*/
        }
        
        public void WriteZoomFirstCPCImage(string dataFolder, string outFolder)
        {
            var bitmap = new Bitmap(System.Drawing.Image.FromFile(outFolder + "\\Simulation_WithMappingTable\\Simu_00.png"));
            //var bitmap = new Bitmap(System.Drawing.Image.FromFile(dataFolder + "\\TestBackground.png"));

            var doubledBitmap = new Bitmap(bitmap.Width * 2, bitmap.Height);
            for (int y = 0; y < bitmap.Height; y++)
            {
                for (int x = 0; x < bitmap.Width; x++)
                {
                    var pixel = bitmap.GetPixel(x, y);
                    doubledBitmap.SetPixel(x * 2, y, pixel);
                    doubledBitmap.SetPixel(x * 2 + 1, y, pixel);
                }
            }

            var cpcImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = doubledBitmap,
                Mode = 0,
                R1 = 48,
                R4 = 21,
                R9 = 0,
                StartR12 = 0x00
            });

            int iPage = 0;
            int iY = 0;
            var cpcBytes = new byte[0x800];
            foreach (var cpcLine in CPCLines_Screen0)
            {
                if(iPage!=cpcLine.Page)
                {
                    File.WriteAllBytes(outFolder + "_Background.page" + iPage + ".vlc0", cpcBytes);
                    iPage++;
                    cpcBytes = new byte[0x800];
                }

                var offsetInPage = (ushort)(cpcLine.CPCScreenPtr&0x0FFF);

                for (int iX = 0; iX < cpcImage.CPCRawBuffer_Width; iX++)
                {
                    var b = cpcImage.CPCRawBuffer[(iY * cpcImage.CPCRawBuffer_Width)+iX];
                    cpcBytes[offsetInPage + iX] = b;
                }

                iY++;
            }
            File.WriteAllBytes(outFolder + "_Background.page" + iPage + ".vlc0", cpcBytes);

            File.WriteAllText(
                outFolder + "_Background.pal.z80",
                cpcImage.Palette.ToZ80TableString(Program.FilenameToKeyWordName("Background"))
            );
        }

        private Bitmap CreateFakeBitmapWithText(int width, int height)
        {
            var bitmap = new Bitmap(width, height);
            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    bitmap.SetPixel(x, y, Color.White);
                }
            }

            using (Graphics g = Graphics.FromImage(bitmap))
            {
                using (Font arialFont = new Font("Arial", 10))
                {
                    g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.SingleBitPerPixelGridFit;
                    g.DrawString("abcdefghijklmnopqrstuvwxyz", arialFont, Brushes.Red, new Point(-2, -7));
                    g.DrawString("hello billy jack", arialFont, Brushes.Blue, new Point(-2, 2));
                    g.DrawString("world ernest gontron", arialFont, Brushes.Red, new Point(-2, 9));
                    g.DrawString("poete citron cactus", arialFont, Brushes.Blue, new Point(-2, 18));
                    g.DrawString("arial font draw string", arialFont, Brushes.Red, new Point(-2, 26));
                    g.DrawString("le petit prince fait pipi", arialFont, Brushes.Blue, new Point(-2, 34));
                    g.DrawString("abcdefghijklmnopqdsfa", arialFont, Brushes.Red, new Point(-2, 42));
                    g.DrawString("anikado-agadou-dou-dou", arialFont, Brushes.Blue, new Point(-2, 50));
                    g.DrawString("la zoubida sur sa moto rouge", arialFont, Brushes.Red, new Point(-2, 58));
                }
            }

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    if ((x < ZoomDeltaX) || (x >= ScreenWidth - ZoomDeltaX))
                    {
                        bitmap.SetPixel(x, y, Color.White);
                    }
                    if ((y < ZoomDeltaY) || (y >= ScreenHeight - ZoomDeltaY))
                    {
                        bitmap.SetPixel(x, y, Color.White);
                    }
                }
            }

            return bitmap;
        }

        private Bitmap CreateDataBitmap(int width, int height)
        {
            var bitmap = new Bitmap(width, height);
            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    int red = x;
                    int green = y;
                    int blue = 255;
                    int alpha = 255;

                    bitmap.SetPixel(x, y, Color.FromArgb(alpha, red, green, blue));
                }
            }

            return bitmap;
        }

        private void CreateZoomSimulation(string outFolder)
        {
            var currentBitmap = CreateFakeBitmapWithText(ScreenWidth, ScreenHeight);

            for (int iFrame = 0; iFrame < NbSimulationFrames; iFrame++)
            {
                currentBitmap.Save(outFolder + "Simulation\\Simu_" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var reducedBitmap = CreateReducedBitmap(currentBitmap);

                currentBitmap = new Bitmap(ScreenWidth, ScreenHeight);
                for (int y = 0; y < ScreenHeight; y++)
                {
                    for (int x = 0; x < ScreenWidth; x++)
                    {
                        currentBitmap.SetPixel(x, y, Color.White);
                    }
                }
                for (int y = 0; y < FXHeight; y++)
                {
                    for (int x = 0; x < FXWidth; x++)
                    {
                        var pixel = reducedBitmap.GetPixel(x, y);
                        currentBitmap.SetPixel(x + ZoomDeltaX, y + ZoomDeltaY, pixel);
                    }
                }
            }
        }

        private MappingTable CreateMappingTable(string outFolder)
        {
            MappingTable mappingTable = null;
            var mappingTableFilename = outFolder + "_MappingTable.xml";

            if (File.Exists(mappingTableFilename))
            {
                XmlSerializer serializer = new XmlSerializer(typeof(MappingTable));
                using (FileStream fileStream = new FileStream(mappingTableFilename, FileMode.Open))
                {
                    mappingTable = (MappingTable)serializer.Deserialize(fileStream);
                }

                if ((mappingTable.RotateDegrees == RotateDegrees)
                    && (mappingTable.ScreenWidth == ScreenWidth)
                    && (mappingTable.ScreenHeight == ScreenHeight)
                    && (mappingTable.FXWidth == FXWidth)
                    && (mappingTable.FXHeight == FXHeight)
                    )
                {
                    return mappingTable;
                }
            }

            var screenBitmap = CreateDataBitmap(ScreenWidth, ScreenHeight);
            var reducedBitmap = CreateReducedBitmap(screenBitmap);

            var srcPoints = new List<SrcPoint>();

            for (int y = 0; y < FXHeight; y++)
            {
                Console.WriteLine("[" + y + "/" + (FXHeight-1) + "]");

                for (int x = 0; x < FXWidth; x++)
                {
                    var reducedPixel = reducedBitmap.GetPixel(x, y);

                    var found = false;

                    for (int iSrcY = 0; iSrcY < ScreenHeight; iSrcY++)
                    {
                        for (int iSrcX = 0; iSrcX < ScreenWidth; iSrcX++)
                        {
                            var srcPixel = screenBitmap.GetPixel(iSrcX, iSrcY);

                            if ((srcPixel.R == reducedPixel.R) && (srcPixel.G == reducedPixel.G))
                            {
                                var srcPoint = new SrcPoint { X = iSrcX, Y = iSrcY };
                                srcPoints.Add(srcPoint);
                                found = true;
                                break;
                            }
                        }

                        if (found)
                        {
                            break;
                        }
                    }

                    if (!found)
                    {
                        System.Diagnostics.Debugger.Break();
                    }
                }
            }

            if (srcPoints.Count != (FXWidth * FXHeight))
            {
                System.Diagnostics.Debugger.Break();
            }

            mappingTable = new MappingTable
            {
                ScreenWidth = ScreenWidth,
                ScreenHeight = ScreenHeight,
                FXWidth = FXWidth,
                FXHeight = FXHeight,
                RotateDegrees = RotateDegrees,
                SrcPoints = srcPoints
            };

            FileStream fs = new FileStream(mappingTableFilename, FileMode.OpenOrCreate);
            System.Xml.Serialization.XmlSerializer s = new System.Xml.Serialization.XmlSerializer(typeof(MappingTable));
            s.Serialize(fs, mappingTable);

            return mappingTable;
        }
        
        private void CreateZoomSimulationWithMappingTable(string outFolder, ref MappingTable mappingTable)
        {
            var currentBitmap = CreateFakeBitmapWithText(ScreenWidth, ScreenHeight);

            for (int iFrame = 0; iFrame < NbSimulationFrames; iFrame++)
            {
                currentBitmap.Save(outFolder + "Simulation_WithMappingTable\\Simu_" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                var reducedBitmap = CreateReducedBitmapWithMappingTable(currentBitmap, ref mappingTable);

                currentBitmap = new Bitmap(ScreenWidth, ScreenHeight);
                for (int y = 0; y < ScreenHeight; y++)
                {
                    for (int x = 0; x < ScreenWidth; x++)
                    {
                        currentBitmap.SetPixel(x, y, Color.White);
                    }
                }
                for (int y = 0; y < FXHeight; y++)
                {
                    for (int x = 0; x < FXWidth; x++)
                    {
                        var pixel = reducedBitmap.GetPixel(x, y);
                        currentBitmap.SetPixel(x + ZoomDeltaX, y + ZoomDeltaY, pixel);
                    }
                }
            }
        }

        private void CreateZoomSimulationWithMappingTable_UpdateSameBitmap(string outFolder, ref MappingTable mappingTable)
        {
            var currentBitmap = CreateFakeBitmapWithText(ScreenWidth, ScreenHeight);

            for (int iFrame = 0; iFrame < NbSimulationFrames; iFrame++)
            {
                currentBitmap.Save(outFolder + "Simulation_WithMappingTable_UpdateSameBitmap\\Simu_" + iFrame.ToString("00") + ".png", ImageFormat.Png);

                CreateReducedBitmapWithMappingTable_UpdateSameBitmap(ref currentBitmap, ref mappingTable);
            }
        }
    }
}
