﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;

namespace phX.Parts
{
    public partial class _19_Zoom
    {
        private ushort GetCPCAddress(int page, int x, int y)
        {
            var lines = CPCLines_Screen0;
            if (page < 2)
            {
                lines = CPCLines_Screen0;
            }

            var line = lines[y];
            if ((y < 0) || (y >= lines.Count))
            {
                System.Diagnostics.Debugger.Break();
            }

            ushort cpcScreenPtr = (ushort)(line.CPCScreenPtr + x);
            return cpcScreenPtr;
        }

        private CPCLine CreateCPCLine(ushort cpcScreenPtr, byte lineIndex)
        {
            byte iPage = 0;
            if(cpcScreenPtr>=0xC000)
            {
                iPage = 3;
            }
            else if (cpcScreenPtr >= 0x8000)
            {
                iPage = 2;
            }
            else if (cpcScreenPtr >= 0x4000)
            {
                iPage = 1;
            }

            ushort screenPtrInWords = (ushort)(cpcScreenPtr >> 1);
            byte R12 = (byte)((screenPtrInWords >> 8) & 3);
            R12 = (byte)(R12 + ((byte)(iPage << 4)));
            byte R13 = (byte)(screenPtrInWords & 255);

            var cpcLine = new CPCLine
            {
                CPCScreenPtr = cpcScreenPtr,
                Page = iPage,
                R12 = R12,
                R13 = R13,
                LineIndex = lineIndex
            };

            return cpcLine;
        }

        private void CreateR12R13Screens(string outFolder)
        {
            const int width = ScreenWidth + FXLogoWidth;

            CPCLines_Screen0 = new List<CPCLine>();
            
            for (int iPage=0; iPage<4; iPage++)
            {
                ushort CPCScreenPtr = (ushort)(0x4000 * iPage);

                for (int i = 0; i < 4; i++)
                {
                    CPCLines_Screen0.Add(CreateCPCLine((ushort)(CPCScreenPtr + width * 0), (byte)CPCLines_Screen0.Count));
                    CPCLines_Screen0.Add(CreateCPCLine((ushort)(CPCScreenPtr + width * 1), (byte)CPCLines_Screen0.Count));
                    CPCLines_Screen0.Add(CreateCPCLine((ushort)(CPCScreenPtr + width * 2), (byte)CPCLines_Screen0.Count));
                    CPCLines_Screen0.Add(CreateCPCLine((ushort)(CPCScreenPtr + width * 3), (byte)CPCLines_Screen0.Count));
                    CPCLines_Screen0.Add(CreateCPCLine((ushort)(CPCScreenPtr + width * 4), (byte)CPCLines_Screen0.Count));
                    CPCScreenPtr += 0x200;
                }
            }

            while(CPCLines_Screen0.Count>ScreenHeight)
            {
                CPCLines_Screen0.RemoveAt(CPCLines_Screen0.Count-1);
            }
        }

        private void WriteR12R13Screen(string filename, string label, List<CPCLine> lines)
        {
            var z80 = "";
            z80 += ";; R12,R13\n";
            z80 += ";; " + lines.Count + " lines\n";
            z80 += "" + label + ":\n";

            foreach (var line in lines)
            {
                z80 += "\tdb &" + line.R12.ToString("X2") + ", &" + line.R13.ToString("X2") + " ;; ScreenPtr=&" + line.CPCScreenPtr.ToString("X4") + "  LineIndex=" + line.LineIndex + "\n";
            }

            File.WriteAllText(filename, z80);
        }

        private void WriteScreenPtr(string filename, string label, List<CPCLine> lines)
        {
            var z80 = "";
            z80 += ";; ScreenPtr\n";
            z80 += "" + label + ":\n";

            foreach (var line in lines)
            {
                z80 += "\tdw &" + line.CPCScreenPtr.ToString("X4") + "\n";
            }

            File.WriteAllText(filename, z80);
        }
    }
}
