﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;

namespace phX.Parts
{
    public partial class _19_Zoom
    {
        public Walk ComputeWalkZoom(string outFilename, string z80Label, ref MappingTable mappingTable)
        {
            var w = new Walk
            {
                StartX = FXWidth - 1,
                StartY = FXHeight - 1,
                IncX = -1,
                IncY = -1
            };

            CreateWalkSteps(w, ref mappingTable);

            var z80 = AddR12R13IntoWalkSteps(z80Label, w);
            File.WriteAllText(outFilename, z80);

            return w;
        }

        public byte GetHigher(ushort value)
        {
            return (byte)(value >> 8);
        }

        public byte GetLower(ushort value)
        {
            return (byte)(value & 255);
        }

        public class Walk
        {
            public int StartX;
            public int StartY;
            public int IncX;
            public int IncY;

            public ushort StartHL;
            public ushort StartDE;

            public Listing Listing = new Listing();
        }

        enum R12R13Selection
        {
            R12,
            R13
        };

        public string AddR12R13IntoWalkSteps(string z80Label, Walk walk)
        {
            var listing = walk.Listing;

            int iOpcodeIndex = 0;
            var totalNops = 0;

            var cpcLines = CPCLines_Screen0;
            
            byte R12 = (byte)0xFF;
            byte R13 = (byte)0xFF;
            R12R13Selection r12r13Selection;

            var z80 = "";
            z80 += "\tld bc, &BC00+12 ;3\n";
            z80 += "\tout (c), c ;4\n";
            r12r13Selection = R12R13Selection.R12;
            totalNops += 7;

            z80 += "\tld hl, &" + walk.StartHL.ToString("X") + " ;3\n";
            totalNops += 3;
            z80 += "\tld de, &" + walk.StartDE.ToString("X") + " ;3\n";
            totalNops += 3;

            z80 += "\n";

            const int ChangeR12R13_NopsDelay = 4 * 64; // 4 VLCs per block

            bool enableBRK = false;

            int nextNopsDelayTarget = (64*0+80)+(ChangeR12R13_NopsDelay * 5);// ChangeR12R13_NopsDelay *4; // discard 5 * 6 VLCs (so most processing is done before displaying lines)

            bool isFirst = true;
            for (int iCpcLine=0; iCpcLine < cpcLines.Count; iCpcLine++)
            {
                Console.WriteLine("" + iCpcLine + "/" + (cpcLines.Count - 1));

                var cpcLine = cpcLines[iCpcLine];

                var requiredNops = 0;
                var z80_r12r13 = "";
                z80_r12r13 += "\n";
                z80_r12r13 += "\t;; LINE " + iCpcLine + "/" + (cpcLines.Count-1) +
                    ((r12r13Selection == R12R13Selection.R12) ? " (R12 selected)":" (R13 selected)" )+
                    " R12=&" +cpcLine.R12.ToString("X2") + 
                    " R13=&" + cpcLine.R13.ToString("X2") + 
                    " ScreenPtr=&" + cpcLine.CPCScreenPtr.ToString("X4") + "\n";

                if(iCpcLine==cpcLines.Count-1)
                {
                    if (enableBRK)
                    {
                        z80_r12r13 += "\t_BRK ;2\n";
                    }
                    else
                    {
                        z80_r12r13 += "\t_NOBRK ;2\n";
                    }
                    requiredNops += 2;
                }

                if ((R12 != cpcLine.R12)&&(R13 != cpcLine.R13))
                {
                    R12 = cpcLine.R12;
                    R13 = cpcLine.R13;

                    if (r12r13Selection==R12R13Selection.R13)
                    {
                        z80_r12r13 += "\tld bc, &BD00+&" + R13.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;

                        z80_r12r13 += "\tld bc, &BC00+12 ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        r12r13Selection = R12R13Selection.R12;
                        requiredNops += 7;

                        z80_r12r13 += "\tld bc, &BD00+&" + R12.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;
                    }
                    else
                    {
                        z80_r12r13 += "\tld bc, &BD00+&" + R12.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;

                        z80_r12r13 += "\tld bc, &BC00+13 ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        r12r13Selection = R12R13Selection.R13;
                        requiredNops += 7;

                        z80_r12r13 += "\tld bc, &BD00+&" + R13.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;

                        if(isFirst)
                        {
                            z80_r12r13 += "\tld bc, &BC00+6 ;3\n";
                            z80_r12r13 += "\tout (c), c ;4\n";
                            requiredNops += 7;

                            if (enableBRK)
                            {
                                z80_r12r13 += "\t_BRK ;2\n";
                            }
                            else
                            {
                                z80_r12r13 += "\t_NOBRK ;2\n";
                            }
                            requiredNops += 2;

                            z80_r12r13 += "\tinc b ;1\n";
                            z80_r12r13 += "\tout (c), c ;4\n";
                            requiredNops += 5;

                            z80_r12r13 += "\tld bc, &BC00+13 ;3\n";
                            z80_r12r13 += "\tout (c), c ;4\n";
                            r12r13Selection = R12R13Selection.R13;
                            requiredNops += 7;

                            isFirst = false;
                        }
                    }
                }
                else if ((R12 == cpcLine.R12) && (R13 != cpcLine.R13))
                {
                    R13 = cpcLine.R13;

                    if (r12r13Selection == R12R13Selection.R13)
                    {
                        z80_r12r13 += "\tld bc, &BD00+&" + R13.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;
                    }
                    else
                    {
                        z80_r12r13 += "\tld bc, &BC00+13 ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        r12r13Selection = R12R13Selection.R13;
                        requiredNops += 7;

                        z80_r12r13 += "\tld bc, &BD00+&" + R13.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;
                    }

                }
                else if ((R12 != cpcLine.R12) && (R13 == cpcLine.R13))
                {
                    R12 = cpcLine.R12;

                    if (r12r13Selection == R12R13Selection.R12)
                    {
                        z80_r12r13 += "\tld bc, &BD00+&" + R12.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;
                    }
                    else
                    {
                        z80_r12r13 += "\tld bc, &BC00+12 ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        r12r13Selection = R12R13Selection.R12;
                        requiredNops += 7;

                        z80_r12r13 += "\tld bc, &BD00+&" + R12.ToString("X2") + " ;3\n";
                        z80_r12r13 += "\tout (c), c ;4\n";
                        requiredNops += 7;
                    }
                }

                Opcode opcode;
                do
                {
                    opcode = listing.Opcodes[iOpcodeIndex];
                    iOpcodeIndex++;

                    z80 += "\t" + opcode.GetZ80() + "\n";
                    totalNops += opcode.Nops;
                } while ((totalNops + requiredNops) < (nextNopsDelayTarget - opcode.Nops));

                nextNopsDelayTarget += ChangeR12R13_NopsDelay;

                z80 += z80_r12r13;
                totalNops += requiredNops;
            }

            // flush remaining
            while (iOpcodeIndex < listing.Opcodes.Count)
            {
                var opcode = listing.Opcodes[iOpcodeIndex];
                iOpcodeIndex++;

                z80 += "\t" + opcode.GetZ80() + "\n";
                totalNops += opcode.Nops;
            }

            z80 += "\n";
            z80 += ";\n";
            z80 += "\tJP WalkStep_End\n";
            totalNops += 2;

            z80 = ";; " +
                "Nops=" + totalNops + "\n"
                + z80Label + ":\n"
                + z80
                + "\n";

            return z80;
        }

        private void SpecialCase_Replace_LD_DE_A_by_PUSH_BC(ref int lastSrcVideoPtrRepeatCount, ref Walk w)
        {
            // special case to replace LD (DE),A:DEC E by PUSH BC
            if (lastSrcVideoPtrRepeatCount != 0)
            {
                var toRemoveCount = lastSrcVideoPtrRepeatCount;

                // we want only an even number since we use stack
                bool also_remove_LD_A_HL = true;
                if ((lastSrcVideoPtrRepeatCount & 1) == 1)
                {
                    toRemoveCount--;
                    also_remove_LD_A_HL = false;
                }

                // remove previous LD (DE),A:DEC E
                for (int i = 0; i < toRemoveCount * 2; i++)
                {
                    w.Listing.Opcodes.RemoveAt(w.Listing.Opcodes.Count - 1);
                }
                if(also_remove_LD_A_HL)
                {
                    w.Listing.Opcodes.RemoveAt(w.Listing.Opcodes.Count - 1);
                }

                // add new version using stack
                w.Listing.AddOpcode(OpcodeId.LD_B_HL);
                w.Listing.AddOpcode(OpcodeId.LD_C_B);
                w.Listing.AddOpcode(OpcodeId.EX_DE_HL);
                w.Listing.AddOpcode(OpcodeId.LD_SP_HL);
                w.Listing.AddOpcode(OpcodeId.EX_DE_HL);
                for (int i = 0; i < toRemoveCount / 2; i++) // /2 because PUSH BC treat 2 bytes at same time
                {
                    w.Listing.AddOpcode(OpcodeId.PUSH_BC);
                }
            }
        }

        public void CreateWalkSteps(Walk w, ref MappingTable mappingTable)
        {
            ushort HL = 0;
            ushort DE = 0;
            ushort lastSrcVideoPtr = (ushort)0xFFFF;
            int lastSrcVideoPtrRepeatCount = 0;

            int dstY = w.StartY;
            for (int iY = 0; iY < FXHeight; iY++)
            {
                int dstX = w.StartX;
                for (int iX = 0; iX < FXWidth; iX++)
                {
                    var srcPoint = mappingTable.SrcPoints[(dstY * FXWidth) + dstX];
                    int srcX = srcPoint.X;
                    int srcY = srcPoint.Y;

                    ushort srcVideoPtr = GetCPCAddress(0, srcX, srcY);
                    ushort dstVideoPtr = GetCPCAddress(0, dstX, dstY);

                    var comment = 
                        "SrcX=" + srcX + 
                        ";SrcY=" + srcY + 
                        ";DstX=" + dstX + 
                        ";DstY=" + dstY + 
                        ";SrcVideoPtr=&" + srcVideoPtr.ToString("X4") + 
                        ";DstVideoPtr=&" + dstVideoPtr.ToString("X4")
                        ;

                    if(lastSrcVideoPtr==srcVideoPtr)
                    {
                        if(lastSrcVideoPtrRepeatCount==0)
                        {
                            w.Listing.AddOpcode(OpcodeId.LD_A_HL);
                        }
                        
                        w.Listing.AddOpcode(OpcodeId.LD_DE_A, comment + ";RepeatCount=" + lastSrcVideoPtrRepeatCount);

                        // DE--
                        if (GetLower(DE) == 0x00)
                        {
                            w.Listing.AddOpcode(OpcodeId.DEC_DE);
                            DE--;
                        }
                        else
                        {
                            w.Listing.AddOpcode(OpcodeId.DEC_E);
                            DE--;
                        }

                        dstX += w.IncX;

                        lastSrcVideoPtrRepeatCount++;
                        continue;
                    }

                    // special case to replace LD (DE),A:DEC E by PUSH BC
                    SpecialCase_Replace_LD_DE_A_by_PUSH_BC(ref lastSrcVideoPtrRepeatCount, ref w);

                    lastSrcVideoPtrRepeatCount = 0;

                    // treat HL
                    if ((iX == 0) && (iY == 0))
                    {
                        w.StartHL = srcVideoPtr;
                        HL = w.StartHL;
                    }
                    else
                    {
                        if (HL == srcVideoPtr)
                        {
                            // do nothing, yepee
                        }
                        else if (srcVideoPtr == (HL - 1))
                        {
                            // HL--
                            if (GetLower(HL) == 0x00)
                            {
                                w.Listing.AddOpcode(OpcodeId.DEC_HL);
                                HL--;
                            }
                            else
                            {
                                w.Listing.AddOpcode(OpcodeId.DEC_L);
                                HL--;
                            }
                        }
                        else if (srcVideoPtr == (HL + 1))
                        {
                            // HL++
                            if (GetLower(HL) == 0xFF)
                            {
                                w.Listing.AddOpcode(OpcodeId.INC_HL);
                                HL++;
                            }
                            else
                            {
                                w.Listing.AddOpcode(OpcodeId.INC_L);
                                HL++;
                            }
                        }
                        else
                        {
                            // HL=srcVideoPtr
                            if ((GetHigher(HL) != GetHigher(srcVideoPtr)) && (GetLower(HL) != GetLower(srcVideoPtr)))
                            {
                                // H and L differs
                                /*ushort diff = (ushort)(srcVideoPtr - HL);
                                w.Listing.AddOpcode(OpcodeId.LD_BC, diff);
                                w.Listing.AddOpcode(OpcodeId.ADD_HL_BC);*/
                                w.Listing.AddOpcode(OpcodeId.LD_HL, srcVideoPtr);
                                HL = srcVideoPtr;
                            }
                            else if ((GetHigher(HL) != GetHigher(srcVideoPtr)) && (GetLower(HL) == GetLower(srcVideoPtr)))
                            {
                                // only H differs
                                byte diff = (byte)(GetHigher(srcVideoPtr) - GetHigher(HL));
                                if (diff == (byte)(0xFF))
                                {
                                    w.Listing.AddOpcode(OpcodeId.DEC_H);
                                }
                                else
                                {
                                    /*w.Listing.AddOpcode(OpcodeId.LD_A, diff);
                                    w.Listing.AddOpcode(OpcodeId.ADD_A_H);
                                    w.Listing.AddOpcode(OpcodeId.LD_H_A);*/
                                    byte H = (byte)GetHigher(srcVideoPtr);
                                    w.Listing.AddOpcode(OpcodeId.LD_H, H);
                                }
                                HL = srcVideoPtr;
                            }
                            else if ((GetHigher(HL) == GetHigher(srcVideoPtr)) && (GetLower(HL) != GetLower(srcVideoPtr)))
                            {
                                // only L differs
                                byte L = (byte)GetLower(srcVideoPtr);
                                w.Listing.AddOpcode(OpcodeId.LD_L, L);
                                HL = srcVideoPtr;
                            }
                        }
                    }

                    // treat DE (for new line..)
                    if ((iX == 0) && (iY == 0))
                    {
                        w.StartDE = dstVideoPtr;
                        DE = w.StartDE;
                    }
                    else
                    {
                        if (DE == dstVideoPtr)
                        {
                            // do nothing, yepee
                        }
                        else
                        {
                            // special case to replace LD HL,&A:LDD by LD A,(HL):LD (DE), A
                            var index = w.Listing.Opcodes.Count;
                            if (   (w.Listing.Opcodes[index-1].Id == KnownOpcode.GetKnownOpcode(OpcodeId.LD_HL).Id) 
                                && (w.Listing.Opcodes[index-2].Id == KnownOpcode.GetKnownOpcode(OpcodeId.LDD).Id)
                                && (w.Listing.Opcodes[index-3].Id == KnownOpcode.GetKnownOpcode(OpcodeId.LD_HL).Id)
                                && (w.Listing.Opcodes[index-3].Value == 0x000A)
                                )
                            {
                                var prev_ptr = w.Listing.Opcodes[index - 3].Value;
                                var prev_ld_hl_opcode = w.Listing.Opcodes[index - 1];
                                w.Listing.Opcodes.RemoveAt(w.Listing.Opcodes.Count - 1); // remove LD_HL
                                w.Listing.Opcodes.RemoveAt(w.Listing.Opcodes.Count - 1); // remove LDD
                                w.Listing.Opcodes.RemoveAt(w.Listing.Opcodes.Count - 1); // remove LD_HL

                                w.Listing.AddOpcode(OpcodeId.LD_A_PTR, prev_ptr);
                                w.Listing.AddOpcode(OpcodeId.LD_DE_A);
                                w.Listing.AddOpcode(OpcodeId.LD_HL, prev_ld_hl_opcode.Value);
                            }

                            if ((GetHigher(DE) != GetHigher(dstVideoPtr)) && (GetLower(DE) != GetLower(dstVideoPtr)))
                            {
                                // D and E differs
                                /* ushort diff = (ushort)(dstVideoPtr - DE);
                                w.Listing.AddOpcode(OpcodeId.EX_DE_HL);
                                w.Listing.AddOpcode(OpcodeId.LD_BC, diff);
                                w.Listing.AddOpcode(OpcodeId.ADD_HL_BC);
                                w.Listing.AddOpcode(OpcodeId.EX_DE_HL);*/
                                w.Listing.AddOpcode(OpcodeId.LD_DE, dstVideoPtr);
                                DE = dstVideoPtr;
                            }
                            else if ((GetHigher(DE) == GetHigher(dstVideoPtr)) && (GetLower(DE) != GetLower(dstVideoPtr)))
                            {   
                                // only E differs
                                byte E = (byte)GetLower(dstVideoPtr);
                                w.Listing.AddOpcode(OpcodeId.LD_E, E);
                                DE = dstVideoPtr;
                            }
                        }
                    }

                    if (w.IncX == 1)
                    {
                        w.Listing.AddOpcode(OpcodeId.LDI, comment);
                        HL++;
                        DE++;
                    }
                    else
                    {
                        w.Listing.AddOpcode(OpcodeId.LDD, comment);
                        HL--;
                        DE--;
                    }

                    dstX += w.IncX;

                    lastSrcVideoPtr = srcVideoPtr;
                }

                dstY += w.IncY;
            }

            // special case to replace LD (DE),A:DEC E by PUSH BC
            SpecialCase_Replace_LD_DE_A_by_PUSH_BC(ref lastSrcVideoPtrRepeatCount, ref w);
        }
    }
}
