﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;

namespace phX.Parts
{
    public class _20_FlipScroll
    {
        const int ZoomAmplitude = 6;//15; // can reduce of -4/+4, cf. 8 different levels
        const double ZoomHorizCoef = 2.9;
        const int ScreenWidth = 96; // 96 bytes width
        const int ReferenceWidth = ScreenWidth + ZoomAmplitude * 2; // *2=left and right edges
        const int FontHeight = 20;
        const int NbFrames = 1;
        const int CharWidth = 16;
        const int CharHeight = 20;
        const int NbAnimFrames = 90;

        public _20_FlipScroll()
        {
            try
            {
                CreateR12R13BackgroundPointers(Program.ToZ80Folder("20_FlipScroll/"));

                CreateBackgroundAnim(Program.ToDataFolder("20_FlipScroll/Background_Anim0.png"),
                            Program.ToZ80Folder("20_FlipScroll/"));

                CompressFlipScrollBinaryD000(Program.ToZ80Folder("20_FlipScroll/"));

                CreateRasterSin(Program.ToZ80Folder("20_FlipScroll/"));
                //CreateBarSin(Program.ToZ80Folder("20_FlipScroll/"));
                CreateRasterTable(
                    Program.ToDataFolder("20_FlipScroll/Razter2.png"),
                    Program.ToZ80Folder("20_FlipScroll/_BackgroundRasters0.z80"),
                    "BackgroundRasters0");
                /*CreateRasterTable(
                    Program.ToDataFolder("20_FlipScroll/Bar0.png"),
                    Program.ToZ80Folder("20_FlipScroll/_Bar0.z80"),
                    "Bar0");*/
                //CreateRasterExplodeAnim(Program.ToZ80Folder("20_FlipScroll/"));

                CreateAnimFrames(Program.ToZ80Folder("20_FlipScroll/"));

                CreateWriteZoomLevelCall(Program.ToZ80Folder("20_FlipScroll/"), false);
                CreateWriteZoomLevelCall(Program.ToZ80Folder("20_FlipScroll/"), true);
                CreateDrawScreens(Program.ToZ80Folder("20_FlipScroll/"), false);
                CreateDrawScreens(Program.ToZ80Folder("20_FlipScroll/"), true);

                var fontCharset = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789()!?:.,-x'y "; //x=smiley y=heart
                //var scrollText = "THIS IS NOT THE USUAL ZOOM SCROLLER xxxxx      yyy HAPPY PEOPLE yyy HAPPY DEMO yyy     xxx NEVER JUDGE A SCROLLER BY ITS FONT xxx     ";//  OKAY SO BASICALLY THIS IS A SCROLLER AND I HAVE NOTHING SPECIAL TO SAY... HOW DUMB IS THAT ?! xxx     ";
                var scrollText = "NEVER JUDGE A ySCROLLERy BY ITS FONT xxx         ";//  OKAY SO BASICALLY THIS IS A SCROLLER AND I HAVE NOTHING SPECIAL TO SAY... HOW DUMB IS THAT ?! xxx     ";
                int fontIndexInBitmap = 0;
                for (int paletteIndexInBitmap = 0; paletteIndexInBitmap < 18; paletteIndexInBitmap++)
                //for (int paletteIndexInBitmap = 0; paletteIndexInBitmap < 1; paletteIndexInBitmap++)
                {
                    CreateScrollFont(
                        Program.ToDataFolder("20_FlipScroll/Font_21.png"),
                        Program.ToZ80Folder("20_FlipScroll/"),
                        fontIndexInBitmap,
                        paletteIndexInBitmap,
                        fontCharset,
                        scrollText);
                }

                CreateCharSliceOffset(Program.ToZ80Folder("20_FlipScroll/"));
                CreateScrollDataPtr(Program.ToZ80Folder("20_FlipScroll/"));
                CreateR12R13(Program.ToZ80Folder("20_FlipScroll/"));
                CreateZoomLevels(Program.ToZ80Folder("20_FlipScroll/"));
            }
            catch (Exception e)
            {
                // try deleting _MappingTable.xml
                Log.Fatal(e.ToString());
            }
        }

        public void CreateR12R13BackgroundPointers(string outFolder)
        {
            var z80 = "";
            z80 += ";\n";

            ushort startPtr = (ushort)0x1800;

            for(int iVCC=0; iVCC<6; iVCC++)
            {
                z80 += "\tR12R13_Table_R12_VCC" + iVCC + " equ &" + startPtr.ToString("X4") + "\n";
                startPtr = (ushort)(startPtr + 256);
                z80 += "\tR12R13_Table_R13_VCC" + iVCC + " equ &" + startPtr.ToString("X4") + "\n";
                startPtr = (ushort)(startPtr + 256);
                z80 += "\tR12R13_Table_R12_VCC" + iVCC + "_BASE equ &" + startPtr.ToString("X4") + "\n";
                startPtr = (ushort)(startPtr + 256);
                z80 += "\tR12R13_Table_R13_VCC" + iVCC + "_BASE equ &" + startPtr.ToString("X4") + "\n";
                startPtr = (ushort)(startPtr + 256);

                z80 += "\n";
            }

            File.WriteAllText(outFolder + "_R12R13BackgroundPointers.z80", z80);
        }

        public Bitmap CreateBitmapStep(string bitmapFilename, int frame)
        {
            var bitmap = new Bitmap(bitmapFilename);

            Bitmap refBitmap = (Bitmap)bitmap.Clone();

            int yDiff = frame;

            for(int x=0; x<bitmap.Width; x++)
            {
                for (int y = 0; y < 6; y++)
                {
                    int iY = y + yDiff;
                    if (x < 130)
                    {
                        iY = y - yDiff;
                    }
                    else if (x < 160)
                    {
                        iY = y + yDiff;
                    }
                    else if (x < 300)
                    {
                        iY = y - yDiff;
                    }
                    else if (x < 440)
                    {
                        iY = y + yDiff;
                    }
                    else if (x < 560)
                    {
                        iY = y - yDiff;
                    }
                    else if (x < 630)
                    {
                        iY = y + yDiff;
                    }
                    else
                    {
                        iY = y - yDiff;
                    }

                    while (iY<0)
                    {
                        iY += 6;
                    }
                    while (iY>=6)
                    {
                        iY -= 6;
                    }

                    var pixel = refBitmap.GetPixel(x, iY);
                    bitmap.SetPixel(x, y, pixel);
                }
            }

            return bitmap;
        }

        public void CreateBackgroundAnim(string bitmapFilename, string outFolder)
        {
            var bitmaps = new List<Bitmap>();
            for(int i=0; i<6; i++)
            {
                var bitmap = CreateBitmapStep(bitmapFilename, i);
                bitmap.Save(outFolder + "_BackgroundBitmap" + i + ".png", System.Drawing.Imaging.ImageFormat.Png);
                bitmaps.Add(bitmap);
            }

            var images = new List<Framework.Image>();
            foreach(var bitmap in bitmaps)
            {
                var palette = new Palette();
                palette.AddColor(Color.Black);
                var image = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = bitmap,
                    Mode = 2,
                    R1 = 48,
                    R4 = 0,
                    R9 = 0,
                    Palette = palette
                });

                images.Add(image);
            }
            
            var page0Bytes = new List<byte>();
            page0Bytes.AddRange(images[0].CPCRawBuffer.Take(96 * 6).ToArray());
            page0Bytes.AddRange(images[1].CPCRawBuffer.Take(96 * 6).ToArray());
            page0Bytes.AddRange(images[2].CPCRawBuffer.Take(96 * 6).ToArray());

            var page1Bytes = new List<byte>();
            page1Bytes.AddRange(images[3].CPCRawBuffer.Take(96 * 6).ToArray());
            page1Bytes.AddRange(images[4].CPCRawBuffer.Take(96 * 6).ToArray());
            page1Bytes.AddRange(images[5].CPCRawBuffer.Take(96 * 6).ToArray());

            // empty line
            for(int i=0; i<96; i++)
            {
                page1Bytes.Add((byte)0);
            }

            var page0RawBytes = page0Bytes.ToArray();
            var BBPacked_page0RawBytes = BBPacker.Pack(ref page0RawBytes);
            File.WriteAllBytes(outFolder + "_BackgroundPage0.bin", page0RawBytes);
            File.WriteAllBytes(outFolder + "_BackgroundPage0.bin.bbPacked", BBPacked_page0RawBytes);

            var page1RawBytes = page1Bytes.ToArray();
            var BBPacked_page1RawBytes = BBPacker.Pack(ref page1RawBytes);
            File.WriteAllBytes(outFolder + "_BackgroundPage1.bin", page1RawBytes);
            File.WriteAllBytes(outFolder + "_BackgroundPage1.bin.bbPacked", BBPacked_page1RawBytes);

            var z80 = "";
            for (int iPage = 0; iPage < 2; iPage++)
            {
                int CRTCOffset = 0;

                var maxVCC = 3;
                if(iPage==1)
                {
                    maxVCC = 4; // for empty line
                }

                for (int iVCC = 0; iVCC < maxVCC; iVCC++)
                {
                    var label = "Background_VCC";

                    var maxVLC = 6;
                    if((iPage==1)&&(iVCC==(maxVCC-1)))
                    {
                        maxVLC = 1;
                        label = "EmptyLine_VCC";
                    }
                    for (int iVLC = 0; iVLC < maxVLC; iVLC++)
                    {
                        int page = 2 + iPage; // 0x8000
                        byte overscanBits = (byte)(page * 0x10);

                        byte r12 = (byte)(overscanBits + ((CRTCOffset) >> 8));
                        byte r13 = (byte)((CRTCOffset) & 255);

                        z80 += label + (iVCC+iPage*3) + "_VLC" + iVLC + "_R12 equ &" + r12.ToString("X2") + "\n";
                        z80 += label + (iVCC + iPage * 3) + "_VLC" + iVLC + "_R13 equ &" + r13.ToString("X2") + "\n";
                        CRTCOffset += 48;
                    }
                }

                z80 += "\n";
            }

            File.WriteAllText(outFolder + "_BackgroundAnim_Config.z80", z80);
        }

        public void CompressFlipScrollBinaryD000(string outFolder)
        {
            var allBytes = File.ReadAllBytes(outFolder + "COMPILED_D000_FROM_WINAPE.BIN");
            var BBPackedBytes = BBPacker.Pack(ref allBytes);
            File.WriteAllBytes(outFolder + "COMPILED_D000_FROM_CSHARP.BIN.bbPacked", BBPackedBytes);
            
            var packer = new Packer();
            var packedBytes = packer.Convert(ref allBytes);
            File.WriteAllBytes(outFolder + "COMPILED_D000_FROM_CSHARP.BIN.packed", packedBytes);
        }

        public void CreateCharSliceOffset(string outFolder)
        {
            var z80 = "";
            z80 += ";\n";
            z80 += "FlipScroll_SliceOffset:\n";

            var sliceOffset = 0;

            for (int i = 0; i < CharWidth; i++)
            {
                z80 += "\tdb " + (sliceOffset/2) + "\n";

                sliceOffset += CharHeight;
            }

            File.WriteAllText(outFolder + "_CharSliceOffset.z80", z80);
        }

        public class CharacterInfo
        {
            public string ASCII = "";
            public int CharIndex;
            public Framework.Image CPCImage;
        }

        private List<Color> CollectColorsForLineBitmap(Bitmap bitmap, int y)
        {
            var pixels = new List<Color>();
            for (int yInChar = 0; yInChar < CharHeight; yInChar++)
            {
                for (int x = 0; x < bitmap.Width; x++)
                {
                    var pixel = bitmap.GetPixel(x, y+yInChar);
                    if (!pixels.Contains(pixel))
                    {
                        pixels.Add(pixel);
                    }
                }
            }
            return pixels;
        }

        private void CreateScrollFont(string inFontFilename, string outFolder, int fontIndexInBitmap, int paletteIndexInBitmap, string charset, string scrollText)
        {
            var baseBitmap = new Bitmap(inFontFilename);

            var basePalette = CollectColorsForLineBitmap(baseBitmap, 0+2);
            var wishedPalette = CollectColorsForLineBitmap(baseBitmap, (24*paletteIndexInBitmap)+2);

            int nbChars = charset.Length;

            var fontBitmap = new Bitmap(nbChars * CharWidth, CharHeight);
            for (int iChar = 0; iChar < nbChars; iChar++)
            {
                for (int iY = 0; iY < CharHeight; iY++)
                {
                    for (int iX = 0; iX < CharWidth; iX++)
                    {
                        var srcX = (iChar * 17) + 1 + iX;
                        var srcY = (fontIndexInBitmap * 24) + 2 + iY;
                        var dstX = (CharWidth * iChar) + iX;
                        var dstY = iY;

                        var basePixel = baseBitmap.GetPixel(srcX, srcY);

                        var convertedPixelIndex = basePalette.IndexOf(basePixel);
                        var convertedPixel = wishedPalette[convertedPixelIndex];

                        fontBitmap.SetPixel(dstX, dstY, convertedPixel);
                    }
                }
            }
            fontBitmap.Save(outFolder + "_FontBitmap.png", System.Drawing.Imaging.ImageFormat.Png);

            var emptyPalette = new Palette(2); // discard 2 first colors for MODE 2
            var tempImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
            {
                Bitmap = fontBitmap,
                Mode = 0,
                R1 = 800,
                R4 = 38,
                R9 = 7,
                StartVLC = 0,
                FirstPixelInPaletteX = 0,
                FirstPixelInPaletteY = 0,
                Palette = emptyPalette
            });
            var cpcPalette = tempImage.Palette;

            var z80 = cpcPalette.ToZ80TableString(Program.FilenameToKeyWordName("FontScrollPalette"));
            
            // hack
            z80 = z80.Replace("_PAL_FONTSCROLLPALETTE_PEN1 equ COLOR0", "_PAL_FONTSCROLLPALETTE_PEN1 equ COLOR25");

            File.WriteAllText(
                outFolder + "FontPalettes\\_FontPalette_" + paletteIndexInBitmap + ".z80",
                z80);

            if (paletteIndexInBitmap>0)
            {
                return;
            }

            var chars = new List<CharacterInfo>();
            int indexInCharset = -1;
            foreach (var c in charset)
            {
                indexInCharset++;

                if (!scrollText.Contains(c))
                {
                    continue;
                }

                var newChar = new CharacterInfo()
                {
                    ASCII = "" + c,
                    CharIndex = chars.Count
                };

                var charBitmap = new Bitmap(CharWidth * 2, CharHeight); // charWidth*2 because 1 byte=1 pixel, not 2 pixels
                for (int y = 0; y < CharHeight; y++)
                {
                    for (int x = 0; x < CharWidth; x++)
                    {
                        var color = fontBitmap.GetPixel(x + (indexInCharset * CharWidth), y);
                        charBitmap.SetPixel(x * 2, y, color);
                        charBitmap.SetPixel((x * 2) + 1, y, color);
                    }
                }
                //charBitmap.Save(outFolder + "_DebugLetter.png", ImageFormat.Png);

                var charImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                {
                    Bitmap = charBitmap,
                    Mode = 0,
                    R1 = 48,
                    R4 = 38,
                    R9 = 7,
                    StartVLC = 0,
                    FirstPixelInPaletteX = 0,
                    FirstPixelInPaletteY = 0,
                    Palette = cpcPalette
                });
                newChar.CPCImage = charImage;

                chars.Add(newChar);
            }

            z80 = "";

            var z80Bytes = new List<byte>();
            foreach (var c in chars)
            {
                z80 += ";\n";
                z80 += "FLIPSCROLL_CHAR_" + c.CharIndex + ": ;; " + c.ASCII + "";

                for (int iX = 0; iX < c.CPCImage.CPCRawBuffer_Width; iX++)
                {
                    for (int iY = 0; iY < c.CPCImage.CPCRawBuffer_Height; iY++)
                    {
                        byte b = c.CPCImage.CPCRawBuffer[(c.CPCImage.CPCRawBuffer_Width * iY) + iX];
                        if (iY == 0)
                        {
                            z80 += "\n\tdb ";
                        }
                        else
                        {
                            z80 += ",";
                        }

                        z80 += "&" + b.ToString("X2");
                        z80Bytes.Add(b);
                    }
                }
                z80 += "\n";
            }

            File.WriteAllText(outFolder + "_Font.z80", z80);
            var rawBin = z80Bytes.ToArray();
            File.WriteAllBytes(outFolder + "_Font.bin", rawBin);
            var BBPackedBytes = BBPacker.Pack(ref rawBin);
            File.WriteAllBytes(outFolder + "_Font.bin.bbPacked", BBPackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref rawBin);
            File.WriteAllBytes(outFolder + "_Font.bin.packed", packedBytes);

            z80 = ";\n";
            z80 += "FlipScroll_Text:\n";
            foreach (var iC in scrollText)
            {
                char c = iC;

                int charIndex = -1;
                foreach (var cFont in chars)
                {
                    if (cFont.ASCII == "" + c)
                    {
                        charIndex = cFont.CharIndex;
                        break;
                    }
                }
                if (charIndex == -1)
                {
                    System.Diagnostics.Debugger.Break();
                }

                z80 += "\tdw FLIPSCROLL_CHAR_" + charIndex + " ;; " + chars[charIndex].ASCII + "\n";
            }

            File.WriteAllText(outFolder + "_ScrollText.z80", z80);

            z80 = "";
            z80 += "FlipScroll_CharWidth_Bytes equ &" + chars[0].CPCImage.CPCRawBuffer_Width.ToString("X2") + "\n";
            z80 += "FlipScroll_CharHeight_Bytes equ &" + chars[0].CPCImage.CPCRawBuffer_Height.ToString("X2") + "\n";
            z80 += "FlipScroll_ScrollTextLength equ &" + scrollText.Length.ToString("X2") + "\n";
            z80 += "FlipScroll_CharHeightWithEmptyLine equ FlipScroll_CharHeight_Bytes + 1\n";
            z80 += "FlipScroll_NbAnimFrames equ " + NbAnimFrames + "\n";
            z80 += "FlipScroll_ScrollDataRAM equ &8800\n";

            File.WriteAllText(outFolder + "_ScrollText_Config.z80", z80);

            z80 = "";
            z80 += "Last_FlipScroll_AnimFrame_ZoomLevels equ FlipScroll_AnimFrame" + (NbAnimFrames - 1) + "_ZoomLevels\n";
            z80 += "Last_FlipScroll_AnimFrame_PosY equ FlipScroll_AnimFrame" + (NbAnimFrames - 1) + "_PosY\n";
            File.WriteAllText(outFolder + "_ScrollText_Config_D000.z80", z80);
        }

        private void CreateScrollDataPtr(string outFolder)
        {
            var z80 = "";
            ushort scrollData_LinePtr = 0x0000;

            //for (int i=0; i<FontScreenHeight+1; i++) // +1 because integrate black line
            for (int i = 0; i < FontHeight; i++)
            {
                z80 += "\tFlipScroll_ScrollData_Line" + i + " equ FlipScroll_ScrollDataRAM + &" + scrollData_LinePtr.ToString("X4") + "\n";
                scrollData_LinePtr += 256;
            }

            File.WriteAllText(outFolder + "_ScrollData_Config.z80", z80);
        }

        private void CreateRasterTable(string imageFilename, string destZ80, string label)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";
            z80Content += ";\n";
            //z80Content += "FlipScroll_BackgroundRasters" + rasterIndex + ":\n";

            var bin = new List<byte>();

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var gateArrayColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + gateArrayColor + "\n";

                bin.Add(palette.ToGateArrayByteArray()[0]);
            }

            var binArray = bin.ToArray();

            File.WriteAllText(destZ80, z80Content);

            File.WriteAllBytes(destZ80 + ".bin", binArray);

            var bbPackedBytes = BBPacker.Pack(ref binArray);
            File.WriteAllBytes(destZ80 + ".bin.bbPacked", bbPackedBytes);
            
            var packer = new Packer();
            var packedBytes = packer.Convert(ref binArray);
            File.WriteAllBytes(destZ80 + ".bin.packed", packedBytes);
        }

        private void CreateRasterExplodeAnim(string outFolder)
        {
            var z80 = ";\n";

            const int BaseRasterHeight = 4;
            const int NbFrames = 64;

            var random = new Random(666);

            var baseFramePosY = new List<double>();
            var baseFrameAcceleration = new List<double>();
            for (int i=0; i< BaseRasterHeight; i++)
            {
                baseFramePosY.Add(0);
                var acceleration = 1.55 + random.NextDouble() * 2.4; // NextDouble between 0 and 1
                baseFrameAcceleration.Add(acceleration);
            }

            baseFrameAcceleration.Sort();

            for (int iFrame=0; iFrame<NbFrames; iFrame++)
            {
                z80 += "RasterExplodeFrame" + iFrame + ":\n";

                z80 += "\tdb ";

                for (int i=0; i<baseFramePosY.Count; i++)
                {
                    if (i!=0)
                    {
                        z80 += ",";
                    }
                    int value = (int)baseFramePosY[i];
                    if(value>=134)
                    {
                        value = 134;
                    }
                    z80 += value;
                    baseFramePosY[i] += baseFrameAcceleration[i];
                }

                z80 += "\n";
                z80 += "\n";
            }
            
            File.WriteAllText(outFolder + "_RasterExplodeAnim.z80", z80);

            z80 = ";\n";
            z80 += "RasterExplodeAnim_NbFrames equ " + NbFrames + "\n";
            z80 += "RasterExplodeAnim_RasterHeight equ " + BaseRasterHeight + "\n";
            File.WriteAllText(outFolder + "_RasterExplodeAnim_Config.z80", z80);
        }

        private void CreateWriteZoomLevelCall(string outFolder, bool isScreen1)
        {
            var z80 = ";\n";
            
            int screen = 0;
            if (!isScreen1)
            {
                //z80 += "FlipScroll_WriteZoomLevelCall_Screen0:\n";
            }
            else
            {
                screen = 1;
                //z80 += "FlipScroll_WriteZoomLevelCall_Screen1:\n";
            }

            for (int y = 0; y < FontHeight; y++)
            {
                z80 += "\tpop hl\n";
                z80 += "\tld (FlipScroll_CalcScreen" + screen + "_Line" + y + "_Call+1), hl\n";
            }

            var filename = "_WriteZoomLevelCallScreen0.z80";
            if (isScreen1)
            {
                filename = "_WriteZoomLevelCallScreen1.z80";
            }
            File.WriteAllText(outFolder + filename, z80);
        }
        
        private void CreateDrawScreens(string outFolder, bool isScreen1)
        {
            var z80 = "";
            
            var value_prefix = "FlipScroll_Screen0_Line";
            if(isScreen1)
            {
                value_prefix = "FlipScroll_Screen1_Line";
            }
            
            var label_prefix = "FlipScroll_CalcScreen0";
            if (isScreen1)
            {
                label_prefix = "FlipScroll_CalcScreen1";
            }

            for (int y = 0; y < FontHeight; y++)
            {
                z80 += ";\n";
                z80 += label_prefix + "_Line" + y + ":\n";
                z80 += "\t_NOPS 1\n";
                //z80 += "\tld a, (Frame_ZoomLevels_Table + " + (y * 2) + ")\n";
                //z80 += "\tld h, a\n";
                //z80 += "\tld a, (Frame_ZoomLevels_Table + " + (y * 2) + " + 1)\n";
                //z80 += "\tld l, a\n";
                //z80 += "\tld (" + label_prefix + "_Line" + y + "_Call+1), hl\n";
                z80 += "\tld de, " + value_prefix + "" + y + "_ScreenPtr\n";
                z80 += "\tld h, FlipScroll_ScrollData_Line" + y + "/256\n";
                //z80 += "\t_BRK\n";
                z80 += label_prefix + "_Line" + y + "_Call:\n";
                z80 += "\tcall 0\n";
                z80 += "\n";
            }

            z80 += ";\n";
            z80 += "\tjp FlipScroll_EndCalc\n";

            var filename = "_CalcScreen0.z80";
            if(isScreen1)
            {
                filename = "_CalcScreen1.z80";
            }
            File.WriteAllText(outFolder + filename, z80);
        }

        private void CreateR12R13(string outFolder)
        {
            ushort cpcScreenPtr = 0x0000;

            var z80_Screen0_Lines = ";\n";
            var z80_Screen1_Lines = ";\n";

            for (int y = 0; y < 21; y++)
            {
                ushort screenPtrInWords = (ushort)(cpcScreenPtr >> 1);
                byte R12 = (byte)((screenPtrInWords >> 8) & 3);
                byte R13 = (byte)(screenPtrInWords & 255);
                R12 = (byte)(R12 + ((byte)(3 << 2))); // overscan bits

                z80_Screen0_Lines += "FlipScroll_Screen0_Line" + y + "_ScreenPtr\tequ &" + cpcScreenPtr.ToString("X4") + "\n";
                z80_Screen0_Lines += "FlipScroll_Screen0_Line" + y + "_R12\t\tequ &" + R12.ToString("X2") + "\n";
                z80_Screen0_Lines += "FlipScroll_Screen0_Line" + y + "_R13\t\tequ &" + R13.ToString("X2") + "\n";
                z80_Screen0_Lines += "\n";

                cpcScreenPtr += 0x4000;
                int iPage = 1;
                R12 = (byte)(R12 + ((byte)(iPage << 4)));

                z80_Screen1_Lines += "FlipScroll_Screen1_Line" + y + "_ScreenPtr\tequ &" + cpcScreenPtr.ToString("X4") + "\n";
                z80_Screen1_Lines += "FlipScroll_Screen1_Line" + y + "_R12\t\tequ &" + R12.ToString("X2") + "\n";
                z80_Screen1_Lines += "FlipScroll_Screen1_Line" + y + "_R13\t\tequ &" + R13.ToString("X2") + "\n";
                z80_Screen1_Lines += "\n";

                cpcScreenPtr -= 0x4000;
                cpcScreenPtr += 96;
            }

            File.WriteAllText(outFolder + "_Screen0_Config.z80", z80_Screen0_Lines);
            File.WriteAllText(outFolder + "_Screen1_Config.z80", z80_Screen1_Lines);

            var z80_Screen0_R12R13_Table = ";\n";
            var z80_Screen1_R12R13_Table = ";\n";
            var z80_Screen0_ScreenPtr_Table = ";\n";
            var z80_Screen1_ScreenPtr_Table = ";\n";

            z80_Screen0_R12R13_Table += "Screen0_R12R13_Table:\n";
            z80_Screen1_R12R13_Table += "Screen1_R12R13_Table:\n";

            z80_Screen0_ScreenPtr_Table += "Screen0_ScreenPtr_Table:\n";
            z80_Screen1_ScreenPtr_Table += "Screen1_ScreenPtr_Table:\n";

            for (int y = 0; y < 21; y++)
            {
                z80_Screen0_R12R13_Table += "\tdb FlipScroll_Screen0_Line" + y + "_R12, FlipScroll_Screen0_Line" + y + "_R13\n";
                z80_Screen1_R12R13_Table += "\tdb FlipScroll_Screen1_Line" + y + "_R12, FlipScroll_Screen1_Line" + y + "_R13\n";

                z80_Screen0_ScreenPtr_Table += "\tdw FlipScroll_Screen0_Line" + y + "_ScreenPtr\n";
                z80_Screen1_ScreenPtr_Table += "\tdw FlipScroll_Screen1_Line" + y + "_ScreenPtr\n";
            }

            File.WriteAllText(outFolder + "_Screen0_R12R13_Table.z80", z80_Screen0_R12R13_Table);
            File.WriteAllText(outFolder + "_Screen1_R12R13_Table.z80", z80_Screen1_R12R13_Table);

            //File.WriteAllText(outFolder + "_Screen0_ScreenPtr_Table.z80", z80_Screen0_ScreenPtr_Table);
            //File.WriteAllText(outFolder + "_Screen1_ScreenPtr_Table.z80", z80_Screen1_ScreenPtr_Table);
        }

        private void CreateZ80ToInject(out string z80_pair, out string z80_impair, out int nops_pair, out int nops_impair)
        {   
            nops_pair = 0;
            z80_pair = "";
            z80_pair += "\texx ;1\n"; nops_pair += 1;
            z80_pair += "\tld a, l ;1\n"; nops_pair += 1;
            z80_pair += "\tinc a ;1\n"; nops_pair += 1; // important to set index-1 at beginning
            z80_pair += "\tcp ixh ;2\n"; nops_pair += 2; // repeat management with R12/R13
            z80_pair += "\tjr nz, $+3 ;2\n"; nops_pair += 2;
            z80_pair += "\txor a ;1\n"; nops_pair += 1;
            z80_pair += "\tld l, a ;1\n"; nops_pair += 1;
            z80_pair += "\tld a, (hl) ;2\n"; nops_pair += 2; // R12 value
            z80_pair += "\tor a ;1\n"; nops_pair += 1; // is R13=0 ?
            z80_pair += "\tjp z, ###SkipR12R13_Ptr### ;3\n"; nops_pair += 3; // is R12=0 ? if yes, then we skip R12/R13. (R12 can never be &00, cf. %00001100 for page 0)
            z80_pair += "\tld bc, &BC00+12 ;3\n"; nops_pair += 3;
            z80_pair += "\tout (c), c ;4\n"; nops_pair += 4;
            z80_pair += "\t;; R12 is selected\n";
            z80_pair += "\tinc b ;1\n"; nops_pair += 1;
            z80_pair += "\tout (c), a ;4\n"; nops_pair += 4;
            z80_pair += "\tdec b ;1\n"; nops_pair += 1;
            z80_pair += "\tinc c ;1\n"; nops_pair += 1;
            z80_pair += "\tout (c), c ;4\n"; nops_pair += 4;
            z80_pair += "\t;; R13 is selected\n";
            z80_pair += "\tinc h ;1\n"; nops_pair += 1;
            z80_pair += "\tld a, (hl) ;2\n"; nops_pair += 2; // R13 value
            z80_pair += "\tdec h ;1\n"; nops_pair += 1;
            z80_pair += "\tinc b ;1\n"; nops_pair += 1;
            z80_pair += "\tout (c), a ;4\n"; nops_pair += 4;
            z80_pair += "\tld a, (de) ;2\n"; nops_pair += 2; // raster
            z80_pair += "\tinc e ;1\n"; nops_pair += 1; // raster
            z80_pair += "\texx ;1\n"; nops_pair += 1;
            z80_pair += "\tout (c), a ;4\n"; nops_pair += 4; // raster
            z80_pair += "###SkippedR12R13_Label###:\n";

            // same implementation for both pair/impair
            nops_impair = nops_pair;
            z80_impair = z80_pair;
        }

        private void CreateZoomLevels(string outFolder)
        {
            var minZoomLevel = 0;
            var maxZoomLevel = ZoomAmplitude * 2;

            var z80_ZoomLevelPtr = ";\n";
            z80_ZoomLevelPtr += "ZoomLevels_TablePtr:\n";

            var referenceBitmap = CreateReferenceBitmap();

            string injectZ80_Pair, injectZ80_Impair;
            int injectNops_Pair, injectNops_Impair;
            CreateZ80ToInject(out injectZ80_Pair, out injectZ80_Impair, out injectNops_Pair, out injectNops_Impair);

            const int injectNops_Delay = 128; // change R12/R13 every 128 nops

            // 1ST PASS: create Z80 opcodes
            int maxNops = 0;
            var zoomLevelNops_1stPass = new List<int>();
            var overviewTxt = "";

            for (var iZoomLevel = minZoomLevel; iZoomLevel < maxZoomLevel; iZoomLevel++)
            {
                var zoomBitmap = CreateZoomBitmap(referenceBitmap, (int)(((double)iZoomLevel)*ZoomHorizCoef));
                var zoomIndices = CreateZoomIndices(zoomBitmap);

                z80_ZoomLevelPtr += "\tdw ZoomLevel_" + iZoomLevel + "\n";

                overviewTxt += CreateZoomLevelOverview(zoomIndices, iZoomLevel);

                int nops = 0;
                CreateZoomLevelZ80(
                    zoomIndices,
                    iZoomLevel,
                    injectZ80_Pair,
                    injectZ80_Impair,
                    injectNops_Pair,
                    injectNops_Impair,
                    injectNops_Delay,
                    0,
                    out nops);

                zoomLevelNops_1stPass.Add(nops);
                if (nops > maxNops)
                {
                    maxNops = nops;
                }
            }

            // 2ND pass: ensure all lines have same exact time processing
            var z80_0 = "";
            var z80_1 = "";
            var z80_2 = "";
            var z80 = "";
            var zoomLevelNops_2ndPass = new List<int>();
            int maxIndexInTexture = 0;

            int iWriteCount = 0;
            for (var iZoomLevel = minZoomLevel; iZoomLevel < maxZoomLevel; iZoomLevel++)
            {
                var zoomBitmap = CreateZoomBitmap(referenceBitmap, (int)(((double)iZoomLevel) * ZoomHorizCoef));
                var zoomIndices = CreateZoomIndices(zoomBitmap);

                maxIndexInTexture = Math.Max(maxIndexInTexture, zoomIndices.Max());

                int nops = 0;
                z80 += CreateZoomLevelZ80(
                    zoomIndices,
                    iZoomLevel,
                    injectZ80_Pair,
                    injectZ80_Impair,
                    injectNops_Pair,
                    injectNops_Impair,
                    injectNops_Delay,
                    maxNops - zoomLevelNops_1stPass[iZoomLevel], // nops to add to make all lines stable
                    out nops);
                zoomLevelNops_2ndPass.Add(nops);

                if (iWriteCount==ZoomAmplitude-3)
                {
                    z80_0 = z80;
                    z80 = "";
                }
                if (iWriteCount == (ZoomAmplitude*2 -5))
                {
                    z80_1 = z80;
                    z80 = "";
                }
                iWriteCount++;
            }

            var nopsReport = "";
            nopsReport += "\n";
            nopsReport += ";; MAX Nops=" + maxNops + "\n";
            for (int iZoomLevel = 0; iZoomLevel < zoomLevelNops_1stPass.Count; iZoomLevel++)
            {
                int nbLines = ((zoomLevelNops_2ndPass[iZoomLevel] / 64) + 1);
                nopsReport += ";; ZoomLevel_" + iZoomLevel + " is " + zoomLevelNops_2ndPass[iZoomLevel] + " nops (" + nbLines + " lines=MAX " + nbLines * 64 + " nops) (first pass was " + zoomLevelNops_1stPass[iZoomLevel] + " nops)\n";
            }

            z80_0 = nopsReport + z80_0;
            z80_1 = nopsReport + z80_1;
            z80_2 = nopsReport + z80;

            //File.WriteAllText(outFolder + "_ZoomLevelPtr.z80", z80_ZoomLevelPtr);
            File.WriteAllText(outFolder + "_ZoomLevels_0.z80", z80_0);
            File.WriteAllText(outFolder + "_ZoomLevels_1.z80", z80_1);
            File.WriteAllText(outFolder + "_ZoomLevels_2.z80", z80_2);
            File.WriteAllText(outFolder + "_Overview.txt", overviewTxt);
        }

        private Bitmap CreateReferenceBitmap()
        {
            var bitmap = new Bitmap(ReferenceWidth, 1);

            for (int x = 0; x < ReferenceWidth; x++)
            {
                int red = x;
                int green = 255;
                int blue = 255;
                int alpha = 255;

                bitmap.SetPixel(x, 0, Color.FromArgb(alpha, red, green, blue));
            }

            return bitmap;
        }

        private Bitmap CreateZoomBitmap(Bitmap referenceBitmap, int amplitude)
        {
            var newBitmap = new Bitmap(ScreenWidth, 1);

            using (Graphics g = Graphics.FromImage((System.Drawing.Image)newBitmap))
            {
                g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
                g.DrawImage(referenceBitmap, -amplitude, 0, ScreenWidth + amplitude * 2, 1);
            }

            return newBitmap;
        }

        private List<byte> CreateZoomIndices(Bitmap zoomBitmap)
        {
            var zoomIndices = new List<byte>();

            for (int iX = 0; iX < zoomBitmap.Width; iX++)
            {
                var srcPixel = zoomBitmap.GetPixel(iX, 0);
                var zoomIndex = srcPixel.R;
                zoomIndices.Add(zoomIndex);
            }

            return zoomIndices;
        }

        private string CreateZoomLevelOverview(List<byte> zoomIndices, int amplitude)
        {
            var z80 = "" + amplitude.ToString("X2") + " ";
            foreach (var zoomIndex in zoomIndices)
            {
                z80 += zoomIndex.ToString("X2") + ",";
            }

            z80 += "\n";
            return z80;
        }

        private string CreateZoomLevelZ80(
            List<byte> zoomIndices,
            int zoomLevel,
            string injectZ80_Pair,
            string injectZ80_Impair,
            int injectNops_Pair,
            int injectNops_Impair,
            int injectNops_Delay, // 128..
            int injectMaxNops, // if 0 (cf. PASS 1) then discard
            out int nops)
        {
            bool isPair = true;
            int nextNopsTarget = injectNops_Delay;
            nops = 0;

            var z80 = "";

            z80 += ";;H=scroll line (multiple of 256)\n";
            z80 += ";;DE=screen ptr (no need for alignment)\n";
            z80 += ";;PEN 0 must be selected\n";
            z80 += ";;R12 must be selected\n";
            z80 += ";;BC' must equal &BD00+12\n";
            z80 += ";;HL' must point to R12/R13 table\n";
            z80 += ";;DE' must point to raster table\n";
            z80 += ";;IXL=offset shifting for dynamic scroll management\n";
            z80 += ";;IXH=repeat management with R12/R13\n";
            z80 += "ZoomLevel_" + zoomLevel + ":\n";

            byte L = zoomIndices[0];

            z80 += "\t_NOBRK ;2\n"; nops += 2;
            z80 += "\tld a, " + L + " ;2\n"; nops += 2;
            z80 += "\tadd a, ixl ;2\n"; nops += 2; // offset shifting for dynamic scroll management
            z80 += "\tld l, a ;1\n"; nops += 1;
            z80 += "\tld c, b ;1\n"; nops += 1; // B will always be &7F (used for background raster)
            z80 += "\tldi ;5\n"; nops += 5;
            L++; // LDI incremented L
            byte prevL = L;

            var z80_skipR12R13 = "";
            int injectZ80Count = 0;

            for (int iZoomIndex = 1; iZoomIndex < zoomIndices.Count; iZoomIndex++)
            {
                L = zoomIndices[iZoomIndex];
                if ((prevL - 1) == L)
                {
                    z80 += "\tdec l ;1\n"; nops += 1;
                }
                else if (prevL == L)
                {
                    // do nothing, yepee!                    
                }
                else if ((prevL + 1) == L)
                {
                    z80 += "\tinc l ;1\n"; nops += 1;
                }
                else
                {
                    // missing inc
                    System.Diagnostics.Debugger.Break();
                }

                z80 += "\tldi ;5\n"; nops += 5;
                L++; // LDI incremented L
                prevL = L;

                int maxPixelProcessingNops = 6; // 6=inc l*1+ldi

                if (injectMaxNops > 0)
                {
                    if (iZoomIndex == zoomIndices.Count - 1)
                    {
                        z80 += "\t_NOPS " + injectMaxNops + " ;" + injectMaxNops + "\n"; nops += injectMaxNops;
                    }
                }

                var skipLabel = "SkipR12R13_" + zoomLevel + "_" + injectZ80Count;
                var skippedLabel = "SkippedR12R13_" + zoomLevel + "_" + injectZ80Count;

                if (isPair)
                {
                    if (nops >= (nextNopsTarget - injectNops_Pair - maxPixelProcessingNops))
                    {
                        z80 += injectZ80_Pair
                            .Replace("###SkipR12R13_Ptr###", skipLabel)
                            .Replace("###SkippedR12R13_Label###", skippedLabel)
                            ;

                        z80_skipR12R13 += skipLabel + ":\n";
                        z80_skipR12R13 += "\tld iy, " + skippedLabel + "\n";
                        z80_skipR12R13 += "\tjp SkipR12R13\n";

                        nops += injectNops_Pair;
                        nextNopsTarget += injectNops_Delay;
                        injectZ80Count++;
                        isPair = false;
                    }
                }
                else
                {
                    if (nops >= (nextNopsTarget - injectNops_Impair - maxPixelProcessingNops))
                    {
                        z80 += injectZ80_Impair
                            .Replace("###SkipR12R13_Ptr###", skipLabel)
                            .Replace("###SkippedR12R13_Label###", skippedLabel)
                            ;

                        z80_skipR12R13 += skipLabel + ":\n";
                        z80_skipR12R13 += "\tld iy, " + skippedLabel + "\n";
                        z80_skipR12R13 += "\tjp SkipR12R13\n";

                        nops += injectNops_Impair;
                        nextNopsTarget += injectNops_Delay;
                        injectZ80Count++;
                        isPair = true;
                    }
                }
            }

            z80 += "\tret\n"; nops += 3;
            
            z80 += z80_skipR12R13;

            z80 = "\n" + ";" + nops + "\n" + z80;

            return z80;
        }

        private double GetDistance(double x0, double y0, double x1, double y1)
        {
            return Math.Sqrt(Math.Pow((x1 - x0), 2) + Math.Pow((y1 - y0), 2));
        }
        
        private void CreateRasterSin(string outFolder)
        {
            var sinCurve1 = SinCurve.CreateSin(NbAnimFrames*2, 0, 256-134, 2);
            File.WriteAllBytes(outFolder + "_RasterSinCurve.bin", sinCurve1);
            
            var bbPackedBytes = BBPacker.Pack(ref sinCurve1);
            File.WriteAllBytes(outFolder + "_RasterSinCurve.bin.bbPacked", bbPackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref sinCurve1);
            File.WriteAllBytes(outFolder + "_RasterSinCurve.bin.packed", packedBytes);
        }

        private void CreateBarSin(string outFolder)
        {
            var sinCurve1 = SinCurve.CreateCos(256, 0, 134-50-6, 2);
            File.WriteAllBytes(outFolder + "_BarSinCurve.bin", sinCurve1);

            var bbPackedBytes = BBPacker.Pack(ref sinCurve1);
            File.WriteAllBytes(outFolder + "_BarSinCurve.bin.bbPacked", bbPackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref sinCurve1);
            File.WriteAllBytes(outFolder + "_BarSinCurve.bin.packed", packedBytes);
        }

        private void CreateAnimFrames(string outFolder)
        {
            var z80_ZoomLevels = "";
            var z80_PosY = "";
            var bin_posY = new List<byte>();

            var baseVerticalLines = new List<int>();
            for(int i=0; i<FontHeight+1; i++)
            {
                baseVerticalLines.Add(i);
            }

            const double Amplitude = 180; // +/- 25 degrees

            var sinCurve1 = SinCurve.CreateSin(NbAnimFrames, 0, 50);
            
            for (int iFrame = 0; iFrame < NbAnimFrames; iFrame++)
            {
                double degrees = ((double)iFrame) / ((double)NbAnimFrames);
                degrees -= 0.5;
                degrees *= Amplitude*2;

                //degrees = 0;

                var maxDegrees = -Amplitude;
                Math3D.Vector3D maxOriginalVector = new Math3D.Vector3D(0, (double)(FontHeight+1), 0);
                var maxRotatedVector = Math3D.RotateX(maxOriginalVector, maxDegrees);
                maxRotatedVector.Y = maxOriginalVector.Y + (maxOriginalVector.Y - maxRotatedVector.Y);
                //maxRotatedVector.Y *= 2;
                var maxDistance = GetDistance(maxOriginalVector.Z, maxOriginalVector.Y, maxRotatedVector.Z, maxRotatedVector.Y);

                var frameVerticalLines_posY = new int[baseVerticalLines.Count];
                var frameVerticalLines_labels = new string[baseVerticalLines.Count];

                for (int iY=0; iY < baseVerticalLines.Count; iY++)
                {
                    int iLine = baseVerticalLines[iY];
                    int iCenteredLine = iLine - (FontHeight / 2);

                    Math3D.Vector3D originalVector = new Math3D.Vector3D(0, (double)iCenteredLine, 0);
                    var rotatedVector = Math3D.RotateX(originalVector, degrees);
                    rotatedVector.Y = originalVector.Y + (originalVector.Y - rotatedVector.Y);

                    frameVerticalLines_posY[iY] = (int)(rotatedVector.Y+ (FontHeight / 2));

                    frameVerticalLines_posY[iY] += (int)(sinCurve1[iFrame] + 30 + degrees/6.5); // center Y

                    var distance = GetDistance(originalVector.Z, originalVector.Y, rotatedVector.Z, rotatedVector.Y);
                    var coefZoomLevel = distance/maxDistance;

                    var label = "";
                    var zoomLevel = coefZoomLevel * ZoomAmplitude * 2.0;
                    if (rotatedVector.Z>=0)
                    {
                        label = "ZoomLevel_" + (int)(ZoomAmplitude - zoomLevel);
                    }
                    else
                    {
                        label = "ZoomLevel_" + (int)(ZoomAmplitude + zoomLevel);
                    }

                    frameVerticalLines_labels[iY] = label;
                }

                z80_ZoomLevels += ";\n";
                z80_ZoomLevels += ";; degrees=" + degrees + "\n";

                z80_ZoomLevels += "FlipScroll_AnimFrame" + iFrame + "_ZoomLevels:\n";
                int iCount = 0;
                int state = 0;
                foreach (var label in frameVerticalLines_labels)
                {
                    if (label == null)
                    {
                        z80_ZoomLevels += "\tdw ZoomLevel_0 ;; hidden line\n";
                        frameVerticalLines_posY[iCount] = 255; // hidden live
                    }
                    else
                    {
                        z80_ZoomLevels += "\tdw " + label + "\n";
                        state = 1;
                    }

                    iCount++;
                }

                z80_PosY += ";\n";
                z80_PosY += ";; degrees=" + degrees + "\n";

                z80_PosY += "FlipScroll_AnimFrame" + iFrame + "_PosY:\n";
                z80_PosY += ";; PosY = index dans la table R12/R13 sur CPC (celle utilisee par le runtime)\n";

                z80_PosY += "\tdb ";
                bool first = true;
                int lastKnownPosY = 0;
                for(int iPosY=0; iPosY<frameVerticalLines_posY.Length; iPosY++)
                {
                    var posY = frameVerticalLines_posY[iPosY];

                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        z80_PosY += ",";
                    }

                    var value = posY;
                    if(value!=255)
                    {
                        lastKnownPosY = posY;
                    }

                    if((value==255)&&(iPosY==frameVerticalLines_posY.Length-1))
                    {
                        z80_PosY += lastKnownPosY;
                        bin_posY.Add((byte)lastKnownPosY);
                    }
                    else
                    {
                        z80_PosY += value;
                        bin_posY.Add((byte)value);
                    }
                }
                z80_PosY += "\n";
            }

            File.WriteAllText(outFolder + "_AnimFrames_ZoomLevels.z80", z80_ZoomLevels);

            var binPosY = bin_posY.ToArray();
            File.WriteAllBytes(outFolder + "_AnimFrames_PosY.z80.bin", binPosY);
            
            var BBPackedBytes = BBPacker.Pack(ref binPosY);
            File.WriteAllBytes(outFolder + "_AnimFrames_PosY.z80.bin.bbPacked", BBPackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref binPosY);
            File.WriteAllBytes(outFolder + "_AnimFrames_PosY.z80.bin.packed", packedBytes);
        }
    }
}
