﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using phX.Framework;
using System.Drawing.Imaging;
using System.Xml.Serialization;
using System.Diagnostics;

namespace phX.Parts
{
    public class _21_Balls
    {
        private const int BallWidth = 19;
        private const int BallWidthCPC = (BallWidth+1)/2;
        private const int BallHeight = 38;

        private const int LineWidthBytes = 128; // not 96 because of left/right clipping
        private const int LeftClippingBytes = 16; // 96+16+16 = 128
        private const int LeftClippingWords = LeftClippingBytes / 2;

        private const int R9 = 5;

        public _21_Balls()
        {
            try
            {
                CreateMoveSinCurves(Program.ToZ80Folder("21_Balls/"));

                CreateRasterOpcodes(
                    Program.ToDataFolder("21_Balls/BouncingRaster_Blue32.png"),
                    Program.ToZ80Folder("21_Balls/BouncingRasters/_Blue32.z80"),
                    "RasterBlue32");
                CreateRasterOpcodes(
                    Program.ToDataFolder("21_Balls/BouncingRaster_Green32.png"),
                    Program.ToZ80Folder("21_Balls/BouncingRasters/_Green32.z80"),
                    "RasterGreen32");
                CreateRasterOpcodes(
                    Program.ToDataFolder("21_Balls/BouncingRaster_Yellow32.png"),
                    Program.ToZ80Folder("21_Balls/BouncingRasters/_Yellow32.z80"),
                    "RasterYellow32");
                CreateRasterOpcodes(
                    Program.ToDataFolder("21_Balls/BouncingRaster_Red32.png"),
                    Program.ToZ80Folder("21_Balls/BouncingRasters/_Red32.z80"),
                    "RasterRed32");
                CreateRasterOpcodes(
                    Program.ToDataFolder("21_Balls/BouncingRaster_Pink32.png"),
                    Program.ToZ80Folder("21_Balls/BouncingRasters/_Pink32.z80"),
                    "RasterPink32");

                CreateRasterTable(
                    Program.ToDataFolder("21_Balls/0.png"),
                    Program.ToZ80Folder("21_Balls/_raster0.z80"));

                CreateRasterTable(
                    Program.ToDataFolder("21_Balls/1.png"),
                    Program.ToZ80Folder("21_Balls/_raster1.z80"));

                CreateRasterTable(
                    Program.ToDataFolder("21_Balls/2.png"),
                    Program.ToZ80Folder("21_Balls/_raster2.z80"));

                CreateRasterTable(
                    Program.ToDataFolder("21_Balls/3.png"),
                    Program.ToZ80Folder("21_Balls/_raster3.z80"));

                CreateR12R13(
                    Program.ToZ80Folder("21_Balls/"));

                List<Ball> balls;
                Palette globalPalette;
                CreateBalls(
                    Program.ToDataFolder("21_Balls/ballzV3.png"),
                    Program.ToZ80Folder("21_Balls/"),
                    out balls,
                    out globalPalette);

                CreatePackedBalls(Program.ToZ80Folder("21_Balls/"), balls);
            }
            catch (Exception e)
            {
                Log.Fatal(e.ToString());
            }
        }

        private void CreateMoveSinCurves(string outFolder)
        {
            var xSinCurve = SinCurve.CreateCos(256, (16)*2, (96+8)*2);
            File.WriteAllBytes(outFolder + "_XMoveCurve.bin", xSinCurve);

            var ySinCurve = SinCurve.CreateSin(256, 0+8, 232);
            File.WriteAllBytes(outFolder + "_YMoveCurve.bin", ySinCurve);

            var raster32SinCurve = SinCurve.CreateReverseCos(256, 0, 268-32+1);
            File.WriteAllBytes(outFolder + "_BouncingRaster32Curve.bin", raster32SinCurve);
        }

        private string GetRegister(Palette rasterPalette, Color color)
        {
            int index = -1;
            for (int i = 0; i < rasterPalette.Pens.Count; i++)
            {
                if (rasterPalette.Pens[i].Color == color)
                {
                    index = i;
                    break;
                }
            }

            if (index == -1)
            {
                System.Diagnostics.Debugger.Break();
            }

            if (index == 0)
            {
                return "b";
            }
            if (index == 1)
            {
                return "c";
            }
            if (index == 2)
            {
                return "d";
            }
            if (index == 3)
            {
                return "e";
            }
            if (index == 4)
            {
                return "h";
            }
            if (index == 5)
            {
                return "l";
            }

            System.Diagnostics.Debugger.Break();
            return "";
        }

        private void CreateRasterOpcodes(string imageFilename, string destZ80, string label)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";
            z80Content += ";\n";
            z80Content += label + ":\n";

            // gather color list
            var rasterPalette = new Palette();
            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);
                rasterPalette.AddColor(pixel);
            }

            if (rasterPalette.Pens.Count != 6)
            {
                System.Diagnostics.Debugger.Break();
            }
            /*z80Content += "\tld bc, &"
                        + rasterPalette.ToGateArrayByteArray()[0].ToString("X2")
                        + rasterPalette.ToGateArrayByteArray()[1].ToString("X2")
                        + "\n";
            z80Content += "\tld de, &"
                          + rasterPalette.ToGateArrayByteArray()[2].ToString("X2")
                          + rasterPalette.ToGateArrayByteArray()[3].ToString("X2")
                          + "\n";
            z80Content += "\tld hl, &"
                          + rasterPalette.ToGateArrayByteArray()[4].ToString("X2")
                          + rasterPalette.ToGateArrayByteArray()[5].ToString("X2")
                          + "\n";*/

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                //var register = GetRegister(rasterPalette, pixel);
                //z80Content += "\tld (hl), " + register + "\n";

                var palette = new Palette();
                palette.AddColor(pixel);

                var firmwareColor = palette.ToFirmwareByteArray()[0];
                
                z80Content += "\tdb COLOR" + firmwareColor + "\n";
            }

            File.WriteAllText(destZ80, z80Content);
        }

        private void CreateRasterTable(string imageFilename, string destZ80)
        {
            Bitmap bmp = new Bitmap(System.Drawing.Image.FromFile(imageFilename));

            var z80Content = "";
            z80Content += ";\n";

            var bin = new List<byte>();

            for (int y = 0; y < bmp.Height; y++)
            {
                var pixel = bmp.GetPixel(0, y);

                var palette = new Palette();
                palette.AddColor(pixel);

                var firmwareColor = palette.ToFirmwareByteArray()[0];

                z80Content += "\tdb COLOR" + firmwareColor + "\n";

                bin.Add(palette.ToGateArrayByteArray()[0]);
            }

            var binArray = bin.ToArray();

            File.WriteAllText(destZ80, z80Content);

            File.WriteAllBytes(destZ80 + ".bin", binArray);

            var bbPackedBytes = BBPacker.Pack(ref binArray);
            File.WriteAllBytes(destZ80 + ".bin.bbPacked", bbPackedBytes);

            var packer = new Packer();
            var packedBytes = packer.Convert(ref binArray);
            File.WriteAllBytes(destZ80 + ".bin.packed", packedBytes);
        }

        public class Line
        {
            public int Index;
            public ushort cpcScreenPtr;
            public byte R12;
            public byte R13;
            public byte Page;
        }

        private void CreateR12R13(string outFolder)
        {
            var lines = new List<Line>();

            for (int iPage = 1; iPage < 4; iPage++)
            {
                ushort pageScreenPtr = (ushort)(iPage * 0x4000);

                byte pageR12 = (byte)(iPage * 0x10);

                ushort cpcScreenPtr = 0x0000;
                while (cpcScreenPtr < 0x800)
                {
                    ushort screenPtrInWords = (ushort)(cpcScreenPtr >> 1);
                    byte R12 = (byte)(pageR12 + ((screenPtrInWords >> 8) & 3));
                    R12 = (byte)(R12 + ((byte)(0 << 4)));
                    byte R13 = (byte)((screenPtrInWords) & 255); // LeftClippingWords = show line without left clipping

                    var line = new Line();
                    line.Index = lines.Count;
                    line.cpcScreenPtr = (ushort)(pageScreenPtr + cpcScreenPtr);
                    line.R12 = R12;
                    line.R13 = R13;
                    line.Page = (byte)iPage;
                    lines.Add(line);

                    cpcScreenPtr += LineWidthBytes; // 128;
                }
            }

            var z80_ScreenPtr = ";" + (lines.Count*(R9+1)) + " lines\n";
            var z80_R12R13 = "; " + lines.Count + " VCCs\n";
            foreach (var line in lines)
            {
                z80_ScreenPtr += "\n\t;; line " + line.Index + "\n";
                for(int iVLC=0; iVLC<=R9; iVLC++)
                {
                    z80_ScreenPtr += "\tdw &" + (line.cpcScreenPtr + iVLC*0x800).ToString("X4") + " ;; vlc " + iVLC + "\n";
                }
                z80_R12R13 += "\tdb &" + line.R12.ToString("X2") + ", &" + (line.R13 + LeftClippingWords).ToString("X2") + " ;; VCC " + line.Index + " (Page " + line.Page + ")\n";
            }

            File.WriteAllText(outFolder + "_ScreenPtr.z80", z80_ScreenPtr);
            File.WriteAllText(outFolder + "_R12R13.z80", z80_R12R13);
        }

        public class Ball
        {
            public Bitmap LeftBitmap;
            public Bitmap RightBitmap;
            public byte[] CPCLeftBytes;
            public byte[] CPCRightBytes;
            public byte[] BinLeft;
            public byte[] BinRight;

            public string Z80Left = "";
            public int NopsLeft = 0;
            public string Z80Right = "";
            public int NopsRight = 0;
        }

        private void CreateBalls(string imageFilename, string outFolder, out List<Ball> outBalls, out Palette globalPalette)
        {
            outBalls = new List<Ball>();
            globalPalette = new Palette();

            var baseBitmap = new Bitmap(imageFilename);

            int maxNops = 0;

            for (int iPass = 0; iPass < 2; iPass++)
            {
                outBalls.Clear();
        
                for (int iBall = 0; iBall < 8; iBall++)
                {
                    Color pixel;

                    var leftBitmap = new Bitmap(BallWidth + 1, BallHeight);
                    for (int iY = 0; iY < BallHeight; iY++)
                    {
                        for (int iX = 0; iX < BallWidth + 1; iX++)
                        {
                            var x = (iBall * (BallWidth + 1)) + iX;
                            if (x < baseBitmap.Width)
                            {
                                pixel = baseBitmap.GetPixel(x, iY);
                                leftBitmap.SetPixel(iX, iY, pixel);
                            }
                        }
                    }
                    if (iPass == 0)
                    {
                        leftBitmap.Save(outFolder + "DebugBalls\\_Ball" + iBall + "_Left.png", System.Drawing.Imaging.ImageFormat.Png);
                    }

                    var rightBitmap = new Bitmap(BallWidth + 1, BallHeight);
                    pixel = baseBitmap.GetPixel(0, 0); // black pixel
                    for (int iY = 0; iY < BallHeight; iY++)
                    {
                        rightBitmap.SetPixel(0, iY, pixel);
                    }
                    for (int iY = 0; iY < BallHeight; iY++)
                    {
                        for (int iX = 0; iX < BallWidth + 1; iX++)
                        {
                            var x = (iBall * (BallWidth + 1)) + iX;
                            if (x < baseBitmap.Width)
                            {
                                pixel = baseBitmap.GetPixel(x, iY);
                                if ((iX + 1) < rightBitmap.Width)
                                {
                                    rightBitmap.SetPixel(iX + 1, iY, pixel);
                                }
                            }
                        }
                    }
                    if (iPass == 0)
                    {
                        rightBitmap.Save(outFolder + "DebugBalls\\_Ball" + iBall + "_Right.png", System.Drawing.Imaging.ImageFormat.Png);
                    }

                    var leftImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                    {
                        Bitmap = leftBitmap,
                        Mode = 0,
                        R1 = 800,
                        R4 = 0,
                        R9 = 0,
                        StartVLC = 0,
                        FirstPixelInPaletteX = 0,
                        FirstPixelInPaletteY = 0,
                        Palette = globalPalette
                    });
                    globalPalette = leftImage.Palette;

                    var leftBytes = new byte[BallWidthCPC * BallHeight];
                    for (int i = 0; i < BallWidthCPC * BallHeight; i++)
                    {
                        leftBytes[i] = leftImage.CPCRawBuffer[i];
                    }

                    string z80Left;
                    int nopsLeft;
                    byte[] binLeft = null;
                    CreateBallZ80(maxNops, leftBitmap, leftBytes, out z80Left, out nopsLeft);
                    if (iPass == 1)
                    {
                        var z80Filename = outFolder + "Z80Balls\\_Ball" + iBall + "_Left.z80";
                        File.WriteAllText(z80Filename, z80Left);
                        binLeft = Compile(z80Filename);
                    }

                    var rightImage = Framework.Image.CreateImage(new Framework.Image.CreateImageStruct
                    {
                        Bitmap = rightBitmap,
                        Mode = 0,
                        R1 = 800,
                        R4 = 0,
                        R9 = 0,
                        StartVLC = 0,
                        FirstPixelInPaletteX = 0,
                        FirstPixelInPaletteY = 0,
                        Palette = globalPalette
                    });
                    globalPalette = rightImage.Palette;

                    var rightBytes = new byte[BallWidthCPC * BallHeight];
                    for (int i = 0; i < BallWidthCPC * BallHeight; i++)
                    {
                        rightBytes[i] = rightImage.CPCRawBuffer[i];
                    }

                    string z80Right;
                    int nopsRight;
                    byte[] binRight = null;
                    CreateBallZ80(maxNops, rightBitmap, rightBytes, out z80Right, out nopsRight);
                    if (iPass == 1)
                    {
                        var z80Filename = outFolder + "Z80Balls\\_Ball" + iBall + "_Right.z80";
                        File.WriteAllText(z80Filename, z80Right);
                        binRight = Compile(z80Filename);
                    }

                    var ball = new Ball();
                    ball.LeftBitmap = leftBitmap;
                    ball.RightBitmap = rightBitmap;
                    ball.CPCLeftBytes = leftBytes;
                    ball.CPCRightBytes = rightBytes;
                    ball.Z80Left = z80Left;
                    ball.Z80Right = z80Right;
                    ball.NopsLeft = nopsLeft;
                    ball.NopsRight = nopsRight;
                    ball.BinLeft = binLeft;
                    ball.BinRight = binRight;
                    outBalls.Add(ball);
                }

                if (iPass == 0)
                {
                    maxNops = 0;
                    foreach (var ball in outBalls)
                    {
                        if (ball.NopsLeft > maxNops)
                        {
                            maxNops = ball.NopsLeft;
                        }

                        // right image not supported
                        /*if (ball.NopsRight > maxNops)
                        {
                            maxNops = ball.NopsRight;
                        }*/
                    }
                }
            }

            File.WriteAllText(
                outFolder + "_GlobalPalette.z80",
                globalPalette.ToZ80TableString(Program.FilenameToKeyWordName("BallsPalette")));
        }

        private void IncNopsAndInjectZ80(
            int nopsCount,
            ref string z80, 
            ref int globalNops, 
            ref int rasterCount, 
            ref int rasterNops)
        {
            globalNops += nopsCount;

            var z80_pair = "";
            int pair_nops = 0;
            z80_pair += "\t;; RASTER BEGIN  -- LINE " + rasterCount + "\n";
            z80_pair += "\texx ;1\n"; pair_nops += 1;

            if (rasterCount == R9)
            {
                z80_pair += "\t;; R12/R13 BEGIN\n";

                // every 6 lines, we set new R12/R13 - R12/R13 is aligned 2
                z80_pair += "\tex af, af'\n"; pair_nops += 1;
                z80_pair += "\tld bc, &BC00+12 ;3\n"; pair_nops += 3;
                z80_pair += "\tout (c), c ;4\n"; pair_nops += 4;
                z80_pair += "\tinc b ;1\n"; pair_nops += 1;
                z80_pair += "\tld a, (de) ;2\n"; pair_nops += 2;
                z80_pair += "\tinc e ;1\n"; pair_nops += 1; // R12/R13 is aligned 256
                z80_pair += "\tout (c), a ;4\n"; pair_nops += 4;
                z80_pair += "\tdec b ;1\n"; pair_nops += 1;
                z80_pair += "\tinc c ;1\n"; pair_nops += 1;
                z80_pair += "\tout (c), c ;4\n"; pair_nops += 4;
                z80_pair += "\tinc b ;1\n"; pair_nops += 1;
                z80_pair += "\tld a, (de) ;2\n"; pair_nops += 2;
                z80_pair += "\tinc e ;1\n"; pair_nops += 1; // R12/R13 is aligned 256
                z80_pair += "\tout (c), a ;4\n"; pair_nops += 4;
                z80_pair += "\tld b, &7F+1\n"; pair_nops += 2; // +1= inc B before OUTI !
                z80_pair += "\tex af, af'\n"; pair_nops += 1;

                z80_pair += "\t;; R12/R13 END\n";
            }
            else
            {
                z80_pair += "\tinc b ;1\n"; pair_nops += 1;
            }
            z80_pair += "\touti ;5\n"; pair_nops += 5;
            z80_pair += "\texx ;1\n"; pair_nops += 1;
            z80_pair += "\t;; RASTER END\n";

            rasterNops += nopsCount;
            if(rasterNops >= (64-pair_nops))
            {
                z80 += z80_pair;
                globalNops += pair_nops;
                rasterNops -= (64 - pair_nops);

                rasterCount++;
                if(rasterCount==(R9+1))
                {
                    rasterCount = 0;
                }
            }
        }

        private void CreateBallZ80(int maxNops, Bitmap bitmap, byte[] bytes, out string z80, out int nops)
        {
            z80 = ";\n";
            nops = 0;

            int rasterCount = R9;
            int rasterNops = 0;

            int index = 0;
            byte b;

            byte RegB = 0x00;
            
            var blackPixel = bitmap.GetPixel(0, 0);

            for(int iY=0; iY<BallHeight; iY++)
            {
                int firstByteIndex = 0;
                while (bytes[index + firstByteIndex] == 0x00)
                {
                    firstByteIndex++;
                }

                var z80_data = "";
                for (int iX = 0; iX < BallWidthCPC; iX++)
                {
                    b = bytes[index+iX];
                    if(b==0x00)
                    {
                        continue;
                    }
                    if(z80_data!="")
                    {
                        z80_data += ",";
                    }
                    z80_data += "&"+b.ToString("X2");
                }

                z80 += ";; line " + iY + "  data=" + z80_data + "\n";
                z80 += "\tpop hl ;3\n"; IncNopsAndInjectZ80(3, ref z80, ref nops, ref rasterCount, ref rasterNops);
                z80 += "\tld a, l ;1\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);
                z80 += "\tadd a, c ;1 c=sprite pos x\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);

                if (firstByteIndex == 0)
                {
                }
                else if (firstByteIndex == 1)
                {
                    z80 += "\tinc a ;1\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);
                }
                else if (firstByteIndex>1)
                {
                    z80 += "\tadd a, " + firstByteIndex + " ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                }

                z80 += "\tld l, a ;1\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);

                var occurenceCount = new Dictionary<byte, int>();
                for (int i = 0; i < 256; i++)
                {
                    occurenceCount[(byte)i] = 0;
                }

                //A=0x8C  B=0x30
                //var data0 = new byte[] { 0x84, 0x98, 0x98, 0x30, 0x30, 0x30, 0xCC, 0x8C, 0x8C, 0x80 };

                //A=0x12  B=0x30
                var data0 = new byte[] { 0xF4, 0x12, 0x12, 0x30, 0x30, 0x30, 0x03, 0x56, 0x56, 0xA0 };
                for (int i = 0; i < data0.Length; i++)
                {
                   // bytes[index + i] = data0[i];
                }

                for (int iX = 0; iX < BallWidthCPC; iX++)
                {
                    b = bytes[index + iX];
                    if (b == 0x00)
                    {
                        continue;
                    }
                    occurenceCount[(byte)b]++;
                }

                // discard second and more
                for (int repeat = 1; repeat <= 255; repeat++)
                {
                    var values = new List<byte>();

                    for (int i = 0; i <= 255; i++)
                    {
                        if (occurenceCount[(byte)i] == repeat)
                        {
                            values.Add((byte)i);
                        }
                    }

                    int foundFirst = -1;
                    for (int iX = 0; iX < BallWidthCPC; iX++)
                    {
                        b = bytes[index + iX];
                        if (values.Contains(b))
                        {
                            if (foundFirst==-1)
                            {
                                foundFirst = b;
                            }
                            else
                            {
                                if (b != foundFirst)
                                {
                                    occurenceCount[b] = 0;
                                }
                            }
                        }
                    }
                }

                occurenceCount = occurenceCount.OrderByDescending(x => x.Value).ToDictionary(x => x.Key, x => x.Value);

                var repeatCount0 = occurenceCount.Values.ElementAt(0);
                var repeatValue0 = occurenceCount.Keys.ElementAt(0);
                var repeatCount1 = occurenceCount.Values.ElementAt(1);
                var repeatValue1 = occurenceCount.Keys.ElementAt(1);

                bool RegA_NeedZ80Init = false;
                bool RegB_NeedZ80Init = false;
                byte RegA = 0x00;
                if (repeatCount0>=2)
                {
                    if (RegB != repeatValue0)
                    {
                        RegB = repeatValue0;
                        RegB_NeedZ80Init = true;
                    }
                    if(repeatCount1>=2)
                    {
                        // RegA must be always set since it's used or OR computations
                        RegA_NeedZ80Init = true;
                        RegA = repeatValue1;
                    }
                }
                
                for (int iX=0; iX<BallWidthCPC; iX++)
                {
                    b = bytes[index];

                    var leftPixel = bitmap.GetPixel(iX * 2, iY);
                    var rightPixel = bitmap.GetPixel((iX * 2)+1, iY);

                    bool doNothing = false;
                    if((leftPixel==blackPixel)&&(rightPixel==blackPixel))
                    {
                        // all black
                        doNothing = true;
                    }
                    else if ((leftPixel == blackPixel) && (rightPixel != blackPixel))
                    {
                        // left pixel is black, right pixel is plain (OR)
                        z80 += "\tld a, (hl) ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        z80 += "\tand e ;1 ;;e=&AA mise a zero du pixel droit\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        z80 += "\tor &" + b.ToString("X2") + "\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        z80 += "\tld (hl), a ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                    }
                    else if ((leftPixel != blackPixel) && (rightPixel == blackPixel))
                    {
                        // left pixel is plain, right pixel is black (OR)
                        z80 += "\tld a, (hl) ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        z80 += "\tand d ;1 ;;d=&55 mise a zero du pixel gauche\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        z80 += "\tor &" + b.ToString("X2") + "\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        z80 += "\tld (hl), a ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                    }
                    else if ((leftPixel != blackPixel) && (rightPixel != blackPixel))
                    {
                        // left pixel is plain, right pixel is plain (no OR, direct write)

                        if (RegA_NeedZ80Init)
                        {
                            z80 += "\tld a, &" + RegA.ToString("X2") + " ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                            RegA_NeedZ80Init = false;
                        }
                        if (RegB_NeedZ80Init)
                        {
                            z80 += "\tld b, &" + RegB.ToString("X2") + " ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                            RegB_NeedZ80Init = false;
                        }

                        if(b==RegA)
                        {
                            z80 += "\tld (hl), a ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        }
                        else if(b == RegB)
                        {
                            z80 += "\tld (hl), b ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        }
                        else
                        {
                            z80 += "\tld (hl), &" + b.ToString("X2") + " ;3\n"; IncNopsAndInjectZ80(3, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        }
                    }

                    // increment HL. Do not increment if end of line OR next byte is 2 black pixels
                    if (!doNothing)
                    {
                        if ((iX != (BallWidthCPC - 1)) && (bytes[index + 1] != 0x00))
                        {
                            z80 += "\tinc l ;1\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);
                        }
                    }
                    
                    index++;
                }

                z80 += "\n";
            }

            if (maxNops > 0) // pass 1
            {
                while (nops < maxNops)
                {
                    z80 += "\tnop ;1\n"; IncNopsAndInjectZ80(1, ref z80, ref nops, ref rasterCount, ref rasterNops);

                    if(nops>maxNops)
                    {
                        // while waiting nops, we got raster insertion that delayed the nops waiting
                        System.Diagnostics.Debugger.Break();
                    }
                }
                z80 += "\n";
            }

            z80 += "\tjp (ix) ;2\n"; IncNopsAndInjectZ80(2, ref z80, ref nops, ref rasterCount, ref rasterNops);
            
            var z80_header = ";; nops=" + nops + " rasterLines=" + (nops / 64) + "   6 VLC*5=30 raster lines\n";
            z80 = z80_header + z80;
        }
        
        private byte[] Compile(string z80Filename)
        {
            var binFilename = z80Filename.Replace(".z80", ".z80.bin");
            var arguments = z80Filename + " " + binFilename;
            ExecThirdParty("pasmo", arguments);
            return File.ReadAllBytes(binFilename);
        }

        private void ExecThirdParty(string execFilename, string arguments)
        {
            var thirdPartyFolder = Program.GetRootDirectory() + "VS2015\\phx\\bin\\Debug\\ThirdParty\\";
            var workingDirectory = thirdPartyFolder;
            StartProcess(workingDirectory, workingDirectory + execFilename, arguments);
        }
        
        public static bool StartProcess(string workingDirectory, string filename, string arguments)
        {
            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.CreateNoWindow = true;
            startInfo.UseShellExecute = false;
            startInfo.WorkingDirectory = workingDirectory;
            startInfo.FileName = filename;
            startInfo.Arguments = arguments;
            try
            {
                using (Process exeProcess = Process.Start(startInfo))
                {
                    exeProcess.WaitForExit();
                }
            }
            catch (Exception e)
            {
                throw e;
            }

            return true;
        }

        private void CreateDiffBytes(Ball ball0, Ball iBall, out byte[] diffOffsets, out byte[] diffValues)
        {
            if (ball0.BinLeft.Length != iBall.BinLeft.Length)
            {
                System.Diagnostics.Debugger.Break();
            }

            var offset = new List<ushort>();
            var values = new List<byte>();
            for (int i = 0; i < ball0.BinLeft.Length; i++)
            {
                byte b0 = ball0.BinLeft[i];
                byte bX = iBall.BinLeft[i];

                if (b0 != bX)
                {
                    offset.Add((ushort)i); // TODO: mem ptr instead
                    values.Add(bX);
                }
            }
            // make sure to create pairs
            if ((offset.Count & 1) == 1)
            {
                offset.Add(offset[0]);
                values.Add(values[0]);
            }
            // create CPC array
            diffOffsets = new byte[offset.Count*2];
            for (int i = 0; i < offset.Count; i++)
            {
                diffOffsets[i * 2 + 0] = (byte) (offset[i] & 255);
                diffOffsets[i * 2 + 1] = (byte) (offset[i] >> 8);
            }
            diffValues = new byte[offset.Count];
            for (int i = 0; i < offset.Count; i++)
            {
                diffValues[i] = values[i];
            }
        }

        public void CreatePackedBalls(string outFolder, List<Ball> balls)
        {
            // FILE 0
            var filteredBalls = new List<Ball>();
            filteredBalls.Add(balls[0]);
            filteredBalls.Add(balls[1]);
            filteredBalls.Add(balls[3]);
            filteredBalls.Add(balls[4]);
            filteredBalls.Add(balls[5]);
            filteredBalls.Add(balls[6]);

            var z80 = ";\n";
            //z80 += "BallsPart0 equ &0040\n";
            //z80 += "BallsPart1 equ &4000\n";

            var allBytes = new List<byte>();
            for(int iBall=0; iBall<filteredBalls.Count; iBall++)
            {
                var ball = filteredBalls[iBall];
                var ball0 = filteredBalls[0];
                //z80 += "BallLeft" + iBall + " equ BallsPart0 + &" + (ball0.BinLeft.Length*iBall).ToString("X4") + "\n";
                z80 += "BallLeft" + iBall + " equ BallsPart0 + &" + (allBytes.Count).ToString("X4") + "\n";

                byte[] diffOffsets = null;
                byte[] diffValues = null;
                CreateDiffBytes(ball0, filteredBalls[iBall], out diffOffsets, out diffValues);

                allBytes.AddRange(ball.BinLeft);
                /*if (iBall == 0)
                {
                    allBytes.AddRange(ball0.BinLeft);
                }
                else
                {
                    if (iBall == 1)
                    {
                        allBytes.AddRange(diffOffsets);
                    }

                    z80 += "BallLeft" + iBall + "_DiffBytes equ BallsPart0 + &" + (ball0.BinLeft.Length * iBall).ToString("X4") + "\n";
                    allBytes.AddRange(diffValues);
                }*/
            }

            // original allBytes size = 7980

            z80 += "Balls_Part0_Size equ &" + allBytes.Count.ToString("X4") + "\n";
            z80 += "\n";

            var rawBytes = allBytes.ToArray();
            var binFilename = outFolder + "_BallPart0.bin";
            File.WriteAllBytes(binFilename, rawBytes);
            var packedBytes = BBPacker.Pack(ref rawBytes);
            File.WriteAllBytes(binFilename + ".bbPacked", packedBytes);
            
            var packer = new Packer();
            packedBytes = packer.Convert(ref rawBytes);
            File.WriteAllBytes(binFilename + ".packed", packedBytes);

            // FILE 1
            int startIndex = filteredBalls.Count;
            filteredBalls.Clear();
            filteredBalls.Add(balls[7]);

            allBytes = new List<byte>();
            for (int iBall = 0; iBall < filteredBalls.Count; iBall++)
            {
                var ball = filteredBalls[iBall];
                z80 += "BallLeft" + (startIndex+iBall) + " equ BallsPart1 + &" + allBytes.Count.ToString("X4") + "\n";
                allBytes.AddRange(ball.BinLeft);

                //z80 += "BallRight" + (startIndex+iBall) + " equ BallsPart1 + &" + allBytes.Count.ToString("X4") + "\n";
                //allBytes.AddRange(ball.BinRight);
            }

            z80 += "\n";
            z80 += "Balls_Part1_Size equ &" + allBytes.Count.ToString("X4") + "\n";

            rawBytes = allBytes.ToArray();
            binFilename = outFolder + "_BallPart1.bin";
            File.WriteAllBytes(binFilename, rawBytes);
            packedBytes = BBPacker.Pack(ref rawBytes);
            File.WriteAllBytes(binFilename + ".bbPacked", packedBytes);
            
            packedBytes = packer.Convert(ref rawBytes);
            File.WriteAllBytes(binFilename + ".packed", packedBytes);

            File.WriteAllText(outFolder + "_BallRoutinesOffset.z80", z80);
        }
    }
}
