﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;
using Shared;

namespace RVIMaker
{
	public class CPCExport
	{
		public static int NbCharLines = 21;
		public static int R1 = 48;
        public static int nbVLC = 3;

        FrameMaker frameMaker;

		public CPCExport(FrameMaker InFrameMaker)
		{
			frameMaker = InFrameMaker;
		}

		public void ExportVRAMBlocks(int mode)
		{
			var palette = new Palette();

			var cpcLines = new List<byte>();

			for (int iLine = 0; iLine < frameMaker.UniqueLines.Count; iLine++)
			{
				var line = frameMaker.UniqueLines[iLine];

				line.CPCRaw.Clear();

				if(mode==2)
				{
					for (int iX = 0; iX < line.Pixels.Count; iX += 8)
					{
						Color p0 = line.Pixels[iX];
						Color p1 = line.Pixels[iX + 1];
						Color p2 = line.Pixels[iX + 2];
						Color p3 = line.Pixels[iX + 3];
						Color p4 = line.Pixels[iX + 4];
						Color p5 = line.Pixels[iX + 5];
						Color p6 = line.Pixels[iX + 6];
						Color p7 = line.Pixels[iX + 7];

						var paletteEntry0 = palette.AddColor(p0);
						var paletteEntry1 = palette.AddColor(p1);
						var paletteEntry2 = palette.AddColor(p2);
						var paletteEntry3 = palette.AddColor(p3);
						var paletteEntry4 = palette.AddColor(p4);
						var paletteEntry5 = palette.AddColor(p5);
						var paletteEntry6 = palette.AddColor(p6);
						var paletteEntry7 = palette.AddColor(p7);

						var cpcByte = PixelHelper.ToMode2Byte(
							paletteEntry0.IndexInPalette,
							paletteEntry1.IndexInPalette,
							paletteEntry2.IndexInPalette,
							paletteEntry3.IndexInPalette,
							paletteEntry4.IndexInPalette,
							paletteEntry5.IndexInPalette,
							paletteEntry6.IndexInPalette,
							paletteEntry7.IndexInPalette);

						line.CPCRaw.Add(cpcByte);
					}
				}
				else if (mode == 1)
				{
					for (int iX = 0; iX < line.Pixels.Count; iX += 4)
					{
						Color p0 = line.Pixels[iX];
						Color p1 = line.Pixels[iX + 1];
						Color p2 = line.Pixels[iX + 2];
						Color p3 = line.Pixels[iX + 3];

						var paletteEntry0 = palette.AddColor(p0);
						var paletteEntry1 = palette.AddColor(p1);
						var paletteEntry2 = palette.AddColor(p2);
						var paletteEntry3 = palette.AddColor(p3);

						var cpcByte = PixelHelper.ToMode1Byte(
							paletteEntry0.IndexInPalette,
							paletteEntry1.IndexInPalette,
							paletteEntry2.IndexInPalette,
							paletteEntry3.IndexInPalette);

						line.CPCRaw.Add(cpcByte);
					}
				}
				else if (mode == 0)
				{
					for (int iX = 0; iX < line.Pixels.Count; iX += 2)
					{
						Color p0 = line.Pixels[iX];
						Color p1 = line.Pixels[iX + 1];

						var paletteEntry0 = palette.AddColor(p0);
						var paletteEntry1 = palette.AddColor(p1);

						var cpcByte = PixelHelper.ToMode0Byte(
							paletteEntry0.IndexInPalette,
							paletteEntry1.IndexInPalette);

						line.CPCRaw.Add(cpcByte);
					}
				}				
			}

			// write palette
			File.WriteAllText(FrameMaker.CreateOutputFilename("_palette.macro.z80"), palette.ToZ80MacroString("FX"));

			// write VRAM blocks (&0000, &4000, etc)
			var nbBlocks = 4;
			var nbLinesPerBlocks = NbCharLines * nbVLC; // 3 VLC per charlines
			int lineIndex = 0;
			bool skip_0038 = true;
			for(var iBlock=0; iBlock<nbBlocks; iBlock++)
			{
				byte[] cpcBytes = new byte[nbVLC * 0x800];

				var iVLC = 0;

				ushort iR12R13Offset = (ushort)(iBlock<<(4+8));
				var iCharLine = 0;
				
				for(var iLine=0; iLine<nbLinesPerBlocks; iLine++)
				{
					if((skip_0038)&&(iBlock==0)&&(iLine==0))
					{
						// skip for &0038 case
					}
					else
					{
						if (lineIndex < frameMaker.UniqueLines.Count)
						{
							var line = frameMaker.UniqueLines[lineIndex];
							lineIndex++;

							int cpcPtr = (iCharLine * R1 * 2) + (iVLC * 0x800);
							for (int i = 0; i < line.CPCRaw.Count; i++)
							{
								cpcBytes[cpcPtr+i] = line.CPCRaw[i];
							}   

							line.VLC = iVLC;
							line.R12R13Offset = iR12R13Offset;   
						}
					}

					iVLC++;
					if(iVLC==3)
					{
						iVLC = 0;
						iCharLine++;
						iR12R13Offset = (ushort)(iR12R13Offset+R1);
					}
				}

				var vramStart = iBlock * 0x4000;
				var vramEnd = vramStart + (0x800 * 3);

				if ((vramEnd - vramStart) != cpcBytes.Length)
				{
					System.Diagnostics.Debugger.Break();
				}

				var skipStart = 0;
				if (skip_0038 && (iBlock == 0))
				{
					skipStart = R1 * 2;
					vramStart += R1*2;
				}

				var blockFilename = String.Format("_block_{0:X4}_{1:X4}.bin", vramStart, vramEnd);
				var array = cpcBytes.Skip(skipStart).ToArray();
				File.WriteAllBytes(FrameMaker.CreateOutputFilename(blockFilename), array);
			}
		}

		public void ExportLines()
		{
			var cpcLinesR12 = new List<byte>();
			var cpcLinesR13 = new List<byte>();

			for (int iLine = 0; iLine<frameMaker.UniqueLines.Count; iLine++)
			{
				var line = frameMaker.UniqueLines[iLine];

				byte r12 = (byte)(line.R12R13Offset >> 8);
				byte r13 = (byte)(line.R12R13Offset & 255);

				r12 = (byte)(r12 + (byte)(line.VLC << 2));

				cpcLinesR12.Add(r12);
				cpcLinesR13.Add(r13);
			}

			File.WriteAllBytes(FrameMaker.CreateOutputFilename("_linesR12.bin"), cpcLinesR12.ToArray());
			File.WriteAllBytes(FrameMaker.CreateOutputFilename("_linesR13.bin"), cpcLinesR13.ToArray());
        }

        public void ExportFrameLines()
        {
            var cpcFrames = new List<byte>();

            int count = frameMaker.Frames.Count;
            for (int iFrame = 0; iFrame < count; iFrame++)
            {
                var frame = frameMaker.Frames[iFrame];

                var cpcFrame = new List<byte>();

                for (int iSrcY = 0; iSrcY < frame.SrcY.Count; iSrcY++)
                {
                    byte srcY = frame.SrcY[iSrcY];
                    cpcFrame.Add((byte)srcY);
                }

                cpcFrames.AddRange(cpcFrame);
            }

            File.WriteAllBytes(FrameMaker.CreateOutputFilename("_srcY.bin"), cpcFrames.ToArray());
        }

        public void ExportFrames()
		{
			var cpcFrames = new List<byte>();

			for (int iFrame = 0; iFrame < frameMaker.Frames.Count; iFrame++)
			{
				var frame = frameMaker.Frames[iFrame];

				var cpcFrame = new List<byte>();

				var clamp = 0; // 25; // 7 for height=256
				var start = clamp;
				var end = frame.LinesIndex.Count - clamp;
				var size = end - start; // must be 256 for FX without any changes

				for (int iLineIndex = start; iLineIndex < end; iLineIndex++)
				{
					int lineIndex = frame.LinesIndex[iLineIndex];
					cpcFrame.Add((byte)lineIndex);
				}

				cpcFrames.AddRange(cpcFrame);

				for(int i=end-start; i<256; i++)
				{
					cpcFrames.Add((byte)255);
				}
			}

			File.WriteAllBytes(FrameMaker.CreateOutputFilename("_frames.bin"), cpcFrames.ToArray());
		}

		public void Export(int mode)
		{
			int nbFrames = frameMaker.Frames.Count;
			int nbUniqueLines = frameMaker.UniqueLines.Count;
			
			ExportVRAMBlocks(mode);
			ExportLines();
			ExportFrames();
            ExportFrameLines();
        }
	}
}
