﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using SharpGL;
using SharpGL.SceneGraph.Assets;
using Shared;

namespace RVIMaker
{
    public class FXCube : FXOpenGLBase
    {
        OpenGL _gl;

        float _w = 1.6f * 0.6f; // width
        float _d = 1.6f; // depth
        float _h = 1.6f; // height

        float _degRotation1 = 0.0f;

        uint[] _textures = new uint[2];

        public FXCube()
        {
        }

        public int GetMode()
        {
            return 0;
        }

        public string GetCompiledFolder()
        {
            return DirHelper.SourceFolder + "Demo/FXCube/Compiled/";
        }

        public string GetSourceFolder()
        {
            return DirHelper.SourceFolder + "Demo/FXCube/Raw/";
        }

        public void SetGL(OpenGL gl, int Width, int Height)
        {
            _gl = gl;

            _gl.MatrixMode(OpenGL.GL_PROJECTION);
            _gl.LoadIdentity();

            _gl.Perspective(60.0f, (double)Width / (double)Height, 0.0001, 1000.0);
            _gl.LookAt(0, 0, -5, 0, 0, 0, 0, 1, 0);
            _gl.MatrixMode(OpenGL.GL_MODELVIEW);

            if(GetMode()==1)
            {
                _w *= 2;
            }

            var textureImage = new Bitmap(GetSourceFolder()+"texture.png");     

            _gl.Enable(OpenGL.GL_TEXTURE_2D);
            _gl.GenTextures(2, _textures);
            _gl.BindTexture(OpenGL.GL_TEXTURE_2D, _textures[0]);
            _gl.TexImage2D(OpenGL.GL_TEXTURE_2D, 0, 3, textureImage.Width, textureImage.Height, 0, OpenGL.GL_BGR, OpenGL.GL_UNSIGNED_BYTE,
                textureImage.LockBits(new Rectangle(0, 0, textureImage.Width, textureImage.Height),
                ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb).Scan0);
            _gl.TexParameter(OpenGL.GL_TEXTURE_ENV, OpenGL.GL_TEXTURE_ENV_MODE, OpenGL.GL_REPLACE);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_MIN_FILTER, OpenGL.GL_NEAREST);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_MAG_FILTER, OpenGL.GL_NEAREST);


            var bgImage = new Bitmap(GetSourceFolder() + "bgtexture.png");

            _gl.Enable(OpenGL.GL_TEXTURE_2D);
            //gl.GenTextures(1, textures);
            _gl.BindTexture(OpenGL.GL_TEXTURE_2D, _textures[1]);
            _gl.TexImage2D(OpenGL.GL_TEXTURE_2D, 0, 3, bgImage.Width, bgImage.Height, 0, OpenGL.GL_BGR, OpenGL.GL_UNSIGNED_BYTE,
                bgImage.LockBits(new Rectangle(0, 0, bgImage.Width, bgImage.Height),
                ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb).Scan0);
            _gl.TexParameter(OpenGL.GL_TEXTURE_ENV, OpenGL.GL_TEXTURE_ENV_MODE, OpenGL.GL_REPLACE);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_MIN_FILTER, OpenGL.GL_NEAREST);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_MAG_FILTER, OpenGL.GL_NEAREST);
        }

        public bool Draw()
        {
            float NbMaxFrames = 80.0f*2;

            _gl.ClearColor(0.6f, 0.5f, 0.4f, 1);
            _gl.Clear(OpenGL.GL_COLOR_BUFFER_BIT | OpenGL.GL_DEPTH_BUFFER_BIT);

            float x = 0;
            float y = 0;
            float z = 0;
            DrawShape(x, y, z, -_degRotation1);

            _degRotation1 += 360.0f / NbMaxFrames;

            bool isEnded = false;
            if (_degRotation1 >= 90.0f)
            {
                isEnded = true;
            }
            return isEnded;
        }

        private void DrawShape(float x, float y, float z, float degRotation)
        {
            _gl.LoadIdentity();

            _gl.BindTexture(OpenGL.GL_TEXTURE_2D, _textures[1]);
            _gl.Begin(OpenGL.GL_QUADS);
            var s = 100.0f;
            var s2 = 1800.0f;
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_WRAP_S, OpenGL.GL_REPEAT);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_WRAP_T, OpenGL.GL_REPEAT);
            _gl.TexCoord(0.0f, 0.0f); _gl.Vertex(-s, s, -0.5);
            _gl.TexCoord(s2, 0.0f); _gl.Vertex(s, s, -0.5);
            _gl.TexCoord(s2, s2); _gl.Vertex(s, -s, -0.5);
            _gl.TexCoord(0.0f, s2); _gl.Vertex(-s, -s, -0.5);
            _gl.End();

            _gl.Clear(OpenGL.GL_DEPTH_BUFFER_BIT);

            _gl.BindTexture(OpenGL.GL_TEXTURE_2D, _textures[0]);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_WRAP_S, OpenGL.GL_CLAMP);
            _gl.TexParameter(OpenGL.GL_TEXTURE_2D, OpenGL.GL_TEXTURE_WRAP_T, OpenGL.GL_CLAMP);

            //_gl.LoadIdentity();
            _gl.Translate(x, y, z);
            _gl.Rotate(degRotation, 1.0f, 0.0f, 0.0f);

            _gl.Begin(OpenGL.GL_QUADS);

            _gl.TexCoord(0.0f, 0.0f); _gl.Vertex(-_w, _h, -_d);
            _gl.TexCoord(1.0f, 0.0f); _gl.Vertex(_w, _h, -_d);
            _gl.TexCoord(1.0f, 1.0f); _gl.Vertex(_w, -_h, -_d);
            _gl.TexCoord(0.0f, 1.0f); _gl.Vertex(-_w, -_h, -_d);

            _gl.TexCoord(0.0f, 1.0f); _gl.Vertex(-_w, _h, _d);
            _gl.TexCoord(1.0f, 1.0f); _gl.Vertex(_w, _h, _d);
            _gl.TexCoord(1.0f, 0.0f); _gl.Vertex(_w, -_h, _d);
            _gl.TexCoord(0.0f, 0.0f); _gl.Vertex(-_w, -_h, _d);

            _gl.TexCoord(0.0f, 0.0f); _gl.Vertex(-_w, _h, _d);
            _gl.TexCoord(1.0f, 0.0f); _gl.Vertex(_w, _h, _d);
            _gl.TexCoord(1.0f, 1.0f); _gl.Vertex(_w, _h, -_d);
            _gl.TexCoord(0.0f, 1.0f); _gl.Vertex(-_w, _h, -_d);

            _gl.TexCoord(0.0f, 1.0f); _gl.Vertex(-_w, -_h, _d);
            _gl.TexCoord(1.0f, 1.0f); _gl.Vertex(_w, -_h, _d);
            _gl.TexCoord(1.0f, 0.0f); _gl.Vertex(_w, -_h, -_d);
            _gl.TexCoord(0.0f, 0.0f); _gl.Vertex(-_w, -_h, -_d);

            _gl.End();
        }
    }
}
