﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Threading.Tasks;
using System.IO;
using Shared;
using ImageMaker;

namespace RVIMaker
{
    public class Step
    {
        public int penIndex;
        public Color color;
        public int X1;
        public int Y1;
        public int X2;
        public int Y2;
        public int Y1Raster;
        public int Y2Raster;
        public int size;
    }
        
    public class R12R13Item
    {
        public byte r12;
        public byte r13;
        public ushort scrPtr;
        public string debugName;
    }

    public partial class FXTunnel : FXSoftwareBase
    {
        public FXTunnel()
        {
            CreateR12R13StaticTable();
        }

        #region Window Inits
        Graphics g;
        int width = 0;
        int height = 0;

        public void SetWindow(int w, int h)
        {
            width = w;
            height = h;

            WIDTH = w;
            HEIGHT = h;
        }

        public int GetMode()
        {
            return 0;
        }

        public string GetSourceFolder()
        {
            return DirHelper.SourceFolder + "Data\\04_VerticalBars\\";
        }
        public string GetCompiledFolder()
        {
            return DirHelper.SourceFolder + "Z80\\04_VerticalBars\\";
        }
        #endregion

        List<byte> RasterAllBytes = new List<byte>();
        List<byte> R12R13AllBytes = new List<byte>();
        List<byte> DrawFXAllBytes = new List<byte>();
        List<byte> CumulatedAllBytes = new List<byte>(); // RasterAllBytes+R12R13AllBytes+DrawFXAllBytes
        List<byte> BarXPosSpriteIndexAllBytes = new List<byte>();

        Bitmap VerticalBar;
        public List<R12R13Item> R12R13Items = new List<R12R13Item>();

        const int barClipping = 0;//12; // 12 cpc bytes left and right, must be div 2

        int iFrame = 0;

        int WIDTH = 64;
        int HEIGHT = 128;
        List<Step> Steps;
        const bool EnableTunnelMovement = true;
        const int FRAME_SPEEDX = 0;		// must be < MAXSINCOS
        const int FRAME_SPEEDY1 = 10;		// must be < MAXSINCOS
        const int FRAME_SPEEDY2 = -16;		// must be < MAXSINCOS
        const int STEPCOUNT = 16;
        const int middleZoneHeight = -18;//-24;        
        int frameX = -FRAME_SPEEDX;
        int frameY1 = -FRAME_SPEEDY1;
        int frameY2 = -FRAME_SPEEDY2;
        private int minYBitmap = 696969;
        private int maxYBitmap = -696969;
        private Palette barPalette = new Palette(2);
        private byte[] barCPCBytes;
        private int barCPCWidthMode0;
        private int barWidth;
        private int NbBarsInBitmap = 5;
        private int NbVerticalStepsInBitmap = 8;
        private int SizeVerticalStepInBitmap = 2;
        const bool AlignXPosMultipleOf2 = true;
        int iTunnelFrame = 0;
        const int TunnelNbFrames = 128;
        //private const ushort TunnelJPTablePtrCPC = 0xBB80;        
        private const bool DoDrawVerticalBars = true; // make rendering faster - does not affect data
        private double AnimTimeBars = 1.1f; // anim start time for Vertical Bars
        private const bool RecordBarsOnly = false;
        private const int NbBars = 5;
        private const float TimeDeltaBetweenBars = 3.14159265f/2.5f; // space between bars
        private const float TimeDelta = (235.0f * 0.0108f) / (float)(TunnelNbFrames); // speed for rotate movement
        private const float ScalePosX = 1.23f*1.0f; // plane width
        private const float ScalePosY = 0.315f*1.0f; // plane height
        private const float FrameTimeCoefX = 2.0f;
        private const float FrameTimeCoefY = 1.0f;
        private const double DepthBar = 1.0f;
        private const int TUNNEL_CURVEY1 = 15;       // must be < MAXSINCOS
        private const float TunnelSpeed = (235.0f* 0.00536f) / (float)(TunnelNbFrames*2);
        private const float Amplitude = 28f * 2.8f;

        class BarItem
        {
            public int XPos = -1;
            public int cpcXPos = -1;
            public int YBitmap = -1;
            public int Y1 = -1;
            public int Y2 = -1;
            public double Z = -1;
            public int CPCSpriteIndex = -1;

            // width is screen-width
            public List<byte> CPCScanline = new List<byte>();
        };
        List<BarItem> BarItems = new List<BarItem>();

        private void InitTunnelMovement()
        {
            if (EnableTunnelMovement)
            {
                const int ratioInit = 1;
                frameX = -FRAME_SPEEDX* ratioInit;
                frameY1 = -FRAME_SPEEDY1* ratioInit;
                frameY2 = -FRAME_SPEEDY2* ratioInit;
            }
            else
            {
                frameX = 0;
                frameY1 = 0;
                frameY2 = 0;
            }
        }

        public class BarInfo
        {
            public int Y1;
            public int Y2;
            public int CPCXPos;
            public List<byte> CPCData = new List<byte>();
        }

        Dictionary<double, double> SplineLeft = new Dictionary<double, double>();
        Dictionary<double, double> SplineRight = new Dictionary<double, double>();
        Dictionary<double, double> SplineTop = new Dictionary<double, double>();
        Dictionary<double, double> SplineBottom = new Dictionary<double, double>();

        private Bitmap LastBitmap;
        public Bitmap GetLastBitmap()
        {
            return LastBitmap;
        }

        private List<byte> LastSrcY = new List<byte>();
        public List<byte> GetLastSrcY()
        {
            return LastSrcY;
        }
    }
}
