﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Threading.Tasks;
using System.IO;
using Shared;
using ImageMaker;

namespace RVIMaker
{
    public partial class FXTunnel : FXSoftwareBase
    {
        private void WriteBarCPCSprite(string filename)
        {
            byte emptyPixel = barCPCBytes[0];

            for (int iBar = 0; iBar < NbBarsInBitmap; iBar++)
            {
                var barCpcBytes = new List<byte>();

                int leftX = iBar * barCPCWidthMode0;
                int rightX = (iBar + 1) * barCPCWidthMode0;

                for (int yStep = 0; yStep < NbVerticalStepsInBitmap; yStep++)
                {
                    int yBase = (yStep * SizeVerticalStepInBitmap);

                    for (int y = yBase; y <= yBase + 1; y++)
                    {
                        for (int iX = leftX; iX < rightX; iX++)
                        {
                            var cpcByte = barCPCBytes[(y * (barCPCWidthMode0 * NbBarsInBitmap)) + iX];
                            barCpcBytes.Add(cpcByte);
                        }                        
                    }
                }

                File.WriteAllBytes(filename + ".bar" + iBar + ".left", barCpcBytes.ToArray());
            }
        }

        private int ClipY(int y)
        {
            if (y < 0)
            {
                y = 0;
            }
            else if (y >= 272 / 2)
            {
                y = 270 / 2;
            }

            return y;
        }

        // for a frame, write Y1/Y2 pos .. (R12/R13)
        private byte[] CreateR12R13ItemsFrame()
        {
            var bytes = new List<byte>();

            // already Y sorted ! index=0 equals top item

            for (int iBarItem = 0; iBarItem < BarItems.Count; iBarItem++)
            {
                var barItem = BarItems[iBarItem];
                int y = ClipY(barItem.Y1 / 2);

                int delta = -1000;
                if (iBarItem < (BarItems.Count - 1))
                {
                    delta = ClipY(BarItems[iBarItem + 1].Y1 / 2);
                }

                if (delta == y)
                {
                    bytes.Add((byte)(137)); // dummy line
                }
                else
                {
                    bytes.Add((byte)(y));
                }
            }
            
            BarItems = new List<BarItem>(BarItems.OrderBy(item => item.Y2));
            //for (int iBarItem = BarItems.Count -1; iBarItem >= 0; iBarItem--)
            for (int iBarItem = 0; iBarItem < BarItems.Count; iBarItem++)
            {
                var barItem = BarItems[iBarItem];
                int y = ClipY(barItem.Y2 / 2);

                int delta = -1000;
                if (iBarItem < (BarItems.Count - 1))
                {
                    delta = ClipY(BarItems[iBarItem + 1].Y2 / 2);
                }

                if (delta == y)
                {
                    bytes.Add((byte)(137)); // dummy line
                }
                else
                {
                    bytes.Add((byte)(y));
                }
            }

            BarItems = new List<BarItem>(BarItems.OrderBy(item => item.Y1));

            return bytes.ToArray();
        }

        // for a frame, write bar SpriteIndex + Xpos
        private byte[] CreateBarCPCFrame()
        {
            byte emptyPixel = barCPCBytes[0];

            var bytes = new List<byte>();

            //for (int iBarItem = BarItems.Count - 1; iBarItem >= 0; iBarItem--)
            for (int iBarItem = 0; iBarItem < BarItems.Count; iBarItem++)
            {
                var barItem = BarItems[iBarItem];

                int x = barItem.cpcXPos; // pixel precision not managed

                x += barClipping;
                x -= 4; // because scanline is 96-4-4 long
                if (x < 0)
                {
                    x = 0;
                }
                else if (x > 96-24-4-4)
                {
                    x = 96 - 24 - 4 - 4;
                }

                int y = barItem.YBitmap;                
                int sprite = y - minYBitmap;
                if (sprite < 0)
                {
                    sprite = 0;
                }
                sprite /= 2;
                if (sprite > 16)
                {
                    sprite = 16;
                }

                //sprite += iBarItem * 8;
                sprite += barItem.CPCSpriteIndex;

                //BarsFrameTable:
                //; Sprite line Index, posX
                // db 0, 0
                // db 8 + 4, 24
                bytes.Add((byte)sprite);
                bytes.Add((byte)(x-1));
            }

            return bytes.ToArray();
        }

        // x between -1 and 1.. -1=left  1=right
        // z between -1 and 1.. -1=back  1=front
        // if g==null then no rendering
        private double DrawVerticalBar(Graphics g, int iBar, double x, double y, int barSpriteIndex)
        {
            var left = new SplineInterpolator(SplineLeft);
            var right = new SplineInterpolator(SplineRight);
            var top = new SplineInterpolator(SplineTop);
            var bottom = new SplineInterpolator(SplineBottom);

            const int MinStep = 2;
            const int MaxStep = STEPCOUNT - 1;

            // Y clamp
            for (int iStep = MinStep; iStep < STEPCOUNT - 1; iStep++)
            {
                var vCurrent = top.GetValue(iStep);
                var vNext = top.GetValue(iStep + 1);

                if (vNext < vCurrent)
                {
                    SplineTop[iStep + 1] = vCurrent;

                    top = new SplineInterpolator(SplineTop);
                    iStep = MinStep;
                }
            }
            for (int iStep = MinStep; iStep < STEPCOUNT - 1; iStep++)
            {
                var vCurrent = bottom.GetValue(iStep);
                var vNext = bottom.GetValue(iStep + 1);

                if (vNext > vCurrent)
                {
                    SplineBottom[iStep + 1] = vCurrent;

                    bottom = new SplineInterpolator(SplineBottom);
                    iStep = MinStep;
                }
            }

            int XCenter = (WIDTH / 2) - (barWidth / 2);

            double iKeyZ = (double)(MinStep + ((double)(MaxStep - MinStep - 1)) * ((-y * 0.5f) + 0.5f));

            iKeyZ -= DepthBar;

            int xPos = (int)((x * 1.0f) * (double)(((double)WIDTH / 2) * (((y * 0.5f) * 0.5f) + 0.5f)));

            if (AlignXPosMultipleOf2)
            {
                xPos = (int)xPos;
                xPos /= 2;
                xPos *= 2;

                XCenter = (int)XCenter;
                XCenter /= 2;
                XCenter *= 2;
            }

            xPos += XCenter;

            var x1 = top.GetValue(iKeyZ);
            var x2 = bottom.GetValue(iKeyZ);
            int y1 = (int)top.GetValue(iKeyZ);
            int y2 = (int)bottom.GetValue(iKeyZ);

            if (g != null)
            {
                // make all bars unique
                /*for (int iBarItem = 0; iBarItem < BarItems.Count; iBarItem++)
                {
                    if (BarItems[iBarItem].Y1 == y1)
                    {
                        y1--;
                        y2++;
                        iBarItem = -1;
                    }
                }*/

                BarItem barItem = new BarItem();
                barItem.XPos = xPos;
                barItem.cpcXPos = xPos / 2;
                barItem.Y1 = y1;
                barItem.Y2 = y2;
                barItem.Z = iKeyZ;
                barItem.CPCSpriteIndex = iBar * 8; // required here because of OrderBy below

                var yBitmap = DrawVerticalBar(g, iBar, xPos, y1, y2, barSpriteIndex);
                barItem.YBitmap = yBitmap;

                BarItems.Add(barItem);
                BarItems = new List<BarItem>(BarItems.OrderBy(item => item.Y1));
            }

            return iKeyZ;
        }

        private void CalcPos(int iBar, ref double x, ref double y, ref int barSpriteIndex)
        {
            double frameTime = (double)AnimTimeBars + ((double)iBar) * TimeDeltaBetweenBars;
            x = (double)Math.Sin(frameTime*FrameTimeCoefX)*ScalePosX;
            y = -(double)Math.Cos(frameTime * FrameTimeCoefY) *ScalePosY;

            double zBar = 0.5f + (-(double)Math.Cos(frameTime)/2);

            double xRatio = 0.8f + zBar*0.2f;

            zBar *= (NbVerticalStepsInBitmap);

            //Console.WriteLine("xRatio="+xRatio);
            x *= xRatio;

            //x = (x*1.09f)* (zBar*0.05f);

            barSpriteIndex = (int)zBar;
            //Console.WriteLine("" + barSpriteIndex);

            if (barSpriteIndex < 0)
            {
                barSpriteIndex = 0;
            }
            if (barSpriteIndex >= NbVerticalStepsInBitmap)
            {
                barSpriteIndex = NbVerticalStepsInBitmap - 1;
            }            
        }

        private void DrawVerticalBars(Graphics g)
        {
            var barDictionary = new Dictionary<int, float>();

            for (int iBar = 0; iBar < NbBars; iBar++)
            {
                double x = 0;
                double y = 0;
                int barSpriteIndex = 0;
                CalcPos(iBar, ref x, ref y, ref barSpriteIndex);

                var projectedZ = (float)DrawVerticalBar(null, iBar, x, y, barSpriteIndex);

                barDictionary.Add(iBar, -projectedZ);
            }

            var orderedDictionary = barDictionary.OrderBy(x => x.Value).ToDictionary(x => x.Key, x => x.Value);

            foreach (var key in orderedDictionary.Keys)
            {
                double x = 0;
                double y = 0;
                int barSpriteIndex = 0;
                CalcPos(key, ref x, ref y, ref barSpriteIndex);

                var iBar = key;
                if (DoDrawVerticalBars)
                {
                    DrawVerticalBar(g, iBar, x, y, barSpriteIndex);
                }
            }

            AnimTimeBars += TimeDelta;
        }

        private int DrawVerticalBar(Graphics g, int iBar, int x, int y1, int y2, int z)
        {
            if (y1 > y2)
            {
                int y = y2;
                y2 = y1;
                y1 = y;
            }
            if (z < 0)
            {
                z = 0;
            }
            if (z >= NbVerticalStepsInBitmap)
            {
                z = NbVerticalStepsInBitmap - 1;
            }

            z = NbVerticalStepsInBitmap-1 - z;

            int yBitmap = (int)((float)z);
            yBitmap *= SizeVerticalStepInBitmap;

            if (yBitmap < minYBitmap)
            {
                minYBitmap = yBitmap;
            }
            if (yBitmap > maxYBitmap)
            {
                maxYBitmap = yBitmap;
            }

            int barWith = VerticalBar.Width / NbBarsInBitmap;

            switch(iBar)
            {
                case 0: iBar = 0; break;
                case 1: iBar = 1; break;
                case 2: iBar = 2; break;
                case 3: iBar = 3; break;
                case 4: iBar = 4; break;
                default: System.Diagnostics.Debugger.Break(); break;
            }

            int xOffset = iBar * barWidth;

            int yOffset = 0; 
            
            yOffset = 0;
            for (int y = y1; y < y2; y++)
            {
                int iY = yBitmap + yOffset;
                yOffset++;
                if (yOffset > 1)
                {
                    yOffset = 0;
                }

                for (int iX = 0; iX < barWith; iX++)
                {
                    var pixel = VerticalBar.GetPixel(iX + xOffset, iY);

                    if ((pixel.R == 0) && (pixel.G == 0) && (pixel.B == 0))
                    {
                        continue;
                    }

                    var brush = new System.Drawing.SolidBrush(pixel);
                    g.FillRectangle(brush, iX + x, y, 1, 1);
                }
            }

            return yBitmap;
        }
    }
}
