﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Threading.Tasks;
using System.IO;
using Shared;
using ImageMaker;

namespace RVIMaker
{
    public partial class FXTunnel : FXSoftwareBase
    {
        private bool DrawTunnel(Graphics g)
        {
            bool ended = false;

            if (EnableTunnelMovement)
            {
                frameX += FRAME_SPEEDX;
                frameY1 += FRAME_SPEEDY1;
                frameY2 += FRAME_SPEEDY2;
            }

            var brush = new System.Drawing.SolidBrush(Steps[Steps.Count-1].color);
            g.FillRectangle(brush, 0, 0, WIDTH, HEIGHT);

            SplineLeft.Clear();
            SplineRight.Clear();
            SplineTop.Clear();
            SplineBottom.Clear();

            int iSinX = frameX;
            int iSinY1 = frameY1;
            int iSinY2 = frameY2;
            int prevX1 = 0;
            int prevY1 = 0;
            int prevX2 = WIDTH;
            int prevY2 = HEIGHT;
            int w = 0;
            int h = 0;
            int X1 = 0;
            int Y1 = 0;
            int X2 = 0;
            int Y2 = 0;
            
            for (int iStep = 0; iStep < STEPCOUNT; iStep++)
            {
                var step = Steps[iStep];
                if(step.size==0)
                {
                    continue;
                }

                iSinY1 += TUNNEL_CURVEY1;
                
                int addX = 0;// (int)(Math.Cos(((float)iSinX) * 0.003f) * 32);

                double fSinY1 = Math.Cos(((float)iSinY1) * TunnelSpeed);
                int addY = (int)(fSinY1 * 28f *2.8f);

                if (!EnableTunnelMovement)
                {
                    addY = 0;
                }

                X1 = step.X1 + addX;
                Y1 = step.Y1 + addY;
                X2 = step.X2 + addX;
                Y2 = step.Y2 + addY;

                if( (Y1&1)==1 )
                {
                    Y1--;
                }
                if ((Y2 & 1) == 1)
                {
                    Y2--;
                }

                if (Y1 > Y2)
                {
                    int t = Y2;
                    Y1 = Y2;
                    Y2 = t;
                }

                if (X1 > X2)
                {
                    int t = X2;
                    X1 = X2;
                    X2 = t;
                }

                SplineLeft.Add((double)iStep, (double)X1);
                SplineRight.Add((double)iStep, (double)X2); 
                SplineTop.Add((double)iStep, (double)Y1);
                SplineBottom.Add((double)iStep, (double)Y2);

                Y1 = step.Y1Raster + addY;
                Y2 = step.Y2Raster + addY;

                if ((Y1 & 1) == 1)
                {
                    Y1--;
                }
                if ((Y2 & 1) == 1)
                {
                    Y2--;
                }

                // make flat horizon
                X1 -= 500;
                X2 += 500;

                if (X1 < prevX1)
                {
                    X1 = prevX1;
                }
                if (X1 > prevX2)
                {
                    X1 = prevX2;
                }
                if (X2 < prevX1)
                {
                    X2 = prevX1;
                }
                if (X2 > prevX2)
                {
                    X2 = prevX2;
                }
                if (Y1 < prevY1)
                {
                    Y1 = prevY1;
                }
                if (Y1 > prevY2)
                {
                    Y1 = prevY2;
                }
                if (Y2 < prevY1)
                {
                    Y2 = prevY1;
                }
                if (Y2 > prevY2)
                {
                    Y2 = prevY2;
                }

                brush = new System.Drawing.SolidBrush(step.color);

                if(prevX2>prevX1)
                {
                    w = prevX2 - prevX1;

                    if(Y1>prevY1)
                    {
                        g.FillRectangle(brush, prevX1, prevY1, w, Y1 - prevY1);
                    }
                    if (prevY2 > Y2)
                    {
                        g.FillRectangle(brush, prevX1, Y2, w, prevY2 - Y2);
                    }
                }

                if(Y2>Y1)
                {
                    if (X1 > prevX1)
                    {
                        w = X1 - prevX1;
                        g.FillRectangle(brush, prevX1, Y1, w, Y2 - Y1);
                    }
                    if(prevX2>X2)
                    {
                        w = prevX2-X2;
                        g.FillRectangle(brush, X2, Y1, w, Y2 - Y1);
                    }
                }

                prevX1 = X1;
                prevY1 = Y1;
                prevX2 = X2;
                prevY2 = Y2;
            }

            w = X2 - X1;
            if(w>0)
            {
                h = Y2 - Y1;
                if(h>0)
                {
                    brush = new System.Drawing.SolidBrush(Steps[Steps.Count - 1].color);
                    g.FillRectangle(brush, X1, Y1, w, h);
                }
            }

            return ended;
        }

        private int GetPenIndex(Color color)
        {
            for (int iStep=0; iStep<this.Steps.Count; iStep++)
            {
                var step = Steps[iStep];

                if(step.color == color)
                {
                    return step.penIndex;
                }   
            }

            return -1;
        }

        private int[] penIndices = new int[] { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 0 };
        
        private List<byte> CreateRasters(Bitmap bitmap)
        {
            var bytes = new List<byte>();

            // top part of Y-pos per pen
            // if value is -1, then the pen is not used for this frame (hidden line)
            var firsts = new List<int>();

            // bottom part of Y-pos
            // if value is -1, then the pen is not used for this frame (hidden line)
            var seconds = new List<int>();

            for (int iPen = 0; iPen < this.Steps.Count; iPen++)
            {
                byte iScanlinePen = 69;

                int first = -1;
                int second = -1;

                for (int y = 0; y < bitmap.Height; y++)
                {
                    var color = bitmap.GetPixel(0, y);

                    int pen = GetPenIndex(color);// +FramePenOffset;
                    if(pen>11)
                    {
                        pen = 11;
                    }
                    /*while(pen>10)
                    {
                        pen -= 11;
                    } */                                       
                    byte penIndex = (byte)pen;

                    if (iScanlinePen != penIndex)
                    {
                        iScanlinePen = penIndex;

                        if(penIndex==iPen)
                        {
                            if(first==-1)
                            {
                                first = y;
                            }
                            else
                            {
                                second = y;
                            }
                        }
                    }
                }

                firsts.Add(first);
                seconds.Add(second);
            }

            //ushort TunnelJPTable = (ushort)(TunnelJPTablePtrCPC);
            //ushort TunnelJPTable_DummyEntry = (ushort)(TunnelJPTablePtrCPC + 272+272 + 272+272 + 2);
            byte dummyEntry = 136+1; // +0=end of FX +1=dummy
            
            if(this.Steps.Count!=16)
            {
                System.Diagnostics.Debugger.Break();
            }

            for (int iPen = 3 /* skip 3 first items */; iPen < 12 /*this.Steps.Count*/; iPen++)
            {
                int first = firsts[iPen];
                int second = seconds[iPen];

                if(first==-1)
                {
                    //bytes.Add((byte)(TunnelJPTable_DummyEntry & 255));
                    //bytes.Add((byte)(TunnelJPTable_DummyEntry >> 8));
                    bytes.Add((byte)(dummyEntry));
                }
                else
                {
                    //ushort ptr = (ushort)(TunnelJPTable + first*4);
                    //bytes.Add((byte)(ptr & 255));
                    //bytes.Add((byte)(ptr >> 8));
                    bytes.Add((byte)(first/2)); // Y / 2 in data!
                }

                if (second == -1)
                {
                    //bytes.Add((byte)(TunnelJPTable_DummyEntry & 255));
                    //bytes.Add((byte)(TunnelJPTable_DummyEntry >> 8));
                    bytes.Add((byte)(dummyEntry));
                }
                else
                {
                    //ushort ptr = (ushort)(TunnelJPTable + second * 4);
                    //bytes.Add((byte)(ptr & 255));
                    //bytes.Add((byte)(ptr >> 8));
                    bytes.Add((byte)(second/2)); // Y / 2 in data!
                }
            }

            // retrieve middle zone
            /*int middleZoneY1 = -1;
            int middleZoneY2 = -1;
            var backgroundColor = Steps[12 - 1].color;//Steps[Steps.Count - 1].color;
            for (int y = 0; y < bitmap.Height; y++)
            {
                var color = bitmap.GetPixel(0, y);
                
                if ((middleZoneY1 == -1) && (color == backgroundColor))
                {
                    middleZoneY1 = y;
                }

                if ((middleZoneY1 != -1) && (middleZoneY2 == -1) && (color != backgroundColor))
                {
                    middleZoneY2 = y;
                }
            }
            
            // 54 lines
            int middleZoneSize = middleZoneY2 - middleZoneY1;
            
            ushort ptrMiddleZone = (ushort)(TunnelJPTable + 2 + middleZoneY1 * 4);
            bytes.Add((byte)(ptrMiddleZone & 255));
            bytes.Add((byte)(ptrMiddleZone >> 8));*/

            return bytes;
        }
    }
}
