﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Threading.Tasks;
using System.IO;
using Shared;
using ImageMaker;

namespace RVIMaker
{
    public partial class FXTunnel : FXSoftwareBase
    {
        private void CreateR12R13StaticTable()
        {
            string r12r13Table = String.Empty;

            int index = 0;

            ushort ptr = 0x104 / 2; // End of Bobs VRAM is &104
            while (ptr < 0x400)
            {
                byte r12 = (byte)(ptr >> 8);
                byte r13 = (byte)(ptr & 255);

                ushort iScreenPtr = (ushort)((ptr * 2));
                ushort iNextPtr = (ushort)((ptr + 48 + barClipping) * 2);

                if ((iNextPtr & 255) < (iScreenPtr & 255)) // skip to ensure INC L
                {
                    ushort skip = (ushort)(256 - (byte)(iScreenPtr & 255));
                    ptr += (ushort)(skip / 2);
                    r12 = (byte)(ptr >> 8);
                    r13 = (byte)(ptr & 255);
                    iScreenPtr = (ushort)((ptr * 2));
                    iNextPtr = (ushort)((ptr + 48 + barClipping) * 2);
                }

                if ((ptr + (48 + barClipping)) < 1024)
                {
                    r12r13Table += ";\n";
                    r12r13Table += "R12_Line" + index + "_Flip0\t\t\t\tequ " + r12 + "\n";
                    r12r13Table += "R13_Line" + index + "_Flip0\t\t\t\tequ " + (r13 + (barClipping / 2)) + "\t\t; &" + (r13 + (barClipping / 2)).ToString("X2") + "\n";
                    r12r13Table += "ScrPtr_HiddenLine" + index + "_Flip0\tequ &" + (iScreenPtr).ToString("X4") + "\t; end at &" + (iNextPtr - 1).ToString("X4") + "\n";
                    r12r13Table += "ScrPtr_VisibleLine" + index + "_Flip0\tequ &" + (iScreenPtr + barClipping).ToString("X4") + "\t; end at &" + (iNextPtr - 1 - barClipping).ToString("X4") + "\n";
                    r12r13Table += "R12_Line" + index + "_Flip1\t\t\t\tequ " + (r12+0x10) + "\n"; // &4000
                    r12r13Table += "R13_Line" + index + "_Flip1\t\t\t\tequ " + (r13 + (barClipping / 2)) + "\t\t; &" + (r13 + (barClipping / 2)).ToString("X2") + "\n";
                    r12r13Table += "ScrPtr_HiddenLine" + index + "_Flip1\tequ &" + (iScreenPtr+0x4000).ToString("X4") + "\t; end at &" + (iNextPtr - 1 + 0x4000).ToString("X4") + "\n";
                    r12r13Table += "ScrPtr_VisibleLine" + index + "_Flip1\tequ &" + (iScreenPtr + 0x4000 + barClipping).ToString("X4") + "\t; end at &" + (iNextPtr - 1 - barClipping + 0x4000).ToString("X4") + "\n";
                    r12r13Table += "\n";

                    var r12r13item = new R12R13Item();
                    r12r13item.debugName = "Line" + index;
                    r12r13item.r12 = r12;
                    r12r13item.r13 = r13;
                    r12r13item.scrPtr = iScreenPtr;
                    R12R13Items.Add(r12r13item);

                    index++;
                }

                ptr += 48 + (barClipping);
            }

            File.WriteAllText(GetCompiledFolder() + "_R12R13Const.z80", r12r13Table);
        }
    }
}
