﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Threading.Tasks;
using System.IO;
using Shared;
using ImageMaker;

namespace RVIMaker
{
    public partial class FXTunnel : FXSoftwareBase
    {
        //int FramePenOffset = 0;
        int iBeatFrame = 0;

        private void CreateTunnel()
        {
            Steps = new List<Step>();

            Step prevStep = new Step();
            prevStep.size = 384;
            prevStep.X1 = (WIDTH / 2) - prevStep.size;
            prevStep.Y1 = (HEIGHT / 2) - prevStep.size;
            prevStep.X2 = (WIDTH / 2) + prevStep.size;
            prevStep.Y2 = (HEIGHT / 2) + prevStep.size;
            prevStep.Y1Raster = (HEIGHT / 2) - prevStep.size;
            prevStep.Y2Raster = (HEIGHT / 2) + prevStep.size;

            //float iProgress = ((float)iFrame) / ((float)(TunnelNbFrames - 1));

            //FramePenOffset = 0;
           /* iProgress *= -11.0f; // nb pens
            while(iProgress<0.0f)
            {
                iProgress += 1.0f;
                FramePenOffset++;
            }*/

            float z = 0.1001f; // do not use 0.0f
            for (int iStep = 0; iStep < STEPCOUNT; iStep++)
            {
                var step = new Step();

                step.penIndex = iStep;

                int r = (step.penIndex * 8);
                int g = r;
                int b = 250;
                step.color = Color.FromArgb(255, r, g, b);

                if(RecordBarsOnly)
                {
                    step.color = Color.FromArgb(255, 0, 0, 0);
                }

                step.size = 0;

                if (iStep != 0)
                {   
                    if(iBeatFrame < 32)
                    {
                       // middleZoneHeight -= iBeatFrame/3;
                    }

                    iBeatFrame++;
                    if(iBeatFrame > (5*4*4))
                    {
                        iBeatFrame = 0;
                    }

                    step.size = (int)((5.445f * (float)(WIDTH)) / (6.0f * z) - (float)middleZoneHeight);

                    step.X1 = (WIDTH / 2) - step.size;
                    step.X2 = (WIDTH / 2) + step.size;
                    step.Y1 = (HEIGHT / 2) - step.size;
                    step.Y2 = (HEIGHT / 2) + step.size;

                    float prev = (float)prevStep.size;
                    float current = (float)step.size;
                    float interpolated = prev - (prev - current);// * iProgress;
                    int stepSize = (int)interpolated;

                    step.Y1Raster = (HEIGHT / 2) - stepSize;
                    step.Y2Raster = (HEIGHT / 2) + stepSize;

                    prevStep = step;
                }

                z += 0.3f + (z*0.35f);
                //z *= 1.01f;

                Steps.Add(step);
            }
        }

        public struct NopZone
        {
            public byte Y;
            public byte NbScanlines;
        };

        int iTunelReset = 0;

        public bool Draw()
        {
            if (Steps == null)
            {
                InitFX();
            }
            
            LastBitmap = new Bitmap(width, height, PixelFormat.Format24bppRgb);
            g = Graphics.FromImage(LastBitmap);

            CreateTunnel();
            DrawTunnel(g);
            g.Flush();
                        
            var rastersFrame = CreateRasters(LastBitmap);

            BarItems.Clear();

            DrawVerticalBars(g);
            g.Flush();

            var r12r13Frame = CreateR12R13ItemsFrame();

            // rasters + R12R13
            var tempMergedFrameBytes = new List<byte>();
            tempMergedFrameBytes.AddRange(rastersFrame); // rasters
            tempMergedFrameBytes.AddRange(r12r13Frame); // for a frame, write Y1/Y2 R12/R13 pos .. (R12/R13, same size than tunnel frames)

            // FINAL RasterBytes
            var rasterBytes = new List<byte>();
            rasterBytes.AddRange(rastersFrame); // rasters

            // FINAL BarBytes R12+R13/Xpos/NopZone
            var r12r13Bytes = new List<byte>();
            r12r13Bytes.AddRange(r12r13Frame);

            // find available zones that have more than 
            int ScanlinesPerNopZone = 12; // must be multiple of 2 (R9=1)
            var nopZones = new List<NopZone>();
            int nbScanlines = -1;
            ScanlinesPerNopZone /= 2; // R9=1
            for (byte y=0; y<272/2; y++)
            {
                if (tempMergedFrameBytes.Contains(y))
                {
                    nbScanlines = -1; // -1 because we want the line AFTER to be processed
                }
                if (nbScanlines == (int)(ScanlinesPerNopZone))
                {
                    var nopZone = new NopZone();
                    nopZone.Y = (byte)(y - (ScanlinesPerNopZone));
                    nopZone.NbScanlines = (byte)nbScanlines;
                    nopZones.Add(nopZone);

                    nbScanlines = -1;
                }
                nbScanlines++;
            }

            // inject zones into drawFXBytes
            var drawFXBytes = new List<byte>();
            const int NbZones = 10; // 10 calls to manage all frame lines - 10*10 scanlines = 100 rasterlines
            if(nopZones.Count<NbZones)
            {
                Console.WriteLine("FAILED");
                Environment.Exit(0);
                //System.Diagnostics.Debugger.Break();
            }
            for(int iZone=0; iZone<NbZones; iZone++)
            {
                var nopZone = nopZones[iZone];
                drawFXBytes.Add(nopZone.Y);
            }
            Console.WriteLine("=== New Frame [" + iFrame + "/" + TunnelNbFrames + "] ===> Found " + 
                nopZones.Count + " blocks (remaining " + (nopZones.Count - NbZones) + " blocks for sprites)");

            // inject scanline info (sprite index [byte] + posX [byte])
            var SpriteIndex_XPos = CreateBarCPCFrame();
            BarXPosSpriteIndexAllBytes.AddRange(SpriteIndex_XPos);

            // populate final buffers (those containers will be written down to disc when ended)
            RasterAllBytes.AddRange(rasterBytes);
            R12R13AllBytes.AddRange(r12r13Bytes);
            DrawFXAllBytes.AddRange(drawFXBytes);

            CumulatedAllBytes.AddRange(rasterBytes);
            CumulatedAllBytes.AddRange(r12r13Bytes);
            CumulatedAllBytes.AddRange(drawFXBytes);

            if (iFrame== TunnelNbFrames - 1)
            {
                InitTunnelMovement();
            }

            if(iFrame==0)
            {
                this.GetLastBitmap().Save(GetCompiledFolder() + "_Frame0.bmp", ImageFormat.Bmp);
            }
            if (iFrame == TunnelNbFrames-2)
            {
                this.GetLastBitmap().Save(GetCompiledFolder() + "_Frame" + (TunnelNbFrames - 2) + ".bmp", ImageFormat.Bmp);
            }
            if (iFrame == TunnelNbFrames-1)
            {
                this.GetLastBitmap().Save(GetCompiledFolder() + "_Frame"+ (TunnelNbFrames-1) +".bmp", ImageFormat.Bmp);
            }
            if (iFrame == TunnelNbFrames-1)
            {
                WriteBarCPCSprite(GetCompiledFolder() + "_BarSprite.bin");
                //File.WriteAllBytes(GetCompiledFolder() + "_RasterFrames.bin", RasterAllBytes.ToArray());
                //File.WriteAllBytes(GetCompiledFolder() + "_R12R13Frames.bin", R12R13AllBytes.ToArray());
                //File.WriteAllBytes(GetCompiledFolder() + "_DrawFXFrames.bin", DrawFXAllBytes.ToArray());
                File.WriteAllBytes(GetCompiledFolder() + "_RasterR12R13DrawFXFrames.bin", CumulatedAllBytes.ToArray());
                File.WriteAllBytes(GetCompiledFolder() + "_BarXPosSpriteIndexFrames.bin", BarXPosSpriteIndexAllBytes.ToArray());

                Console.WriteLine("FILES WRITTEN : iFrame="+ iFrame);
                Environment.Exit(0);
            }

            iFrame++;
            return false; // never ended
        }
    }
}
