﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using SharpGL;

namespace RVIMaker
{
    public class Line
    {
        public int Index = -1;
        public byte SrcY = 0;
        public List<Color> Pixels = new List<Color>();
        public List<byte> CPCRaw = new List<byte>();

        public int VLC = 0; 
        public ushort R12R13Offset = 0;        
    }

    public class Frame
    {
        public int Index = -1;
        public List<int> LinesIndex = new List<int>();
        public List<byte> SrcY = new List<byte>();
    }

    public class FrameMaker
    {
        public List<Line> UniqueLines = new List<Line>();
        public List<Frame> Frames = new List<Frame>();
        static public string _compiledFolder;

        static public string CreateOutputFilename(string filename)
        {
            var directoryName = _compiledFolder;
            if (!Directory.Exists(directoryName))
            {
                Directory.CreateDirectory(directoryName);
            }
            var dirInfo = new DirectoryInfo(directoryName);

            var absolutePath = dirInfo.FullName + "\\" + filename;

            return absolutePath;
        }
        
        OpenGL gl;
        int frameWidth;
        int frameHeight;

        public FrameMaker(string compiledFolder, OpenGL InGL, int width, int height)
        {
            _compiledFolder = compiledFolder;
            gl = InGL;
            frameWidth = width;
            frameHeight = height;
        }

        private Bitmap CreateBitmapFromScreenBuffer()
        {
            int x = 0;
            int y = 0;
            int w = frameWidth;
            int h = frameHeight;

            byte[] pixelData = new byte[w * h * 4];
            gl.ReadPixels(x, y, w, h, OpenGL.GL_RGBA, OpenGL.GL_UNSIGNED_BYTE, pixelData);

            var bitmap = new Bitmap(w, h, PixelFormat.Format32bppArgb);
            int offsetInPixelData = 0;
            for (y = h - 1; y >= 0; y--)
            {
                for (x = 0; x < w; x++)
                {
                    byte red = pixelData[offsetInPixelData + 0];
                    byte green = pixelData[offsetInPixelData + 1];
                    byte blue = pixelData[offsetInPixelData + 2];

                    var color = Color.FromArgb((byte)255, red, green, blue);
                    bitmap.SetPixel(x, y, color);

                    offsetInPixelData += 4;
                }
            }

            return bitmap;
        }

        private void WriteBitmapToFile(Bitmap bitmap, int frameIndex)
        {
            var filename = String.Format("Img{0:000}.bmp", frameIndex);

            bitmap.Save(FrameMaker.CreateOutputFilename(filename), ImageFormat.Bmp);
        }

        public void ProcessLines(Bitmap bitmap, List<byte> srcY)
        {
            var frame = new Frame();
            frame.Index = Frames.Count;
            if(srcY!=null)
            {
                frame.SrcY = new List<byte>(srcY);
            }

            for (int y=0; y<bitmap.Height; y++)
            {
                List<Color> pixels = new List<Color>();
                for (int x=0; x<bitmap.Width; x++)
                {
                    pixels.Add(bitmap.GetPixel(x, y));
                }

                var index = -1;
                for(int iLine=0; iLine<UniqueLines.Count; iLine++)
                {
                    var iPixels = UniqueLines[iLine].Pixels;

                    if(pixels.Count!=iPixels.Count)
                    {
                        continue;
                    }

                    int similar = 0;
                    for(int i=0;i<pixels.Count; i++)
                    {
                        if(pixels[i]==iPixels[i])
                        {
                            similar++;
                        }
                    }

                    if(similar==pixels.Count)
                    {
                        index = iLine;
                    }
                }

                if(index==-1)
                {
                    var line = new Line();
                    line.Index = UniqueLines.Count;
                    if (srcY != null)
                    {
                        line.SrcY = srcY[y];
                    }
                    line.Pixels = pixels;
                    UniqueLines.Add(line);

                    int maxLines = CPCExport.NbCharLines * 3 * 4;
                    if (UniqueLines.Count >= maxLines)
                    {
                        System.Diagnostics.Debugger.Break();
                    }

                    index = line.Index;
                }

                frame.LinesIndex.Add(index);
            }

            Frames.Add(frame);            
        }

        public void AddFrame(int frameIndex)
        {
            var bitmap = CreateBitmapFromScreenBuffer();

            AddFrame(frameIndex, bitmap);
        }

        public void AddFrame(int frameIndex, Bitmap bitmap)
        {
            ProcessLines(bitmap, null);

            WriteBitmapToFile(bitmap, frameIndex);
        }

        public void AddFrame(int frameIndex, Bitmap bitmap, List<byte> srcY)
        {
            ProcessLines(bitmap, srcY);

            WriteBitmapToFile(bitmap, frameIndex);
        }
    }
}
