﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;

namespace Shared
{
    public class CompositeDataItem
    {
        public string Filename;
        public byte CPCBank = 0xC2;
        public ushort CPCAddress = 0x0000;
        public int DuplicateCount = 0;

        public CompositeDataItem(string filename, byte cpcBank, ushort cpcAddress, int duplicateCount=0)
        {
            Filename = DirHelper.SourceFolder + filename;
            CPCBank = cpcBank;
            CPCAddress = cpcAddress;
            DuplicateCount = duplicateCount;
        }
    }

    public static class CompositeHelper
    {
        static public void CreateCompositeData(
            string destFilename,
            List<CompositeDataItem> data)
        {
            string z80 = "";

            var bin = new List<byte>();

            Console.WriteLine("Creating Composite file '" + destFilename + "'");

            foreach (var compositeDataItem in data)
            {
                AddCompositeDataItem(-1, compositeDataItem, ref z80, ref bin);

                if(compositeDataItem.DuplicateCount>0)
                {
                    for(int iDuplicate=0; iDuplicate<compositeDataItem.DuplicateCount+1;iDuplicate++)
                    {
                        AddCompositeDataItem(iDuplicate, compositeDataItem, ref z80, ref bin);
                    }
                }
            }

            bin.Add((byte)0);

            Console.WriteLine("Composite file '" + destFilename +"' is completed");

            File.WriteAllBytes(destFilename, bin.ToArray());
            File.WriteAllText(destFilename + ".z80", z80);
        }

        static private void AddCompositeDataItem(int iDuplicate, CompositeDataItem compositeDataItem, ref string z80, ref List<byte> bin)
        {
            var fileInfo = new FileInfo(compositeDataItem.Filename);

            Shared.ToolHelper.ExecOriginalBBPack(
                fileInfo.Directory.FullName,
                fileInfo.Directory.FullName,
                fileInfo.Name);

            var binFile = File.ReadAllBytes(fileInfo.FullName + ".pck");

            ushort packedSize = (ushort)(new FileInfo(fileInfo.FullName + ".pck").Length);

            Console.WriteLine("..added '" + fileInfo.FullName + "' (raw size=&" + fileInfo.Length.ToString("X4") + " packed size=&" + packedSize.ToString("X4") + ")");

            var resourceName = fileInfo.Name.ToUpper().Replace(".", "_");
            ushort cpcAddress = compositeDataItem.CPCAddress;

            if (iDuplicate!=-1)
            {
                resourceName += "_" + iDuplicate.ToString("D2");

                cpcAddress += (ushort)(fileInfo.Length * (1 + iDuplicate));
            }

            z80 += "\n";

            if ((compositeDataItem.CPCBank == 0xC3) && (cpcAddress >= 0x4000) && (cpcAddress < 0x8000))
            {
                z80 += resourceName + "_BANK_C1    equ &C1\n";
                z80 += resourceName + "_PTR_C1     equ &" + ((ushort)((cpcAddress + 0x8000))).ToString("X4") + "\n";
                z80 += resourceName + "_BANK_C3    equ &C3\n";
                z80 += resourceName + "_PTR_C3     equ &" + ((ushort)((cpcAddress))).ToString("X4") + "\n";
            }
            else
            {
                z80 += resourceName + "_BANK       equ &" + ((byte)(compositeDataItem.CPCBank)).ToString("X2") + "\n";
                z80 += resourceName + "_PTR        equ &" + ((ushort)(cpcAddress)).ToString("X4") + "\n";
            }
            z80 += resourceName + "_RAWSIZE    equ &" + ((ushort)(fileInfo.Length)).ToString("X4") + "\n";
            z80 += resourceName + "_PACKEDSIZE equ &" + ((ushort)(packedSize)).ToString("X4") + "\n";

            bin.Add(compositeDataItem.CPCBank);
            bin.Add((byte)(cpcAddress & 255));
            bin.Add((byte)(cpcAddress >> 8));
            bin.Add((byte)(packedSize & 255));
            bin.Add((byte)(packedSize >> 8));
            bin.AddRange(binFile);
        }
    }
}
