﻿using System;
using System.Collections.Generic;
using System.Drawing;
using Shared;

namespace Shared
{
    public class PaletteEntry
    {
        public byte IndexInPalette = 0; // between 0 and 16
        public byte FirmwareColor = 255; // between 0 and 26
        public byte GateArrayColor = 255; // between 0 and 26

        public Color Color = Color.Black; // PC Color

        public bool IsDummyEntry = false;

        public PaletteEntry()
        {
        }
    }

    public class Palette
    {
        public List<PaletteEntry> PaletteEntries = new List<PaletteEntry>();

        public Palette()
        {
        }

        public Palette(int firstPensSkip)
        {
            var color = Color.Black;
            var firmwareColor = PixelHelper.GetFirmwareColor(color);
            var gateArrayColor = PixelHelper.CPCGateArrayPalette[firmwareColor];

            for (int indexInPalette = 0; indexInPalette < firstPensSkip; indexInPalette++)
            {
                var dummyPaletteEntry = new PaletteEntry()
                {
                    IsDummyEntry = true,
                    IndexInPalette = (byte)indexInPalette,
                    Color = color,
                    FirmwareColor = firmwareColor,
                    GateArrayColor = gateArrayColor
                };
                PaletteEntries.Add(dummyPaletteEntry);
            }
        }

        public PaletteEntry GetPaletteEntry(Color color)
        {
            foreach (var paletteEntry in PaletteEntries)
            {
                if (paletteEntry.IsDummyEntry)
                {
                    continue;
                }

                if (paletteEntry.Color == color)
                {
                    return paletteEntry;
                }
            }

            return null;
        }

        public PaletteEntry AddColor(Color color)
        {
            PaletteEntry paletteEntry = null;

            var firmwareColor = PixelHelper.GetFirmwareColor(color);
            var gateArrayColor = PixelHelper.CPCGateArrayPalette[firmwareColor];
            
            paletteEntry = GetPaletteEntry(color);
            if (paletteEntry != null)
            {
                return paletteEntry;
            }

            for (int iPal = 0; iPal < PaletteEntries.Count; iPal++)
            {
                if (PaletteEntries[iPal].IsDummyEntry)
                {
                    continue;
                }

                if (PaletteEntries[iPal].FirmwareColor == firmwareColor)
                {
                    return PaletteEntries[iPal];
                }
            }

            paletteEntry = new PaletteEntry()
            {
                IndexInPalette = (byte)PaletteEntries.Count,
                FirmwareColor = firmwareColor,
                GateArrayColor = gateArrayColor,
                Color = color
            };
            PaletteEntries.Add(paletteEntry);

            return paletteEntry;
        }

        public string ToZ80TableString()
        {
            var content = ";\n";

            int remaining = 16;
            foreach (var paletteEntry in PaletteEntries)
            {
                content += "\tdb COLOR" + paletteEntry.FirmwareColor + "\n";
                remaining--;
            }

            while (remaining > 0)
            {
                content += "\tdb COLOR0 ; unused\n";
                remaining--;
            }

            return content;
        }

        public string ToZ80MacroString(string upperName)
        {
            var content = ";\n";

            content += "" + upperName.Replace("-", "_") + "_COLOR0 equ COLOR" + PaletteEntries[0].FirmwareColor + "\n";

            content += "\n ; 178 NOPS (=2 lines+50 nops)\n";
            content += "MACRO _PAL_" + upperName.Replace("-","_") + "\n";

            bool first = true;

            int remaining;
            if (PaletteEntries.Count<=4)
            {
                remaining = 4;
                content += "\t_PAL_MODE1 ";
            }
            else
            {
                remaining = 16;
                content += "\t_PAL_MODE0 ";
            }
            foreach (var paletteEntry in PaletteEntries)
            {
                if (first)
                {
                    first = false;
                }
                else
                {
                    content += ", ";
                }

                content += "COLOR" + paletteEntry.FirmwareColor;
                remaining--;
            }

            while (remaining > 0)
            {
                if (first)
                {
                    first = false;
                }
                else
                {
                    content += ", ";
                }

                content += "COLOR0";
                remaining--;
            }

            content += "\n";

            content += "MEND\n";

            return content;
        }

        public byte[] ToFirmwareByteArray()
        {
            var palette = new byte[16];

            for (int i = 0; i < 16; i++)
            {
                palette[i] = 0;
            }

            int offset = 0;
            foreach (var paletteEntry in PaletteEntries)
            {
                palette[offset] = paletteEntry.FirmwareColor;
                offset++;
            }

            return palette;
        }

        public byte[] ToGateArrayByteArray()
        {
            var palette = new byte[16];

            for (int i = 0; i < 16; i++)
            {
                palette[i] = PixelHelper.CPCGateArrayPalette[0];
            }

            int offset = 0;
            foreach (var paletteEntry in PaletteEntries)
            {
                palette[offset] = PixelHelper.CPCGateArrayPalette[paletteEntry.FirmwareColor];
                offset++;
            }

            return palette;
        }
    }
}
