﻿using System;
using System.Drawing;

namespace Shared
{
	public static class PixelHelper
	{
		public static int[] CPCAsicPalette = new int[]
		{ 
			0x020702, 0x050663, 0x0507f1, 0x670600,
			0x680764, 0x6807F1,	0xFD0704, 0xFF0764,
			0xFD07F2, 0x046703,	0x046764, 0x0567f1,
			0x686704, 0x686764, 0x6867f1, 0xfd6704,
			0xfd6763, 0xfd67f1,	0x04f502, 0x04f562,
			0x04f5f1, 0x68f500,	0x68f564, 0x68f5f1,
			0xfef504, 0xfdf563,	0xfdf5f0
		};

		public static byte[] CPCGateArrayPalette = new byte[]
		{
			0x54, 0x44, 0x55, 0x5C,
			0x58, 0x5D,	0x4C, 0x45,
			0x4D, 0x56,	0x46, 0x57,
			0x5E, 0x40,	0x5F, 0x4E,
			0x47, 0x4F,	0x52, 0x42,
			0x53, 0x5A,	0x59, 0x5B,
			0x4A, 0x43,	0x4B
		};

		public static byte GetFirmwareColor(Color color)
		{
			var CPCcolor = new Color();
			int tolerance = 95;

			int colorR = color.R;
			int colorG = color.G;
			int colorB = color.B;

			int bestScore = 10000;
			byte index = 26;
			for (byte iCPCPal = 0; iCPCPal <= 26; iCPCPal++)
			{
				CPCcolor = Color.FromArgb(CPCAsicPalette[iCPCPal]);
				int CPCcolorR = CPCcolor.R;
				int CPCcolorG = CPCcolor.G;
				int CPCcolorB = CPCcolor.B;

				if ((colorR > (CPCcolorR - tolerance)) && (colorR <= (CPCcolorR + tolerance)))
				{
					if ((colorG > (CPCcolorG - tolerance)) && (colorG <= (CPCcolorG + tolerance)))
					{
						if ((colorB > (CPCcolorB - tolerance)) && (colorB <= (CPCcolorB + tolerance)))
						{
							int diffR = CPCcolorR - colorR;
							if (diffR < 0)
							{
								diffR = -colorR;
							}

							int diffG = CPCcolorG - colorG;
							if (diffG < 0)
							{
								diffG = -colorG;
							}

							int diffB = CPCcolorB - colorB;
							if (diffB < 0)
							{
								diffB = -colorB;
							}

							int score = diffR + diffG + diffB;
							if (score <= bestScore)
							{
								bestScore = score;
								index = iCPCPal;
							}

							break;
						}
					}
				}
			}

			return index;
		}

		public static byte ToMode0Byte(byte p0, byte p1)
		{
			byte v;
			v = (byte)(
				(((p0) & 1) << 7) + (((p1) & 1) << 6)
				+ (((p0 >> 2) & 1) << 5) + (((p1 >> 2) & 1) << 4)
				+ (((p0 >> 1) & 1) << 3) + (((p1 >> 1) & 1) << 2)
				+ (((p0 >> 3) & 1) << 1) + (((p1 >> 3) & 1)));

			return v;
		}

		public static byte ToMode1Byte(byte p0, byte p1, byte p2, byte p3)
		{
			byte v;
            v = (byte)(
                    (((p0) & 1) << 7) + (((p1) & 1) << 6)
                + (((p2) & 1) << 5) + (((p3) & 1) << 4)
                + (((p0 >> 1) & 1) << 3) + (((p1 >> 1) & 1) << 2)
                + (((p2 >> 1) & 1) << 1) + (((p3 >> 1) & 1)));

			return v;
		}

		public static byte ToMode2Byte(byte p0, byte p1, byte p2, byte p3, byte p4, byte p5, byte p6, byte p7)
		{
			byte v;
			v = (byte)(
				(((p0) & 1) << 7) + (((p1) & 1) << 6)
				+ (((p2) & 1) << 5) + (((p3) & 1) << 4)
				+ (((p4) & 1) << 3) + (((p5) & 1) << 2)
				+ (((p6) & 1) << 1) + (((p7) & 1)));

			return v;
		}
	}
}

