﻿using System;
using System.Threading.Tasks;
using System.IO;
using System.Diagnostics;
using System.Reflection;

namespace Shared
{
	static public class ToolHelper
	{
        public static bool StartProcessNoWait(string workingDirectory, string filename, string arguments)
        {
            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.CreateNoWindow = true;
            startInfo.UseShellExecute = false;
            //startInfo.RedirectStandardOutput = true;
            startInfo.WorkingDirectory = DirHelper.FormatDirectory(workingDirectory);
            startInfo.FileName = filename;
            startInfo.Arguments = arguments;
            try
            {
                Process.Start(startInfo);
            }
            catch (Exception e)
            {
                throw e;
                //Console.WriteLine(e.ToString());
            }

            return true;
        }
        
        public static bool StartProcess(string workingDirectory, string filename, string arguments)
		{
			ProcessStartInfo startInfo = new ProcessStartInfo();
			startInfo.CreateNoWindow = true;
			startInfo.UseShellExecute = false;
            //startInfo.RedirectStandardOutput = true;
            startInfo.WorkingDirectory = DirHelper.FormatDirectory(workingDirectory);
            startInfo.FileName = filename;
			startInfo.Arguments = arguments;
			try
			{
				using (Process exeProcess = Process.Start(startInfo))
				{
					exeProcess.WaitForExit();

                    //Console.WriteLine(exeProcess.StandardOutput.ReadToEnd());
				}
			}
			catch(Exception e)
			{
				throw e;
				//Console.WriteLine(e.ToString());
			}

			return true;
		}

		private static bool ExecMusicalLoaderTool(string toolName, string arguments)
		{
			string relativePath = "Framework/MusicalLoader/ThirdParty/";
			string workingDirectory = DirHelper.SourceFolder + relativePath;

			string binFilename = String.Empty;

			var osVersion = System.Environment.OSVersion;
			if(osVersion.VersionString.Contains("Microsoft"))
			{
				binFilename = DirHelper.GetBinaryFullName(relativePath, toolName);
			}
			else
			{
				binFilename = "/Applications/Wine.app/Contents/Resources/bin/wine";
				arguments = toolName + ".exe " + arguments;
			}

			var result = StartProcess(workingDirectory, binFilename, arguments);
			if(!result)
			{
				return false;
			}

			return true;
		}

		public static bool ExecDiscsys(string arguments)
		{
			return ExecMusicalLoaderTool("discsys", arguments);
		}

		public static bool ExecAddhead(string arguments)
		{
			return ExecMusicalLoaderTool("addhead", arguments);
		}

		public static bool ExecCpcxfs(string arguments)
		{
			return ExecMusicalLoaderTool("cpcxfs", arguments);
		}

		public static bool ExecHFEConvert(string dskFilename, string hfeFilename)
		{
			string relativePath = "../Emulation/HxCFloppyEmulator/Windows/";
			string toolName = "flopimage_convert";
			string workingDirectory = DirHelper.SourceFolder + relativePath;

			var tempFilename = workingDirectory+"image.dsk";
			File.Copy(dskFilename, tempFilename, true);

			var arguments = ". . -HFE";

			string binFilename = String.Empty;

			var osVersion = System.Environment.OSVersion;
			if(osVersion.VersionString.Contains("Microsoft"))
			{
				binFilename = DirHelper.GetBinaryFullName(relativePath, toolName);
			}
			else
			{
				binFilename = "/Applications/Wine.app/Contents/Resources/bin/wine";
				arguments = toolName + ".exe " + arguments;
			}

			var result = StartProcess(workingDirectory, binFilename, arguments);
			if(!result)
			{
				return false;
			}

			var genHfeFilename = tempFilename.Replace(".dsk", "_dsk.hfe");
			File.Copy(genHfeFilename, hfeFilename, true);
			File.Delete(tempFilename);
			File.Delete(genHfeFilename);

			return true;
		}

        public static bool ExecOriginalBBPack(string rawFolder, string compiledFolder, string filename)
        {
            var binFilename = DirHelper.GetBinaryFullName("Framework/BBStreamer/Original/BBPackOriginal/bin/", "BBPackOriginal");

            var result = StartProcess(rawFolder, binFilename, filename);
            if (!result)
            {
                return false;
            }

            DirHelper.CreateDirectory(compiledFolder);

            if (rawFolder != compiledFolder)
            {
                File.Move(rawFolder + filename + ".orig_pck", compiledFolder + filename + ".orig_pck");
            }

            return true;
        }

		public static bool ExecBBPack(string rawFolder, string compiledFolder, string filename)
		{
			var binFilename = DirHelper.GetBinaryFullName("Framework/BBStreamer/GetXBytesPerCalls/BBPack/bin/", "BBPack");

			var result = StartProcess(rawFolder, binFilename, filename);
			if(!result)
			{
				return false;
			}

			DirHelper.CreateDirectory(compiledFolder);

            if (rawFolder != compiledFolder)
            {
                File.Move(rawFolder + filename + ".pck", compiledFolder + filename + ".pck");
            }

			return true;
		}

		public static bool ExecBBPack256(string rawFolder, string compiledFolder, string filename)
		{
			var binFilename = DirHelper.GetBinaryFullName("Framework/BBStreamer/GetXBytesPerCalls_DestInCycling256Buffer/BBPack256/bin/", "BBPack256");

			var result = StartProcess(rawFolder, binFilename, filename);
			if(!result)
			{
				return false;
			}

			DirHelper.CreateDirectory(compiledFolder);

            if (rawFolder != compiledFolder)
            {
                File.Move(rawFolder + filename + ".pck", compiledFolder + filename + ".pck");
            }

			return true;
		}
	}
}

